/*****************************************************************************
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT")


  File:   alertd.c
  Author: Andrew Lambeth <andrew@nfr.net>

 *****************************************************************************/

#include "sys-include.h"
/* #define DEBUG_LEV 1 */
#include "libnfr.h"
#include "nfr_res.h"
#include "alertqueue.h"
#include "fac.h"
#include "alertd.h"

PAlertd gpAlertd = NULL;

int Alertd_AlertdErr(char *File, int Line, ULONG AlertCode, ...)
{
  /* must be reentrant! */
  va_list ap = NULL;
  BYTE Buf[MAXSIZE_ALERT];

  va_start(ap, AlertCode);
  Alert_Assemble(Buf, sizeof(Buf), File, Line, 
		 NFR_SRC_ALERTD, AlertCode, NULL, ap);
  va_end(ap);

  return Alertd_ProcAlert((PAlert)Buf);
}

int Alertd_OnData(PDGram pDGram, PDGServer pDGSrv)
{
  PAlert pAlert = NULL;
  ULONG Result;

  pAlert = (PAlert) DGram_GetData(pDGram);
  Result = Alertd_ProcAlert(pAlert);
  if(Result != NFR_ERR_SUCCESS)
    return Result;

  /* check if we need to wake up each queue so it can possibly flush itself */
  if (pAlert->Timestamp >= gpAlertd->NextFlushTime)
    Alertd_OnIdle(pDGSrv);
  
  return NFR_ERR_SUCCESS;
}

void Alertd_Bail(ULONG ErrCode)
{
  /* go here as a last resort, when nothing's working */
  time_t Time = 0;
  char TimeStr[32];

  Time = time(0);
  strcpy(TimeStr, ctime(&Time));
  TimeStr[19] = 0;

  /* this may not be that useful except that it looks scary and may
     attract the required attention */
  fprintf(stderr, "%s alertd[%ld]: FATAL ERROR 0x%.8lX, alertd exiting.\n",
	     TimeStr + 4, getpid(), ErrCode);
  ConsoleMsg("%s alertd[%ld]: FATAL ERROR 0x%.8lX, alertd exiting.",
	     TimeStr + 4, getpid(), ErrCode);

  exit(ErrCode);
}

int Alertd_ProcAlert(PAlert pAlert)
{
  PAlertRes pAlertRes = NULL;
  PAlertSrc pAlertSrc = NULL;
  int i = 0, j = 0, k = 0;
  time_t TimeNow = 0;
  static FILE *SerialFile = NULL;
  ULONG NBOSerial = 0;
  static char *Path = NULL;
  static int recursions = 0; 
  ULONG Result = 0;
  PfnFacCall FacAlternates[] = { Fac_NfrLog, Fac_Console, NULL };

  if(recursions > MAX_ALERT_RECURSIONS + 1)
    {
      /* the call below must have caused more problems */
      /* Alertd_Bail(NFR_ALERTD_RECURSION); */
      recursions = 0;
      return NFR_ALERTD_RECURSION;
    }

  if(recursions++ == MAX_ALERT_RECURSIONS)
    {
      /* once more, to attempt to report the fact, and then bail, this
         will usually fail and be caught above, but hey, we tried */
      Result = Alertd_AlertdErr(__FILE__,__LINE__,
				NFR_ALERTD_RECURSION, NULL);
      if(Result != NFR_ERR_SUCCESS)
	{
	  recursions = 0;
	  return NFR_ALERTD_RECURSION;
	}
      /* Alertd_Bail(NFR_ALERTD_RECURSION); */
    }

  TimeNow = time(NULL);

  if(TimeNow > gpAlertd->DefLog->nextRotation)
    {
      Log_Rotate(gpAlertd->DefLog);
    }
      

  gpAlertd->Serial++;
  

  /* since this came up out of the fifo from another process the address
     pAlert->Inserts needs to be reset. we know where the inserts
     start 'cause they immediately follow the TAlert struct in memory */
  /* NOTE WELL:  this will bite the hell out of us if we ever do alert 
     forwarding by just sending the uncooked alert over the wire,
     since the sizeof(TAlert) may well be different on different 
     on different architectures :( */
  pAlert->Inserts = (BYTE *) pAlert + sizeof(TAlert); 
  
  
  if(gpAlertd->TraceOn)
    pAlertRes = Alert_ToAlertResEx(ALERT_PRINT_SOURCE_INFO, pAlert);
  else 
    pAlertRes = Alert_ToAlertRes(pAlert);
  
  if(pAlertRes->ResCode == DEF_RESCODE)
    pAlertRes->Severity = SEV_ERROR;
  pAlertSrc = Alert_ToAlertSrc(pAlert);

  
  if(gpAlertd->FacCount < 1 || pAlertRes->ResCode == DEF_RESCODE)
    {
      /* there are no facilities loaded, or this is the default alert
         resource, so we default to just using the log */
      Fac_NfrLog(gpAlertd->Serial,
		 pAlert->Timestamp,
		 pAlertRes,
		 pAlertSrc,
		 NULL);
    }
  
  else
    {
      /* send the alert to each facility listed in the AlertRes */
      i = pAlertRes->FacCodeCount;
      while (i--)
	{
	  j = pAlertRes->FacCodes[i];
	  k = 0;
	  while(TRUE)
	    {
	      /* try to send the alert to the requested facility, if
                 that fails, then reroute the facility and try again */
	      if((j <= gpAlertd->FacCount) && 
		 (gpAlertd->FacHandles[j].FacCall))
		{
		  Result = 
		    gpAlertd->FacHandles[j].FacCall(gpAlertd->Serial,
						 pAlert->Timestamp,
						 pAlertRes,
						 pAlertSrc,
						 &(gpAlertd->FacHandles[j]));
		  if(Result == NFR_ERR_SUCCESS)
		    break; /* that one works */

		  /* bug: need to report that a facility has been
                     rerouted */
		  if(k >= gpAlertd->FacCount)
		    {
		      DBG_OUT("couldn't reroute facility\n", i, k);
		      break; /* we're out of facilities to try */
		    }
		  DBG_OUT("rerouting facility %d to %d\n", i, k);
		  if(gpAlertd->FacHandles[j].AlertQueue)
		    {
		      free(gpAlertd->FacHandles[j].AlertQueue);
		      gpAlertd->FacHandles[j].AlertQueue = NULL;
		    }
		  gpAlertd->FacHandles[j] = gpAlertd->AltFacHandles[k];
		  k++;
		}
	      else
		break;
	    }
	}
    }

  /* now we check to see if this is one of the "magic" alerts that 
     triggers some action */
  switch(pAlert->ResCode)
    {
    case ALERTD_CTRL_STOP:
      exit(Alertd_OnStop(gpAlertd->Daemon));
      break;

    case ALERTD_CTRL_CONFIG:
      Alertd_Unconfig();
      Alertd_Config();
      ResFiles_Refresh();
      break;

    case ALERTD_CTRL_ROTATE:
      Log_Rotate(gpAlertd->DefLog);
      break;

    default:
      break;
    }

  AlertRes_Close(pAlertRes);
  AlertSrc_Close(pAlertSrc);
  
  /* now checkpoint the serial number */
  if(!SerialFile)
    {
      if(!Path)
        Path = nfrPath("", "data/alerts/", gpAlertd->Hostname, 
		       "/serial.dat", NULL);
      SerialFile = fopen(Path, "w");
    }
  if(!SerialFile)
    {
      if(Path)
	{
	  free(Path);
	  Path = NULL;
	}
      fprintf(stderr, "alertd: Warning: can't write file: %s error: %s\n", 
	      "data/alerts/serial.dat", strerror(errno));      
      ConsoleMsg("alertd: Warning: can't write file: %s error: %s", 
		 "data/alerts/serial.dat", strerror(errno));      
    }
  else
    {
      rewind(SerialFile);
      NBOSerial = htonl(gpAlertd->Serial);
      fwrite(&NBOSerial, 1, sizeof(NBOSerial), SerialFile);
    }
  
  recursions--;
  return NFR_ERR_SUCCESS;
}
  
/* this gets called roughly every Alertd->FlushGrain usecs */
int Alertd_OnIdle(PDGServer pDGSrv)
{
  int i = 0;
  time_t TimeNow = 0;

  TimeNow = time(NULL);
  /* DBG_OUT("\n%s - checking for flushable queues\n", ctime(&TimeNow)); */
  i = gpAlertd->FacCount;
  while (i--)
    {
      if(gpAlertd->FacHandles[i].AlertQueue != NULL)
	AlertQueue_Touch(TimeNow, gpAlertd->FacHandles[i].AlertQueue);
    }
  gpAlertd->NextFlushTime = TimeNow + (gpAlertd->FlushGrain / 1000);
  return NFR_ERR_SUCCESS;
}

/* this gets called on any error condition raised during the daemon's
   sleep/read loop */
int Alertd_OnErr(ULONG Error, PDGServer pDGSrv)
{
  switch(Error)
    {
    case LIBNFR_ERR_EOF:
    case LIBNFR_ERR_AGAIN:
      /* if we just got an EOF in the pipe and nothing else, that's
	 ok.  it happens whenever a client closes its end of the pipe,
	 also under bsdi we get EAGAIN sometimes */
      return NFR_ERR_SUCCESS;

    case NFR_ALERTD_RECURSION:
      /* nothing to do, it's already been reported, if possible */
      return NFR_ERR_SUCCESS;	
      
    default:
      Alertd_OnBadFifo(Error);
      return NFR_ERR_SUCCESS;
    }
}

int Alertd_OnBadFifo(ULONG Error)
{
  ULONG Result = NFR_ERR_SUCCESS;

  Alertd_AlertdErr(__FILE__,__LINE__, NFR_ALERTD_BADFIFO, 
		   gpAlertd->InFifo, NULL);

  /* bug: this function doesn't work */

  Fifo_Destroy(gpAlertd->InFifo);
  /* bug: should try to dump anything that's in there somewhere for a
     little light postmortem reading */

  Result = Fifo_Create(gpAlertd->InFifo);
  if(Result != NFR_ERR_SUCCESS)
    return Result;

  return NFR_ERR_SUCCESS;
}

int Alertd_Main(PDaemon pDaemon)
{
  int Result = 0;
  TDGServer *pDGSrv = NULL;
  FILE *PidFile = NULL;
  char *Path = NULL;

  Result = Fifo_Create(gpAlertd->InFifo);
  if (Result != LIBNFR_ERR_SUCCESS)
    {
      Alertd_AlertdErr(__FILE__,__LINE__,
		       NFR_ERR_CANTOPENFIFO, gpAlertd->InFifo, NULL);
      return Result;
    }

  Path = nfrPath("", ALERTD_PIDFILE, NULL);
  PidFile = fopen(Path, "w");
  if(!PidFile)
    Alertd_AlertdErr(__FILE__,__LINE__, NFR_ERR_CANTWRITE, Path, NULL);
  else
    {
      fprintf(PidFile, "%d", getpid());
      fclose (PidFile);
    }
  free(Path);

  /* now set this so that we don't do fopen()s for every alert that
     comes in */
  gKeepOpenFlag = TRUE;

  pDGSrv = DGServer_New();

  pDGSrv->fdRead = Fifo_OpenRead(gpAlertd->InFifo);
  pDGSrv->DataProc = Alertd_OnData;
  pDGSrv->IdleProc = Alertd_OnIdle;
  pDGSrv->ErrProc = Alertd_OnErr;
  pDGSrv->MaxWait = gpAlertd->FlushGrain;

  Result = DGServer_Run(pDGSrv);

  close(pDGSrv->fdRead);
  DGServer_Delete(pDGSrv);
  return Result;
}

int Alertd_OnStart(PDaemon pDaemon)
{
  int Result = 0;
  char *Path = NULL;
  FILE *SerialFile = NULL;
  ULONG NBOSerial = 0;
  int pid = 0;
  FILE *PidFile = NULL;
  char pidstr[12];

  gpAlertd->InFifo = ALERTD_FIFO;

  /* check to be sure that there's not already one running */
  
  Path = nfrPath("", ALERTD_PIDFILE, NULL);
  PidFile = fopen(Path, "r");
  if(PidFile)
    {
      fscanf(PidFile, "%d", &pid);
      fclose (PidFile);
      if (kill(pid,0) == 0)
	{
	  sprintf(pidstr, "%ld", pid);
	  Alert_Args_GenerateEx(__FILE__,__LINE__, 
				NFR_SRC_ALERTD, NFR_ALERTD_PROCEXIST, 
				pidstr, NULL);
	  ConsoleMsg("alertd: Error: "
		     "another alertd process is already running, pid: %d\n", 
		     pid);
	  fprintf(stderr, "alertd: Error: "
		  "another alertd process is already running, pid: %d\n", pid);
	  return(1);
	}
    }
  free(Path);

  remove(gpAlertd->InFifo);

  gpAlertd->FlushGrain = ALERTD_FLUSHGRAIN;

  Path = nfrPath("", "data/alerts/", gpAlertd->Hostname, "/serial.dat", NULL);
  SerialFile = fopen(Path, "r");
  if(!SerialFile)
    {
      gpAlertd->Serial = ntohl(NBOSerial);
      Alertd_AlertdErr(__FILE__,__LINE__,NFR_ALERTD_STARTED, NULL); 
      ConsoleMsg("alertd: Warning: unable to read file \"%s\"\n"
		 "initializing alert serial number to %d\n", 
		 Path, ntohl(NBOSerial));
      fprintf(stderr, "alertd: Warning: unable to read file \"%s\"\n"
	      "initializing alert serial number to %d\n", 
	      Path, ntohl(NBOSerial));
    }
  else
    {
      fread(&NBOSerial, 1, sizeof(NBOSerial), SerialFile);
      gpAlertd->Serial = ntohl(NBOSerial);
      Alertd_AlertdErr(__FILE__,__LINE__,NFR_ALERTD_STARTED, NULL); 

      fclose(SerialFile);
    }
  free(Path);

  return Alertd_Config();
}

int Alertd_Config()
{
  int Result = 0;
  char *Path = NULL;
  PCf pCf = NULL;

  pCf = Cf_New(0);

  Result = Cf_Read("etc/resources/facilities.cf", pCf);
  /* bug fix these error messages */
  if(Result != NFR_ERR_SUCCESS)
    {
      fprintf(stderr, "\nalertd: Warning: unable to read file \"%s\", "
	      "\nall alerts will be handled by log facility only\n", 
	      "etc/resources/facilities.cf");  
    }
  else if (pCf->SymTab->SymList->Count < 1)
    {
      fprintf(stderr, "\nalertd: Warning: no facilities found in file \"%s\", "
	      "\nall alerts will be handled by log facility only\n", 
	      "etc/resources/facilities.cf");  
    }
  else /* success */
    {
      gpAlertd->FacCount = pCf->SymTab->SymList->Count + 1;
      gpAlertd->FacHandles = (PFacHandle) calloc(gpAlertd->FacCount, 
						 sizeof(TFacHandle));
      Cf_ForEach(Alertd_LoadFac, NULL, pCf);
    }
  
  /* bug: replace the following block with something user configable */
  {
    char *Num = NULL;

    gpAlertd->AltFacCount = 3;
    gpAlertd->AltFacHandles = (PFacHandle) calloc(gpAlertd->AltFacCount,
						  sizeof(TFacHandle));
    Num = Cf_GetVal("NFR_FAC_NFRLOG", pCf);
    if(Num)
      memcpy(&(gpAlertd->AltFacHandles[0]), 
	     &(gpAlertd->FacHandles[atoi(Num)]), 
	     sizeof(TFacHandle));
    Num = Cf_GetVal("NFR_FAC_CONSOLE", pCf);
    if(Num)
      memcpy(&(gpAlertd->AltFacHandles[1]), 
	     &(gpAlertd->FacHandles[atoi(Num)]), 
	     sizeof(TFacHandle));

    /* leave the last one blank, since nothing is the only thing
       that's truly guaranteed to work */
  }

  Cf_Delete(pCf);

  return Result;
}

void Alertd_Unconfig()
{
  int i = 0;

  i = gpAlertd->FacCount;
  while (i--)
    {
      if (gpAlertd->FacHandles[i].AlertQueue)
	AlertQueue_Delete(gpAlertd->FacHandles[i].AlertQueue);
    }

  if(gpAlertd->FacHandles)
    {
      free(gpAlertd->FacHandles);
      gpAlertd->FacHandles = NULL;
    }

  i = gpAlertd->AltFacCount;
  while (i--)
    {
      if (gpAlertd->AltFacHandles[i].AlertQueue)
	AlertQueue_Delete(gpAlertd->AltFacHandles[i].AlertQueue);
    }

  if(gpAlertd->AltFacHandles)
    {
      free(gpAlertd->AltFacHandles);
      gpAlertd->AltFacHandles = NULL;
    }
}

int Alertd_OnStop(PDaemon pDaemon)
{
  Fifo_Destroy(gpAlertd->InFifo);

  Alertd_AlertdErr(__FILE__,__LINE__,NFR_ALERTD_STOPPED, NULL); 

  Alertd_Unconfig();

  free(gpAlertd);
  return NFR_ERR_SUCCESS;
}

int Alertd_LoadFac(PNVPair pNV, void *PArg)
{
  int i = 0;

  if(!pNV->Name || !pNV->Val)
    return NFR_ERR_UNDEFINED;
  DBG_OUT("%s - %s\n", pNV->Name, pNV->Val);

  i = atoi(pNV->Val);
  if(i > gpAlertd->FacCount)
    return NFR_ERR_UNDEFINED;

  if(!strcasecmp(pNV->Name, "NFR_FAC_NULL"))
    gpAlertd->FacHandles[i].FacCall = NULL;
  else if(!strcasecmp(pNV->Name, "NFR_FAC_CONSOLE"))
    gpAlertd->FacHandles[i].FacCall = Fac_Console;
  else if(!strcasecmp(pNV->Name, "NFR_FAC_SYSLOG"))
    gpAlertd->FacHandles[i].FacCall = Fac_SysLog;
  else if(!strcasecmp(pNV->Name, "NFR_FAC_NFRLOG"))
    gpAlertd->FacHandles[i].FacCall = Fac_NfrLog;
  else if(!strcasecmp(pNV->Name, "NFR_FAC_ACK"))
    gpAlertd->FacHandles[i].FacCall = Fac_Ack;
  else if(!strcasecmp(pNV->Name, "NFR_FAC_TRACE"))
    {
      if(gpAlertd->TraceOn)
	gpAlertd->FacHandles[i].FacCall = Fac_NfrLog;
      else
	gpAlertd->FacHandles[i].FacCall = NULL;
    }
  else
    {
      gpAlertd->FacHandles[i].FacCall = Fac_Generic;	  
      gpAlertd->FacHandles[i].AlertQueue = AlertQueue_New();
      AlertQueue_Init(pNV->Name, gpAlertd->FacHandles[i].AlertQueue);
    }
  return NFR_ERR_SUCCESS;
}

int Usage(void)
{
  fprintf(stderr, 
	  "\nUsage: alertd [-tf | -krR]\n"
	  "\nOptions:\n"
	  "\tt - run in trace mode\n"
	  "\tf - run in forward mode\n"
	  "\tk - tell currently running alertd process to stop\n"
	  "\tr - tell currently running alertd process to "
	  "reread config files\n"
	  "\tR - tell currently running alertd process to "
	  "reopen output files\n\n");
  return 1;
}

int main(int argc, char **argv)
{
  PDaemon pDaemon = NULL;
  int Result = 0, i = 0;
  PCf pCf = NULL;
  char *p = NULL;
  BOOL TraceOn = FALSE;
  BOOL Forward = FALSE;
  char buf[256];
  struct hostent *he;


  i = 1;			/* parse cmd line */
  while(i < argc)
    {
      p = argv[i];
      if (*p == '-')
	{
	  while (*++p != 0)
	    {
	      switch (*p)
		{
		case 't':
		  TraceOn = TRUE;
		  break;

		case 'f':
		  Forward = TRUE;
		  break;

		case 'k':
		  exit( Alert_Args_GenerateEx(__FILE__,__LINE__,
					      NFR_SRC_ALERTD,
					      ALERTD_CTRL_STOP, 
					      NULL) );
		  
		case 'r':
		  exit( Alert_Args_GenerateEx(__FILE__,__LINE__,
					      NFR_SRC_ALERTD,
					      ALERTD_CTRL_CONFIG, 
					      NULL) );
		  
		case 'R':
		  exit( Alert_Args_GenerateEx(__FILE__,__LINE__,
					      NFR_SRC_ALERTD,
					      ALERTD_CTRL_ROTATE, 
					      NULL) );

		default:
		  exit(Usage());		  
		}
	    }
	}
      i++;
    }


  pDaemon = Daemon_New();
  pDaemon->Main = Alertd_Main;
  pDaemon->OnStop = Alertd_OnStop;
  pDaemon->OnStart = Alertd_OnStart;
  gpAlertd = (PAlertd) calloc(1, sizeof(TAlertd));
  gpAlertd->TraceOn = TraceOn;
  gpAlertd->Forward = Forward;
  gpAlertd->DefLog = calloc(1, sizeof(TLog));
  gpAlertd->Daemon = pDaemon;
  gethostname(buf, sizeof(buf) - 1);
  he = gethostbyname(buf);
  if(he)
    {
      gpAlertd->Hostname = calloc(1, strlen(he->h_name) + 1);
      strcpy(gpAlertd->Hostname, he->h_name);
    }
  else
    {
      fprintf(stderr, "alertd: Warning: can't get fully qualified name for host \"%s\"\n", buf);
      gpAlertd->Hostname = calloc(1, strlen(buf) + 1);
      strcpy(gpAlertd->Hostname, buf);
    }

  Result = Daemon_Start(pDaemon, 30000);

  /* parent only from here down */

  Alertd_Unconfig();

  free(gpAlertd->DefLog);
  free(gpAlertd);
  Daemon_Delete(pDaemon);
  exit(0);
}
