/*
        Copyright(C) 1997 Network Flight Recorder, Inc.  
        All rights reserved.

        Use and distribution of this software and its source code
        are governed by the terms and conditions of the
        Network Flight Recorder Software License ("LICENSE.TXT")

By Mark Sienkiewicz / NFR

*/

/*

Simple sparse arrays

See interface definition in array.h

You can replace this function set with any compatible array
implementation as you like.

This particular implementation is based on "Skip Lists".  Some
interesting papers on skip lists have been published.  Some are 
available from the University of Maryland Institute for Advanced
Computer Studies.  (UMIACS)  I did a web search for "skip list"
and "university of maryland" and found some good stuff.

Don't take this code as an example of how to do skip lists, since
I am now writing it from memory of how I think they work from
reading a paper on extending their functionality.

*/

#include "sys-include.h"
#include "array.h"

#define MAX_FORWARD_POINTERS 8


/*
* some performance monitoring
*/
static int lookups = 0, lookupcompares = 0 , adds = 0, addcompares = 0;


array_dump_performance(FILE *f)
{
fprintf(f,"skiplist: %d %d %d %d",lookups,lookupcompares,adds,addcompares);
fprintf(f,"\n");
}

struct array_node
        {
        int subscript_size;
	struct array_node *forward_pointers[MAX_FORWARD_POINTERS];
        int data_size;
        void * data;
        unsigned char subscript[1];
        };

struct array
	{
	int level_subscript;	
			/* what subscript is the highest interesting level */
	struct array_node dummy_first;
	};

struct array_scan
	{
	struct array_node * node;
	};

array_t
array_create(array_tag_t tag)
{
array_t a;
a=array_tag_alloc(tag,sizeof(struct array));
a->level_subscript=0;
return a;
}

int array_binary_cmp(void *first, int first_size, 
		void *second, int second_size)
{
	/*
	* if equal size, just compare them
	*/
if (first_size == second_size)
	return memcmp(first,second, first_size);

	/*
	* if we are comparing "AAA" to "AAAAA"
	* then compare the shorter string.  If unequal, it is something
	* like "BBB" and "CCCC" or "DDD" and "CCCC", so return the
	* result.  Otherwise, it is like "AAA" and "AAAA" so the
	* smaller one is less.
	*/
if (first_size < second_size)
	{
	int x = memcmp(first, second, first_size);
	if (x == 0)
		return -1;
	else
		return x;
	}
else
	{
	int x = memcmp(first, second, second_size);
	if (x == 0)
		return 1;
	else
		return x;
	}
}


static int getrandomlevel()
{
int l = 0;
#ifdef RAND_MAX
#define portion (RAND_MAX/4)
#else
#define portion (65536/4)
#endif
while ((rand() < portion ) && (l < MAX_FORWARD_POINTERS))
	l++;
return l;
}


enum array_return
array_find(array_t a, array_node *rn, 
	void *subscript, int subscript_size, int create, array_tag_t tag,
	array_subscript_compare comparefn)
{
array_node n, prev = (void *)0, newn;
enum array_return rval = array_ok;
int level;
array_node update[MAX_FORWARD_POINTERS];
int cmp,x,lvl;
int comparecount = 0;

*rn = (void *)0;


/*
* walk down the list looking for it
*/
n = &a->dummy_first;

for (level = a->level_subscript; level >= 0; level -- )
	{
	array_node t;
	int foo=1;
	t=n->forward_pointers[level];
	for (;;)
		{
		if (!t)
			break;
		comparecount++;
		cmp = (*comparefn)(subscript,subscript_size,
			t->subscript, t->subscript_size);
		if (cmp <= 0)
			break;
		n = t;
		t= n->forward_pointers[level];
		}
	if (t && cmp == 0)
		{
		lookups++;
		lookupcompares += comparecount;
		*rn = t;
		return array_ok;
		}
	update[level] = n;
	}

fflush(stdout);

/*
* if we get here, we passed it
*/

if (! array_create)
	{
	/*
	* didn't find it and not going to create it
	*/
	lookups++;
	lookupcompares += comparecount;
	*rn= (array_node)0;
	return array_notfound;
	}

/*
* found it and going to create
*
* first: create the node and fill it.
*/

adds++;
addcompares += comparecount;

newn=array_tag_alloc(tag,sizeof(struct array_node)+subscript_size);
if (!newn)
	{
	*rn= (array_node)0;
	return array_error;
	}
memcpy(newn->subscript,subscript,subscript_size);
newn->subscript_size = subscript_size;
newn->data= (void *) 0;
newn->data_size = 0;
*rn = newn;

/*
* next: insert it in the list
*/
lvl = getrandomlevel();
if (lvl > a->level_subscript)
	{
	lvl = a->level_subscript+1;
	if (lvl > MAX_FORWARD_POINTERS-1)
		lvl = MAX_FORWARD_POINTERS - 1;
	a->level_subscript = lvl;
	update[lvl] = &a->dummy_first;
	}

for (x = 0; x <= lvl; x++)
	newn->forward_pointers[x] = update[x]->forward_pointers[x];
for (x = 0 ; x <= lvl; x++)
	update[x]->forward_pointers[x] = newn;

return array_new;
}


enum array_return
array_get(array_node n,
	void **subscript, int *actual_subscript_size,
	void *data, int max_data_size, int *actual_data_size)
{
enum array_return rval = array_ok;
if (actual_subscript_size)
	*actual_subscript_size = n->subscript_size;
if (actual_data_size)
	*actual_data_size = n->data_size;
if (subscript)
	{
	*subscript = n->subscript;
	}
if (data)
	{
	if (n->data_size > max_data_size)
		rval=array_error;
	else
		memcpy(data,n->data,n->data_size);
	}
return rval;
}

enum array_return
array_set(struct array_node *n, void *data, int data_size, 
	array_tag_t tag)
{
void *t;
if (n->data_size != data_size)
	{
	t=array_tag_alloc(tag,data_size);
	if (!t)
		return array_error;
	/* free(n->data); bug: leak! */
	n->data=t;
	n->data_size=data_size;
	}
memcpy(n->data, data, data_size);
return array_ok;
}

struct array_node *
array_scan(struct array_scan *s)
{
struct array_node *n;
if (!s || !s->node)
	return (void *)0;
n=s->node;
if (n)
	s->node=n->forward_pointers[0];
return n;
}

struct array_scan *
array_start_scan(struct array *a)
{
struct array_scan *t;
t=calloc(1,sizeof(struct array_scan));
if (!t) return t;
t->node= a->dummy_first.forward_pointers[0];
return t;
}

void array_end_scan(struct array_scan *a)
{
free(a);
}

#define MAX_CHUNKS 500
struct array_tag
	{
	struct array_tag_sub *list;
	};

struct array_tag_sub
	{
	int used;
	struct array_tag_sub *next;
	void *chunks[MAX_CHUNKS];
	};

static struct array_tag_sub *
array_tag_new_chunk()
{
struct array_tag_sub *t;
t=calloc(1,sizeof(struct array_tag_sub));
if (!t)
	return (void *)0;	/* bug: handle error?  leave to caller? */
t->used=0;
t->next=(void *)0;
return t;
}

array_tag_t array_tag_create()
{
array_tag_t t;
t=calloc(1,sizeof(struct array_tag));
if (!t)
	return (void *)0;	/* bug: handle error?  leave to caller? */
t->list=array_tag_new_chunk();
if (!t->list)
	{ free(t); return (void *)0; }	/* bug: handle error? */
else
	return t;
}


void *array_tag_alloc(array_tag_t tag, size_t size)
{
void *t;
struct array_tag_sub *u;
t= calloc(1,size);
if (!t)
	return t;		/* bug: ? */
if (!tag)
	return t;
u=tag->list;
if (u->used >= MAX_CHUNKS)
	{
	struct array_tag_sub *new=array_tag_new_chunk();
	if (!new)
		{ free(t); return (void *)0; }
	new->next = tag->list;
	tag->list=u=new;
	}
u->chunks[u->used++] = t;
return t;
}

void array_tag_free(array_tag_t tag, void *ptr)
{
/* bug: should really free it; find it in the list too */
}


void array_tag_purge(array_tag_t tag)
{
struct array_tag_sub *t,*tp;
int x;
if (!tag)
	return;

for (t=tag->list; t; )
	{
	for (x=t->used - 1; x>=0; x--)
		free(t->chunks[x]);
	tp=t;
	t=t->next;
	free(tp);
	}
free(tag);
}

