/*
        Copyright(C) 1997 Network Flight Recorder, Inc.  
        All rights reserved.

        Use and distribution of this software and its source code
        are governed by the terms and conditions of the
        Network Flight Recorder Software License ("LICENSE.TXT")

By Mark Sienkiewicz / NFR

*/

/*

Simple sparse arrays

*/

/*
* tags are used internally to associate memory allocated by the array.
* create a tag with array_tag_create();  Then pass it to all the
* array calls.  Then call array_tag_purge() to free all memory
* allocated in association with that tag.
*/
typedef struct array_tag *array_tag_t;

array_tag_t array_tag_create() ;
void * array_tag_alloc(array_tag_t tag, size_t size);
void array_tag_free(array_tag_t tag, void *ptr);
void array_tag_purge(array_tag_t tag);


/*
* these are return values from some of the array functions
*/
enum array_return 
	{
	array_ok,	/* success */
	array_new,	/* success, added new node to array */
	array_error, 	/* failure */
	array_notfound	/* no element at requested subscript */
	};

/*
* array_t is the actual array
*/
typedef struct array * array_t;

/*
* array_node is an element in the array
*/
typedef struct array_node * array_node;

/*
* array_scan_t is a descriptor for sequentially stepping through the
* array fetching successive elements
*/
typedef struct array_scan * array_scan_t;

/*
* array_create creates an empty array
*/
array_t array_create(array_tag_t);

/*
* array_find locates an element in the array given a subscript.  Note
* you must always give the subscript size because it is not necessarily
* the same for all elements.
*
* if create is non-zero, the element is created.
*
* returns
*	array_ok	found element, *n is the element
*	array_new	did not find element, but created it.
*	array_notfound	did not find element, did not create. 
*	array_error	could not find/add element for some reason...
*			- currently only because malloc failed
*
* The value of *n will be NULL if the element was not found or created.
* 
*/

typedef int (*array_subscript_compare)(
	void *first, int first_size, void *second, int second_size);

enum array_return array_find(array_t a, array_node *n, void *subscript, 
	int subscript_size, int create, array_tag_t, array_subscript_compare c);


/*
* array_get fetches values from an array_node.
*
* subscript		address of a pointer.  After the call, subscript
*			will point to the subscript field in the array node.
*			DON'T change the subscript - you will break it.  
*			This feature exists so you can read the value of 
*			the subscript.
*			into (NULL if you don't want to fetch it)
* actual_subscript_size	pointer to int to store actual size of subscript in
*			This is how many bytes long the subscript is
*
* data			pointer to buffer to fetch the data value
*			into (NULL if you don't want to fetch it)
* max_data_size		size of buffer
* actual_data_size	pointer to int to store actual size of data in
*
* Any of the pointers can be NULL if you don't want to fetch the
* associated data.
*
* Returns:
*	array_error	the subscript or data would not fit into the buffer
*			specified. Note: all values which could be 
*			successfully fetched were. The actual_..._size 
*			values are always fetched.
*	array_ok	values fetched ok
*/
enum array_return array_get(array_node n,
        void **subscript, int *actual_subscript_size,
        void *data, int max_data_size, int *actual_data_size);

/*
* array_set sets the data value of a particular node.
*
* returns:
*	array_error	malloc failed trying to grow the data element
*			to the requested size.  The old value remains.
*	array_ok	data set sucessfully
*/
enum array_return array_set(array_node n, void *data, int data_size,
	array_tag_t tag);


/*
* To step through an array and see all the contents, use this sequence:
*
* array_scan_t x;
* array_node n;
* x=array_start_scan(a);
* while (n=array_scan(x))
*	look at n
*
* array_start_scan returns NULL on malloc error.
*
* The scan takes place in order of increasing subscript value.
*
* You can have multiple concurrent scans of the same array.
*/
array_scan_t array_start_scan(array_t a);
array_node array_scan(struct array_scan *s);
void array_end_scan(struct array_scan *a);


/*
* comparison routines take this form:
*
*/
int array_binary_cmp( void *first, int first_size,
	void *second, int second_size );
