/*****************************************************************************
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT")


  File:   alertsrc.h
  Author: Andrew Lambeth <andrew@nfr.net>

 *****************************************************************************/

/******************************************************************************
 * alertsrc.h - functions and structures used in handling alert source 
 * resources.
 * Resource code diagram:
 *
 *  3 3 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1
 *  1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0
 * + - - - - - - - - - +-----------+-------------------------------+
 * | res-specific use  | Res Type  |         Res ID                |
 * + - - - - - - - - - +-----------+-------------------------------+
 *
 * AlertSrc:          
 * +-------------------+                                            
 * |      unused       |                                            
 * +-------------------+
 *                      
 *****************************************************************************/
#ifndef ALERTSRC_H_
#define ALERTSRC_H_

#include "xtypes.h"
#include "rescodes.h"
#include "resid.h"

#define DEF_ALERTSRC_ID_STRING "DEF_ALERTSRC_ID"
#define DEF_ALERTSRC_NAME "<rescode 0x%.8lX>"
#define DEF_ALERTSRC_DESC "<resource unavailable for alert source 0x%.8lX>"

typedef struct TAlertSrc
{
  RESID_REQ_FIELDS;		/* see resid.h */
  pid_t Pid;
  char *Name;
  USHORT NameLen;
  char *Desc;
  USHORT DescLen;
} TAlertSrc, *PAlertSrc;

ULONG AlertSrc_WriteBody(PAlertSrc , ULONG *,  FILE *);

ULONG AlertSrc_ReadBody(PAlertSrc , ULONG *, FILE *);

/******************************************************************************
 * AlertSrc_OpenEx(ULONG AlertSrcCode, BOOL KeepOpen) 
 * - creates a TAlertSrc struct and
 * initializes it with the resource information corresponding to
 * AlertSrcCode. On success the a pointer to the new TAlertSrc struct is
 * returned.  If KeepOpen is TRUE the resource file is kept open */
PAlertSrc AlertSrc_OpenEx(ULONG, BOOL);
#define AlertSrc_Open(U) AlertSrc_OpenEx(U, FALSE)

/******************************************************************************
 * AlertSrc_Close(PAlertSrc pAlertSrc) - performs cleanup and destroys the
 * TAlertSrc struct pointed to by pAlertSrc.  */
void AlertSrc_Close(PAlertSrc);

/******************************************************************************
 * AlertSrc_SetName(char *Text, USHORT Size, PAlertSrc) - First frees
 * the storage pointed to by pAlertSrc->Name (if any) and then copies
 * Name to a newly allocated buffer and sets the pAlertSrc->Name to
 * point to it.  AlertSrc_SetDesc() and AlertSrc_SetId() behave
 * similarly */
int AlertSrc_SetName(char *, USHORT, PAlertSrc);
int AlertSrc_SetDesc(char *, USHORT, PAlertSrc);
int AlertSrc_SetId(char *, USHORT, PAlertSrc);

/******************************************************************************
 * AlertSrc_GetName(char *Buf, USHORT Size, PAlertSrc pAlertSrc) - puts
 * the contents (up to Size bytes) into Buf.  Procedures should use
 * this rather than accessing pAlertSrc->Text directly.  The pointer to
 * Buf is returned on success.  AlertSrc_GetDesc() and
 * AlertSrc_GetId() behave similarly */
char *AlertSrc_GetName(char *, USHORT, PAlertSrc);
char *AlertSrc_GetDesc(char *, USHORT, PAlertSrc);
char *AlertSrc_GetId(char *, USHORT, PAlertSrc);

/******************************************************************************
 * AlertSrc_New() - allocates memory for a TAlertSrc and returns a
 * pointer to the new structure */
PAlertSrc AlertSrc_New(void);

/******************************************************************************
 * AlertSrc_Delete(PAlertSrc pAlertSrc) - frees the space pointed to by
 * pAlertSrc. */
void AlertSrc_Delete(PAlertSrc);

#define AlertSrc_Close(P) AlertSrc_Delete(P)

#endif
