/***************************************************************************** 
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT") 

  File:   archive.h 
  Author: Kent Landfield <kent@nfr.net>
 
 *****************************************************************************/ 
#ifndef _ARCHIVE_H_
#define _ARCHIVE_H_ 1

#ifndef MAXNAMELEN
#define MAXNAMELEN 256
#endif 

#define CFVERSION_LEN 128
#define DTYPE_LEN      10

#define ARCHIVE_WRITE               0
#define ARCHIVE_READ                1
#define ARCHIVE_READ_CONTENT_TYPE   2

#define ONE_K    1024
#define ONE_MEG  1000000
#define ONE_GIG  1000000000

#define INDEXDIR "data/index"

/*
** Format of archived files
** ========================
**
**  ----------------------
**  |  struct header     |
**  ----------------------
**  |  recorder blob     |
**  ----------------------
**  |    Data records    |
**  |        ....        |
**  |    Data records    |
**  ----------------------
*/

/*
** Data Record Format 
** ==================
**
**  -----------------------------------
**  | struct rec | ID |    Data       |
**  -----------------------------------
**
**  ID is a unique-id that is variable to 1024 bytes. 1024 bytes is the
**  largest an ID can be. 
**
**  Data is variable in size and has no real limits except those "supplied"
**  by the native operating system.
*/

struct rec {
    time_t tstamp; 
    int id_len; 
    long data_len;  
    long next_rec;  
    long last_rec;  
};  

struct header {
    char   backend[MAXNAMELEN];      /* backend name             */
    char   cfversion[CFVERSION_LEN]; /* Version of CF used       */
    char   ctype[DTYPE_LEN];         /* Data content type        */
    time_t starttime;                /* Start time delimiter     */
    time_t endtime;                  /* End time delimeter       */
    size_t archive_size;             /* File size                */
    long   datastart;                /* Data Start Seek position */ 
    long   maxreclen;                /* Size of largest record   */ 
};

typedef struct {
    struct header head;
    char   *info;                  /* Recorder specific information blob. */
    int    info_read;              /* boolean indicating the application  */
                                   /* has not read the current recorder   */
                                   /* specific info.  For use by back     */
                                   /* ends that need to keep track of the */
                                   /* changes in the recorder information */
    int    op;                     /* Type:  ARCHIVE_WRITE / ARCHIVE_READ */
    char   *path;                  /* diskpath to archive                 */
    char   *indexpath;             /* Path to associated index file       */
    char   *pathfmt;               /* Path format string from config file */
    FILE   *fp;                    /* File pointer to operate on          */
    FILE   *ifp;                   /* Index file pointer to operate on    */
    long   rec1len;                /* size of record 1 with data          */
    long   last_rec_pos;           /* position of last record written     */ 
    long   bytesread;              /* # of bytes read                     */ 
    long   byteswritten;           /* # of bytes written                  */
    long   rollsize;               /* byte size of file before rollover   */
                                   /* -1 disabled, 0 initialized          */
    long   rolltime;               /* Seconds before rollover of file     */
                                   /* -1 disabled, 0 initialized          */
    long   rtimeval;               /* Config file time before rollover    */
    time_t lasttime;               /* Time of last write if ARCHIVE_WRITE */
                                   /* or ARCHIVE_READ endtime             */
    char   *external_archive_cmd;  /* For use with archiving to a process */
                                   /* Such as a database import process.  */
} Archive;

/*
** Index File Record Format
**
**  ----------------------------------------------
**  | backend | filename |  starttime  | endtime |
**  ----------------------------------------------
*/

struct indx_rec {
    char   backend[MAXNAMELEN];    /* backend name          */
    char   ctype[DTYPE_LEN];       /* Data content type     */
    char   path[MAXPATHLEN];       /* path to file          */
    time_t starttime;              /* Start time delimiter  */
};

/*
** archive_errno Error defines
*/

#define ARCHIVE_NO_ERRS                    0
#define ARCHIVE_NO_MEMORY                  1
#define ARCHIVE_CONFIG_READ_FAILED         2
#define ARCHIVE_INVALID_ROLLOVER_SIZE      3
#define ARCHIVE_INVALID_ROLLOVER_SIZE_VAL  4
#define ARCHIVE_INVALID_ROLLOVER_TIME      5
#define ARCHIVE_INVALID_ROLLOVER_TIME_VAL  6
#define ARCHIVE_BAD_PATH_FORMAT            7
#define ARCHIVE_MISSING_CF_VERSION         8
#define ARCHIVE_MISSING_ARCHIVE_PATH       9
#define ARCHIVE_MISSING_ROLLOVER_SIZE     10
#define ARCHIVE_MISSING_ROLLOVER_SIZE_VAL 11
#define ARCHIVE_MISSING_ROLLOVER_TIME     12
#define ARCHIVE_MISSING_ROLLOVER_TIME_VAL 13
#define ARCHIVE_CANT_MKDIR                14
#define ARCHIVE_CANT_OPEN_ARCHIVE         15
#define ARCHIVE_CANT_OPEN_INDEX           16
#define ARCHIVE_CANT_STAT                 17
#define ARCHIVE_CANT_CLOSE_ARCHIVE        18
#define ARCHIVE_CANT_CLOSE_INDEX          19
#define ARCHIVE_CANT_SEEK                 20
#define ARCHIVE_HEADER_READFAILED         21
#define ARCHIVE_BLOB_READFAILED           22
#define ARCHIVE_HEADER_WRITEFAILED        23
#define ARCHIVE_BLOB_WRITEFAILED          24
#define ARCHIVE_INDEX_WRITEFAILED         25
#define ARCHIVE_INDEX_READFAILED          26
#define ARCHIVE_NO_STARTTIME_MATCH        27
#define ARCHIVE_NO_CTYPE_MATCH            28
#define ARCHIVE_RECHDR_WRITE_ERR          29
#define ARCHIVE_RECID_WRITE_ERR           30
#define ARCHIVE_RECDATA_WRITE_ERR         31
#define ARCHIVE_RECHDR_READ_ERR           32
#define ARCHIVE_RECID_READ_ERR            33
#define ARCHIVE_RECDATA_READ_ERR          34
#define ARCHIVE_RECDATA_BUF_SHORT         35

extern int archive_errno;

int archive_write( Archive *rkive, size_t nbytes, char *buf, 
	time_t tstamp, char *unique_id );

Archive *archive_openwrite( char *backend, char *content_type, 
	time_t starttime, char *recorder_blob);

int archive_close( Archive *rkive);

Archive *archive_openread( char *backend, char *content_type, 
	time_t starttime, time_t endtime);

struct rec *archive_read( Archive *rkive, size_t nbytes, char *buf,
	time_t endtime, char *unique_id);

int archive_shortfall;	/* how many more bytes you need if archive_read */
			/* returns ARCHIVE_RECDATA_BUF_SHORT */

int archive_flush( Archive *rkive);

char *astrerror( int num);

void adebug();

#endif /* _ARCHIVE_H_ */
