/*****************************************************************************
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT")


  File:   symtab.h
  Author: Andrew Lambeth <andrew@nfr.net>

 *****************************************************************************/

/******************************************************************************
* symtab.h - a simple symbol table with typed symbols and a table-wide
* linked list of symbols
******************************************************************************/
#ifndef _SYMTAB_H_
#define _SYMTAB_H_

#include "xtypes.h"
#include "list.h"

/* Symbol Types */
#define ALL_TYPES 0

typedef struct TSymbol
{
  struct TSymbol        *Next;
  char			*String;
  USHORT                StrLen;
  ULONG			Data;
  BYTE			Type;
} TSymbol, *PSymbol;

typedef struct TSymTab
{
  PNIList	SymList;
  PSymbol 	*HashTab;
  ULONG         HashTabSize;
} TSymTab, *PSymTab;

/* SymTab_LookEx() return values */
#define SYM_NOT_FOUND 0
#define SYM_FOUND 1
#define SYM_INSERTED 2

/* SymTab_LookEx() flags */
#define AUTO_INSERT 1

/* wrappers for SymTab_LookEx() */
#define SymTab_Look(String, ppSymbol, pSymTab) \
  SymTab_LookEx(AUTO_INSERT, String, ppSymbol, pSymTab)

#define SymTab_Peek(String, ppSymbol, pSymTab) \
  SymTab_LookEx(0, String, ppSymbol, pSymTab)


/* wrapper macros and functions for use with SymTab.SymList */
PSymbol SymList_Insert(PSymbol, PNIList);

#define SymList_Remove(pSym, pList) \
  (PSymbol) NIList_Remove((void *)pSym, pList);

#define SymList_RemoveFirst(pList) \
  (PSymbol) NIList_RemoveHead(pList);

#define SymList_RemoveLast(pList) \
  (PSymbol) NIList_RemoveTail(pList);

#define SymList_GetFirst(pList) \
  (PSymbol) NIList_GetFirst(pList);

#define SymList_GetNext(pList) \
  (PSymbol) NIList_GetNext(pList);

/******************************************************************************
* SymTab_New(ULONG Size) - returns a pointer to an initialized TSymtab with a
* hash table of size Size
*/
PSymTab	SymTab_New(ULONG);

/******************************************************************************
* SymTabLookEx(ULONG Flags, char *String, PSymbol *ppSymbol, 
*              PSymTab pSymTab) -
* checks for String in the table pointed to by pSymtab and inserts it
* if it wasn't there and Flags contains AUTO_INSERT. ppSymbol is set
* to point to the symbol used by String.  Return value is either
* SYM_FOUND, SYM_NOT_FOUND, or SYM_INSERTED
******************************************************************************/
int SymTab_LookEx(ULONG, char *, PSymbol *, PSymTab);

/******************************************************************************
* SymTabRemove() - removes the symbol for String from the pSymTab.
* Returns the pointer to the symbol's Data member if the symbol was
* found in the table.  Otherwise NULL is returned
******************************************************************************/
void *SymTab_Remove(char *, PSymTab);

/******************************************************************************
* SymTab_Delete(PSymTab pSymTab) - call this to free the space
* allocated by SymTab_New() and destroy the symbol table */
void SymTab_Delete(PSymTab);

/******************************************************************************
* SymTab_DeleteAll(PSymTab pSymTab, PfnDelete pDeleteData) - call this
* to free the space allocated by SymTab_New() and destroy the symbol
* table AND free whatever space is pointed to each symbol's Data
* member.  If pDeleteData is specified it is called for each symbol in
* the table, otherwise the Data pointer is simply passed to free().  */

typedef void (*PfnDelete)(PSymbol);
void SymTab_DeleteAll(PfnDelete, PSymTab);

/******************************************************************************
* SymTab_ForEach(BYTE Type, PfnSymIterate pfnIterate, int IArg,
*                void *PArg, PSymTab pSymTab) 
* - Calls pfnIterate for each symbol of type Type in the table
* pSymTab.  pfnIterate is called with three arguments: a PSymbol, and
* the provided int IArg and void *PArg.  The number of Symbols
* operated on is returned on success, so if you just want to count the
* number of symbols of a particular type set pfnIterate to NULL */
typedef int (PfnSymIterate)(PSymbol, int, void *);
ULONG SymTab_ForEach(BYTE, PfnSymIterate, int, void *, PSymTab);

#endif /* sentry */
