/*
        Copyright(C) 1997 Network Flight Recorder, Inc.  
        All rights reserved.

        Use and distribution of this software and its source code
        are governed by the terms and conditions of the
        Network Flight Recorder Software License ("LICENSE.TXT")

By Mark Sienkiewicz / NFR

*/

/*
* This function is basically the interface for a class of recorders
* to fetch data coming in from nfrd.
*
* call read_record().  The vsch_record parameter describes the
* format of the record you are expecting.  read_record assumes that you
* have already read the record type from the stream;  It assumes
* the first record element is a time stamp and that the time stamp
* IS NOT included as one of the fields of the vsch_record.
*
* On return, the global variable read_record_buffer points at
* a buffer that contains the timestamp and the fields described
* by the vsch_record struct.
*
* The offset (and size, if appropriate) fields of the vsch_record
* data structure will be updated to reflect the record in read_record_buffer.
*
* The return value is the number of bytes in read_record_buffer, or
* -1 for EOF or -2 for other error.
*/

#include "sys-include.h"

#include "counter.h"
#include "vsch.h"
#include "cf.h"

/*
* bug: you have to refit this a bit to make it handle different
* endian-ness of the input streams.  For now I am going to wave
* my hands and say that nfrd "normally doesn't" run on a different
* architecture than the recorder.
*/

/*
* this is a buffer that accumulates the record as it arrives.
* it gets bigger as necessary.
*/
int record_size = 0;
unsigned char *read_record_buffer = (void *)0;


void *
buffer_grow(int off, int size)
{
int add;
add = record_size;
if (add > 1000)
	add = 1000;
if (size > add)
	add = size;
read_record_buffer=realloc(read_record_buffer,record_size + add);
if (! read_record_buffer)
	return (void *)0;
record_size += add;
return read_record_buffer + off;
}

/*
* reads a record coming in from nfrd.
*/
read_record(FILE *f, struct vsch_record *r, int data_flag )
{
int x;
int type;
int tyme;
int i;
counter c;
time_t t;

unsigned char *p = read_record_buffer;
int remaining=record_size;

/*
build up a buffer adding each object to it in a form suitable
to write to the data file; realloc it if it gets bigger
*/

if (! read_record_buffer )
	{
	p = read_record_buffer = malloc(10240);
	remaining = record_size = 10240;
	}

memset(read_record_buffer,0,10240);


#define BUFFER_USE(size) if (remaining < size) if (!(p=buffer_grow(p-read_record_buffer,size))) return -2;
	
#define BUFFER_PUT(ptr,size) memcpy(p,ptr,size); p+= size;

/*
* the first element is always the time of the record
*/
BUFFER_USE(sizeof(int))
if (!intDecode(f,p))
	return -1;

memcpy(&r->time,p,sizeof(int));

p += sizeof(int);

for (x=0; x<r->num_fields; x++)
	{
	switch (r->f[x].type)
		{
	case VSCH_COUNTER:
		BUFFER_USE(sizeof(int))	/* use an int size from the input */
		if (!intDecode(f,&i))
			abort();
		r->f[x].offset= p - read_record_buffer;
		c = counter_from_int(i);
		BUFFER_PUT(&c,sizeof(counter))	/* but output a counter size */
		break;
	case VSCH_TIME_T:
	case VSCH_INT:
		BUFFER_USE(sizeof(int))
		if (!intDecode(f,&i))
			abort();
		r->f[x].offset= p - read_record_buffer;
		BUFFER_PUT(&i,sizeof(int))
		break;
	case VSCH_IP:
		BUFFER_USE(4);
		if (!ipaddrDecode(f,p))
			abort();
		r->f[x].offset= p - read_record_buffer;
		p += 4;
		break;
	case VSCH_ETHERMAC:
		BUFFER_USE(6);
		if (fread(p, sizeof(char), 6, f ) != 6)
			abort();
		r->f[x].offset= p - read_record_buffer;
		p += 6;
		break;
	case VSCH_STR:
	case VSCH_BLOB:
		BUFFER_USE(4);
		if (!intDecode(f,&i))
			abort();
		BUFFER_PUT(&i,sizeof(int));
		BUFFER_USE(i);
		r->f[x].offset = p - read_record_buffer;
		r->f[x].size = i;
		if (fread(p, sizeof(char), i, f ) != i)
			abort();
		p += i;
		break;
	default:
		abort();	/* "never happens" */
		}
	}

if (data_flag)
	{
	BUFFER_USE(sizeof(int))
	if (!intDecode(f,&i))
		abort();
	r->data_offset= p - read_record_buffer;
	BUFFER_PUT(&i,sizeof(int))
	}

return p - read_record_buffer;
}

