/*
*        Copyright(C) 1997 Network Flight Recorder, Inc. 
*        All rights reserved. 
* 
*        Use and distribution of this software and its source code 
*        are governed by the terms and conditions of the 
*        Network Flight Recorder Software License ("LICENSE.TXT") 
* 
* By Mark Sienkiewicz / NFR 
* 
*/ 


/*
* general query support functions
*/

#include "sys-include.h"

#include "counter.h"
#include "vsch.h"
#include "query.h"


#define YYSTYPE union query_private_yystype

struct query_match
	{
	struct query_match *next;
	YYSTYPE	u;
	};

/*

For each type
	IP
	int
	ethernet mac
	etc.
be able to answer the question
	is AAA in the set BBB-DDD, EEE, GGG-HHH, MMM, OOO
*/


/*
* this is the text we are parsing for this query.  Whichever of
* the several lexical analyzers we use will consume it.
*/
char *yacc_input_string;

/*
* this is where we construct the compiled query.  it is a list
* of range pairs.
*/
static struct query_match * the_list, *the_list_end;


/*
* compiles a query string.  The return value can be used with
* query_compare() to see if a value matches.
*
* This function really just looks at what you are asking for
* passes the buck to a more specific query parser
*/

int had_yacc_error;

struct query_match * 
query_compile(int type, char *string)
{
int result=1;
had_yacc_error=0;
if (type == VSCH_INT)
	type=VSCH_COUNTER;
fprintf(stderr,"compiling for type %d -%s-\n",type,string);
yacc_input_string=string;
the_list=(void *)0;
the_list_end=(void *)0;
switch (type)
	{
case VSCH_COUNTER:
	result=yy_counter_parse();
	break;
case VSCH_IP:
	result=yy_ip_parse();
	break;
case VSCH_STR:
	result=yy_str_parse();
	break;
case VSCH_ETHERMAC:
	result=yy_ether_parse();
	break;
	}
fprintf(stderr,"compiled\n");
query_debug_printer(the_list,type);
if (result)
	return (void *)0;
if (had_yacc_error)
	return (void *)0;
return the_list;
}


/*
*
* query_compare() matches a compiled query against a value.
* It returns true if the value is in one of the ranges specified
* in the query.
*
*/

static ip_cmp(unsigned char *x, unsigned char *y)
{
int t;
for (t=0; t<4; t++)
	{
	if (x[t] - y[t] > 0) return 1;
	else if (x[t] - y[t] < 0) return -1;
	}
return 0;
}

query_compare(struct query_match *n, void *d, int type, int size )
{
counter cval;
int x,y;
if (!n)
	return 1;
switch (type)
	{
case VSCH_INT:
	memcpy(&x,d,sizeof(int));
	cval=counter_from_int(x);
	goto counter_compare;
case VSCH_COUNTER:
	memcpy(&cval, d, sizeof(counter));
	goto counter_compare;
case VSCH_IP:
	for (; n; n=n->next)
		if ((ip_cmp(d, n->u.ip_range.first) >= 0) &&
		    (ip_cmp(d, n->u.ip_range.second) <= 0))
			return 1;
	return 0;
case VSCH_STR:
	for (; n; n=n->next)
		{
		if (n->u.string_range.count == 1)
			{
			char *s = d;
			int count = size;
			char *cm = n->u.string_range.first;
			int cml = strlen(cm);
	str1_again:
			/* ! the parameter is NOT null terminated */
			while ((count > 0) && (*cm != *s))
				s++, count--;
			if (count > 0)
				{
				if (count < cml)
					goto str_next;
				if (strncmp(s,cm,cml) == 0)
					return 1;
				s++;
				goto str1_again;
				}
			goto str_next;
			}
		else
			{
			char *s = d;
			int count = size;
			char *cmf = n->u.string_range.first;
			int cmlf = strlen(cmf);
			char *cms = n->u.string_range.second;
			int cmls = strlen(cms);
	str2_again:
			if (count > 0)
				{
				int ccs;
				if (count < cmlf)  /* "ab" never >= "abc" */
					return 1;
				if (count < ccs)   /* "xy" could be <= "xyz" */
					ccs = count;
				if ((strncmp(s,cmf,cmlf) >= 0) &&
					(strncmp(s,cms,ccs) <= 0))
					return 1;
				s++;
				count--;
				goto str2_again;
				}
			goto str_next;
			}
	str_next: ;
		}
	return 0;
case VSCH_ETHERMAC:
	for (; n; n=n->next)
		{
		int i;
		for (i=0; i<6; i++)
			{
			if (n->u.ether.wild[i])
				continue;
			if (((unsigned char *)d)[i] < n->u.ether.first[i])
				goto ether_nomatch;
			if (((unsigned char *)d)[i] > n->u.ether.second[i])
				goto ether_nomatch;
			}
		return 1;

ether_nomatch:	;
		}

	return 0;

default:
	abort();	/* "never happens" */
	}

counter_compare:

for (; n; n=n->next)
	{
	if ((counter_cmp(cval, n->u.counter_range.first) >= 0) &&
   	    (counter_cmp(cval, n->u.counter_range.second) <= 0))
		{
		return 1;
		}
	}
return 0;
}

/*
* frees a compiled query.  It isn't absolutely essential that you do
* this in most NFR query processors because they process one query
* and 
*/
query_free(struct query_match *n)
{
struct query_match *t;
while (n)
	{
	if (n->u.type == VSCH_STR)
		{
		if (n->u.string_range.first)
			free(n->u.string_range.first);
		if (n->u.string_range.second &&
			(n->u.string_range.second != n->u.string_range.first))
			free(n->u.string_range.second);
		}
	t=n->next;
	free(n);
	n=t;
	}
}

/*
/*
* used by the individual parsers - adds a node onto the query_list.
* Note we want to keep the list in the order that the parser finds
* things because the user may be entering the most common match
* first to try to speed the query along.
*/
query_add_list(YYSTYPE y)
{
struct query_match *n;
fprintf(stderr,"query_add_list\n");
n=malloc(sizeof(struct query_match));
if (!n)
	abort(); 	/* bug: handle error */
n->u=y;
n->next=NULL;
if (! the_list)
	the_list=n;
else
	the_list_end->next = n;
the_list_end=n;
}

static struct internal_query_types
	{
	char *primary;
	char *secondary;
	int time_magic;
	int vsch_type;
	struct internal_query_types *next;
	} *query_types_table = (struct internal_query_types *)0;


struct query_types *query_primary_types, *query_secondary_types;
enum vsch_type *query_primary_vsch, *query_secondary_vsch;

/*
* bug: obsoleted by the type mapping method used in the java code
* This should be replaced with similar code to read the new file.
*/
void query_load_types()
{
FILE *f;
char b[200],s[100],t[100],u[100],v[100];
struct internal_query_types *n,*scan_n;
int x,y,time_magic;
if (query_types_table)
	return;
f=fopen("etc/type-mapping.map","r");
if (!f)
	goto error;
t[0]='p'; t[1]='_';
u[0]='s'; u[1]='_';
while (fgets(b,sizeof(b),f))
	{
	if (b[0] == '#') continue;
	if (sscanf(b,"%s%s%s%s",
		s,t+2,u+2,v) != 4)
		continue;
	n=malloc(sizeof(struct internal_query_types));
	if (!n)
		nomem();
	n->primary=strdup(t);
	n->secondary=strdup(u);
	n->vsch_type=vsch_string_to_type(v);
	if (strcmp(t,"max_time") == 0)	/* bug: international */
		n->time_magic=2;
	if (strcmp(t,"min_time") == 0)	/* bug: international */
		n->time_magic=1;
	n->next=query_types_table;
	query_types_table=n;
	}
fclose(f);


x=1;
for (n=query_types_table; n; n=n->next)
	x++;
query_primary_types=calloc(x,sizeof(struct query_types));
if (! query_primary_types ) nomem();
query_primary_vsch=calloc(x,sizeof(enum vsch_type));
if (!query_primary_vsch) nomem();
query_secondary_types=calloc(x,sizeof(struct query_types));
if (!query_secondary_types) nomem();
query_secondary_vsch=calloc(x,sizeof(enum vsch_type));
if (!query_secondary_vsch) nomem();

for (n=query_types_table, x=0; n; n=n->next)
	{
	char *s;
	s=query_primary_types[x].name= n->primary;
	query_primary_types[x].type= n->vsch_type;
	query_primary_types[x].time_magic= n->time_magic;
	for (scan_n=query_types_table; scan_n; scan_n=scan_n->next)
		if (strcmp(scan_n->primary,s) == 0)
			scan_n->primary=s;
	x++;
	}
for (n=query_types_table, x=0; n; n=n->next)
	{
	char *s;
	for (y=0; query_secondary_types[y].name; y++)
		{
		if (strcmp(query_secondary_types[y].name,n->secondary) == 0)
			goto skip_me;
		}
	s=query_secondary_types[x].name=n->secondary;
	query_secondary_types[x].type=n->vsch_type;
	query_secondary_types[x].time_magic= n->time_magic;
	for (scan_n=query_types_table; scan_n; scan_n=scan_n->next)
		if (strcmp(scan_n->secondary,s) == 0)
			scan_n->secondary=s;
	x++;
	skip_me: ;
	}
/* should we bother reallocing down query_secondary_types? */
return;
error:
	fprintf(stderr,"error in query_types.cfg\n");
	abort();
}

enum vsch_type
query_expand_type(char *primary,char **secondary)
{
struct internal_query_types *t;
if (! query_types_table)
	query_load_types();
for (t=query_types_table; t; t=t->next)
	{
	if (strcmp(primary,t->primary) == 0)
		{
		if (secondary) *secondary = t->secondary;
		return t->vsch_type;
		}
	}
if (secondary) *secondary="";
return VSCH_ERROR;
}

nomem()
{
printf("no memory\n");
abort();
}

query_debug_printer(struct query_match * x, int type)
{
char b[200],b1[200];
fprintf(stderr,"begin query_debug_printer\n");
while (x)
	{
	switch (type)
		{
	case VSCH_INT:
	case VSCH_COUNTER:
		vsch_print_type(b,&x->u.counter_range.first,type,0);
		vsch_print_type(b1,&x->u.counter_range.second,type,0);
		fprintf(stderr," %s - %s\n",b,b1);
		break;
	case VSCH_IP:
		vsch_print_type(b,&x->u.ip_range.first,type,0);
		vsch_print_type(b1,&x->u.ip_range.second,type,0);
		fprintf(stderr," %s - %s\n",b,b1);
		break;
	case VSCH_STR:
		fprintf(stderr," '%s' - '%s'\n",
			x->u.string_range.first, 
			x->u.string_range.second);
		break;
	case VSCH_ETHERMAC:
		{
		int i;
		fprintf(stderr," ");
		for (i=0; i<6; i++)
			if (x->u.ether.wild[i])
				fprintf(stderr,"*:");
			else
				fprintf(stderr,"%02x:",
					x->u.ether.first[i]);
		fprintf(stderr," - ");
		for (i=0; i<6; i++)
			if (x->u.ether.wild[i])
				fprintf(stderr,"*:");
			else
				fprintf(stderr,"%02x:",
					x->u.ether.second[i]);
		fprintf(stderr,"\n");
		}
		}

	x=x->next;
	}
fprintf(stderr,"end query_debug_printer\n");
}


/*
* optionally used by the type-specific parsers - these functions
* build up a dynamically allocated string, once character at a time.
* You would use them in the lexical analyzer of query parsers that
* need to keep the token strings.
*
* For example, a string query needs to know what the token was.
* A numeric query only needs the numeric value.
*
*/

char *query_lex_string;

static int query_lex_size;
static int query_lex_index;

query_lex_init()
{
query_lex_size=5;
query_lex_string=malloc(query_lex_size);
query_lex_index=0;
}

query_lex_add(int c)
{
if (query_lex_index >= query_lex_size-1)
	{
	if (query_lex_size > 1000)
		query_lex_size = query_lex_size + 1000;
	else
		query_lex_size = query_lex_size * 2;
	query_lex_string=realloc(query_lex_string,query_lex_size);
	if (!query_lex_string)
		nomem();
	}
query_lex_string[query_lex_index++] = c;
query_lex_string[query_lex_index] = 0;
}
