/*****************************************************************************
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT")


  File:   alert.c
  Author: Andrew Lambeth <andrew@nfr.net>

 *****************************************************************************/

/*****************************************************************************
 * alert.c 
 * functions for sending alerts to the alert daemon
 *****************************************************************************/
#include "sys-include.h"

#include "memsafe.h"
#include "home.h"
#include "fifo.h"
#include "resource.h"
#include "alertres.h"
#include "alertsrc.h"
#include "alert.h"

#include "libnfr_err.h"


/* set this true to keep from reopening the resource files for every
   call to Alert_ToAlertSrc(), Alert_ToAlertRes(), and
   Alert_ToAlertResEx()  - gross hack, blech :( */
BOOL gKeepOpenFlag = FALSE;


/*****************************************************************************
 * Alert_Generate() - takes an alert source resource code and an alert
 * res code. A TAlert struct containing the resource codes is sent to
 * the alert daemon for processing.  No source file or line info is
 * included .
 * UsrAlert_Generate() and UsrAlert_GenerateEx() work
 * similarly except that they take the string identifiers of the alert
 * and alert source instead of the numeric ids */

int Alert_Generate(ULONG AlertSrcCode, ULONG AlertCode)
{
  TAlert Alert;

  Alert.Timestamp = time(NULL);
  Alert.InsCount = 0;
  Alert.InsLen = 0;
  Alert.Inserts = NULL;
  Alert.SrcCode = AlertSrcCode;
  Alert.SrcPid = getpid();
  Alert.ResCode = AlertCode;
  return Alert_Send(&Alert);
}

int Alert_GenerateEx(char *File, int Line, ULONG AlertSrcCode, ULONG AlertCode)
{
  return Alert_Args_GenerateEx(File, Line, AlertSrcCode, AlertCode, NULL);
}

int UsrAlert_Generate(char *AlertSrcId, char *AlertId)
{
  TAlert Alert;
  ULONG AlertCode = DEF_RESCODE;
  ULONG AlertSrcCode = DEF_RESCODE;

  UsrAlert_GetCodes(AlertSrcId, &AlertSrcCode, AlertId, &AlertCode);

  Alert.Timestamp = time(NULL);
  Alert.InsCount = 0;
  Alert.InsLen = 0;
  Alert.Inserts = NULL;
  Alert.SrcCode = AlertSrcCode;
  Alert.SrcPid = getpid();
  Alert.ResCode = AlertCode;

  return Alert_Send(&Alert);
}

int UsrAlert_GenerateEx(char *File, int Line, char *SrcId, char *ResId)
{
  return UsrAlert_Args_GenerateEx(File, Line, SrcId, ResId, NULL);
}


/*****************************************************************************
 * Alert_Args_Generate() - takes an alert source res code and an alert
 * resource code and any number of char* arguments.  The last argument
 * must be NULL. A TAlert struct containing the alert resource code
 * and the insertion strings is sent to the alert daemon for
 * processing. Alert_Args_GenerateEx() is the same except source file
 * and linenumber info is included as the first two arguments.
 * UsrAlert_ArgsGenerate() and UsrAlert_Args_GenerateEx() work
 * similarly except that they take the string identifiers of the alert
 * and alert source instead of the numeric ids 
 * VAlert_Args_Generate() and VAlert_Args_GenerateEx() are the same
 * as Alert_Args_Generate() and Alert_Args_GenerateEx() respectively,
 * except that they take a NULL terminated array of char * instead of
 * a variable number of arguments. */

int Alert_Args_Generate(ULONG AlertSrcCode, ULONG AlertCode, ...)
{
  va_list ap = NULL;
  BYTE Buf[MAXSIZE_ALERT];

  va_start(ap, AlertCode);
  Alert_Assemble(Buf, sizeof(Buf), NULL, 0, 
		 AlertSrcCode, AlertCode, NULL, ap);
  va_end(ap);
  return Alert_Send((PAlert)Buf);
}

int Alert_Args_GenerateEx(char *File, int Line, 
			  ULONG AlertSrcCode, ULONG AlertCode, ...)
{
  va_list ap = NULL;
  BYTE Buf[MAXSIZE_ALERT];

  va_start(ap, AlertCode);
  Alert_Assemble(Buf, sizeof(Buf), File, Line, 
		 AlertSrcCode, AlertCode, NULL, ap);
  va_end(ap);
  return Alert_Send((PAlert)Buf);
}


int UsrAlert_Args_Generate(char *AlertSrcId, char *AlertId, ...)
{
  ULONG AlertCode = DEF_RESCODE;
  ULONG AlertSrcCode = DEF_RESCODE;
  va_list ap = NULL;
  BYTE Buf[MAXSIZE_ALERT];

  UsrAlert_GetCodes(AlertSrcId, &AlertSrcCode, AlertId, &AlertCode);

  va_start(ap, AlertId);
  Alert_Assemble(Buf, sizeof(Buf), NULL, 0, 
		 AlertSrcCode, AlertCode, NULL, ap);
  va_end(ap);
  return Alert_Send((PAlert)Buf);
}

int UsrAlert_Args_GenerateEx(char *File, int Line,
			     char *AlertSrcId, char *AlertId, ...)
{
  ULONG AlertCode = DEF_RESCODE;
  ULONG AlertSrcCode = DEF_RESCODE;
  va_list ap = NULL;
  BYTE Buf[MAXSIZE_ALERT];

  UsrAlert_GetCodes(AlertSrcId, &AlertSrcCode, AlertId, &AlertCode);

  va_start(ap, AlertId);
  Alert_Assemble(Buf, sizeof(Buf), File, Line, 
		 AlertSrcCode, AlertCode, NULL, ap);
  va_end(ap);
  return Alert_Send((PAlert)Buf);
}

int VAlert_Args_Generate(ULONG AlertSrcCode, ULONG AlertCode, char **Inserts)
{
  return VAlert_Args_GenerateEx(NULL, 0, AlertSrcCode, AlertCode, Inserts);
}

int VAlert_Args_GenerateEx(char *File, int Line, 
			   ULONG AlertSrcCode, ULONG AlertCode, char **Inserts)
{
  BYTE Buf[MAXSIZE_ALERT];

  Alert_Assemble(Buf, sizeof(Buf), File, Line, 
		 AlertSrcCode, AlertCode, Inserts, NULL);
  return Alert_Send((PAlert)Buf);
}


/* Alert_Assemble() creates a TAlert struct in a contiguous block of
   memory suitable for use as a datagram msg body. this is ugly but if
   this code isn't localized here it will be in seven different places
   in two files, blech */
int Alert_Assemble(BYTE *Buf, int BufSize, char *File, int Line,
		   ULONG AlertSrcCode, ULONG AlertCode, 
		   char **InsertsArray, va_list InsertsVaList)
{
  ULONG Result = LIBNFR_ERR_SUCCESS;
  int s = 0;
  int i = 0;
  PAlert pAlert;
  char LineStr[21];
  char *pArg = NULL;

  memset(Buf, 0, BufSize);
  pAlert = (PAlert) Buf;
  pAlert->Timestamp = time(NULL);
  pAlert->SrcCode = AlertSrcCode;
  pAlert->SrcPid = getpid();
  pAlert->ResCode = AlertCode;
  pAlert->Inserts = (BYTE *) pAlert + sizeof(TAlert);

  BufSize -= sizeof(TAlert);

  if(File)
    {
      /* add the source file name */
      s = strlen(File);
      
      if ( s > BufSize - 1 - pAlert->InsLen)
	{
	  /* don't truncate filename, just leave it out if no room */
	  (pAlert->Inserts)[++(pAlert->InsLen)] = 0; /* empty string */
	  pAlert->InsCount++;
	  Result = LIBNFR_ERR_UNDEFINED;
	}
      else 
	{
	  memcpy(pAlert->Inserts + pAlert->InsLen, File, s);
	  pAlert->InsLen += s;
	  (pAlert->Inserts)[++(pAlert->InsLen)] = 0; /* terminate string */
	  pAlert->InsCount++;
	}

      /* add the source file linenumber string */
      sprintf(LineStr, "%d", Line);	/* do this so we know the length */
      s = strlen(LineStr);

      if (s > BufSize - 1 - pAlert->InsLen )
	{
	  /* don't truncate linenumber, just leave it out if no room */
	  (pAlert->Inserts)[++(pAlert->InsLen)] = 0; /* empty string */
	  pAlert->InsCount++;
	  Result = LIBNFR_ERR_UNDEFINED;
	}
      else 
	{
	  memcpy(pAlert->Inserts + pAlert->InsLen, LineStr, s);
	  pAlert->InsLen += s;
	  (pAlert->Inserts)[++(pAlert->InsLen)] = 0; /* terminate string */
	  pAlert->InsCount++;
	}
    }
  else
    {
      (pAlert->Inserts)[++(pAlert->InsLen)] = 0; /* empty str for file*/
      (pAlert->Inserts)[++(pAlert->InsLen)] = 0; /* empty str for linenum */
      pAlert->InsCount += 2;
    }

  /* add the inserts from the array, if any */
  if(InsertsArray)
    {
      for(i = 0; InsertsArray[i] != NULL; i++)
	{
	  if (pAlert->InsLen >= BufSize)
	    {
	      /* out of buffer so we're done */
	      return LIBNFR_ERR_UNDEFINED;
	    }
	  s = strlen(InsertsArray[i]);
	  if ( s > BufSize - 1 - pAlert->InsLen)
	    {
	      s = BufSize - 1 - pAlert->InsLen; /* must truncate this */
	      Result = LIBNFR_ERR_UNDEFINED;
	    }
	  memcpy(pAlert->Inserts + pAlert->InsLen, InsertsArray[i], s);
	  pAlert->InsLen += s;
	  (pAlert->Inserts)[++(pAlert->InsLen)] = 0; /* terminate string */
	  pAlert->InsCount++;
	}
    }



  /* add the inserts from the va_list, if any */
  if(InsertsVaList)
    {
      while( (pArg = va_arg(InsertsVaList, char *)) != NULL)
	{
	  if (pAlert->InsLen >= BufSize)
	    {
	      /* out of buffer so we're done */
	      return LIBNFR_ERR_UNDEFINED;
	    }
	  s = strlen(pArg);
	  if ( s > BufSize - 1 - pAlert->InsLen)
	    {
	      s = BufSize - 1 - pAlert->InsLen; /* must truncate this */
	      Result = LIBNFR_ERR_UNDEFINED;
	    }
	  memcpy(pAlert->Inserts + pAlert->InsLen, pArg, s);
	  pAlert->InsLen += s;
	  (pAlert->Inserts)[++(pAlert->InsLen)] = 0; /* terminate string */
	  pAlert->InsCount++;
	}
    }

  return Result;
}

/* UsrAlert_GetCodes looks up the resource codes for the identifiers
   provided */
int UsrAlert_GetCodes(char *AlertSrcId, ULONG *pAlertSrcCode,
		      char *AlertId, ULONG *pAlertCode)
{
  FILE *Map = NULL;
  char linebuf[256];
  char *Id = NULL;
  char *Code = NULL;
  BOOL FoundAlertCode = FALSE;
  BOOL FoundAlertSrcCode = FALSE;

  Map = fopen(RESFILE_PATH "/usr/" RES_MAPFILE, "r");
  if(Map == NULL)
    return LIBNFR_ERR_UNDEFINED;
  while(fgets(linebuf, sizeof(linebuf) - 1, Map))
    {
      /* go through the file 'til we find both rescodes */
      Id = strtok(linebuf, "=");
      if(Id == NULL)
	break;
      if(!FoundAlertCode && (strcmp(AlertId, Id) == 0))
	{
	  FoundAlertCode = TRUE;
	  Code = strtok(NULL, "\n");
	  *pAlertCode = strtoul(Code, NULL, 0);
	  if(FoundAlertSrcCode)
	    break;
	}
      else if(!FoundAlertSrcCode && (strcmp(AlertSrcId, Id) == 0))
	{
	  FoundAlertSrcCode = TRUE;
	  Code = strtok(NULL, "\n");
	  *pAlertSrcCode = strtoul(Code, NULL, 0);
	  if(FoundAlertCode)
	    break;
	}
    }
  fclose (Map);

  if(!(FoundAlertCode && FoundAlertSrcCode))
    return LIBNFR_ERR_UNDEFINED;

  return LIBNFR_ERR_SUCCESS;
}


/*****************************************************************************
 * Alert_Send() - just stuffs an assembled alert down the pipe to the daemon
 */

/****
if you are wondering how/why/if these things are atomic and when/why/if
they block:

the way the fifo actually works in alertd is that a write to the
fifo blocks until the data from the previous write has been read. 
That's why it gets so slow when you saturate alertd.  The first two
writes happen fast, but then if third write happens before alertd
finshes with processing first and reads the second, it blocks. it acts
like it has a queue with a size of 2.  

So you always get an atomic alert, unless the alert happens to be bigger
than the fifo buffer. POSIX defines this buffer as 512 bytes.  On BSDI
it's 1024 and on SOlaris it's 5120.  An alert with no insertion strings
is 28 bytes.  The alert sending functions will truncate anything over
the size of the fifo buffer on the host system, though, the
documentation should say 512.

The problem with the current system is that if you send more than two
alerts in rapid succession, you have to wait for alertd to finsh
processing each alert.  I plan to add a way to buffer alerts on the
receiving end to alleviate this when I get some time.  Prolly won't be
in beta1, unless someone really thinks it's a priority.  The feedback
I've gotten thus far is that it's not.
******/

int Alert_Send(PAlert pAlert)
{
  static BYTE Buf[DGRAM_MAXSIZE];
  static int fdWrite = -1;
  int Result = 0;
  char *fifoPath = NULL;

  /* bug: prolly only really need to initialize the header portion of
     Buf[], since the data portion gets memcpy()'d to */
  memset(Buf, 0, sizeof(Buf));

  DGram_SetData(pAlert, ( sizeof(TAlert) + pAlert->InsLen ), ((PDGram) Buf));

  if(fdWrite < 0)
    {
      /* bug: this doesn't work because cgi programs don't have NFRHOME
	 in their env */
      /****
	fifoPath = nfrPath("~", "/", ALERTD_FIFO, NULL);
	fdWrite = Fifo_OpenWrite(fifoPath);
	free(fifoPath);
      ****/

      fdWrite = Fifo_OpenWrite(ALERTD_FIFO);
      if(fdWrite < 0)
        {
        switch(errno)
          {
          case ENOENT:
            return LIBNFR_ERR_FNOEXIST;
          default:
            return LIBNFR_ERR_UNDEFINED;
          }
        }
    }
  Result = DGram_Send_fd(fdWrite, ((PDGram) Buf), ALERT_SEND_TIMEOUT);

  /* bug: this is lame to close this and open it every time, but if we 
     don't, then we get a SIG_PIPE when alertd recreates the fifo */
  close (fdWrite);
  fdWrite = -1;

  return Result;

}

PAlertSrc Alert_ToAlertSrc(PAlert pAlert)
{
  PAlertSrc pAlertSrc = NULL;

  if(pAlert->SrcCode != DEF_RESCODE)
    pAlertSrc = AlertSrc_OpenEx(pAlert->SrcCode, gKeepOpenFlag);

  if(!pAlertSrc)
    {
      /* something is hosed with the resource files or the rescode was
         bogus, so we'll construct a default resource */
      
      pAlertSrc = AlertSrc_New();
      pAlertSrc->ResCode = DEF_RESCODE;	
      pAlertSrc->IdLen  = strlen(DEF_ALERTSRC_ID_STRING);
      pAlertSrc->Id = (char *) SCalloc(1, pAlertSrc->IdLen + 1);
      strcpy(pAlertSrc->Id, DEF_ALERTSRC_ID_STRING);

      /* leave room for insertion of the res number */
      pAlertSrc->Name = (char *) SCalloc(1, strlen(DEF_ALERTSRC_NAME) + 11);

      sprintf(pAlertSrc->Name, DEF_ALERTSRC_NAME, pAlert->SrcCode);
      pAlertSrc->NameLen  = strlen(pAlertSrc->Name);


      /* leave room for insertion of the res number */
      pAlertSrc->Desc = (char *) SCalloc(1, strlen(DEF_ALERTSRC_DESC) + 11);

      sprintf(pAlertSrc->Desc, DEF_ALERTSRC_DESC, pAlert->SrcCode);
      pAlertSrc->DescLen  = strlen(pAlertSrc->Desc);
    }

  pAlertSrc->Pid = pAlert->SrcPid;
  return pAlertSrc;
}

PAlertRes Alert_ToAlertRes(PAlert pAlert)
{
  return Alert_ToAlertResEx(0, pAlert);
}

PAlertRes Alert_ToAlertResEx(ULONG Flags, PAlert pAlert)
{
  /* must be reentrant! */
  PAlertRes pAlertRes = NULL;
  char Buf[ALERTRES_MAXSIZE];
  char *Expanded = NULL;
  int ExpSize = 0;
  char **Inserts = NULL;
  char *p = NULL;
  int i = 0;
  char DefSourceFile[] = "<no file>";
  char *SourceFile = NULL;
  int SourceFileLen = 0;
  char DefSourceLine[] = "<no linenumber>";
  char *SourceLine = NULL;
  int SourceLineLen = 0;

  if(pAlert->ResCode != DEF_RESCODE)
    pAlertRes = AlertRes_OpenEx(pAlert->ResCode, gKeepOpenFlag);

  if(!pAlertRes)
    {
      /* something is hosed with the resource files or the rescode was
         bogus, so we'll construct a default resource */

      pAlertRes = AlertRes_New();
      pAlertRes->ResCode = DEF_RESCODE;
      pAlertRes->IdLen  = strlen(DEF_ALERT_ID_STRING);
      pAlertRes->Id = (char *) SCalloc(1, pAlertRes->IdLen + 1);
      strcpy(pAlertRes->Id, DEF_ALERT_ID_STRING);

      /* leave room for insertion of the res number */
      pAlertRes->Text = SCalloc(1, strlen(DEF_ALERT_TEXT)+11);
      
      sprintf(pAlertRes->Text, DEF_ALERT_TEXT, pAlert->ResCode);
      pAlertRes->TextLen = strlen(pAlertRes->Text);
      /* leave the FacCodes member empty so that if this is going to
         alertd, the default target facility will get used */
    }


  /* now go through the inserts byte array, leaving a pointer at the
     beginning of each string */
  i = 0;
  p = (char *) pAlert->Inserts;


  /* the first insert is the source file name for tracing */
  if(pAlert->InsCount > 0)
    {
      SourceFile = p;
      while(*p++)
	SourceFileLen++;
    }
  if(SourceFileLen == 0) /* didn't get it :( */
    {
      SourceFile = DefSourceFile;
      SourceFileLen = sizeof(DefSourceFile);
    }


  /* the second insert is the source file linenumber for tracing */
  if(pAlert->InsCount > 1)
    {
      SourceLine = p;
      while(*p++)
	SourceLineLen++;
    }
  if(SourceLineLen == 0) /* didn't get it :( */
    {
      SourceLine = DefSourceLine;
      SourceLineLen = sizeof(DefSourceLine);
    }




  /* now get the rest of the inserts into the Inserts array */
  Inserts = (char **) SCalloc(pAlert->InsCount - 2 + 1, sizeof(char *));
  while(i < (pAlert->InsCount - 2)) /* take off 2 for the file and linenum */
    {
      Inserts[i++] = p;
      while(*p++)
	;
    }

  ExpSize = Res_ASNPrintf(Buf, pAlertRes->Text, Inserts, pAlert->InsCount - 2, 
			  sizeof(Buf));

  SFree(Inserts);

  if (ExpSize >= 0)
    {
      Expanded = (char *) SCalloc(1, 
				  ExpSize + 
				  SourceFileLen + SourceLineLen + 4 +
				  1);
      memcpy(Expanded, Buf, ExpSize);
      if(Flags & ALERT_PRINT_SOURCE_INFO)
	{
	  sprintf(Expanded + ExpSize," (%s:%s)",  SourceFile, SourceLine);
	  ExpSize += SourceFileLen + SourceLineLen + 4;
	}
      SFree(pAlertRes->Text);
      pAlertRes->Text = Expanded;
      pAlertRes->TextLen = ExpSize;
    }
  return pAlertRes;
}

