/*****************************************************************************
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT")


  File:   alertres.c
  Author: Andrew Lambeth <andrew@nfr.net>

 *****************************************************************************/

/******************************************************************************
 * alertres.c
 * functions used in processing nfr alert resources
 *****************************************************************************/
#include "sys-include.h"

#include "memsafe.h"
#include "alertres.h"
#include "resource.h"

#include "libnfr_err.h"


PAlertRes AlertRes_New(void)
{
  PAlertRes pNewAlertRes = NULL;

  pNewAlertRes = SCalloc(1, sizeof(TAlertRes));
  return pNewAlertRes;
}

void AlertRes_Delete(PAlertRes pAlertRes)
{
  if (pAlertRes->Text != NULL)
    SFree(pAlertRes->Text);
  if (pAlertRes->FacCodes != NULL)
    SFree(pAlertRes->FacCodes);
  ResId_Delete((PResId)pAlertRes);
}

/* AlertRes_Open() is a macro that calls AlertRes_OpenEx() with
   KeepOpen set FALSE */

PAlertRes AlertRes_OpenEx(ULONG ResCode, BOOL KeepOpen)
{
  static PResFile pUsrResFile = NULL;
  static PResFile pNfrResFile = NULL;
  PResFile pResFile = NULL;
  PAlertRes pNewAlertRes = NULL;

  pNewAlertRes = AlertRes_New();
  pNewAlertRes->ResCode = ResCode;
  if(ResCode_IsUser(ResCode))
    {
      if(!pUsrResFile)
	pUsrResFile = ResFile_Open(RESFILE_PATH "/usr/" RESFILE_ALERTS);
      pResFile = pUsrResFile;
    }
  else
    {
      if(!pNfrResFile)
	pNfrResFile = ResFile_Open(RESFILE_PATH "/nfr/" RESFILE_ALERTS);
      pResFile = pNfrResFile;
    }
  if(pResFile == NULL)
    {
      SFree(pNewAlertRes);
      return NULL;
    }
  ResFile_ReadRes((TResId *)pNewAlertRes, pResFile);

  if(!KeepOpen)
    {
      if(pUsrResFile)
	{
	  ResFile_Close(pUsrResFile);
	  pUsrResFile = NULL;
	}
      if(pNfrResFile)
	{
	  ResFile_Close(pNfrResFile);
	  pNfrResFile = NULL;
	}
    }


  return pNewAlertRes;
}
  
int AlertRes_SetText(char *Text, USHORT Size, PAlertRes pAlertRes)
{
  if (pAlertRes->Text != NULL)
    SFree(pAlertRes->Text);

  pAlertRes->Text = SCalloc(1, Size + 1);
  memcpy(pAlertRes->Text, Text, Size + 1);
  pAlertRes->TextLen = Size;
  return LIBNFR_ERR_SUCCESS;
}

char *AlertRes_GetText(char *Buf, USHORT Size, PAlertRes pAlertRes)
{
  strncpy(Buf, pAlertRes->Text, Size);
  return Buf;
}

int AlertRes_SetId(char *Id, USHORT Size, PAlertRes pAlertRes)
{
  if (pAlertRes->Id != NULL)
    SFree(pAlertRes->Id);
  pAlertRes->Id = SCalloc(1, Size + 1);
  memcpy(pAlertRes->Id, Id, Size + 1);
  pAlertRes->IdLen = Size;
  return LIBNFR_ERR_SUCCESS;
}

char *AlertRes_GetId(char *Buf, USHORT Size, PAlertRes pAlertRes)
{
  strncpy(Buf, pAlertRes->Id, Size);
  return Buf;
}

int AlertRes_SetFacCodes(FacCode *FacCodes, USHORT Size, PAlertRes pAlertRes)
{
  if (pAlertRes->FacCodes != NULL)
    SFree(pAlertRes->FacCodes);
  if (Size == 0)
    {
      pAlertRes->FacCodes = NULL;
      return LIBNFR_ERR_SUCCESS;
    }
  pAlertRes->FacCodes = SCalloc(Size, sizeof(FacCode));
  memcpy(pAlertRes->FacCodes, FacCodes, Size * sizeof(FacCode));
  pAlertRes->FacCodeCount = Size;
  return LIBNFR_ERR_SUCCESS;
}

BYTE *AlertRes_GetFacCodes(FacCode *Buf, USHORT Size, PAlertRes pAlertRes)
{
  if (Size > pAlertRes->FacCodeCount)
    Size = pAlertRes->FacCodeCount;
  memcpy(Buf, pAlertRes->FacCodes, Size);
  return Buf;
}


ULONG AlertRes_WriteBody(PAlertRes pAlertRes, ULONG *pByteCounter, FILE *Out)
{
  ULONG Result = 0;
  ULONG NBOTextLen = 0;
  ULONG NBOFacCodeCount = 0;
  ULONG NBOSeverity = 0;
  /* always store numeric things as network byte order ulongs */

  NBOSeverity = htonl((ULONG)pAlertRes->Severity);
  Result = fwrite(&NBOSeverity, 1, sizeof(NBOSeverity), Out);
  if (Result != sizeof(NBOSeverity))
    return LIBNFR_ERR_UNDEFINED;
  if(pByteCounter != NULL)
    *pByteCounter += Result;

  NBOTextLen = htonl(pAlertRes->TextLen);
  Result = fwrite(&NBOTextLen, 1, sizeof(NBOTextLen), Out);
  if (Result != sizeof(NBOTextLen))
    return LIBNFR_ERR_UNDEFINED;
  if(pByteCounter != NULL)
    *pByteCounter += Result;

  Result = fwrite(pAlertRes->Text, 1, pAlertRes->TextLen, Out);
  if (Result != pAlertRes->TextLen)
    return LIBNFR_ERR_UNDEFINED;
  if(pByteCounter != NULL)
    *pByteCounter += Result;

  NBOFacCodeCount = htonl(pAlertRes->FacCodeCount);
  Result = fwrite(&NBOFacCodeCount, 1, sizeof(NBOFacCodeCount), Out);
  if (Result != sizeof(NBOFacCodeCount))
    return LIBNFR_ERR_UNDEFINED;
  if(pByteCounter != NULL)
    *pByteCounter += Result;

  Result = fwrite(pAlertRes->FacCodes, 1, pAlertRes->FacCodeCount, Out);
  if (Result != pAlertRes->FacCodeCount)
    return LIBNFR_ERR_UNDEFINED;
  if(pByteCounter != NULL)
    *pByteCounter += Result;

  return LIBNFR_ERR_SUCCESS;
}


ULONG AlertRes_ReadBody(PAlertRes pAlertRes, ULONG *pByteCounter, FILE *In)
{
  ULONG Result = 0;
  ULONG NBOTextLen = 0;
  ULONG NBOFacCodeCount = 0;
  ULONG NBOSeverity = 0;

  Result = fread(&NBOSeverity, 1, sizeof(NBOSeverity), In);
  if (Result != sizeof(NBOSeverity))
    return LIBNFR_ERR_UNDEFINED;
  if(pByteCounter != NULL)
    *pByteCounter += Result;
  pAlertRes->Severity = (BYTE) ntohl(NBOSeverity);

  Result = fread(&NBOTextLen, 1, sizeof(NBOTextLen), In);
  if (Result != sizeof(NBOTextLen))
    return LIBNFR_ERR_UNDEFINED;
  if(pByteCounter != NULL)
    *pByteCounter += Result;
  pAlertRes->TextLen = (USHORT) ntohl(NBOTextLen);

  pAlertRes->Text = (char *) SCalloc(1, pAlertRes->TextLen + 1);
  Result = fread(pAlertRes->Text, 1, pAlertRes->TextLen, In);
  if (Result != pAlertRes->TextLen)
    return LIBNFR_ERR_UNDEFINED;
  if(pByteCounter != NULL)
    *pByteCounter += Result;

  Result = fread(&NBOFacCodeCount, 1, sizeof(NBOFacCodeCount), In);
  if (Result != sizeof(NBOFacCodeCount))
    return LIBNFR_ERR_UNDEFINED;
  if(pByteCounter != NULL)
    *pByteCounter += Result;
  pAlertRes->FacCodeCount = (USHORT) ntohl(NBOFacCodeCount);

  pAlertRes->FacCodes = (BYTE *) SCalloc(1, pAlertRes->FacCodeCount + 1);
  Result = fread(pAlertRes->FacCodes, 1, pAlertRes->FacCodeCount, In);
  if (Result != pAlertRes->FacCodeCount)
    return LIBNFR_ERR_UNDEFINED;
  if(pByteCounter != NULL)
    *pByteCounter += Result;

  return LIBNFR_ERR_SUCCESS;
}


