/*****************************************************************************
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT")


  File:   alertsrc.c
  Author: Andrew Lambeth <andrew@nfr.net>

 *****************************************************************************/

/******************************************************************************
 * alertsrc.c
 * functions used in processing nfr alert source resources
 *****************************************************************************/
#include "sys-include.h"

#include "memsafe.h"
#include "alertsrc.h"
#include "alertres.h"
#include "resource.h"

#include "libnfr_err.h"


PAlertSrc AlertSrc_New(void)
{
  PAlertSrc pNewAlertSrc = NULL;

  pNewAlertSrc = SCalloc(1, sizeof(TAlertSrc));
  return pNewAlertSrc;
}

void AlertSrc_Delete(PAlertSrc pAlertSrc)
{
  if (pAlertSrc->Desc != NULL)
    SFree(pAlertSrc->Desc);
  if (pAlertSrc->Name != NULL)
    SFree(pAlertSrc->Name);
  ResId_Delete((PResId)pAlertSrc);
}

/* AlertRes_Open() is a macro that calls AlertRes_OpenEx() with
   KeepOpen set FALSE */

PAlertSrc AlertSrc_OpenEx(ULONG ResCode, BOOL KeepOpen)
{
  static PResFile pUsrResFile = NULL;
  static PResFile pNfrResFile = NULL;
  PResFile pResFile = NULL;
  PAlertSrc pNewAlertSrc = NULL;
  
  pNewAlertSrc = AlertSrc_New();
  pNewAlertSrc->ResCode = ResCode;
  if(ResCode_IsUser(ResCode))
    {
      if(!pUsrResFile)
	pUsrResFile = ResFile_Open(RESFILE_PATH "/usr/" RESFILE_ALERT_SRCS);
      pResFile = pUsrResFile;
    }
  else
    {
      if(!pNfrResFile)
	pNfrResFile = ResFile_Open(RESFILE_PATH "/nfr/" RESFILE_ALERT_SRCS);
      pResFile = pNfrResFile;
    }
  if(pResFile == NULL)
    {
      SFree(pNewAlertSrc);
      return NULL;
    }
  ResFile_ReadRes((TResId *)pNewAlertSrc, pResFile);

  if(!KeepOpen)
    {
      if(pUsrResFile)
	{
	  ResFile_Close(pUsrResFile);
	  pUsrResFile = NULL;
	}
      if(pNfrResFile)
	{
	  ResFile_Close(pNfrResFile);
	  pNfrResFile = NULL;
	}
    }

  return pNewAlertSrc;
}

int AlertSrc_SetName(char *Name, USHORT Size, PAlertSrc pAlertSrc)
{
  if (pAlertSrc->Name != NULL)
    SFree(pAlertSrc->Name);
  pAlertSrc->Name = SCalloc(1, Size + 1);
  memcpy(pAlertSrc->Name, Name, Size + 1);
  pAlertSrc->NameLen = Size;
  return LIBNFR_ERR_SUCCESS;
}

char *AlertSrc_GetName(char *Buf, USHORT Size, PAlertSrc pAlertSrc)
{
  strncpy(Buf, pAlertSrc->Name, Size);
  return Buf;
}

int AlertSrc_SetDesc(char *Desc, USHORT Size, PAlertSrc pAlertSrc)
{
  if (pAlertSrc->Desc != NULL)
    SFree(pAlertSrc->Desc);
  pAlertSrc->Desc = SCalloc(1, Size + 1);
  memcpy(pAlertSrc->Desc, Desc, Size + 1);
  pAlertSrc->DescLen = Size;
  return LIBNFR_ERR_SUCCESS;
}

char *AlertSrc_GetDesc(char *Buf, USHORT Size, PAlertSrc pAlertSrc)
{
  strncpy(Buf, pAlertSrc->Desc, Size);
  return Buf;
}

int AlertSrc_SetId(char *Id, USHORT Size, PAlertSrc pAlertSrc)
{
  if (pAlertSrc->Id != NULL)
    SFree(pAlertSrc->Id);
  pAlertSrc->Id = SCalloc(1, Size + 1);
  memcpy(pAlertSrc->Id, Id, Size + 1);
  pAlertSrc->IdLen = Size;
  return LIBNFR_ERR_SUCCESS;
}

char *AlertSrc_GetId(char *Buf, USHORT Size, PAlertSrc pAlertSrc)
{
  strncpy(Buf, pAlertSrc->Id, Size);
  return Buf;
}


ULONG AlertSrc_WriteBody(PAlertSrc pAlertSrc, ULONG *pByteCounter, 
			  FILE *Out)
{
  ULONG Result = 0;
  ULONG NBONameLen = 0;
  ULONG NBODescLen = 0;
  /* always store numeric things as network byte order ulongs */

  NBONameLen = htonl(pAlertSrc->NameLen);
  Result = fwrite(&NBONameLen, 1, sizeof(NBONameLen), Out);
  if (Result != sizeof(NBONameLen))
    return LIBNFR_ERR_UNDEFINED;
  if(pByteCounter != NULL)
    *pByteCounter += Result;

  Result = fwrite(pAlertSrc->Name, 1, pAlertSrc->NameLen, Out);
  if (Result != pAlertSrc->NameLen)
    return LIBNFR_ERR_UNDEFINED;
  if(pByteCounter != NULL)
    *pByteCounter += Result;

  NBODescLen = htonl(pAlertSrc->DescLen);
  Result = fwrite(&NBODescLen, 1, sizeof(NBODescLen), Out);
  if (Result != sizeof(NBODescLen))
    return LIBNFR_ERR_UNDEFINED;
  if(pByteCounter != NULL)
    *pByteCounter += Result;

  Result = fwrite(pAlertSrc->Desc, 1, pAlertSrc->DescLen, Out);
  if (Result != pAlertSrc->DescLen)
    return LIBNFR_ERR_UNDEFINED;
  if(pByteCounter != NULL)
    *pByteCounter += Result;

  return LIBNFR_ERR_SUCCESS;
}


ULONG AlertSrc_ReadBody(PAlertSrc pAlertSrc, ULONG *pByteCounter, FILE *In)
{
  ULONG Result = 0;
  ULONG NBONameLen = 0;
  ULONG NBODescLen = 0;

  Result = fread(&NBONameLen, 1, sizeof(NBONameLen), In);
  if (Result != sizeof(NBONameLen))
    return LIBNFR_ERR_UNDEFINED;
  if(pByteCounter != NULL)
    *pByteCounter += Result;
  pAlertSrc->NameLen = (USHORT) ntohl(NBONameLen);

  pAlertSrc->Name = (char *) SCalloc(1, pAlertSrc->NameLen + 1);
  Result = fread(pAlertSrc->Name, 1, pAlertSrc->NameLen, In);
  if (Result != pAlertSrc->NameLen)
    return LIBNFR_ERR_UNDEFINED;
  if(pByteCounter != NULL)
    *pByteCounter += Result;

  Result = fread(&NBODescLen, 1, sizeof(NBODescLen), In);
  if (Result != sizeof(NBODescLen))
    return LIBNFR_ERR_UNDEFINED;
  if(pByteCounter != NULL)
    *pByteCounter += Result;
  pAlertSrc->DescLen = (USHORT) ntohl(NBODescLen);

  pAlertSrc->Desc = (char *) SCalloc(1, pAlertSrc->DescLen + 1);
  Result = fread(pAlertSrc->Desc, 1, pAlertSrc->DescLen, In);
  if (Result != pAlertSrc->DescLen)
    return LIBNFR_ERR_UNDEFINED;
  if(pByteCounter != NULL)
    *pByteCounter += Result;

  return LIBNFR_ERR_SUCCESS;
}
