/*****************************************************************************
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT")


  File:   daemon.c
  Author: Andrew Lambeth <andrew@nfr.net>

 *****************************************************************************/

/*****************************************************************************
 * daemon.c - functions to isolate daemon code from OS specifics
 *****************************************************************************/
#include "sys-include.h"

#include "daemon.h"
#include "memsafe.h"

#include "libnfr_err.h"

/* uncomment this to prevent fork()ing for debugging startup code */
/* #define NO_FORK_DAEMON */

/* (gross) - the child process will use this global as it's "this" ptr
 * in SigHandler() so it can give the callbacks a pointer to itself and
 * so it can delete itself on termination */
PDaemon Daemon_This;

PDaemon Daemon_New(void)
{
  PDaemon pNewDaemon = NULL;

  pNewDaemon = SCalloc(1, sizeof(TDaemon));
  return pNewDaemon;
}

void Daemon_Delete(PDaemon pDaemon)
{
  if (pDaemon->OnDelete != NULL)
    pDaemon->OnDelete(pDaemon);
  SFree(pDaemon);
}

int Daemon_Start(PDaemon pDaemon, int Timeout)
{
  int Result = 0;

  /* do this first so we can bail out if it fails */
  if (pDaemon->OnStart != NULL)
    {
      Result = pDaemon->OnStart(pDaemon);
      if (Result != LIBNFR_ERR_SUCCESS)
	return Result;
    }

#ifndef NO_FORK_DAEMON
  pDaemon->DaemonId = fork();
  if (pDaemon->DaemonId < 0)
    return LIBNFR_ERR_UNDEFINED;
  if (pDaemon->DaemonId > 0)
    {
      signal(SIGCHLD, SIG_IGN);
      return LIBNFR_ERR_SUCCESS;
    }

  /* now we're alone with a healthy child */
#endif /* NO_FORK_DAEMON */

  Daemon_This = pDaemon;	/* blech. we set this only in the child, */
				/* it's meaningless in the parent and  */
				/* the parent can keep a pointer for each */
				/* daemon if it wants. OOC, oook! :( */

  pDaemon->DaemonId = getpid();

  /* Set OnTermSigs as the default handler for all signals whose
     default action is termination, so that we ensure that the cleanup
     routines get called unless we get an uncatchable */

  /* bug: do something about os specific signals here */
  signal(SIGHUP, OnTermSigs);
  signal(SIGABRT, OnTermSigs);
  signal(SIGTERM, OnTermSigs);
  signal(SIGINT, OnTermSigs);

#ifdef CATCH_ALL_SIGS
  signal(SIGALRM, OnTermSigs);
  signal(SIGBUS, OnTermSigs);
  signal(SIGFPE, OnTermSigs);
  signal(SIGILL, OnTermSigs);
  signal(SIGIO, OnTermSigs);
  signal(SIGIOT, OnTermSigs);
  signal(SIGPIPE, OnTermSigs);
  signal(SIGPOLL, OnTermSigs);
  signal(SIGPROF, OnTermSigs);
  signal(SIGQUIT, OnTermSigs);
  signal(SIGSEGV, OnTermSigs);
  signal(SIGTRAP, OnTermSigs);
  signal(SIGUSR1, OnTermSigs);
  signal(SIGUSR2, OnTermSigs);
  signal(SIGXCPU, OnTermSigs);
  signal(SIGXFSZ, OnTermSigs);
  signal(SIGVTALRM, OnTermSigs);
  /* signal(SIGSYS, OnTermSigs); */
  /* signal(SIGEMT, OnTermSigs); */
#endif

  if (pDaemon->Main == NULL)
    return LIBNFR_ERR_UNDEFINED;
  return (pDaemon->Main(pDaemon));
}

int Daemon_Stop(PDaemon pDaemon, int Timeout)
{
  int Result;

  Result = kill(pDaemon->DaemonId, SIGTERM);
  pDaemon->Status = NFR_STAT_STOPPED;
  return LIBNFR_ERR_SUCCESS;
}

int Daemon_Kill(PDaemon pDaemon, int Timeout)
{
  int Result;

  Result = kill(pDaemon->DaemonId, SIGKILL);
  pDaemon->Status = NFR_STAT_STOPPED;
  return LIBNFR_ERR_SUCCESS;
}

void OnTermSigs(int SigNo)
{
  int Result;
  
  switch (SigNo)
    {
      /* bug: should add something here for unexpected signals that we
         caught, like SIGSEGV */
    default:
      Result = Daemon_This->OnStop(Daemon_This);
      Daemon_Delete(Daemon_This);
      exit(Result);
      break;
    }
}

