/*****************************************************************************
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT")


  File:   dgram.c
  Author: Andrew Lambeth <andrew@nfr.net>

 *****************************************************************************/

#include "sys-include.h"

#include "dgram.h"
#include "debug.h"
#include "memsafe.h"

#include "libnfr_err.h"

int DGram_SetData(void *Data, int DataLen, PDGram pDGram) 
{
  if( (DataLen + DGRAM_HDR_SIZE) > DGRAM_MAXSIZE)
    DataLen = DGRAM_MAXSIZE - DGRAM_HDR_SIZE; /* keep 'em all within spec */

  /* bug: if the above code truncates a string, we may cause a buffer
     overrun on the recieving end if they expect a null terminator in
     the data */

  memcpy ((BYTE *)pDGram + DGRAM_HDR_SIZE, Data, DataLen);
  pDGram->Len = DataLen;
  return LIBNFR_ERR_SUCCESS;
}

int DGram_Recv_fd(int fdRead, PDGram pDGram, int DGramSize, int Timeout)
{
  int Result = 0;
  fd_set fdsRead;
  struct timeval Tv;
  struct timeval *pTv = NULL;

  FD_ZERO(&fdsRead);
  FD_SET(fdRead, &fdsRead);  

  if (Timeout != TIMEOUT_NEVER) 
    {
      Tv.tv_sec = Timeout / 1000;
      Tv.tv_usec = (Timeout % 1000) * 1000;
      pTv = &Tv;
    }
  
  /* bug: need to fix this for when the whole msg is not in the pipe */
 
  Result = select(fdRead+1, &fdsRead, NULL, NULL, pTv);  
  if (Result > 0)
    { /* we've got somthin, see if it's a valid DGram */
      Result = read(fdRead, pDGram, DGRAM_HDR_SIZE);
      if (Result > 0)
	{ /* got the header, now go ahead and read the data */
	  if (DGramSize > pDGram->Len)
	    DGramSize = pDGram->Len + DGRAM_HDR_SIZE;
	  

	  /* kill(getpid(), SIGSEGV); */ /* test for mark's question */


	  Result = read(fdRead, 
			(BYTE *)pDGram + DGRAM_HDR_SIZE, 
			DGramSize - DGRAM_HDR_SIZE);
	  
	  if(Result == pDGram->Len)
	    return LIBNFR_ERR_SUCCESS; /* yay, we got a DGram */



	  /* bug: fix the following error handling */
	  

	  /* 
	  ** if we got to here, the second call to read() (the DGram
	  ** data) failed OR we didn't get ALL the data 
	  */
	  if (Result < 0)
	    return LIBNFR_ERR_UNDEFINED;
	  else if (Result == 0)
	    return LIBNFR_ERR_EOF;
	  else
	    return LIBNFR_ERR_UNDEFINED;
	}
 


      /* 
      ** if we got to here, the first call to read() failed, we didn't
      ** get a DGram header
      */
      else if (Result < 0)
	{ 
	  /* bug: fill this in with something useful */
	  switch (errno)
	    {
	    case EAGAIN:
	      return LIBNFR_ERR_AGAIN;
	    default:
	      return LIBNFR_ERR_UNDEFINED;
	    }
	}
      else
	return LIBNFR_ERR_EOF;
    }



  /* 
  ** if we got to here, the select() call failed 
  */
  if (Result == 0)
    return LIBNFR_ERR_TIMEOUT;
  else
    return LIBNFR_ERR_UNDEFINED;
}


int DGram_Send_fd(int fdWrite, PDGram pDGram, int Timeout)
{
  int Result;
  int ByteCount;
  fd_set fdsWrite;
  struct timeval Tv;
  struct timeval *pTv = NULL;

  FD_ZERO(&fdsWrite);
  FD_SET(fdWrite, &fdsWrite);  

  if (Timeout != TIMEOUT_NEVER) 
    {
      Tv.tv_sec = Timeout / 1000;
      Tv.tv_usec = (Timeout % 1000) * 1000;
      pTv = &Tv;
    }
  
  Result = select(fdWrite+1, NULL, &fdsWrite, NULL, pTv);
  if (Result > 0)
    {
      /* bug: need to fix this for when the whole msg did not get into
         the pipe */
      ByteCount =  write(fdWrite, pDGram, pDGram->Len + DGRAM_HDR_SIZE);
      return LIBNFR_ERR_SUCCESS;
    }

  else if (Result < 0)
    return LIBNFR_ERR_UNDEFINED;
  else
    return LIBNFR_ERR_TIMEOUT;     
}




TDGServer *DGServer_New()
{
  TDGServer *pNewSrv = (TDGServer *) SCalloc(1, sizeof(TDGServer));
  pNewSrv->MaxWait = TIMEOUT_NEVER; /* block 4ever by default */
  return pNewSrv;
}

void DGServer_Delete(TDGServer *pSrv)
{
  SFree(pSrv);
}

int DGServer_Run(TDGServer *pSrv)
{
  BYTE Buf[DGRAM_MAXSIZE];
  int Result = LIBNFR_ERR_SUCCESS;
  ULONG Wait = 0;
  ULONG State = DGSRV_BUSY_NOIDLE;
  BOOL CallIdleProc = FALSE;

  while (TRUE)
    {
      switch (State)
	{
	case DGSRV_BUSY_NOIDLE:
	  /* this state is entered initially, and whenever there is
             data on the line */

	  /* read data with no waiting, and no calls to pIdleProc */
	  Wait = 0;
	  CallIdleProc = FALSE;
	  break;

	case DGSRV_BUSY:
	  /* this state is entered whenever there is no data on the
             line, and the last call to pIdleProc() returned
             LIBNFR_ERR_AGAIN */

	  /* call pIdleProc() with no delay as long as there isn't any
             data on the line */
	  Wait = 0;
	  if (pSrv->IdleProc != NULL)
	    CallIdleProc = TRUE;
	  break;

	case DGSRV_IDLE:
	  /* this state is entered whenever there is no data on the
             line, and the last call to pIdleProc() returned
             LIBNFR_ERR_SUCCESS */

	  /* check for data for Timeout usecs, and then if there isn't
             any, call pIdleProc() */
	  Wait = pSrv->MaxWait;
	  if (pSrv->IdleProc != NULL)
	    CallIdleProc = TRUE;
	  break;

	case DGSRV_ERROR:
	  /* this state is entered whenever an error is encountered */

	  /* call an pErrProc which sould attempt to resolve the
             problem */
	  Result = pSrv->ErrProc(Result, pSrv);
	  if (Result != LIBNFR_ERR_SUCCESS)
	    return Result;
	  State = DGSRV_BUSY;
	  break;
	}

      Result = DGram_Recv_fd(pSrv->fdRead,
			     (PDGram)Buf,
			     sizeof(Buf), 
			     Wait);

      if(Result == LIBNFR_ERR_SUCCESS)
	{
	  State = DGSRV_BUSY_NOIDLE;
	  Result = pSrv->DataProc((PDGram)Buf, pSrv);
	  if(Result != LIBNFR_ERR_SUCCESS)
	    State = DGSRV_ERROR;
	}

      else if (Result == LIBNFR_ERR_TIMEOUT)
	{
	  if (CallIdleProc)
	    {
	      Result = pSrv->IdleProc(pSrv);
	      if (Result == LIBNFR_ERR_AGAIN)
		State = DGSRV_BUSY;
	      else if (Result == LIBNFR_ERR_SUCCESS)
		State = DGSRV_IDLE;
	      else
		State = DGSRV_ERROR;
	    }
	  else
	    State = DGSRV_IDLE;
 	}
      else
	State = DGSRV_ERROR;
    }
}
