/*****************************************************************************
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT")


  File:   list.c
  Author: Andrew Lambeth <andrew@nfr.net>

 *****************************************************************************/

/******************************************************************************
 * TList - a simple sorted linked list, in two flavors, intrusive and
 * non-intrisive.  These are intended to be used as a base for other
 * dynamically sized structures like stacks and queues.
 *****************************************************************************/
#include "sys-include.h"

#include "xtypes.h"
#include "memsafe.h"
#include "list.h"

PList List_New(PfnCompare Compare)
{
  PList pNewList = NULL;

  pNewList = SCalloc(1, sizeof(TList));
  pNewList->Compare = Compare;
  return pNewList;
}

void List_Delete(PList pList)
{
  SFree(pList);
}

PListNode List_Insert(PListNode pNew, PList pList)
{
  PListNode pCur = pList->Head;

  assert((pNew != NULL) && (pList != NULL));

  /* handle the 2 cases where the only choice is the head */
  if (pList->Compare == NULL || pList->Head == NULL)
    {
      List_AddToHead(pNew, pList);
      if (pList->Tail == NULL)	/* this happens when the list was empty */
	pList->Tail = pNew; 
      return pNew;
    }
  else /* a comparison function has been provided by our diligent coder */
    {
      /* handle the 2 special cases */
      if ( pList->Compare(pNew,pList->Head) <= 0 )
	return List_AddToHead(pNew, pList);
      if ( pList->Compare(pNew,pList->Tail) >= 0 )
	return List_AddToTail(pNew, pList);
      
      /* handle the general case */
      while(pCur->Next != pList->Tail)
	{
	  /* go through the list until pNew is "smaller" than the next one */
	  if (pList->Compare(pNew, pCur->Next) < 0)
	    {
	      pNew->Next = pCur->Next;
	      pCur->Next = pNew;
	      pList->Count++;
	      return pNew;
	    }
	  pCur = pCur->Next;
	}
      assert(TRUE);		/* if we get here something's hosed */
    }
  return NULL; /* shutup cc -Wall */
}

PListNode List_AddToHead(PListNode pNew, PList pList)
{
  assert((pNew != NULL) && (pList != NULL));

  if(pList->Count == 0)
    pList->Tail = pNew;
  pNew->Next = pList->Head;
  pList->Head = pNew;
  pList->Count++;
  return pNew;
}

PListNode List_AddToTail(PListNode pNew, PList pList)
{
  assert((pNew != NULL) && (pList != NULL));

  if(pList->Count == 0)
    pList->Head = pNew;
  else
    pList->Tail->Next = pNew;
  pList->Tail = pNew;
  pList->Count++;
  return pNew;
}

PListNode List_Remove(PListNode pNode, PList pList)
{
  PListNode pCur = NULL;
  
  assert((pNode != NULL) && (pList != NULL));
  
  /* handle the 3 special cases */
  if (pList->Count == 0)
    return NULL;		/* empty list so we know it ain't here */
  if (pNode == pList->Head)
    return List_RemoveHead(pList);
  if (pNode == pList->Tail)
    return List_RemoveTail(pList);

  /* handle the general case */
  pCur = pList->Head;
  while(pCur->Next != pList->Tail)
    {
      if (pCur->Next == pNode)
	{
	  pCur->Next = pCur->Next->Next;
	  pList->Count--;
	  return pNode;
	}
      pCur = pCur->Next;
    }
  return NULL;			/* 'tweren't there */
}

PListNode List_RemoveHead(PList pList)
{
  PListNode pNode = pList->Head;

  assert(pList != NULL);

  if (pList->Count == 0)
    return NULL;
  if (pList->Count == 1)
    {
      pList->Count = 0;
      pList->Head = NULL;
      pList->Tail = NULL;
      return pNode;
    }
  pList->Head = pList->Head->Next;
  pList->Count--;
  return pNode;
}

PListNode List_RemoveTail(PList pList)
{
  PListNode pCur = pList->Head;
  PListNode pPrev = NULL;

  assert(pList != NULL);

  if (pList->Count == 0)
    return NULL;
  if (pList->Count == 1)
    {
      pList->Count = 0;
      pList->Head = NULL;
      pList->Tail = NULL;
      return pCur;
    }
  while (pCur->Next != NULL)
    {
      pPrev = pCur;
      pCur = pCur->Next;
    }
  pList->Tail = pPrev;
  pList->Tail->Next = NULL;
  pList->Count--;
  return pCur;
}

void *NIList_Insert(void *pData, PNIList pList)
{
  PNIListNode pNew;

  pNew = (PNIListNode) SCalloc(1, sizeof(TNIListNode));
  pNew->Data = pData;
  if (List_Insert((PListNode)pNew, (PList) pList) != NULL)
    return pNew->Data;
  return NULL;
}

void *NIList_AddToHead(void *pData, PNIList pList)
{
  PNIListNode pNew;

  pNew = (PNIListNode) SCalloc(1, sizeof(TNIListNode));
  pNew->Data = pData;
  if (List_AddToHead((PListNode)pNew, (PList) pList) != NULL)
    return pNew->Data;
  return NULL;
}

void *NIList_AddToTail(void *pData, PNIList pList)
{
  PNIListNode pNew;

  pNew = (PNIListNode) SCalloc(1, sizeof(TNIListNode));
  pNew->Data = pData;
  if (List_AddToTail((PListNode)pNew, (PList) pList) != NULL)
    return pNew->Data;
  return NULL;
}

void *NIList_Remove(void *pData, PNIList pNIList)
{
  PList pList = (PList)pNIList;
  PListNode pCur = NULL;
  PListNode pTarget = NULL;
  
  /* handle the 3 special cases */
  if (pList->Count == 0)
    return NULL;		/* empty list so we know it ain't here */
  else if (pData == ( (PNIListNode)pList->Head)->Data )
    pTarget = List_RemoveHead(pList);
  else if (pData == ( (PNIListNode)pList->Tail)->Data )
    pTarget = List_RemoveTail(pList);
  else
    {
      /* handle the general case, it's in the middle somewheres */
      pCur = pList->Head;
      while(pCur->Next != pList->Tail)
	{
	  if ( ((PNIListNode)pCur->Next)->Data == pData)
	    {
	      pTarget = pCur->Next;
	      pCur->Next = pCur->Next->Next;
	      pList->Count--;
	      return pData;
	    }
	  pCur = pCur->Next;
	}
    }

  if(pTarget == NULL)
    return NULL;			/* 'tweren't there */

  SFree(pTarget);
  return pData;
}

void *NIList_RemoveHead(PNIList pList)
{
  PNIListNode pNode = NULL;
  void *pData;

  pNode = (PNIListNode) List_RemoveHead((PList) pList);
  if (pNode != NULL)
    {
      pData = pNode->Data;
      SFree (pNode);
      return pData;
    }
  return NULL;
}

void *NIList_RemoveTail(PNIList pList)
{
  PNIListNode pNode = NULL;
  void *pData;

  pNode = (PNIListNode) List_RemoveTail((PList) pList);
  if (pNode != NULL)
    {
      pData = pNode->Data;
      SFree (pNode);
      return pData;
    }
  return NULL;
}

void *NIList_GetFirst(PNIList pList)
{
  /* not reentrant */
  pList->CurNode = (PNIListNode) pList->Head;
  if(pList->CurNode == NULL)
    return NULL;
  return ((PNIListNode) pList->Head)->Data;
}

void *NIList_GetNext(PNIList pList)
{
  /* not reentrant */
  pList->CurNode = pList->CurNode->Next;
  if(pList->CurNode == NULL)
    return NULL;
  return pList->CurNode->Data;
}
