/*****************************************************************************
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT")


  File:   resource.c
  Author: Andrew Lambeth <andrew@nfr.net>

 *****************************************************************************/

/******************************************************************************
 * resource.c
 * functions used in processing nfr resources
 *****************************************************************************/
#include "sys-include.h"

#include "memsafe.h"
#include "rescodes.h"
#include "resource.h"
#include "rangeres.h"
#include "alertres.h"

#include "libnfr_err.h"

/* bug: move this somewhere else */
#define MAX_PATHLEN 4096

BOOL gNeedResFileRefresh = TRUE;

int Res_Dump(TSymbol *pSym, int IArg, void *PArg)
{
  ULONG Result = 0;
  TResId *pResId = (TResId *) pSym->Data;
  FILE *Out = (FILE *) PArg;

  if(pResId == NULL)
    return LIBNFR_ERR_UNDEFINED;

  Result = ResId_Write(pResId, NULL, Out);
  if (Result != LIBNFR_ERR_SUCCESS)
    return Result;

  Result = Res_WriteBody(pResId, NULL, Out);
  if (Result != LIBNFR_ERR_SUCCESS)
    return Result;

  return LIBNFR_ERR_SUCCESS;
}
  
ULONG Res_ReadBody(PResId pResId, ULONG *pByteCounter, FILE *In)
{
  ULONG Result;

  switch(ResCode_GetResType(pResId->ResCode))
    {
    case ID_FAC:
      Result = LIBNFR_ERR_SUCCESS;
      break;
      
    case RES_RANGE:
      Result = RangeRes_ReadBody((PRangeRes)pResId, pByteCounter, In);
      break;
      
    case RES_ALERT:
      Result = AlertRes_ReadBody((PAlertRes)pResId, pByteCounter, In);
      break;
      
    case RES_ALERT_SRC:
      Result = AlertSrc_ReadBody((PAlertSrc)pResId, pByteCounter, In);
      break;
      
    case RES_STRING:
      Result = StringRes_ReadBody((PStringRes)pResId, pByteCounter, In);
      break;
      
    default:
      Result =  LIBNFR_ERR_UNDEFINED;
    }
  
  return Result;
}

ULONG Res_WriteBody(PResId pResId, ULONG *pByteCounter, FILE *Out)
{
  ULONG Result = 0;

  switch(ResCode_GetResType(pResId->ResCode))
    {
    case ID_FAC:
      Result = LIBNFR_ERR_SUCCESS;
      break;
      
    case RES_RANGE:
      Result = RangeRes_WriteBody((PRangeRes)pResId, pByteCounter, Out);
      break;
      
    case RES_ALERT:
      Result = AlertRes_WriteBody((PAlertRes)pResId, pByteCounter, Out);
      break;
      
    case RES_ALERT_SRC:
      Result = AlertSrc_WriteBody((PAlertSrc)pResId, pByteCounter, Out);
      break;
      
    case RES_STRING:
      Result = StringRes_WriteBody((PStringRes)pResId, pByteCounter, Out);
      break;
      
    default:
      Result =  LIBNFR_ERR_UNDEFINED;
    }

  return Result;
}

int Res_ASNPrintf(char *Dest, char *FmtStr, char **Inserts, int InsCount, 
		  int DestSize)
{
  char Buf[64]; /* use this for creating error inserts */
  char DefaultFmtStr[] = 
    "<NULL_FORMAT_STRING> first 3 inserts: $(1), $(2), $(3)";
  char *ThisInsert = NULL;
  char *s = FmtStr;
  char *d = Dest;
  char NumStr[4];
  char *NumStrPtr = NumStr;
  int Num = 0;
  ULONG State = RESEXP_INIT;

  if(!s)
    {
      /* use the default, this should never happen and should probably
         generate some sort of warning or error */
      s = DefaultFmtStr;
    }

  while (*s)
    {
      switch(State)
	{
	case RESEXP_INIT:
	  if (*s == '\\')
	    State = RESEXP_ESC;
	  else if (*s == '$')
	    State = RESEXP_WANTNUM;
	  else
	    *d++ = *s;
	  break;

	case RESEXP_ESC:
	  State = RESEXP_INIT;	
	  /* we only enter this state from RESEXP_INIT so that's what
             we'll always return to */

	  if (*s == 'n')
	    *d++ = '\n';
	  else if (*s == 't')
	    *d++ = '\t';
	  else if (*s == '\"')
	    *d++ = '\"';
	  else if (*s == '\'')
	    *d++ = '\'';
	  else
	    *d++ = *s;
	  break;

	case RESEXP_WANTNUM:
	  if(*s == '(')
	    State = RESEXP_GETNUM;
	  else
	    {
	      State = RESEXP_INIT;
	      *d++ = *s;
	    }
	  break;

	case RESEXP_GETNUM:
	  if(isdigit(*s) && ((NumStrPtr - NumStr) < sizeof(NumStr)) )
	    *NumStrPtr++ = *s;
	  else if(*s == ')')
	    {
	      *NumStrPtr = 0;
	      Num = atoi(NumStr);
	      NumStrPtr = NumStr; /* reset the pointer */
	      State = RESEXP_INIT;
	      if(Num-- < 1)
		{
		  sprintf(Buf, "<bad insert token number: $(%d)>", Num + 1);
		  ThisInsert = Buf;
		}
	      else if( (Num > InsCount) || !Inserts[Num] )	
		{
		  /* sorry, we don't have that one */
		  sprintf(Buf, "<null insert: $(%d)>", Num + 1);
		  ThisInsert = Buf;
		}
	      else
		ThisInsert = Inserts[Num];
	      
	      while(*ThisInsert && ((d - Dest) < (DestSize - 1)))
		*d++ = *ThisInsert++;

	      break;
	    }
	}
      if( (d - Dest) >= (DestSize - 1) )
	break;
      s++;
    }
  *d = 0;
  return (d - Dest);
}

PResFile ResFile_Create(char *ResName, ULONG Size)
{
  PResFile pResFile = NULL;

  pResFile = SCalloc(1, sizeof(TResFile));
  pResFile->File = fopen(ResName, "w");
  if(!pResFile->File)
    return NULL;
  pResFile->DataPtr = RES_DATA_START(Size);
  pResFile->IndexPtr = 0;
  pResFile->Size = Size;
  pResFile->Dirty = TRUE;
  return pResFile;
}

PResFile ResFile_Open(char *ResName)
{
  PResFile pResFile = NULL;

  pResFile = SCalloc(1, sizeof(TResFile));
  pResFile->File = fopen(ResName, "r");
  if(!pResFile->File)
    return NULL;
  fread(&(pResFile->Size), sizeof(pResFile->Size), 1, pResFile->File);
  pResFile->DataPtr = RES_DATA_START(pResFile->Size);
  pResFile->IndexPtr = 0;
  pResFile->Dirty = TRUE;  
  return pResFile;
}

void ResFile_Close(PResFile pResFile)
{
  if (pResFile->Dirty)
    ResFile_WriteHdr(pResFile);
  fclose(pResFile->File);
  SFree(pResFile);
}

void ResFiles_Refresh()
{
  PAlertSrc pNewAlertSrc = NULL;
  PAlertRes pNewAlertRes = NULL;

  /* bug: this is pretty gross, fix later */
  pNewAlertRes = AlertRes_OpenEx(0, FALSE);
  if(pNewAlertRes)
    AlertRes_Delete(pNewAlertRes);
  pNewAlertSrc = AlertSrc_OpenEx(0, FALSE);
  if(pNewAlertSrc)
    AlertSrc_Delete(pNewAlertSrc);
}

int ResFile_WriteHdr(PResFile pResFile)
{
  rewind(pResFile->File);
  fwrite(&(pResFile->Size), sizeof(pResFile->Size), 1, pResFile->File);
  pResFile->Dirty = FALSE;
  return LIBNFR_ERR_SUCCESS;
}


ULONG ResFile_WriteRes(PResId pResId, PResFile pResFile)
{
  ULONG Result = 0;

  if (pResFile->IndexPtr >= pResFile->Size)
    return LIBNFR_ERR_RESFILE_FULL;

  pResFile->Dirty = TRUE;

  ResFile_IdxAddEntry(pResId->ResCode, pResFile);

  fseek(pResFile->File, pResFile->DataPtr, SEEK_SET);

  Result = ResId_Write(pResId, &(pResFile->DataPtr), pResFile->File);
  if (Result != LIBNFR_ERR_SUCCESS)
    return Result;

  Result = Res_WriteBody(pResId, &(pResFile->DataPtr), pResFile->File);

  return Result;
}


ULONG ResFile_ReadRes(PResId pResId, PResFile pResFile)
{
  ULONG Result;

  Result = ResFile_Seek(pResId->ResCode, pResFile);
  if (Result != LIBNFR_ERR_SUCCESS)
    return Result;

  Result = ResId_Read(pResId, NULL, pResFile->File);
  if (Result != LIBNFR_ERR_SUCCESS)
    return Result;

  Result = Res_ReadBody(pResId, NULL, pResFile->File);

  return Result;
}

int ResFile_IdxAddEntry(ULONG ResCode, PResFile pResFile)
{
  TResNo CurResNo = 0;

  CurResNo = ResCode_GetResNo(ResCode);
  fseek(pResFile->File, RES_INDEX_OFFSET(pResFile->IndexPtr), SEEK_SET);
  fwrite(&CurResNo, sizeof(TResNo), 1, pResFile->File);
  fwrite(&(pResFile->DataPtr), sizeof(pResFile->DataPtr), 1, pResFile->File);
  pResFile->IndexPtr++;
  return LIBNFR_ERR_SUCCESS;
}

int ResFile_Seek(ULONG ResCode, PResFile pResFile)
{
  TResNo CurResNo = 0;
  int Result = LIBNFR_ERR_RESFILE_BADINDEX;

  fseek(pResFile->File, RES_INDEX_OFFSET(0), SEEK_SET);
  pResFile->IndexPtr = 0;
  while(pResFile->IndexPtr < pResFile->Size)
    {
      fread(&CurResNo, sizeof(TResNo), 1, pResFile->File);
      fread(&(pResFile->DataPtr), 
	    sizeof(pResFile->DataPtr), 1, pResFile->File);
      if (CurResNo == ResCode_GetResNo(ResCode))
	{
	  Result = LIBNFR_ERR_SUCCESS;
	  fseek(pResFile->File, pResFile->DataPtr, SEEK_SET);
	  break;
	}
      pResFile->IndexPtr++;
    }
  return Result;
}

