/*****************************************************************************
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT")


  File:   stringres.c
  Author: Andrew Lambeth <andrew@nfr.net>

 *****************************************************************************/

/******************************************************************************
 * stringres.c
 * functions used in processing nfr string resources
 *****************************************************************************/
#include "sys-include.h"

#include "memsafe.h"
#include "stringres.h"
#include "alertres.h"
#include "resource.h"

#include "libnfr_err.h"

PStringRes StringRes_New(void)
{
  PStringRes pNewStringRes = NULL;

  pNewStringRes = SCalloc(1, sizeof(TStringRes));
  return pNewStringRes;
}

void StringRes_Delete(PStringRes pStringRes)
{
  if (pStringRes->String != NULL)
    SFree(pStringRes->String);
  ResId_Delete((PResId)pStringRes);
}

PStringRes StringRes_Open(ULONG ResCode)
{
  PResFile pResFile;
  PStringRes pNewStringRes;
  
  pNewStringRes = StringRes_New();
  pNewStringRes->ResCode = ResCode;
  if(ResCode_IsUser(ResCode))
    pResFile = ResFile_Open(RESFILE_PATH "/usr/" RESFILE_STRINGS);
  else
    pResFile = ResFile_Open(RESFILE_PATH "/nfr/" RESFILE_STRINGS);
  if(pResFile == NULL)
    {
      SFree(pNewStringRes);
      return NULL;
    }
  ResFile_ReadRes((TResId *)pNewStringRes, pResFile);
  ResFile_Close(pResFile);

  return pNewStringRes;
}

int StringRes_SetString(char *String, USHORT Size, PStringRes pStringRes)
{
  if (pStringRes->String != NULL)
    SFree(pStringRes->String);
  pStringRes->String = SCalloc(1, Size + 1);
  memcpy(pStringRes->String, String, Size + 1);
  pStringRes->StringLen = Size;
  return LIBNFR_ERR_SUCCESS;
}

char *StringRes_GetString(char *Buf, USHORT Size, PStringRes pStringRes)
{
  strncpy(Buf, pStringRes->String, Size);
  return Buf;
}

char *StringRes_Load(ULONG ResCode)
{
  /* bug: this is a really inefficient way to do this, we shouldn't
     be using the other api calls 'cause we malloc our brains out */
  PStringRes pStringRes = NULL;
  char *String = NULL;

  pStringRes = StringRes_Open(ResCode);
  if (pStringRes != NULL)
    {
      String = SCalloc(1, pStringRes->StringLen + 1);
      StringRes_GetString(String, pStringRes->StringLen + 1, pStringRes);
    }
  StringRes_Close(pStringRes);
  return String;
}

char *StringRes_Build(ULONG ResCode, ...)
{
  /* bug: this is a really inefficient way to do this, we shouldn't
     be using the other api calls 'cause we malloc our brains out */
  char *pArg = NULL;
  va_list ap = NULL;
  char *Inserts[RES_MAX_INSERTS];
  int i = 0;
  ULONG TotalLen = 0;
  PStringRes pStringRes = NULL;
  char *String = NULL;

  pStringRes = StringRes_Open(ResCode);
  if (pStringRes != NULL)
    {
      TotalLen += pStringRes->StringLen;
      va_start(ap, ResCode);
      while( (pArg = va_arg(ap, char *)) != NULL)
	{
	  Inserts[i++] = pArg;
	  TotalLen += strlen(pArg);
	  if (i >= RES_MAX_INSERTS)
	    break;
	}
      va_end(ap);

      String = SCalloc(1, TotalLen  + 1);
      Res_ASNPrintf(String, pStringRes->String, Inserts, i, TotalLen);
    }
  StringRes_Close(pStringRes);
  return String;
}

int StringRes_SetId(char *Id, USHORT Size, PStringRes pStringRes)
{
  if (pStringRes->Id != NULL)
    SFree(pStringRes->Id);
  pStringRes->Id = SCalloc(1, Size + 1);
  memcpy(pStringRes->Id, Id, Size + 1);
  pStringRes->IdLen = Size;
  return LIBNFR_ERR_SUCCESS;
}

char *StringRes_GetId(char *Buf, USHORT Size, PStringRes pStringRes)
{
  strncpy(Buf, pStringRes->Id, Size);
  return Buf;
}

ULONG StringRes_WriteBody(PStringRes pStringRes, ULONG *pByteCounter, 
			  FILE *Out)
{
  ULONG Result = 0;
  ULONG NBOStringLen = 0;
  /* always store numeric things as network byte order ulongs */

  NBOStringLen = htonl(pStringRes->StringLen);
  Result = fwrite(&NBOStringLen, 1, sizeof(NBOStringLen), Out);
  if (Result != sizeof(NBOStringLen))
    return LIBNFR_ERR_UNDEFINED;
  if(pByteCounter != NULL)
    *pByteCounter += Result;

  Result = fwrite(pStringRes->String, 1, pStringRes->StringLen, Out);
  if (Result != pStringRes->StringLen)
    return LIBNFR_ERR_UNDEFINED;
  if(pByteCounter != NULL)
    *pByteCounter += Result;

  return LIBNFR_ERR_SUCCESS;
}


ULONG StringRes_ReadBody(PStringRes pStringRes, ULONG *pByteCounter, FILE *In)
{
  ULONG Result = 0;
  ULONG NBOStringLen = 0;

  Result = fread(&NBOStringLen, 1, sizeof(NBOStringLen), In);
  if (Result != sizeof(NBOStringLen))
    return LIBNFR_ERR_UNDEFINED;
  if(pByteCounter != NULL)
    *pByteCounter += Result;
  pStringRes->StringLen = (USHORT) ntohl(NBOStringLen);

  pStringRes->String = (char *) SCalloc(1, pStringRes->StringLen + 1);
  Result = fread(pStringRes->String, 1, pStringRes->StringLen, In);
  if (Result != pStringRes->StringLen)
    return LIBNFR_ERR_UNDEFINED;
  if(pByteCounter != NULL)
    *pByteCounter += Result;

  return LIBNFR_ERR_SUCCESS;
}
