/*****************************************************************************
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT")


  File:   symtab.c
  Author: Andrew Lambeth <andrew@nfr.net>

 *****************************************************************************/

/******************************************************************************
* symtab.c - a simple symbol table with typed symbols and a table-wide
* linked list of symbols
******************************************************************************/
#include "sys-include.h"

#include "list.h"
#include "debug.h"
#include "xtypes.h"
#include "memsafe.h"
#include "symtab.h"

/******************************************************************************
* SymTab_New() - returns a pointer to an initialized TSymtab with a
* hash table of size Size
******************************************************************************/
PSymTab SymTab_New(ULONG Size)
{
  PSymTab pSymTab = NULL;

  pSymTab = (PSymTab) SCalloc(1, sizeof(TSymTab));
  pSymTab->HashTab = (PSymbol *) SCalloc (Size, sizeof (PSymbol));
  pSymTab->HashTabSize = Size;
  pSymTab->SymList = (PNIList) SCalloc (1, sizeof (TNIList));
  return (pSymTab);
}

/******************************************************************************
* SymTab_Delete() - call this to free the space allocated by NewSymTab
* and destroy the symbol table
******************************************************************************/
void SymTab_Delete(PSymTab pSymTab)
{
  PSymbol pCurSymbol = NULL;

  pCurSymbol = SymList_RemoveLast(pSymTab->SymList);
  while(pCurSymbol != NULL)
  {
    SFree(pCurSymbol->String);
    SFree(pCurSymbol);
    pCurSymbol = SymList_RemoveLast(pSymTab->SymList);
  }
  SFree(pSymTab->HashTab);
  SFree(pSymTab->SymList);
  SFree(pSymTab);
}

/******************************************************************************
* SymTab_DeleteAll() - call this to free the space allocated by
* NewSymTab and destroy the symbol table AND free whatever space is
* pointed to each symbol's Data member.  If pDeleteData is specified
* it is called for each symbol in the table, otherwise only the Symbol
* struct and the String buffer are deleted.
******************************************************************************/
void SymTab_DeleteAll(PfnDelete pDeleteData, PSymTab pSymTab)
{
  PSymbol pCurSymbol = NULL;

  pCurSymbol = SymList_RemoveLast(pSymTab->SymList);
  while(pCurSymbol != NULL)
  {
    if((pDeleteData != NULL) && ((void *)(pCurSymbol->Data) != NULL))
      pDeleteData(pCurSymbol);
    SFree(pCurSymbol->String);
    SFree(pCurSymbol);
    pCurSymbol = SymList_RemoveLast(pSymTab->SymList);
  }
  SFree(pSymTab->HashTab);
  SFree(pSymTab->SymList);
  SFree(pSymTab);
}

/******************************************************************************
* SymTabRemove() - removes the symbol for String from the pSymTab.
* Returns the pointer to the symbol's Data member if the symbol was
* found in the table.  Otherwise NULL is returned
******************************************************************************/
void *SymTab_Remove(char *String, PSymTab pSymTab)
{
  /* bug: this hoses the hash table */

  PSymbol pSym = NULL;
  int Result = 0;
  void *Data = NULL;
  
  Result = SymTab_Peek(String, &pSym, pSymTab);
  if (Result != SYM_FOUND)
    return NULL;

  Data = (void *)pSym->Data;
  SymList_Remove(pSym, pSymTab->SymList);  
  SFree(pSym->String);
  pSym->String = NULL;
  SFree(pSym);
  return Data;
}

/******************************************************************************
* SymTabLookEx() - checks for String in the table pointed to by
* pSymtab and inserts it if it wasn't there and Flags contains
* AUTO_INSERT. ppSymbol is set to point to the symbol used by String.
* Return value is either SYM_FOUND, SYM_NOT_FOUND, or SYM_INSERTED
******************************************************************************/
int SymTab_LookEx(ULONG Flags, 
		  char *String, PSymbol *ppSymbol, PSymTab pSymTab)
{
  PSymbol pCurSymbol = NULL;
  long Hash = 0;
  char *p = String;


  while (*p)
  {
    Hash <<= 1;
    Hash ^= *p++;
  }
  if (Hash < 0)
    Hash = -Hash;
  Hash %= pSymTab->HashTabSize;
  for (pCurSymbol = pSymTab->HashTab[Hash]; 
       pCurSymbol != NULL;
       pCurSymbol = pCurSymbol->Next)
  {
    if (!strcasecmp(String, pCurSymbol->String))
    {
      if (ppSymbol != NULL)
	*ppSymbol = pCurSymbol;
      return SYM_FOUND; 
    }
  }

  if(!(Flags & AUTO_INSERT))
    {
      if (ppSymbol != NULL)
	*ppSymbol = NULL;      
      return SYM_NOT_FOUND;
    }

  /* string ain't be dere, best put it, sho 'nuff */
  pCurSymbol = (PSymbol) SCalloc(1, sizeof(TSymbol));
  pCurSymbol->StrLen = strlen(String);
  pCurSymbol->String = (char *) SCalloc(1, pCurSymbol->StrLen + 1);
  memcpy(pCurSymbol->String, String, pCurSymbol->StrLen + 1);
  pCurSymbol->Next = pSymTab->HashTab[Hash];
  pSymTab->HashTab[Hash] = pCurSymbol;
  if (ppSymbol != NULL)
    *ppSymbol = SymList_Insert(pCurSymbol, pSymTab->SymList);
  return SYM_INSERTED;
}

PSymbol SymList_Insert(PSymbol pSym, PNIList pList)
{
  if (pList->Compare == NULL)	/* no sorting */
    return (PSymbol) NIList_AddToTail((void *)pSym, pList);
  return (PSymbol) NIList_Insert((void *)pSym, pList); /* use sort function */
}

/******************************************************************************
* SymTab_ForEach() - Calls pfnIterate for each symbol of type Type in
* the table pSymTab.  pfnIterate is called with three arguments: a
* PSymbol, and the provided int IArg and void *PArg.  The number of
* Symbols operated on is returned on success, so if you just want to
* count the number of symbols of a particular type, set pfnIterate to
* NULL
******************************************************************************/
ULONG SymTab_ForEach(BYTE Type, PfnSymIterate pfnIterate, int IArg, void *PArg,
                  PSymTab pSymTab)
{
  PSymbol pCurSymbol = NULL;
  ULONG	  i=0;

  pCurSymbol = SymList_GetFirst(pSymTab->SymList);
  while (pCurSymbol != NULL)
  {
    if ((Type == ALL_TYPES) | (pCurSymbol->Type == Type))
    {
      if(pfnIterate != NULL)
	pfnIterate(pCurSymbol, IArg, PArg);
      i++;
    }
    pCurSymbol = SymList_GetNext(pSymTab->SymList);
  }
  return i;
}





















