/*****************************************************************************
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT")


  File:   restab.c
  Author: Andrew Lambeth <andrew@nfr.net>

 *****************************************************************************/

#include "sys-include.h"

#include "libnfr.h"
#include "rc.h"

/* #define DEBUG_LEV 1 */
#include "debug.h"

TResTab *ResTab_New(void)
{
  TResTab *pResTab = NULL;

  pResTab = SCalloc(1, sizeof(TResTab));
  pResTab->SymTab = SymTab_New(1024);
  pResTab->RangeList = SCalloc(1, sizeof(TNIList));
  return pResTab;
}

void ResTab_Delete(TResTab *pResTab)
{
  TRangeRes *pRangeRes = NULL;

  if(pResTab == NULL)
    return;

  SymTab_ForEach(ID_FAC, ResId_DeleteSym, 0, NULL, 
		 pResTab->SymTab);
  SymTab_ForEach(RES_ALERT_SRC, AlertSrc_DeleteSym, 0, NULL, 
		 pResTab->SymTab);
  SymTab_ForEach(RES_ALERT, AlertRes_DeleteSym, 0, NULL, pResTab->SymTab);
  SymTab_ForEach(RES_STRING, StringRes_DeleteSym, 0, NULL, pResTab->SymTab);
  pRangeRes = (TRangeRes *) NIList_RemoveHead(pResTab->RangeList);
  while(pRangeRes != NULL)
    {
      RangeRes_Delete(pRangeRes);
      pRangeRes = (TRangeRes *) NIList_RemoveHead(pResTab->RangeList);
    }
  SFree(pResTab->RangeList);
  SymTab_Delete(pResTab->SymTab);
  SFree(pResTab);
}

ULONG ResTab_DefineRange(char *RangeId, USHORT Min, USHORT Max,
			 char *DispName, char MemberType, 
			 TRangeRes **ppNewRange, TResTab *pResTab)
{
  ULONG Result = 0;
  TRangeRes *pRange = NULL;
  TSymbol *pSym = NULL;

  if ( ppNewRange != NULL)
    *ppNewRange = NULL;

  Result = SymTab_Look(RangeId, &pSym, pResTab->SymTab);
  if (Result != SYM_INSERTED)
    return RC_ERR_UNDEFINED;


  /* check to be sure that the requested range doesn't overlap any
     existing ranges */
  pRange = NIList_GetFirst(pResTab->RangeList);
  while (pRange != NULL)
    {
      if( (Min <= pRange->Max) && (Max >= pRange->Min) &&
	  (MemberType == pRange->MemberType) )
	{
	  return RC_ERR_UNDEFINED; /* ranges overlap */
	}
      pRange = NIList_GetNext(pResTab->RangeList);
    }

  

  /* create and add the new range */
  pRange = (TRangeRes *) SCalloc(1, sizeof(TRangeRes));

  pRange->Min = Min;
  pRange->Max = Max;

  pRange->IdLen = strlen(RangeId);
  pRange->Id = (char *) SCalloc(1, pRange->IdLen + 1 );
  memcpy (pRange->Id, RangeId, pRange->IdLen);

  pRange->DispNameLen = strlen(DispName);
  pRange->DispName = (char *) SCalloc(1, pRange->DispNameLen + 1 );
  memcpy (pRange->DispName, DispName, pRange->DispNameLen);

  ResCode_SetResNo(pRange->ResCode, pRange->Min);

  pRange->HighNo = pRange->Min;

  ResCode_SetResType(pRange->ResCode, RES_RANGE);

  pRange->MemberType = MemberType;

  if(gFixedResFlag)
    ResCode_SetUser(pRange->ResCode);
    
  NIList_AddToTail ((void *)pRange, pResTab->RangeList);

  pSym->Data = (ULONG) pRange;
  pSym->Type = RES_RANGE;

  if(ppNewRange != NULL)
    *ppNewRange = pRange;

  return NFR_ERR_SUCCESS;
}


ULONG ResTab_AssignResNo(ULONG Preset, char *RangeId, TResId *pNewRes, 
			 TResTab *pResTab)
{
  ULONG Result = 0;
  TRangeRes *pRange = NULL;
  TSymbol *pSym = NULL;
  
  Result = SymTab_Peek(RangeId, &pSym, pResTab->SymTab);
  if (Result != SYM_FOUND)
    return RC_ERR_UNDEFINED;	/* the requested range has not been defined */

  pRange = (TRangeRes *) pSym->Data;

  if(Preset)
    pNewRes->ResCode = Preset;	/* use the preset value if available */
  else
    ResCode_SetResNo(pNewRes->ResCode, (pRange->HighNo)++);

  if (ResCode_GetResType(pNewRes->ResCode) != pRange->MemberType)
    return RC_ERR_UNDEFINED;	/* can't be mix'n types */

  if(ResCode_GetResNo(pNewRes->ResCode) >= pRange->Max)
    return RC_ERR_UNDEFINED;	/* out of range */

  if(gFixedResFlag)
    ResCode_SetUser(pNewRes->ResCode);
    

  return NFR_ERR_SUCCESS;
}

TResId *ResTab_Get(char *ResName, TResTab *pResTab)
{
  TSymbol *pSym = NULL;
  
  if(SymTab_Peek(ResName, &pSym, pResTab->SymTab) != SYM_FOUND)
    return NULL;
  return (TResId *) pSym->Data;
}

TResId *ResTab_Remove(char *ResName, TResTab *pResTab)
{
  return (TResId *) SymTab_Remove(ResName, pResTab->SymTab);
}

ULONG ResTab_Insert(TResId *pResId, TResTab *pResTab)
{
  ULONG Result = 0;
  TSymbol *pSym = NULL;

  Result = SymTab_Look(pResId->Id, &pSym, pResTab->SymTab);
  if (Result != SYM_INSERTED)
    return RC_ERR_UNDEFINED;	/* don't overwrite */
  
  pSym->Data = (ULONG) pResId;
  pSym->Type = ResCode_GetResType(pResId->ResCode);

  if(ResCode_GetResType(pResId->ResCode) == RES_RANGE)
    NIList_AddToTail ((void *)pResId, pResTab->RangeList);

  return NFR_ERR_SUCCESS;
}

ULONG ResTab_Overwrite(TResId *pResId, TResTab *pResTab)
{
  ULONG Result = 0;
  TSymbol *pSym = NULL;

  Result = SymTab_Look(pResId->Id, &pSym, pResTab->SymTab);
  if (pSym == NULL)
    return RC_ERR_UNDEFINED;
  
  pSym->Data = (ULONG) pResId;
  pSym->Type = ResCode_GetResType(pResId->ResCode);

  if(ResCode_GetResType(pResId->ResCode) == RES_RANGE)
    NIList_AddToTail ((void *)pResId, pResTab->RangeList);

  return NFR_ERR_SUCCESS;
}

ULONG ResTab_MergeDiffs(char *SrcName, TResTab *pDiffTab)
{
  ULONG Result = 0;
  char TpltIn[] = "tmp/rcmrginXXXXXX";
  char *TempNameIn = NULL;
  FILE *In = NULL;
  char TpltOut[] = "tmp/rcmrgoutXXXXXX";
  char *TempNameOut = NULL;
  FILE *Out = NULL;

  TempNameIn = mktemp(TpltIn);
  TempNameOut = mktemp(TpltOut);
  rename(SrcName, TempNameIn);

  Result = Rc_Parse(RC_MODE_MERGE_DIFFS, In, Out, pDiffTab);
  
  if(Result != NFR_ERR_SUCCESS)
    {
      rename(TempNameIn, SrcName);
      unlink(TempNameOut);
      return Result;
    }

  rename(TempNameOut, SrcName);
  unlink(TempNameIn);
  return NFR_ERR_SUCCESS;
}

ULONG ResTab_Dump(TResTab *pResTab, FILE *Out)
{
  /* dump the ranges */
  SymTab_ForEach(RES_RANGE, Res_Dump, 0, (void *)Out, pResTab->SymTab);

  /*dump the facilities so they'll be there when the alerts come in */
  SymTab_ForEach(ID_FAC, Res_Dump, 0, (void *)Out, pResTab->SymTab);

  /* dump the rest of the resources */
  SymTab_ForEach(RES_ALERT, Res_Dump, 0, (void *)Out, pResTab->SymTab);
  SymTab_ForEach(RES_ALERT_SRC, Res_Dump, 0, (void *)Out, pResTab->SymTab);
  SymTab_ForEach(RES_STRING, Res_Dump, 0, (void *)Out, pResTab->SymTab);
  return NFR_ERR_SUCCESS;
}

ULONG ResTab_Load(TResTab *pResTab, FILE *In)
{
  ULONG Result = 0;
  ULONG ResCode = 0;
  TResId ResId;
  TResId *pNewResId;
  ULONG ByteCount = 0;
  BOOL NeedFacStrings = FALSE;
  ULONG NBOStrLen = 0;
  int StrLen = 0;
  char Str[256];
  PResId pFacId = NULL;
  int i = 0;

  fprintf(stderr, "in ResTab_Load()\n");
  while (TRUE)
    {
      Result = ResId_Read(&ResId, &ByteCount, In);
      if (Result != NFR_ERR_SUCCESS)
	{
	  if(feof(In))
	    {
	      fprintf(stderr, "got eof after %ld bytes\n", ByteCount);
	      /* bug: maybe need to return err on empty data */
	      return NFR_ERR_SUCCESS;
	    }
	  fprintf(stderr, "bogus ResId\n");
	  return Result;
	}

      NeedFacStrings = FALSE;

      switch(ResCode_GetResType(ResId.ResCode))
	{
	case ID_FAC:
	  fprintf(stderr, "got a facid\n");
	  pNewResId = (TResId *) ResId_New();
	  break;

	case RES_RANGE:
	  fprintf(stderr, "got a rangeres\n");
	  pNewResId = (TResId *) RangeRes_New();
	  break;

	case RES_ALERT:
	  fprintf(stderr, "got an alertres\n");
	  /* NeedFacStrings = TRUE; */
	  pNewResId = (TResId *) AlertRes_New();
	  break;

	case RES_ALERT_SRC:
	  fprintf(stderr, "got an alertsrc\n");
	  pNewResId = (TResId *) AlertSrc_New();
	  break;

	case RES_STRING:
	  fprintf(stderr, "got a stringres\n");
	  pNewResId = (TResId *) StringRes_New();
	  break;

	default:
	  fprintf(stderr, "got something bogus\n");
	  return RC_ERR_UNDEFINED;
	}

      if(pNewResId == NULL)
	return RC_ERR_UNDEFINED;

      memcpy(pNewResId, &ResId, sizeof(TResId));

      Result = Res_ReadBody(pNewResId, &ByteCount, In);
      if (Result != LIBNFR_ERR_SUCCESS)
	return Result;

      /* we need to get the strings for the facilities from the stream
         for TAlertRes structs */
      if(NeedFacStrings)
	{
	  assert(FALSE);
	  /* bug: remove this code block */

	  fprintf(stderr, "trying to read some fac strings\n");
	  for(i=0; i < ((PAlertRes)pNewResId)->FacCodeCount; i++)
	    {  
	      Result = fread(&NBOStrLen, 1, sizeof(NBOStrLen), In);
	      if (Result != sizeof(NBOStrLen))
		return RC_ERR_UNDEFINED;
	      ByteCount += Result;
	      StrLen = (USHORT) ntohl(NBOStrLen);
	      
	      if(StrLen >= sizeof(Str))
		return RC_ERR_UNDEFINED;
	      
	      Result = fread(Str, 1, StrLen, In);
	      if (Result != StrLen)
		return RC_ERR_UNDEFINED;
	      ByteCount += Result;
	      Str[StrLen] = 0;
	          
	      pFacId = ResTab_Get(Str, pResTab);
	      /* bug: need to handle when it's not there */
	      fprintf(stderr, "%s - %d\n", Str, 
		      ResCode_GetResNo(pFacId->ResCode));
	      ((PAlertRes)pNewResId)->FacCodes[i] =  
		ResCode_GetResNo(pFacId->ResCode);
	    }
	  fprintf(stderr, "done trying to read some fac strings\n");
	}
      
      fprintf(stderr, "resid: Id=%s, ResNo=%d, DispName=%s, State=%d\n",
	      pNewResId->Id, ResCode_GetResNo(pNewResId->ResCode),
	      pNewResId->DispName, pNewResId->State);

      Result = ResTab_Overwrite(pNewResId, pResTab);
      if (Result != NFR_ERR_SUCCESS)
	return Result;
    }
  return NFR_ERR_SUCCESS;
}

