/*****************************************************************************
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT")

  File:   aclose.c
  Author: Kent Landfield <kent@nfr.net>

 *****************************************************************************/

#include "sys-include.h"
#include "archive.h"

/*
***************************************************************
**
** _archive_closeread - close the archive opened read-only.
**
** Abstract:   Close the file but do not free the previously allocated
**             memory.  Used in other archive_* functions.
**
** Parameters:
**      archive - pointer to type Archive returned from archive_openread()
**
** Returns:       1 - success, 0 failure and  archive_errno set
**
***************************************************************
*/

int _archive_closeread(Archive *rkive)
{
    fclose(rkive->fp);
    return(1);
}

/*
***************************************************************
**
** _archive_closewrite - close the archive
**
** Abstract:   Close the file opened write but do not free the previously 
**             allocated memory.  Used in other archive_* functions.
**
** Parameters:
**      archive - pointer to type Archive returned from archive_openwrite()
**
** Returns:       1 - success, 0 failure and  archive_errno set
**
***************************************************************
*/

int _archive_closewrite(Archive *rkive)
{
    int archive_write_index(Archive *rkive);
    int archive_notify_spaceman(Archive *rkive);

    struct stat finfo;
    int rc = 1;

adebug("_archive_closewrite(): flushing file pointer \n");

    (void)archive_flush(rkive);

adebug("_archive_closewrite(): fstat archive file\n");

    /*
    ** Get the file size
    */
    if (fstat(fileno(rkive->fp), &finfo) != 0) {
        archive_errno = ARCHIVE_CANT_STAT;
        rc = 0;
    }
    else {

        /*
        ** Update the rec0 with file size and last write time 
        */
        rkive->head.archive_size = finfo.st_size;
        rkive->head.endtime = rkive->lasttime;
        rkive->head.maxreclen++;

adebug("_archive_closewrite(): size = [%ld] endtime = [%ld]\n", 
                 finfo.st_size, rkive->lasttime);

        /*
        ** rewrite rec0 
        */
        rewind(rkive->fp);
adebug("_archive_closewrite(): rewind to %ld\n", ftell(rkive->fp));
        
        if (fwrite(&rkive->head, sizeof(struct header), 1, rkive->fp) == 0) {
            archive_errno = ARCHIVE_HEADER_WRITEFAILED;
            rc = 0;
        }

adebug("_archive_closewrite(): rc = %d \n", rc);
    }

adebug("_archive_closewrite(): flushing file pointer \n");

    /*
    ** Close the file pointers 
    */
    
    fflush(rkive->fp);
    fclose(rkive->fp);

adebug("_archive_closewrite(): archive_write_index()\n");

    fclose(rkive->ifp);                /* close index file          */

adebug("_archive_closewrite(): archive_notify_spaceman()\n");

    /*
    ** Notify spaceman as to the new file
    */
    archive_notify_spaceman(rkive);

adebug("_archive_closewrite(): returning %d\n", rc);

    return(rc);
}

/*
***************************************************************
**
** archive_close - close the archive
** 
** Abstract:   Close a specific archive.
**             This is actually a wrapper to _archive_close(). This
**             function frees memory allocated during the archive_open*().
**             This is the utility that backend writers need to call.
**
** Parameters:
**      archive - pointer to type Archive returned from archive_open*()
**
** Returns:       1 - success, 0 failure and  archive_errno set
**
**                Used in other archive_* functions.
**
***************************************************************
*/

int archive_close(Archive *rkive)
{
    void free_aow_mem(Archive *rkive);
    int rc;

    /*
    ** Close the file and cleanup allocated memory
    */

    if (rkive->op == ARCHIVE_WRITE) 
        rc = _archive_closewrite(rkive);   /* Standard close file stuff */
    else {
        rc = _archive_closeread(rkive);    /* Standard close file stuff */
        fclose(rkive->ifp);                /* close index file          */
    }
    free_aow_mem(rkive);               /* free Archive * memory     */
    return(rc);
}
