/***************************************************************************** 
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT") 

  File:   adirpath.c 
  Author: Kent Landfield <kent@nfr.net>
 
 *****************************************************************************/  
#include "sys-include.h"
#include "archive.h"
#include "libnfr.h"
#include "nfr_res.h"

/*
*****************************************************************
**
** add_char - add supplied character to output string
**
** Returns:    pointer to next position to add to.
**
*****************************************************************
*/

char *add_char(char *ptr, char member_char)
{
    *ptr++ = member_char;
    return(ptr);
}

/*
*****************************************************************
**
** add_string - add supplied string to output string
**
** Returns:    pointer to next position to add to.
**
*****************************************************************
*/

char *add_string(char *ptr, char *member_str)
{
    while(*member_str)
        *ptr++ = *member_str++;
    return(ptr);
}

/*
*****************************************************************
**
** gen_archive_dirpath - fill in values substituting for magic cookies
**
**      Substitution cookies supported
**
**              %p - name of package the file was archived with
**              %b - name of backend the file was archived with
**              %d - two digit day of month (1-28/30/31)
**              %m - two digit month of year (1-12)
**              %y - four digit year (1990,..2001)
**              %D - current date archived as %y/%m/%d
**
** Examples:
**
**      ARCHIVE_PATH: ~/data/%p/%r/%y/%m/%d/
**
**             /cdrom/nfrdata/1997-08/webwatcher/urlhist/28/
**
**      ARCHIVE_PATH: /cdrom/nfrdata/%y-%m/%p/%r/%d
**
**     The actual filename is currently setup as HH:MM:SS
**   
** Parameters:
**             frmptr  - format string to fill 
**             package - package pathstring  
**             backend - recorder pathstring 
**             now     - filled tm structure
**
** Returns:   expanded path string
**
*****************************************************************
*/

char *gen_archive_dirpath(char *frmptr, char *package, char *backend, struct tm *now)
{
    static char format[MAXPATHLEN];
    register char *aptr;
    register char *cp;
    char dtstr[32];
    char c;

    for (cp = format; cp < format + sizeof(format); cp++)
         *cp = '\0';
 
    aptr = frmptr;
    cp = format;

    while ((c = *aptr++)) {
        if (c == '%') {
           switch (*aptr++) {
               case '%':  /* Add the % character */
                   cp = add_char(cp,'%');
                   continue;
               case '~':  /* Convert to NFR Home Directory */
                   cp = add_string(cp, nfrHome("~"));
                   continue;
               case 'p':     /* package name */
                   cp = add_string(cp, package);
                   continue;
               case 'b':     /* backend name  */
                   if (*backend)
                       cp = add_string(cp, backend);
                   continue;
               case 'd':     /* two digit day of month (1-31) */
                   sprintf(dtstr,"%.2d",now->tm_mday);
                   cp = add_string(cp, dtstr);
                   continue;
               case 'm':     /* two digit month of year (1-12) */
                   sprintf(dtstr,"%.2d",now->tm_mon+1);
                   cp = add_string(cp, dtstr);
                   continue;
               case 'y':     /* 4 digit year */
                   sprintf(dtstr,"%.4d",now->tm_year+1900);
                   cp = add_string(cp, dtstr);
                   continue;
               case 'D':     /* current date in %y/%m/%d format */
                   sprintf(dtstr,"%.4d/%.2d/%.2d",now->tm_year+1900,now->tm_mon+1,now->tm_mday);
                   cp = add_string(cp, dtstr);
                   continue;
               default:
                   archive_errno = ARCHIVE_BAD_PATH_FORMAT;
                   return(NULL);
           }    /* end switch */
        }
        *cp++ = c;
     }  /* end while */
     return(format);
}
