/***************************************************************************** 
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT") 

  File:   aopenread.c 
  Author: Kent Landfield <kent@nfr.net>
 
 *****************************************************************************/  
#include "sys-include.h"
#include "archive.h"
#include "libnfr.h"
#include "nfr_res.h"

#ifdef DEBUG
void dump_rkive(Archive *rkive);
#endif
/*
**************************************************************
** archive_openread - open a previously archived file for reading
**
** Parameters:
**       backend -   The unique name of the specific backend whose config 
**                   file is to be read to see what directory it should open 
**                   files in.
**
**    content_type - suffix used to distinquish between different datasets 
**                   created from a single recorder.
**
**       starttime - used to delineate file contents started
**                   if starttime is 0 then read from beginning of 
**                   backend's archive. If starttime is a time, start
**                   reading from that time.
**
**       endtime   - used to delienate when file contents ended
**                   if endtime is 0 then read to the end of the backend's 
**                   archive. If endtime is a time, stop reading when
**                   records are encountered outside that time.
**
** Returns:          Pointer to struct Archive on successful open
**                   Null on error
**************************************************************
*/

Archive *archive_openread(char *backend, char *content_type, 
		time_t starttime, time_t endtime)
{
    Archive *archive_header(char *path);
    void free_aow_mem(Archive *rkive);

    char path[MAXPATHLEN];
    FILE *ifp;
    Archive *rkive;
    struct rec drec;
    struct indx_rec Irec, tmpindex;
    long pos;
    int fnd;
    int opened;

    /*
    ** Build path to index file
    */
    sprintf(path,"%s/%s/index",nfrHome(INDEXDIR),backend);

    /*
    ** - open the index file
    */

    if ((ifp = fopen(path,"rb")) == NULL) {
         archive_errno = ARCHIVE_CANT_OPEN_INDEX;
         return(NULL);
    }

    /*
    ** - locate the first file that falls in the start range
    **   * if start time == 0 then start with the first file listed. 
    **   * else find the first file that falls inside the requested starttime
    */
    /*
    ** - Wrong.  For all cases, find the latest file that is not greater 
    **   than the start time.
    */

    {
    struct indx_rec before, equal, after;
    off_t index_offset;

    opened = 0;
    while (!opened) {
        fnd = 0;
        /* Need to find the right file to open */
        fnd = 0;
	index_offset = ftell(ifp);
adebug("archive_openread(): scan index\n");

        while (fscanf(ifp,"%s %s %s %ld", tmpindex.backend, tmpindex.ctype, tmpindex.path, &tmpindex.starttime) != EOF) {

/* 
* bug: do we have to check tmpindex.backend to be correct? 
*/
            /* Is this an appropriate content_type file ? */
adebug("archive_openread(): look at %s %s %s %d\n",
			tmpindex.backend, tmpindex.ctype,
			tmpindex.path, tmpindex.starttime);

            if ((content_type != NULL) && strcasecmp(tmpindex.ctype,content_type))
                continue;

            if (tmpindex.starttime < starttime) {
		/*
		* we see a file that is earlier than the time we are
		* looking for.  we have to keep looking, just in case
		* another is later, but still before the start time
		*/
adebug("archive_openread(): fnd |= 1\n");
                fnd |= 1;
		before = tmpindex;
		index_offset = ftell(ifp);
		continue;
            } else if (starttime == tmpindex.starttime) {
adebug("archive_openread(): fnd |= 2\n");
		fnd |= 2;
		equal = tmpindex;
		break;
	    } else {
		/*
		* we found one that is later than the time we are looking
		* for.  
		*/
adebug("archive_openread(): fnd |= 4\n");
		fnd |= 4;
		after = tmpindex;
		break;
	    }
        }
	if (fnd & 2)
		{
		/*
		* If we found exactly the right start time, use it.
		*/
		Irec = equal;
		}
	else if (fnd & 1)
		{
		/*
		* otherwise, if we found one BEFORE the desired start
		* time, start looking there.  If we also found one AFTER
		* the desired start time, make sure the index file is
		* positioned just before the AFTER record, so when we
		* read the end of the BEFORE file, we go on to the
		* AFTER one, which is now next to be read in the index file.
		*/
		if (fnd & 4)
			fseek(ifp,index_offset,SEEK_SET);
		Irec = before;
		}
	else if (fnd & 4)
		{
		/*
		* found no before or equal record.  start with earliest
		* after.
		*/
		Irec = after;
		}
	else
		{
		/*
		* bummer.  NO files of the data type we are looking for.
		*/
adebug("archive_openread(): fnd == 0\n");
		archive_errno = ARCHIVE_NO_STARTTIME_MATCH;
		fclose(ifp);
		return(NULL);
		}
 
       /*
       ** - open the archive file, fill the rkive structure
       ** - pass through archive_errno set in archive_header() on error
       */

adebug("archive_openread(): going on with %s\n",Irec.path);

       if ((rkive = archive_header(Irec.path)) != NULL) 
           opened = 1;

       }   /* end while not opened */

    }

    /*
    ** - initialize rkive structure variables.
    */

    if (content_type != NULL)
        rkive->op = ARCHIVE_READ_CONTENT_TYPE; /* Match content type */
    else
        rkive->op = ARCHIVE_READ;              /* Time match only */

    rkive->path = strdup(Irec.path);
    rkive->indexpath = strdup(path);
    rkive->info_read = 0;
    rkive->ifp = ifp;
    rkive->bytesread = 0;
    rkive->lasttime = endtime;

    fseek(rkive->fp, rkive->head.datastart, SEEK_SET);

#ifdef DEBUG
dump_rkive(rkive);
#endif
 
    /*
    ** - if (starttime != 0)
    **   * seek to the first record that has a data in the appropriate starttime
    */

    if (starttime != 0) {
        fnd = 0;

        pos = ftell(rkive->fp);

        while (fread(&drec, sizeof(struct rec), 1, rkive->fp)) {
adebug("Searching: starttime == %ld drec.tstamp == %ld\n", starttime, drec.tstamp);  
            if (starttime <= drec.tstamp) {
                fseek(rkive->fp, pos, SEEK_SET);
                fnd = 1;
                break;
            }
            pos = drec.next_rec;
            fseek(rkive->fp, pos, SEEK_SET);
        }
#if 0
/*
* just let it slide.  It doesn't matter whether we found an appropriate
* record or not.  The cases are:
*	1. We did; the next archive_read() will return it.
*	2. We did not.
*	2A. This file is the latest file before the start time.  We will
*		immediately get EOF on the file and rotate to the next file
*		which is after the start time requested.
*	2B. This file is equal to the start time.  It must be empty.  We
*		immediately get EOF on the file and rotate to the next file
*		which is after the start time requested.
*	2C. This file is later than the start time.  This can't happen
*		because it satisfies condition 1.
*		
*/
        if (!fnd) {
            archive_errno = ARCHIVE_NO_STARTTIME_MATCH;
            fclose(rkive->fp);
            fclose(rkive->ifp);
            free_aow_mem(rkive);
            return(NULL);
        }
#endif
    }

    return(rkive);
}

/*
**********************************************************************
** reopen_archive_read - close and existing file and open a new one.
**
** The idea here is that another potential file needs to be opened.
** At this point the routine is passed an opened Archive stream with
** the Archive * structure filled appropriately.  There is no reason to
** go through the pain of refilling the structure as is done when the
** file is opened. Simply change the appropriate information, reset
** the appropriate variables and return.
**
** Parameters:     rkive  - pointer to Archive struct.
** Returns:        -1 error, 0 end of read time, 1 reopen successful
**********************************************************************
*/

int reopen_archive_read(Archive *rkive)
{
      int _archive_closeread(Archive *rkive);

      long reclen;
      struct indx_rec Irec;
      int fnd;
      int opened;

      archive_errno = 0;

      /* 
      ** close existing archive
      ** if error, return -1
      */ 

adebug("reopen_archive_read(): closing old archive file\n");  

      if (_archive_closeread(rkive) == 0)
          return(-1);

      opened=0;

      while(!opened) {

adebug("reopen_archive_read(): looking for next valid index record\n");  
    
          /* 
          ** The index file pointer is still pointing to an open
          ** index file. Read the next record from the file, see 
          ** if it is within the timeframe requested on the open
          ** and open it if it is, else we are done and the read
          ** needs to return a 0;
          */
          if (rkive->op == ARCHIVE_READ_CONTENT_TYPE) {
              /* 
              ** Locate next file that matches the content_type specified
              */ 
              fnd = 0;
              while (fscanf(rkive->ifp,"%s %s %s %ld", Irec.backend, Irec.ctype, Irec.path, &Irec.starttime) != EOF) {
                  if (!strcasecmp(Irec.ctype,rkive->head.ctype)) {
                      fnd = 1;
                      break;
                  }
              }
              if (!fnd) {
    adebug("reopen_archive_read(): EOF Index file\n");  
                  return(0);
              }
          }
          else {
              if (fscanf(rkive->ifp,"%s %s %s %ld", Irec.backend, Irec.ctype, Irec.path, &Irec.starttime) == EOF) {
    adebug("reopen_archive_read(): EOF Index file\n");  
                  return(0);
              }
          }
    
    adebug("reopen_archive_read(): Checking if in time range\n");  
    
          /* Check to see if we have a file that is in the right time range */
    
    adebug("reopen_archive_read(): rkive->lasttime %d Irec.starttime %d\n",
    		rkive->lasttime, Irec.starttime);
    
          if (rkive->lasttime != 0 && Irec.starttime > rkive->lasttime)
              return(0); /* we be done */
    
    adebug("reopen_archive_read(): opening %s archive\n", Irec.path);  
    
          /*
          ** - open the archive file, fill the rkive structure
          ** - set archive_errno on error
          */
    
          if ((rkive->fp = fopen(Irec.path,"rb")) != NULL) 
               opened = 1;
      }

adebug("reopen_archive_read(): reading file header\n");  

      /* Read the file header structure */

      if (fread(&rkive->head, sizeof(struct header), 1, rkive->fp) == 0) {
           archive_errno = ARCHIVE_HEADER_READFAILED;
           return(-1);
      }    

adebug("reopen_archive_read(): reading recorder header\n");  

      /*
      ** Free any allocated memory
      */

      if (rkive->info)
          free(rkive->info);
      if (rkive->path)
          free(rkive->path);

      /* Read the recorder supplied blob if available */

      if ((reclen = rkive->head.datastart - sizeof(struct header)) > 0) {
          if (fseek(rkive->fp, sizeof(struct header), SEEK_SET) != 0) {
               archive_errno = ARCHIVE_CANT_SEEK;
               return(-1);
          }
          rkive->info = (char *)calloc(1, reclen+1);
 
          if (fread(rkive->info, reclen, 1, rkive->fp) == 0) {
               archive_errno = ARCHIVE_BLOB_READFAILED;
               return(-1);
          }
      }
      else   /* No blob here */
          rkive->info = NULL;
 
adebug("reopen_archive_read(): resetting variables\n");  

      /* 
      ** At this point we are ready to reset the rkive variables.
      */ 

      rkive->path = strdup(Irec.path);
      rkive->info_read = 0;
      rkive->rec1len = reclen;

#ifdef DEBUG
dump_rkive(rkive);
#endif

      return(1);
}

