/***************************************************************************** 
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT") 

  File:   aopenwrite.c 
  Author: Kent Landfield <kent@nfr.net>
 
 *****************************************************************************/  
#include "sys-include.h"
#include "archive.h"
#include "libnfr.h"
#include "nfr_res.h"

#ifdef DEBUG
void dump_rkive();
#endif

/*
**********************************************************************
** archive_wfilehdr - write the file header records
**
** Parameters:     rkive - pointer to Archive struct.
** Returns:        0 for failure, 1 if write successful
**********************************************************************
*/

int archive_wfilehdr(Archive *rkive) 
{
    /*
    ** Write rkive->head as record 0 
    */

adebug("Pos Rec 0 == %d:\n", ftell(rkive->fp));
    
    if (fwrite(&rkive->head, sizeof(struct header), 1, rkive->fp) == 0) {
         archive_errno = ARCHIVE_HEADER_WRITEFAILED;
         return(0);
    }

    /* 
    ** Write rkive->info as record 1 
    */

adebug("Recorder Specific Information:\n");
adebug("tBlob Read:   %s\n",rkive->info_read ? "Yes": "No");
adebug("tBlob: %s\n", rkive->info);
adebug("tBlob Length: %d\n",rkive->rec1len);
adebug("nPos Rec 1 == %d:\n", ftell(rkive->fp));

    if ((rkive->rec1len > 0) 
	&& fwrite(rkive->info, rkive->rec1len, 1, rkive->fp) == 0)
    {
         archive_errno = ARCHIVE_BLOB_WRITEFAILED;
         return(0);
    }

    /*
    ** Initialize for write 
    ** 
    ** - Save the last written record's file position.
    ** - Count the written header in the total bytes written count.
    */

    rkive->last_rec_pos = rkive->head.datastart;
    rkive->byteswritten = rkive->head.datastart-1;          

#ifdef DEBUG
dump_rkive(rkive);
#endif

    /*
    ** Ok... The file exists and need to be registered.
    ** Time to put it in the index.
    */

    return(archive_write_index(rkive));
}

/*
**********************************************************************
** archive_pathopen - Create path with character format substitution,
**                    open the archive file creating any needed directories
**
** Parameters:     rkive  - pointer to Archive struct.
**                 tstamp - time_t
** Returns:        0 for failure, 1 if successful
**********************************************************************
*/

int archive_pathopen(Archive *rkive, time_t tstamp) 
{
    int makedirs(char *path);
    char *gen_archive_dirpath(char *frmtptr, char *package, 
                           char *recorder, struct tm *now);

    char path[MAXPATHLEN];
    char *rptr;
    char *dirpath;
    struct tm *now;

    /*
    **  if memory previously allocated, cleanup 
    */

    if (rkive->path)          
        free(rkive->path);

    if (rkive->indexpath) 
        free(rkive->indexpath);

    /* Now convert the backend name into separate package/recorder strings */

    strcpy(path,rkive->head.backend);
    if ((rptr = strrchr(path,'/')) != NULL)
         *rptr++ = '\0';
    else
         rptr = NULL;

    /* fill the time structure for use in filling in magic cookies */

    now = localtime(&tstamp);

    /*
    ** generate the path to the archive's storage directory 
    */
    dirpath = gen_archive_dirpath(rkive->pathfmt, path, rptr, now);
    if (dirpath == NULL) {
        return(0); /* Pass through the archive_errno value */
    }

    /*
    ** Create complete archive file path 
    */

    if (*rkive->head.ctype) 
        sprintf(path,"%s/%.2d_%.2d_%.2d-%s", 
                dirpath,now->tm_hour,now->tm_min,now->tm_sec,rkive->head.ctype);
    else 
        sprintf(path,"%s/%.2d_%.2d_%.2d", 
                dirpath,now->tm_hour,now->tm_min,now->tm_sec);

adebug("DIRPATH == [%s]\n",dirpath);

    /*
    ** Assure all parent directories are created 
    */
    if (!makedirs(dirpath)) {
        archive_errno = ARCHIVE_CANT_MKDIR;
        return(0);
    }

    rkive->path = strdup(path);

adebug("rkive->path ==[%s]\n",rkive->path);

    /*
    ** Create path to index directory
    */
    sprintf(path,"%s/%s",nfrHome(INDEXDIR),rkive->head.backend);
    
adebug("Indexpath == [%s]\n",path);
    
    /*
    ** Assure all parent directories are created 
    */
    if (!makedirs(path)) {
        archive_errno = ARCHIVE_CANT_MKDIR;
        return(0);
    }

    /*
    ** Create path to index file
    */
    strcat(path,"/index");
    rkive->indexpath = strdup(path);

    /* open the index file and archive files */

    if ((rkive->ifp = fopen(rkive->indexpath,"ab+")) == NULL) {
         archive_errno = ARCHIVE_CANT_OPEN_INDEX;
         return(0);
    }

    if ((rkive->fp = fopen(rkive->path,"wb+")) == NULL) {
perror(rkive->path);
         archive_errno = ARCHIVE_CANT_OPEN_ARCHIVE;
         return(0);
    }

    return(1);
}


/*
**********************************************************************
** reopen_archive_write - close and existing file and open a new one.
**
** The idea here is that the file needs to be rolled over due to time
** or size.  At this point the routine is passed an opened Archive 
** stream with the Archive * structure filled appropriately.  There is 
** no reason to go through the pain of refilling the structure as is 
** done when the file is opened. Simply change the appropriate information,
** reset the appropriate variables and return.
**
** Parameters:     rkive  - pointer to Archive struct.
**                 tstamp - time_t
** Returns:        0 for failure, 1 successful
**********************************************************************
*/
 
int reopen_archive_write(Archive *rkive, time_t tstamp)
{
      int _archive_closewrite();

adebug("reopen_archive_write()\n");
      /*
      ** close existing archive 
      ** if error, return null 
      */

      if (_archive_closewrite(rkive) == 0)
          return(0);

      /*
      ** At this point we are ready to reset the rkive variables.
      */
      rkive->rolltime = tstamp + rkive->rtimeval;
      rkive->head.starttime = tstamp;
      rkive->head.endtime = 0;
      rkive->head.archive_size = 0;
      rkive->bytesread = 0;            
      rkive->info_read = 0;

      if (!archive_pathopen(rkive, tstamp)) 
          return(0);

      return(archive_wfilehdr(rkive));
}

/*
**********************************************************************
** archive_openwrite - Open an archive file for writing.
**
** Parameters:
**       backend     - The unique name of the specific backend whose config 
**                     file is to be read to see what directory it should 
**                     open files in.
**     content_type  - Backend supplied option suffix for filename
**     starttime     - used to build filename
**    recorder_blob  - filled in at archive_openwrite() time
**
** Returns:            Pointer to type Archive 
**
** Notes:              Records do not have uniform sizes.
**********************************************************************
*/

Archive *archive_openwrite(char *backend, char *content_type, 
                        time_t starttime, char *recorder_blob)
{
    PCf pCf;
    char path[MAXPATHLEN];
    char *cp;
    char *rlval;
    long val;
    Archive *rkive;

    void free_aow_mem();

    /* Need to allocate memory for the Archive structure */

    if ((rkive = (Archive *)calloc(1, sizeof(Archive))) == NULL) {
         archive_errno = ARCHIVE_NO_MEMORY;
         return(NULL);
    }

adebug("OPENING\n");
         
    /* Init/fill header 0 record with passed data */

    strcpy(rkive->head.backend, backend);

    if (content_type != NULL)
       strcpy(rkive->head.ctype, content_type);
    else 
       rkive->head.ctype[0] = 0;

    if (starttime == 0)
        starttime = time((time_t *)0);

    rkive->head.starttime = starttime;
    rkive->head.endtime = 0;
    rkive->head.archive_size = 0;

    /* Init/fill recorder specific record 1 with passed data */

    rkive->info = strdup(recorder_blob);
    rkive->rec1len = strlen(recorder_blob);
    rkive->info_read = 0;

    /* Init Archive structure */
 
    rkive->bytesread = 0;
    rkive->external_archive_cmd = NULL;

    /*
    ** Set the operation type of stream
    */
    
    rkive->op = ARCHIVE_WRITE;

    /* Locate the backend configuration file and open it */

adebug("OPENING:packages\n");

    sprintf(path,"packages/%s.cfg", backend);
    pCf = Cf_New(0);
    if (Cf_Read(path, pCf) != NFR_ERR_SUCCESS) {
        archive_errno = ARCHIVE_CONFIG_READ_FAILED;
        free_aow_mem(rkive);
        return(NULL);
    }

    /* Get config file version information */

    if ((rlval = Cf_GetVal("cfversion", pCf)) == NULL) {
        archive_errno = ARCHIVE_MISSING_CF_VERSION;
        free_aow_mem(rkive);
        return(NULL);
    }

    strcpy(rkive->head.cfversion,rlval);

    /*
    ** Get roll over information 
    ** First determine if roll over is to be used 
    ** and initialize accordingly 
    */

    if ((rlval = Cf_GetVal("rollover_size", pCf)) == NULL) {
        archive_errno = ARCHIVE_MISSING_ROLLOVER_SIZE;
        free_aow_mem(rkive);
        return(NULL);
    }

    if ((!strcasecmp(rlval,"DONT")) && (!strcasecmp(rlval,"NO")))
        rkive->rollsize = -1;
    else if (strcasecmp(rlval,"DO") && strcasecmp(rlval,"YES")) {
        archive_errno = ARCHIVE_INVALID_ROLLOVER_SIZE;
        free_aow_mem(rkive);
        return(NULL);
    }
    else {  /* Do rollover based on size - set it up */
        if ((rlval = Cf_GetVal("rollover_size_val", pCf)) == NULL) {
            archive_errno = ARCHIVE_MISSING_ROLLOVER_SIZE_VAL;
            free_aow_mem(rkive);
            return(NULL);
        }
        cp = rlval;
        while(*cp && isdigit(*cp))
             ++cp;

        /*
        ** If no special expansion character specified consider
        ** the value as absolute bytes.
        */

        if (!*cp) 
            rkive->rollsize = atol(rlval);
          
        else {    /* determine the expansion character's value */
            switch (tolower(*cp)) {
                    case 'b': val = 1;       break;   
                    case 'k': val = ONE_K;   break; 
                    case 'm': val = ONE_MEG; break;
                    case 'g': val = ONE_GIG; break; 
                default: archive_errno = ARCHIVE_INVALID_ROLLOVER_SIZE_VAL;
                         free_aow_mem(rkive);
                         return(NULL);
            }
            rkive->rollsize = atol(rlval) * val;;
        }
    }
    
    if ((rlval = Cf_GetVal("rollover_time", pCf)) == NULL) {
        archive_errno = ARCHIVE_MISSING_ROLLOVER_TIME;
        free_aow_mem(rkive);
        return(NULL);
    }

    if ((!strcasecmp(rlval,"DONT")) && (!strcasecmp(rlval,"NO")))
        rkive->rolltime = -1;
    else if (strcasecmp(rlval,"DO") && strcasecmp(rlval,"YES")) {
         archive_errno = ARCHIVE_INVALID_ROLLOVER_TIME;
         free_aow_mem(rkive);
         return(NULL);
    }
    else { /* Do rollover by time - set it up */
        if ((rlval = Cf_GetVal("rollover_time_val", pCf)) == NULL) {
            archive_errno = ARCHIVE_MISSING_ROLLOVER_TIME_VAL;
            free_aow_mem(rkive);
            return(NULL);
        }

        cp = rlval;
        while(*cp && isdigit(*cp))
             ++cp;

        /*
        ** If no special expansion character specified consider
        ** the value as absolute seconds.
        */

        if (!*cp) 
             rkive->rtimeval = atol(rlval);
        else {
            switch (tolower(*cp)) {
                    case 's': val = 1;     break;   
                    case 'm': val = 60;    break; /* seconds per minute */
                    case 'h': val = 3600;  break; /* seconds per hour   */
                    case 'd': val = 86400; break; /* seconds per day   */
                default: archive_errno = ARCHIVE_INVALID_ROLLOVER_TIME_VAL;
                         free_aow_mem(rkive);
                         return(NULL);
            }
            rkive->rtimeval = (atol(rlval) * val);
        }
        rkive->rolltime = starttime + rkive->rtimeval;
    }

    /* Get path to archive format */

    if ((rlval = Cf_GetVal("archive_path", pCf)) == NULL) {
        archive_errno = ARCHIVE_MISSING_ARCHIVE_PATH;
        free_aow_mem(rkive);
        return(NULL);
    }

    rkive->pathfmt = strdup(nfrHome(rlval));

    Cf_Delete(pCf);  /* close the configuration file */

    if (!archive_pathopen(rkive, starttime)) {
        free_aow_mem(rkive);
        return(NULL);
    }

    /*
    ** Calculate the start of record 1, size of record 1 
    ** and start of data records 
    */
    
    rkive->head.datastart = sizeof(struct header) + rkive->rec1len;

    if (!archive_wfilehdr(rkive)) {
        free_aow_mem(rkive);
        return(NULL);
    }

    return(rkive); /* return a pointer to the the Archive structure */
}
