/***************************************************************************** 
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT") 

  File:   aread.c 
  Author: Kent Landfield <kent@nfr.net>
 
 *****************************************************************************/  
#include "sys-include.h"
#include "archive.h"

/*
*****************************************************************************
** archive_read - read next record in the time period 
**
** Parameters:
**      archive  - pointer to type Archive returned from archive_openread()
**
**      nbytes   - maximum number of bytes to read
**
**      buf      - pointer to buffer to read the data into
**
**      endtime  - time to stop reading records. Can be zero read all records.
**
**     unique_id - tcp.hash or some indexable identifier that is unique to 
**                 the content type. Only records matching that are returned.
**
** Returns:       A Pointer to a rec structure containing the the data bytes 
**                read, the record's timestamp and unique-id is returned. 
**                The rec structure is stored in as a static structure that
**                will be overwritten on subsequent calls. If there is a
**                need within the application to keep the information returned
**                in the rec structure, the application must do so itself.
**                Returns NULL when an error is encountered and archive_errno
**                is set.
**
** Notes:         Read the next record from the archive and if it hits EOF 
**                it'll open the next file as long as endtime has not been 
**                encountered.  Deal with the archive on a record basis
**                and will return at most nbytes to the calling application.
*****************************************************************************
*/

int archive_shortfall;

struct rec *archive_read(Archive *rkive, size_t nbytes, char *buf, 
                          time_t endtime, char *unique_id)
{
    int reopen_archive_read(Archive *);

    static struct rec drec;
    long pos;

    pos = ftell(rkive->fp);

    while (fread(&drec, sizeof(struct rec), 1, rkive->fp) == 0) {
        if (feof(rkive->fp)) {
adebug("Going to try for a new file\n");  
            switch(reopen_archive_read(rkive)) {
                case  0: /* Read termination time or end of archive reached */
                         archive_errno = 0;  /* Reset to no error condition */
                         return(NULL);
                case  1: /* Reopen completed, continue reading. */
                         pos = ftell(rkive->fp);
                         break;
                default: /* Error occurred  archive_errno set in reopen */
                         return(NULL);
            }
        }
        else {   /* stream error */
adebug("stream error\n");  
           /* problems - reset to start of record */
           fseek(rkive->fp, pos, SEEK_SET);
           archive_errno = ARCHIVE_RECHDR_READ_ERR;
           return(NULL);
        }
    }

    /* See if we are exceeding endtime */

adebug("compare times\n");

    if (((endtime > 0) && (drec.tstamp > endtime)) ||
        ((rkive->lasttime > 0) && (drec.tstamp > rkive->lasttime))) {
        /* problems - reset to start of record */
        fseek(rkive->fp, pos, SEEK_SET);
        archive_errno = 0;  /* Reset to no error condition */
        return(NULL);
    }

adebug("nbytes [%ld] - drec.data_len [%ld]\n", nbytes, drec.data_len);

    if (nbytes < drec.data_len) {
        /* problems - reset to start of record */
        fseek(rkive->fp, pos, SEEK_SET);
        archive_errno = ARCHIVE_RECDATA_BUF_SHORT;
	archive_shortfall = drec.data_len - nbytes;
        return(NULL);
    }

    if ((drec.id_len != 0) && 
		(fread(unique_id, drec.id_len, 1, rkive->fp) == 0)) {
        /* problems - reset to start of record */
        fseek(rkive->fp, pos, SEEK_SET);
        archive_errno = ARCHIVE_RECID_READ_ERR;
        return(NULL);
    }
    *(unique_id+drec.id_len) = '\0';

adebug("Unique_id [%s]\n", unique_id);

    if ((drec.data_len != 0) &&
	(fread(buf, drec.data_len, 1, rkive->fp) == 0)) {
        /* problems - reset to start of record */
        fseek(rkive->fp, pos, SEEK_SET);
        archive_errno = ARCHIVE_RECDATA_READ_ERR;
        return(NULL);
    }

    /* bug: why are we doing this? if this is necessary then here and in the 
       line above which tests nbytes < drec.data_len, drec.data_len must be
       changed to drec.data_len + 1 */
    /* *(buf+drec.data_len) = '\0'; */

    return(&drec);
}
