/***************************************************************************** 
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT") 

  File:   arewind.c 
  Author: Kent Landfield <kent@nfr.net>
 
 *****************************************************************************/  
#include "sys-include.h"
#include "archive.h"

/* 
***********************************************************************
** archive_rewind - rewind archive pointer to time specified
**
** Parameters: 
**          rkive   - Archive struct returned by archive_openread()
**        starttime - time to rewind to.
**
** Returns:       1 - rewound, 0 - error
**
** Notes:         Reset archive access to the first record in the archive if
**                start_time is 0, else it rewinds back to the first record
**                matching starttime, and if supplied, content_type.
***********************************************************************
*/

int archive_rewind(Archive *rkive, time_t starttime)
{
    int _archive_closeread(Archive *);

    int fnd;
    long reclen;
    long pos;
    struct rec drec;
    struct indx_rec Irec;

    /* Check and make sure this is being used in a read operation. */ 

    if (rkive->op == ARCHIVE_WRITE)
        return(0);

    /* Rewind the index file pointer, */

    rewind(rkive->ifp);  /* This really should be a close and reopen... */

    /* Check if starttime supplied */

    if (starttime != 0) {
        /* Need to find the right file to open */
        fnd = 0;
        while (fscanf(rkive->ifp,"%s %s %s %ld", Irec.backend, Irec.ctype, Irec.path, &Irec.starttime) != EOF) {

            /* Is this an appropriate content_type file ? */

            if ((*rkive->head.ctype) && 
                  strcasecmp(Irec.ctype,rkive->head.ctype))
                continue;

            if (starttime <= Irec.starttime) {
                fnd = 1;
                break;
            }
        }
        if (!fnd) {
            archive_errno = ARCHIVE_NO_STARTTIME_MATCH;
            return(0);
        }
    }
    else if (*rkive->head.ctype) {
        /* 
        ** Starttime == 0
        ** OK, start at begining of archive but only with those 
        ** files that match the appropriate content_type specified
        */ 
        fnd = 0;
        while (fscanf(rkive->ifp,"%s %s %s %ld", Irec.backend, Irec.ctype, Irec.path, &Irec.starttime) != EOF) {
            if (!strcasecmp(Irec.ctype,rkive->head.ctype)) {
                fnd = 1;
                break;
            }
        }
        if (!fnd) {
            archive_errno = ARCHIVE_NO_CTYPE_MATCH;
            return(0);
        }
    }
    else {
        /* Return the first index record, we're not picky. */
        if (fscanf(rkive->ifp,"%s %s %s %ld", Irec.backend, Irec.ctype, Irec.path, &Irec.starttime) == EOF) {
            archive_errno = ARCHIVE_NO_STARTTIME_MATCH;
            return(0);
        }
    }

    /*
    ** Do we currently have the file open ?
    */

    if (!strcmp(Irec.path, rkive->path)) {
       /* Yep */
       rewind(rkive->fp);
    }
    else {
        /* 
        ** close existing archive
        ** if error, return -1
        */ 
   
        if (_archive_closeread(rkive) == 0)
            return(-1);
   
        /*
        ** - open the archive file, fill the rkive structure
        ** - set archive_errno on error
        */
  
        if ((rkive->fp = fopen(Irec.path,"rb")) == NULL) {
           archive_errno = ARCHIVE_CANT_OPEN_ARCHIVE;
           return(-1);
        }    
  
adebug("archive_rewind(): reading file header\n");  
  
        /* Read the file header structure */
  
        if (fread(&rkive->head, sizeof(struct header), 1, rkive->fp) == 0) {
             archive_errno = ARCHIVE_HEADER_READFAILED;
             return(-1);
        }    
  
adebug("archive_rewind(): reading recorder header\n");  
  
        /*
        ** Free any allocated memory
        */
        if (rkive->info)
            free(rkive->info);
        if (rkive->path)
            free(rkive->path);
  
        /* Read the recorder supplied blob if available */
  
        if ((reclen = rkive->head.datastart - sizeof(struct header)) > 0) {
            if (fseek(rkive->fp, sizeof(struct header), SEEK_SET) != 0) {
                 archive_errno = ARCHIVE_CANT_SEEK;
                 return(-1);
            }
            rkive->info = (char *)calloc(1, reclen+1);
   
            if (fread(rkive->info, reclen, 1, rkive->fp) == 0) {
                 archive_errno = ARCHIVE_BLOB_READFAILED;
                 return(-1);
            }
        }
        else   /* No blob here */
            rkive->info = NULL;
   
adebug("archive_rewind(): resetting variables\n");
        /* 
        ** At this point we are ready to reset the rkive variables.
        */ 
  
        rkive->path = strdup(Irec.path);
        rkive->info_read = 0;
        rkive->rec1len = reclen;
    }

    /*
    ** - if (starttime != 0)
    **   seek to the first record that has a data in starttime    
    */
 
    if (starttime != 0) {
        fnd = 0;
 
        pos = ftell(rkive->fp);

        while (fread(&drec, sizeof(struct rec), 1, rkive->fp)) {
adebug("archive_rewind: starttime == %ld drec.tstamp == %ld\n", starttime, drec.tstamp);
            if (starttime <= drec.tstamp) {
                fseek(rkive->fp, pos, SEEK_SET);
                fnd = 1;
                break;
            }
            pos = drec.next_rec;
            fseek(rkive->fp, pos, SEEK_SET);
        }
        if (!fnd) {
            archive_errno = ARCHIVE_NO_STARTTIME_MATCH;
            return(0);
        }
    }
    return(1);
}
