/***************************************************************************** 
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT") 

  File:   awrite.c 
  Author: Kent Landfield <kent@nfr.net>
 
 *****************************************************************************/

#include "sys-include.h"
#include "archive.h"

/*
*************************************************************************
** archive_write
**
** Parameters:
**      archive - pointer to type Archive returned from archive_openwrite()
**     nbytes   - number of data bytes to write
**      buf     - pointer to buffer to read the data into
**     tstamp   - timestamp
**    unique-id - Unique identifier, recorder specific.
**
** Returns:       Number of bytes written, -1 on error
**
** Notes:         Writes to the archive and if it's past the rolltime
**                configured in the backend's configuration, close it, 
**                open a new archive file and write the data. (roll over)
*************************************************************************
*/

int archive_write(Archive *rkive, size_t nbytes, char *buf, 
                   time_t tstamp, char *unique_id)
{
    int reopen_archive_write(Archive *, time_t);

    int idlen;
    long bytesrit;
    long pos;
    struct rec drec;
    int rolled = 0;

adebug(" tstamp == %ld - rkive->rolltime == %ld\n", 
          tstamp, rkive->rolltime);

    /* check if rollover by size needed */

    if (rkive->rollsize != -1) {
        if ((rkive->head.archive_size+rkive->byteswritten) > rkive->rollsize) {
            if (!reopen_archive_write(rkive, tstamp))
                return(-1);

            rolled = 1; 
        }
    }

    /* check if rollover by time needed */

    if (!rolled) {
        if ((rkive->rolltime != -1) && (tstamp >= rkive->rolltime)) {
             if (!reopen_archive_write(rkive, tstamp))
                 return(-1);
        }
    }

    /*
    ** Fill the record header structure
    */

    if (unique_id)
	idlen = strlen(unique_id);
    else
	idlen = 0;

    bytesrit = sizeof(struct rec) + idlen + nbytes;
    pos = ftell(rkive->fp);

    drec.tstamp = tstamp;
    drec.id_len = idlen;
    drec.data_len = nbytes;
    drec.last_rec = rkive->last_rec_pos;
    drec.next_rec = pos + bytesrit;

    /*
    ** save information about the largest record, 
    ** used for read buffer allocation.
    */

    if (nbytes > rkive->head.maxreclen || rkive->head.maxreclen == 0)
        rkive->head.maxreclen = nbytes;

    /*
    ** Write the record header structure
    */

    if (fwrite(&drec, sizeof(struct rec), 1, rkive->fp) == 0) {
         /* problems... */
         archive_errno = ARCHIVE_RECHDR_WRITE_ERR;
         return(-1);
    }

    /*
    ** Write the unique id 
    */

    if ((idlen != 0) && (fwrite(unique_id, idlen, 1, rkive->fp) == 0)) {
         /* problems... */
         archive_errno = ARCHIVE_RECID_WRITE_ERR;
         return(-1);
    }

    /*
    ** Write the data
    */

    if ((nbytes != 0) && (fwrite(buf, nbytes, 1, rkive->fp) == 0)) {
         /* problems... */
         archive_errno = ARCHIVE_RECDATA_WRITE_ERR;
         return(-1);
    }

    /*
    ** Fill in structure info 
    */
    rkive->byteswritten += bytesrit;
    rkive->lasttime = tstamp;
    rkive->last_rec_pos = pos;

    return(bytesrit);
}
