/*****************************************************************************

  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT") 

  File:   lockfile.c
  Author: Kent Landfield <kent@nfr.net>
 

*****************************************************************************/

/****
  c'mon kent, you knew better than this!
#include <sys/types.h>
#include <limits.h>
#include <unistd.h>
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <fcntl.h>
#include <dirent.h>
#include <sys/wait.h>
#include <sys/utsname.h>
#include <sys/stat.h>
#include <signal.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include <sysexits.h>
****/

#include "sys-include.h"
#include "alert.h"
#include "nfr_res.h"

#define UNIQ_PREFIX	'_'
#define charsSERIAL	4
#define UNIQnamelen	(1+charsSERIAL+HOSTNAMElen+1)
#define bitsSERIAL	(6*charsSERIAL)
#define maskSERIAL	((1L<<bitsSERIAL)-1)
#define rotbSERIAL	2
#define irotbSERIAL	(1L<<(bitsSERIAL-rotbSERIAL))
#define mrotbSERIAL	((maskSERIAL&(irotbSERIAL-1))+irotbSERIAL)

#define SLEEPINTERVAL 2
#define nfsTRY      (7+1)    /* # of times+1 to ignore spurious NFS errors */
#define HOSTNAMElen	9

extern int errno;

#ifndef O_SYNC
#define O_SYNC          0
#endif

#ifndef P
#define P(x)    x
#define Q(x)    ()
#endif

int exitflag;
int gotsig;

char *strncpy();
void report ();
extern int debug;

char *fullpath = NULL;

void fakehandler P((void))
{
    gotsig=1;
}

void qsignal(int sig, void(*action)P((void)))
{
  gotsig=0;
  if (SIG_IGN==signal(sig,(void(*)())fakehandler))
      signal(sig,SIG_IGN);
  else { 
      signal(sig,(void(*)())action);
      if (gotsig)
	 (*action)();
  }
}

char *hostname(void)
{
  static char name[HOSTNAMElen+1];
  /* int gethostname(); */ /* what the heck is this doing here? */

  gethostname(name,HOSTNAMElen+1);
  name[HOSTNAMElen]='\0';
  return(name);
}

/* convert to a number within the set [0-9A-Za-z-_] */

void ultoan(unsigned long val, char *dest)
{
  register i;		 
  do
   { i=val&0x3f;
     *dest++=i+(i<10?'0':i<10+26?'A'-10:i<10+26+26?'a'-10-26:
      i==10+26+26?'-'-10-26-26:'_'-10-26-27);
   }
  while(val>>=6);
  *dest='\0';
}

int myrename(char *old, char *newn) /* rename MUST fail if already existent */
{
  int i;
  int save_errno;
  struct stat stbuf;

  if(!link(old,newn)) { 
     unlink(old);
     return (0);
  }
  save_errno = errno;
  i=stat(old,&stbuf);
  unlink(old);
  errno = save_errno;
  return (stbuf.st_nlink==2?i:-1);
}

/* create unique file name */

int unique(char *full, char *p, mode_t mode)
{ 
   char *strcat();

   unsigned long retry=mrotbSERIAL;
   int i;	  

   do { 
     ultoan((maskSERIAL&(retry-=irotbSERIAL))+(long)getpid(),p+1);
     *p=UNIQ_PREFIX;
     strcat(p,hostname());
   } while(((i = open(full,O_WRONLY|O_CREAT|O_EXCL|O_SYNC,mode)) < 0) && 
         errno == EEXIST && retry);

   if (i<0) { 
      if (debug)
          report("Error while writing to \"%s\"\n",full);
      else
          Alert_Args_Generate(NFR_SRC_SPACEMAND, NFR_SPACEMAND_LOCKWRITE, full, 0);
      return (0);
   }
   close(i);
   return (1);
}


int creatlock(char *name, mode_t mode)
{
   char *strpbrk();
   char *p,*q;
   int j= -2;
   int i;

   for (q=name;(p=strpbrk(q,"/"));q=p+1);

   if (!(fullpath=malloc((i=q-name)+UNIQnamelen))) {
      exitflag = 1;            /* merely sets a flag */
      return (j);
   }

   strncpy(fullpath,name,i);
   if (unique(fullpath,fullpath+i,mode))
       /* try and rename it, fails if nonexclusive */
       j=myrename(fullpath,name);	 

   free(fullpath);
   return (j);
}

void failure P ((void))            /* signal trap */
{
  exitflag = 2;            /* merely sets a flag */
}

char *genlockname(char *name)
{
    char *lockname;

    if ((lockname = malloc(strlen(name)+6)) == NULL) 
        return(NULL);
    
    sprintf(lockname,"%s.lock", name);
    return(lockname);
}

int cleanup(char *lockname, int retval)
{
      if (fullpath)
          (void) unlink(fullpath);

      free(lockname);
      return (retval);  
}

int lockfile(char *filename)
{
    int retval = EXIT_SUCCESS;
    int permanent;
    int sleepsec;
    int retries;
    char *lockname;

    if ((lockname = genlockname(filename)) == NULL) {
      if (debug)
          report("Whoa!... Out of Memory allocating for %s\n",filename);
      else
          Alert_Args_Generate(NFR_SRC_SPACEMAND, NFR_SPACEMAND_LOCKMEM, filename, 0);
        return(EX_OSERR);
    }
  
    sleepsec = SLEEPINTERVAL;
    retries = -1;

    signal (SIGPIPE, SIG_IGN);

    qsignal (SIGHUP, failure);
    qsignal (SIGINT, failure);
    qsignal (SIGQUIT, failure);
    qsignal (SIGTERM, failure);

    permanent = nfsTRY;

    while (creatlock(lockname, (mode_t)(S_IRUSR|S_IRGRP|S_IROTH)) < 0) {  
        if (exitflag) {    /* time to stop? */
           if (exitflag == 1) {    /* was it failure() or malloc() */
               if (debug)
                   report("Whoa!... Out of Memory allocating for %s\n",filename);
               else
                   Alert_Generate(NFR_SRC_SPACEMAND, NFR_SPACEMAND_LCKMALLOC);
               retval = EX_OSERR;
           }
           else {
               retval = EX_TEMPFAIL;
               if (debug)
                   report("Signal %s\n", "received");
           }
           return(cleanup(lockname,retval));
        }
        switch (errno) {    /* why did the lock fail? */
          case EEXIST:    /* hmmm..., by force then? */
               switch (retries)  {   /* await your turn like everyone else */
                    case 0: 
                            if (debug)
                                report("Retry limit %s\n", "exceeded");
                            else
                                Alert_Generate(NFR_SRC_SPACEMAND, 
                                               NFR_SPACEMAND_LCKRETRY);
                            retval = EX_CANTCREAT;
                            return(cleanup(lockname,retval));

                    case -1: sleep (sleepsec);  /* wait and see */

                    default: retries--;         /* count sheep */
               }
               break;
          case ENOSPC:
          case ENOENT:
          case ENOTDIR:
          case EIO:  
                   if (!--permanent) {   /* NFS sporadically generates these */
                       sleep (sleepsec); /* unwarranted, so ignore them first */
                       continue;
                   }
                   /*FALLTHROUGH*/
          case ENAMETOOLONG:
                   if (0 < (permanent = strlen (lockname) - 1) &&    
                       !strchr ("/", lockname[permanent - 1])) {
                       /* can we truncate it?  and retry ?*/
                       lockname[permanent] = '\0';
                       break;
                   }        /* otherwise, forget it */
                   if (debug)
                       report("File name \"%s\" too long\n", lockname);
                   else
                       Alert_Generate(NFR_SRC_SPACEMAND, NFR_SPACEMAND_LCKNAME);
                   retval = EX_UNAVAILABLE;

                   return(cleanup(lockname,retval));
          default:        /* but, it seems to persist, so give up */
                   if (debug)
                       report("Lock giving up...%s\n", "sigh...");
                   else
                       Alert_Generate(NFR_SRC_SPACEMAND, NFR_SPACEMAND_LCKNO);
                   retval = EX_UNAVAILABLE;
                   return(cleanup(lockname,retval));
        }        
        permanent = nfsTRY;    /* refresh the NFS-error-ignore count */
      }

      free(lockname);
      return (retval);
}

int unlockfile(char *filename)
{
   char *lockname;

   if ((lockname = genlockname(filename)) == NULL) {
        if (debug)
            report("Whoa!... Out of %s\n","Memory");
        else
            Alert_Generate(NFR_SRC_SPACEMAND, NFR_SPACEMAND_LCKMALLOC);
        return(EX_OSERR);
    }
    unlink(lockname);
    free(lockname);
    return(0);
}


#ifdef TEST
int main()
{
  int lc;

  lc = lockfile("fred");

  fprintf(stderr,"lockfile(fred) returned %d\n",lc);

  sleep(60);

  unlockfile("fred");

  return(0);
}
#endif 
