/*
** Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
** All rights reserved.
**
** Use and distribution of this software and its source code
** are governed by the terms and conditions of the
** Network Flight Recorder Software License ("LICENSE.TXT")
*/

#include <stdio.h>

#include "nfrd.h"

/*
** listlen
**
** compute the number of members.
*/

int
listlen( ListMem * list )
{
    int cnt = 0;
    while( list ) {
	list = list->next;
	cnt++;
    }
    return(cnt);
}

/*
** liststrlen/liststrcat.
**
** listlen - returns the number of characters it would take to represent
** the values in the list (including lists).  It may overestimate the
** sizes of some elements in favor of speed  :^(
**
** listcat - returns a concatenated string of all the members in the
** list.
**
**
**
**
*/

int
liststrlen( ListMem * list )
{
    ListMem * lp = list;
    int len = 0;
    while(lp) {
	MemVal * mp = &lp->data;
	switch (mp->typ) {
	    case NFR_TYPE_INT:
		len += 13; /* estimate */
		break;
	    case NFR_TYPE_IPV4HOST:
		len += 15; /* estimate */
		break;
	    case NFR_TYPE_IPV4NET:
		len += 31; /* estimate */
		break;
	    case NFR_TYPE_ETHMAC:
		len += 17; /* estimate */
		break;
	    case NFR_TYPE_STR:
		len += strlen( mp->val.c );
		break;
	    case NFR_TYPE_LIST:
		len += liststrlen( mp->val.v );
		break;
	}
	lp = lp->next;
    }
    return(len);
}

char *
liststrcat( ListMem * list, char * str )
{
    ListMem * lp = list;
    char * new = str;
    char * t;
    if (new == 0) {
	int len = liststrlen(list);
	if (len == 0) {
	    return(0);
	}
	new = (char *)malloc( len+1 );
    }
    t = new;

    while (lp) {
	MemVal * mp = &lp->data;
	switch(mp->typ) {
	    case NFR_TYPE_INT:
		{
		    char buffer[32];
		    sprintf( buffer, "%d", mp->val.i );
		    {
			char * f = buffer;
			while( *f ) {
			    *t++ = *f++;
			}
		    }
		}
		break;

	    case NFR_TYPE_IPV4HOST:
		{
		    char buff[64], *f;
		    sprintf( buff, "%s", util_ntoa( mp->val.h.s_addr ) );
		    f = buff;
		    while( *f ) {
			*t++ = *f++;
		    }
		}
		break;

	    case NFR_TYPE_IPV4NET:
		{
		    struct in_addr mask;
		    char buff[64];

		    mask.s_addr = mp->val.n->ip_mask;

		    /* see if we can display this in cidr style notation. */
		    {
			int i = 32;
			while (i--) {
			    if (mp->val.n->ip_mask == htonl(0xffffffff << i)) {
				sprintf( buff, "%s/%d", 
					inet_ntoa( mp->val.n->ip_addr ), 32-i );
				break;
			    }
			}
			if (i == -1) {
			    sprintf(buff, "%s:", inet_ntoa( mp->val.n->ip_addr ) );
			    strcat( buff, util_ntoa( mp->val.n->ip_mask ) );
			}
		    }
		    {
			char* f = buff;
			while( *f ) {
			    *t++ = *f++;
			}
		    }
		}
		break;

	    case NFR_TYPE_ETHMAC:
		{
		    char buff[64], *f;
		    sprintf( buff, "%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x",
			mp->val.e.e_addr[0], mp->val.e.e_addr[1],
			mp->val.e.e_addr[2], mp->val.e.e_addr[3],
			mp->val.e.e_addr[4], mp->val.e.e_addr[5] );
		    f = buff;
		    while( *f ) {
			*t++ = *f++;
		    }
		}
		break;
	    case NFR_TYPE_STR:
		{
		    char * f = mp->val.c;
		    while(*f) {
			    *t++ = *f++;
		    }
		}
		break;
	    case NFR_TYPE_LIST:
		{
		    t = liststrcat( mp->val.v, t );
		}
	}
	lp = lp->next;
    }
    /*
    ** if a string was passed in, then return the next location to muck.
    */
    if (str) {
	return(t);
    }
    return new;
}

/*
** listbloblen/listblobcat.
**
** listlen - returns the number of characters it would take to represent
** the values in the list (including lists).  It may overestimate the
** sizes of some elements in favor of speed  :^(
**
** listcat - returns a concatenated string of all the members in the
** list.
**
**
**
**
*/

int
listbloblen( ListMem * list )
{
    ListMem * lp = list;
    int len = 0;
    while(lp) {
	MemVal * mp = &lp->data;
	switch (mp->typ) {
	    case NFR_TYPE_STR:
		if (mp->val.c) {
		    len += strlen( mp->val.c );
		}
		break;
	    case NFR_TYPE_BLOB:
		if (mp->val.b) {
		    len += mp->val.b->b_len;
		}
		break;
	    case NFR_TYPE_LIST:
		len += listbloblen( mp->val.v );
		break;
	}
	lp = lp->next;
    }
    return(len);
}

/*
** Although this routine works with blobs, its passing around
** char*'s.  That's cause its seralizing/pickling the data for
** the blobs into memory...
*/ 

char *
listblobcat( ListMem * list, char * str )
{
    ListMem * lp = list;
    char * new = str;
    char * t;
    if (new == 0) {
	int len = listbloblen(list);
	if (len == 0) {
	    return(0);
	}
	new = (char *)malloc( len+1 );
    }
    t = new;

    while (lp) {
	MemVal * mp = &lp->data;
	switch(mp->typ) {
	    case NFR_TYPE_BLOB:
		{
		    blob_t * b = mp->val.b;
		    if (b) {
			if (b->b_len) {
			    memcpy( t, b->b_data, b->b_len );
			    t += b->b_len;
			}
		    }
		}
		break;
	    case NFR_TYPE_STR:
		{
		    char * f = mp->val.c;
		    while(*f) {
			    *t++ = *f++;
		    }
		}
		break;
	    case NFR_TYPE_LIST:
		{
		    t = listblobcat( mp->val.v, t );
		}
	}
	lp = lp->next;
    }
    /*
    ** if a string was passed in, then return the next location to muck.
    */
    if (str) {
	return(t);
    }
    return new;
}

/*
** listbyteslen/listbytescat.
**
** These two routines are used by blobbytes.
**
** listlen - returns the number of characters it would take to represent
** the values in the list (including lists).  It may overestimate the
** sizes of some elements in favor of speed  :^(
**
** listcat - returns a concatenated string of all the members in the list.
**
**
**
**
*/

int
listbyteslen( ListMem * list )
{
    ListMem * lp = list;
    int len = 0;
    while(lp) {
	MemVal * mp = &lp->data;
	switch (mp->typ) {
	    case NFR_TYPE_STR:
		if (mp->val.c) {
		    len += strlen( mp->val.c );
		}
		break;
	    case NFR_TYPE_BLOB:
		if (mp->val.b) {
		    len += mp->val.b->b_len;
		}
		break;

	    case NFR_TYPE_INT:
		len += 1;
		break;

	    case NFR_TYPE_LIST:
		len += listbyteslen( mp->val.v );
		break;
	}
	lp = lp->next;
    }
    return(len);
}

/*
** Although this routine works with blobs, its passing around
** char*'s.  That's cause its seralizing/pickling the data for
** the blobs into memory...
*/ 

char *
listbytescat( ListMem * list, char * str )
{
    ListMem * lp = list;
    char * new = str;
    char * t;
    if (new == 0) {
	int len = listbyteslen(list);
	if (len == 0) {
	    return(0);
	}
	new = (char *)malloc( len+1 );
    }
    t = new;

    while (lp) {
	MemVal * mp = &lp->data;
	switch(mp->typ) {
	    case NFR_TYPE_INT:
		*t++ = (mp->val.i & 0xff);
		break;

	    case NFR_TYPE_BLOB:
		{
		    blob_t * b = mp->val.b;
		    if (b) {
			if (b->b_len) {
			    memcpy( t, b->b_data, b->b_len );
			    t += b->b_len;
			}
		    }
		}
		break;
	    case NFR_TYPE_STR:
		{
		    char * f = mp->val.c;
		    while(*f) {
			    *t++ = *f++;
		    }
		}
		break;

	    case NFR_TYPE_LIST:
		{
		    t = listbytescat( mp->val.v, t );
		}
	}
	lp = lp->next;
    }
    /*
    ** if a string was passed in, then return the next location to muck.
    */
    if (str) {
	return(t);
    }
    return new;
}


/*
** nfr_blt_sort.
**
** sort the members of the list.
**
** We group all the differet types together, and among members we sort based
** on the following criterea:
**
** 1. integers in increasing order,
** 2. strings lexically,
** 3. lists first by their length (number of members), then by what?
** 4. host numbers in their order.
** 5. net nubmers in their order.
*/

int
listSortCmp( const void * left, const void * right )
{
    MemVal * l = &(*((ListMem **)left))->data;
    MemVal * r = &(*((ListMem **)right))->data;

    /*
    ** different types get sorted by their length.
    */
    if (l->typ != r->typ) {
	return(l->typ - r->typ);
    }
    /*
    ** Same members get different rules.
    */
    switch(l->typ) {
	case NFR_TYPE_INT:
	    return(l->val.i - r->val.i);

	case NFR_TYPE_STR:
	    return(strcmp(l->val.c, r->val.c ) );

	case NFR_TYPE_LIST:
	    /*
	    ** XXX can we do something more here?
	    */
	    return(listlen(l->val.v) - listlen(r->val.v));

	case NFR_TYPE_IPV4HOST:
	    /* for hosts, its just the simple order... */
	    return(memcmp( (char *)l->val.h.s_addr, (char *)r->val.h.s_addr,
		sizeof(l->val.h.s_addr) ) );

	case NFR_TYPE_IPV4NET:
	    /*
	    ** The more zero bits, the bigger the net is.
	    ** So i guess we hafta order based on ascending 
	    ** order and also on containment.  If net:a is
	    ** completely inside net:b, then seems that a<b.
	    */
	    if (l->val.n->ip_mask < r->val.n->ip_mask) {
		return(-1);
	    }
	    else
	    if (l->val.n->ip_mask > r->val.n->ip_mask) {
		return(1);
	    }
	    else {
		/*
		** The masks are equal.  That means that
		** we now need to compare host addresses.
		*/
		if (l->val.n->ip_addr.s_addr < r->val.n->ip_addr.s_addr) {
		    return(-1);
		}
		else
		if (l->val.n->ip_addr.s_addr > r->val.n->ip_addr.s_addr) {
		    return(-1);
		}
		else {
		    return(0);
		}
	    }
	case NFR_TYPE_ETHMAC:
	    /*
	    ** it may be faster to just compare the six bytes.
	    */
	    {
		int d;
		if (d = (l->val.e.e_addr[0] - r->val.e.e_addr[0])) {
			return(d);
		}
		if (d = (l->val.e.e_addr[1] - r->val.e.e_addr[1])) {
			return(d);
		}
		if (d = (l->val.e.e_addr[2] - r->val.e.e_addr[2])) {
			return(d);
		}
		if (d = (l->val.e.e_addr[3] - r->val.e.e_addr[3])) {
			return(d);
		}
		if (d = (l->val.e.e_addr[4] - r->val.e.e_addr[4])) {
			return(d);
		}
		if (d = (l->val.e.e_addr[5] - r->val.e.e_addr[5])) {
			return(d);
		}
		return(0);
	    }

	case NFR_TYPE_PATTERN:
	    return(strcmp(l->val.p->p_str, r->val.p->p_str ));

	default:
	    /* oops trouble.  XXX internal error */
		;
    }
}

/*
** we actually do the sort by collecing all the list members in an array,
** caling qsort, then exploding the entries back into a new list.
*/

void
nfr_blt_lsort(Machine *m,int argc,int firstarg,int result)
{
    MemVal * mem = &m->m_mem[firstarg];
    /*
    ** We could sort multiple lists, but why?
    */
    if (mem->typ != NFR_TYPE_LIST) {
	userError:
	m->m_mem[result].typ = NFR_TYPE_ERR;
	m->m_mem[result].val.i = NFR_ERR_USER;
	return;
    }
    /*
    ** compute the length of the list, make an array, then populate it with
    ** memVals, then pour the result back into the same list..
    */
    {
	ListMem * lp = mem->val.v;
	int listLen = listlen(lp);
	if (listLen) {
	    ListMem ** array = (ListMem **)malloc( sizeof(*array) * listLen );
	    ListMem ** ap = array;
	    while(lp) {
		*ap++ = lp;
		lp = lp->next;
	    }
	    qsort( array, listLen, sizeof(array[0]), listSortCmp );

	    /*
	    ** members are now sorted.
	    ** make a new list.
	    */
	    {
		int i;
		ListMem * new = 0;
		for (i=listLen-1; i>=0; --i) {
		    ListMem * np = (ListMem *)tmpcalloc( sizeof(*np) );
		    np->next = new;
		    new = np;
		    memtmpvalclonedata( &np->data, &array[i]->data );
		}
		free(array);
		m->m_mem[result].typ = NFR_TYPE_LIST;
		m->m_mem[result].val.v = new;
	    }
	} else {
	    m->m_mem[result].typ = NFR_TYPE_LIST;
	    m->m_mem[result].val.v = 0;
	}
    }
}

/*
** Review the list for adjacent elements, and only append new elements onto
** the (returned) list.
*/

nfr_blt_luniq(Machine *m,int argc,int firstarg,int result) 
{
    MemVal * mem = &m->m_mem[firstarg];
    if (mem->typ != NFR_TYPE_LIST) {
	userError:
	m->m_mem[result].typ = NFR_TYPE_ERR;
	m->m_mem[result].val.i = NFR_ERR_USER;
	return;
    }

    /*
    ** Start making the new list...
    ** we make an array to hold the list entries to walk the interstion backwards.
    */
    {
	ListMem * lp = mem->val.v;
	int len = listlen(lp);
	if (len) {
	    ListMem **array = (ListMem **)malloc( sizeof(*array) * len );
	    ListMem **ap = array;
	    while(lp) {
		*ap++ = lp;
		lp = lp->next;
	    }
	    {
		int i;
		ListMem * new = 0;
		ListMem * np = 0;
		for (i=len-1; i>=0; --i) {
		    if ((np == 0) || (is_equal( &np->data, &array[i]->data)==0)) {
			np = (ListMem *)tmpcalloc( sizeof(*np) );
			np->next = new;
			new = np;
			memtmpvalclonedata( &np->data, &array[i]->data );
		    }
		}
		free(array);
		m->m_mem[result].typ = NFR_TYPE_LIST;
		m->m_mem[result].val.v = new;
	    }
	} else {
	    m->m_mem[result].typ = NFR_TYPE_LIST;
	    m->m_mem[result].val.v = 0;
	}
    }
}

/*
** sublist
**
** like substr, but deals in list elements, not string elements.
** it either has two, or three arguments.
*/

nfr_blt_sublist(Machine *m,int argc,int firstarg,int result)
{
    MemVal * listMem = &m->m_mem[firstarg];
    MemVal * indexMem =  &m->m_mem[firstarg+1];
    if (listMem->typ != NFR_TYPE_LIST) {
	userError:
	m->m_mem[result].typ = NFR_TYPE_ERR;
	m->m_mem[result].val.i = NFR_ERR_USER;
	return;
    }
    /*
    **
    */
    if (indexMem->typ != NFR_TYPE_INT) {
	goto userError;
    }
    if ((argc != 2) && (argc != 3)) {
    goto userError;
    }
    {
	/*
	** Now copy the elements we're interested in,
	** and then make the new list.
	*/
	int len = listlen(listMem->val.v);

	if (len) {
	    ListMem ** array = (ListMem **)malloc( len*sizeof(array[0]) );
	    ListMem ** ap = array;
	    ListMem * lp = listMem->val.v;


	    while(lp) {
		*ap++ = lp;
		lp = lp->next;
	    }
	    /*
	    ** now the list is blasted into an array.
	    ** fish out the elements we're interested in... to make the new list.
	    ** again note we traverse backwards, cause we can then
	    ** do head insertion on the list.  In other words, we guarntee
	    ** wwe keep the order of the list elements intact.
	    */

	    {
		int index = indexMem->val.i;	/* index user passed in */
		int length;			/* (optional) length user passed in */

		/* verify low bound on index */
		if (index < 0) {
		    index = 0;
		}
		/* compute length (including default length) */
		if (argc == 3) {
		    MemVal * lenMem = &m->m_mem[firstarg+2];
		    if (lenMem->typ != NFR_TYPE_INT) {
			goto userError;
		    }
		    length = lenMem->val.i;
		    if (length < 0) {
			length = 0;
		    }

		}
		if (argc == 2) {	/* compute default length */
		    length = len - index;
		}
		{
		    ListMem * new = 0;
		    ListMem * np;

		    int i = len;
		    if (index+length < i) {
			i = index+length;
		    }

		    for (--i; i>=index; --i) {
			np = (ListMem *)tmpcalloc( sizeof(*np) );
			np->next = new;
			new = np;
			memtmpvalclonedata( &np->data, &array[i]->data );
		    }
		    m->m_mem[result].typ = NFR_TYPE_LIST;
		    m->m_mem[result].val.v = new;
		}
	    }
	    free(array);
	} else {
	    /* no elements on the list. */
	    m->m_mem[result].typ = NFR_TYPE_LIST;
	    m->m_mem[result].val.v = 0;
	}
    }
}

/*
** nfr_blt_listlen
**
** Compute the length of the list.
*/

nfr_blt_listlen(Machine *m,int argc,int firstarg,int result)
{
    MemVal * listMem = &m->m_mem[firstarg];
    
    if (listMem->typ != NFR_TYPE_LIST) {
	userError:
	m->m_mem[result].typ = NFR_TYPE_ERR;
	m->m_mem[result].val.i = NFR_ERR_USER;
	return;
    }
    m->m_mem[result].typ = NFR_TYPE_INT;
    m->m_mem[result].val.i = listlen( listMem->val.v );
}
