/*
** Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
** All rights reserved.
**
** Use and distribution of this software and its source code
** are governed by the terms and conditions of the
** Network Flight Recorder Software License ("LICENSE.TXT")
*/


/*
** intf.c
*/

#include "nfrd.h"

#include "pcap.h"

int	intfTotalPackets = 0;

/*
** ------------------------------------------------------------
**
** intfPacket.
**
** Target of pcap dispatch routine.
*/

void
intfPacket( v, pkt, pk )
u_char * v; /* ought to be void, but pcap.h missed :^( */
struct pcap_pkthdr * pkt;
const u_char *pk;
{
    intf_t * i = (intf_t *)v;	/* input packet */
    intfTotalPackets++;
    if (i->i_pkts++ > 0) {
	double sq;
	timeval_t d = util_diffTv( &pkt->ts, &i->i_arr );
	double diff = (util_tv2usec(&d)/MILLION);
	util_incrTv( &i->i_arrDiff, &d );
	i->i_arrSumSq += (diff*diff);
	if (i->i_pkts > 1) {
	    i->i_arrStdDev = util_stdDev(
		util_tv2usec(&i->i_arrDiff)/i->i_pkts/MILLION,
		i->i_arrSumSq,
		i->i_pkts );
	}

    }
    i->i_arr = pkt->ts;
    i->i_bytes = counter_add( i->i_bytes, counter_from_int(pkt->len) );
    packetRecv( pkt, pk );
}

/*
** ------------------------------------------------------------
**
** intfProcess.
**
*/

int
intfProcess( fd, v )
int fd;
void * v;
{
    intf_t * i = (intf_t *)v;

    LOGV( 2, ( FL, "intfProcess: fd %d time to dispatch %x\n", fd, v ));
    pcap_dispatch( i->i_in, -1, intfPacket, (u_char *)i );
}

/*
** ------------------------------------------------------------
**
** intfNew
** Make a new intf_t structure... keep track of it on the intfList.
*/

intf_t *
intfNew( name )
char * name;
{
    intf_t * new = (intf_t *)calloc( sizeof(*new), 1 );
    new->i_name = strdup(name);

    return(new);
}

/*
** ------------------------------------------------------------
**
** intfFree.
*/

void
intfFree( it )
intf_t * it;
{
    free(it->i_name);
    if (it->i_in) {
	intfClose(it);
    }
    free(it);
}

/*
** ------------------------------------------------------------
*/

que_t *
intfAdd( it )
intf_t * it;
{
    que_t * q = queNew(it);
    if (queAdd( intfList, q ) == 0) {
	intfList = q;
    }
    return(q);
}

/*
** ------------------------------------------------------------
**
** intfDel
** Remove the entry on the list, then remove the entry.
**
*/

void
intfDel( q )
que_t * q;
{
    intf_t * it = (intf_t *)q->q_data;
    if (intfList == q) {
	intfList = queForw(q);
    }
    queFree(q);
    intfFree(it);
}

/*
** ------------------------------------------------------------
**
** intfOpen
**
** return a handle describing the input interface.
** This would be the place to decide whether we're reading
** an interface, or a saved file..
**
** pass in a name of the interface we're interested in.
**
** Use a timeout of 3 seconds... just to make sure we collect
** data recently acquired.
**
*/

void *
intfOpen( n )
char * n;
{
    intf_t * new = intfNew(n);

    {
	char error[256];

	new->i_in = pcap_open_live( n,
			/* man len (complete packet) */ -1,
			/* promisc */		1,
			/* read timeout */	500, 
			/* error buffer */	error );
	if (new->i_in == 0) {
	    Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_PCAPOPENERR, n, error, 0 );
	    uiCurrPrintf( "intfOpen: intf %s error %s\n", n, error );
	    intfFree(new);
	    return(0);
	}
	intfAdd(new);
    }
    /*
    ** Tell the polling routines what to do...
    */
    pollCall( intf_fd(new->i_in), intfProcess, new );

    return( (void *)new );
}


/*
** ------------------------------------------------------------
**
** intfClose.
**
*/

int
intfClose( v )
void * v;
{
    intf_t * i = (intf_t *)v;

    /*
    ** tell the polling routines no more data...
    */
    pollUnCall( intf_fd(i->i_in) );

    if (i->i_in) {
	pcap_close( i->i_in );
	free(i);
	return(0);
    } else {
	return(1);
    }
}

/*
** ------------------------------------------------------------
**
** intfStats.
**
** Display status information about each known interface.
**
*/

void
intfStats()
{
    que_t * q = intfList;
    if (q) {
	do {
	    /*
	    ** get pcap stats for each interface we know about.
	    */
	    struct pcap_stat s[1];
	    intf_t * it = (intf_t *)q->q_data;

	    if (it->i_in) {
		s->ps_ifdrop = 0;
		if (pcap_stats( it->i_in, s ) == 0) {
		    uiCurrPrintf( "intf: %s: ps_recv %d\n", it->i_name, s->ps_recv );
		    uiCurrPrintf( "intf: %s: ps_drop %d\n", it->i_name, s->ps_drop );
		    uiCurrPrintf( "intf: %s: ps_ifDrop %d\n", it->i_name, s->ps_drop );
		} else {
		    uiCurrPrintf( "intf: %s: no pcap stats\n", it->i_name );
		}
		{
		    char cbuf[COUNTER_TEXT_SIZE];
		    counter_itoa( cbuf, it->i_bytes );
		    uiCurrPrintf( "intf: %s: packets %d\n", it->i_name, it->i_pkts );
		    uiCurrPrintf( "intf: %s: bytes %s\n", it->i_name, cbuf );
		}
		if (it->i_pkts) {
		    char tbuf[64];
		    strftime( tbuf, sizeof(tbuf), "[%D %T]", localtime( &it->i_arr.tv_sec ) );
		    uiCurrPrintf( "intf: Arriv %s secTotal %g secTotal/cnt %g stdDev %g\n",
			tbuf,  (util_tv2usec( &it->i_arrDiff)/MILLION),
			(util_tv2usec( &it->i_arrDiff)/MILLION)/it->i_pkts, it->i_arrStdDev
			);
		}
	    } else {
		uiCurrPrintf( "intf: %s: not configured\n", it->i_name );
	    }
	} while (q != intfList);
    }
    uiCurrPrintf( "intf: totalPackets %d\n", intfTotalPackets );
}



/*
** ------------------------------------------------------------
** intfSample.
**
** review each of the interfaces we know about, and check to
** see what packet loss is like.
*/

void
intfSample()
{
    que_t * q = intfList;
    if (q) {
	do {
	    struct pcap_stat s[1];
	    intf_t * it = (intf_t *)q->q_data;
	    /*
	    ** get another copy of the pcap_stats, and see whether
	    ** we have additional packet loss from last time...
	    */
	    if (it->i_in) {
		if (pcap_stats( it->i_in, s ) == 0) {
		    if (it->i_stat->ps_drop < s->ps_drop) {
			int diff = s->ps_drop - it->i_stat->ps_drop;
			/* Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_INTFLOSSERR,
				it->i_name, 
				format( "%d", diff ), 
				format( "%d", it->i_stat->ps_drop ),
				format( "%d", s->ps_drop ), 0 ); */
			log( FL, "intfSample: %s packet loss %dD (%d/%d)\n",
				it->i_name, diff, it->i_stat->ps_drop,s->ps_drop);
			*it->i_stat = *s;
		    }
		} else {
		    Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_PCAPSTATERR,
			it->i_name, 0 );
		    uiCurrPrintf( "intfSample: can't pcap_stat %s\n", it->i_name );
		}
	    }
	} while (q != intfList);
    }
}

/*
** ------------------------------------------------------------
**
** intfInit
**
** Perform any activity associated with initializing any of the
** inbound streams.  Currently, this only set up the timeout
** interval for intfSample.
**
*/

void
intfInit( )
{
    /*
    ** set up the timeout for intfSample.
    */
    {
	timeval_t when = util_makeTv( /* sec */ 10, /* usec */ 0 );
	to_t * to = toNew( &when, TO_FLAGS_SPECIAL | TO_FLAGS_REPEAT, 0, intfSample );
    }
}

