/*
** Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
** All rights reserved.
**
** Use and distribution of this software and its source code
** are governed by the terms and conditions of the
** Network Flight Recorder Software License ("LICENSE.TXT")
*/


#include <stddef.h>
#include <stdio.h>
#include <string.h>

#include "nfrd.h"


/*
**
** Memory-hauling routines. Due to data-hiding constraints
** not all memory or structure creating routines are here. But
** miscellaneous code for deleting or creating structures, temporary
** memory allocation for runtime, cloning data, etc, are here.
**
**	 Marcus J. Ranum, Network Flight Recorder, Inc. 1996
*/



/*
** temporary data is managed by keeping a list of allocated stuff, which
** can be traversed and freed. this is not too terribly efficient, but
** is not a load-bearing section of code.
*/
struct	tmplist	{
	struct	tmplist	*next;
	void 	(*tmpFree)(void *);	/* to allow decRef's of dat. */
	char	*dat;
};

static	struct	tmplist	*tmpl;		/* currently active temporaries */
static	struct	tmplist	*freetmpl;	/* temp list structures freed and stacked */

void	* tmpCallOnFree( void (*free)(void *), void * );

/*
** ----------------------------------------
*/

Machine *
machinenew()
{
	Machine	*ret;

	if((ret = (Machine *)malloc(sizeof(Machine))) == (Machine *)0)
		return((Machine *)0);
	ret->m_maxfp = 64;
	ret->m_fram = (Frame *)malloc(sizeof(Frame) * ret->m_maxfp);
	if(ret->m_fram == (Frame *)0) {
		free(ret);
		return((Machine *)0);
	}
	ret->m_maxsp = 2048;
	ret->m_mem = (MemVal *)malloc(sizeof(MemVal) * ret->m_maxsp);
	if(ret->m_mem == (MemVal *)0) {
		free(ret->m_fram);
		free(ret);
		return((Machine *)0);
	}
	ret->m_sp = 0;
	return(ret);
}

/*
** ----------------------------------------
*/

void
machinefree(m)
Machine * m;
{
	if (m->m_fram) {
	    free(m->m_fram);
	    m->m_fram = 0;
	}
	if (m->m_mem) {
	    free(m->m_mem);
	    m->m_mem = 0;
	}
	free(m);
}

/*
** ----------------------------------------
*/

blob_t *
blobNew( char * in, int len, int flags )
{
    char * newData = 0;
    if (len) {
	newData = (char *)malloc( len );
	if (newData == 0) {
	    len = 0;
	} else if (in) {
	    memcpy( newData, in, len );
	}
    }
    {
	blob_t * b = (blob_t *)calloc( 1, sizeof(*b) );
	/* XXX assert? */
	b->b_refCnt = 1;
	b->b_flags = BLOB_FLAGS_MALLOC;
	b->b_data = newData;
	b->b_len = len;
	return(b);
    }
}

blob_t *
blobSet( blob_t * b, char * in, int len, int cnt, int flags )
{
    if (b) {
	b->b_data = in;
	b->b_len = len;
	b->b_flags = flags;
	b->b_refCnt = cnt;
    }
    return(b);
}

void
blobFree( v )
void * v;
{
    blob_t * b = (blob_t *)v;
    if (b) {
	if (b->b_flags & BLOB_FLAGS_MALLOC) {
	    if (b->b_data)  {
		char * data = b->b_data;
		b->b_data = 0;
		free(data);
	    }
	}
	b->b_flags = 0;
	b->b_len = 0;
	free(b);
    }
}

blob_t *
blobIncRef( b )
blob_t * b;
{
    if (b) {
	b->b_refCnt++;
    }
    return(b);
}

void
blobDecRef( v )
void * v;
{
    blob_t * b = (blob_t *)v;
    if (b) {
	if (--b->b_refCnt <= 0) {
	    blobFree(b);
	}
    }

}
/*
** ----------------------------------------
** memvalIsNull.
** We need this to tell whether symbols get unset.
*/

int
memvalIsNull(MemVal *m)
{
    if (m) {
	return( (m->typ == NFR_TYPE_ERR) && (m->val.i == NFR_ERR_NULL) );
    }
    return(0);
}

/*
** ----------------------------------------
*/

void
memvalfreedata(MemVal *s)
{
	ListMem	*lp;
	ListMem	*np;

	switch(s->typ) {
	case NFR_TYPE_ERR:
	case NFR_TYPE_INT:
	case NFR_TYPE_IPV4HOST:
	case NFR_TYPE_ETHMAC:
	case NFR_TYPE_INDIRECT:
		break;
	case NFR_TYPE_STR:
		free(s->val.c);
		break;
	/* recursively free a list */
	case NFR_TYPE_LIST:
		lp = (ListMem *)s->val.v;
		while(lp != (ListMem *)0) {
			np = lp->next;
			memvalfreedata(&(lp->data));
			free(lp);
			lp = np;
		}
		break;

	case NFR_TYPE_IPV4NET:
		{
		    ipv4net_t * n = s->val.n;
		    s->val.n = 0;
		    free( n );
		}
		break;

	case NFR_TYPE_RECORDER:
		recorderDecRef( (recorder_t *)s->val.v );
		break;

	case NFR_TYPE_PATTERN:
		{
		    pattern_t * p = s->val.p;
		    s->val.p = 0;
		    patternFree( p );
		}
		break;

	case NFR_TYPE_TRIGGER:
		triggerDecRef( s->val.t );
		break;

	case NFR_TYPE_MAP:
		mapDecRef(s->val.v);
		break;

	case NFR_TYPE_MAPBAG:
		mapBagDecRef(s->val.v);
		break;

	case NFR_TYPE_BLOB:
		if (s->val.b && (s->val.b->b_flags & BLOB_FLAGS_STATIC) == 0) {
		    blobDecRef(s->val.b);
		}
		break;
	default:
		panic( "unknown data type in memvalfreedata(): %d\n",s->typ);
		break;
	}
}


/*
** ----------------------------------------
*/

int
memvalclonedata(MemVal *s, MemVal *v)
{
	ListMem	*lp;
	ListMem	*np;
	ListMem	*xp;

	s->typ = v->typ;

	switch(v->typ) {
	case NFR_TYPE_ERR:
		s->val.i = v->val.i;
		break;
	case NFR_TYPE_INT:
		s->val.i = v->val.i;
		break;
	case NFR_TYPE_STR:
		s->val.c = (char *)malloc(strlen(v->val.c) + 1);
		if(s->val.c == (char *)0)
			return(-1);
		strcpy(s->val.c,v->val.c);
		break;
	/* recursively clone a list */
	case NFR_TYPE_LIST:
		lp = (ListMem *)v->val.v;
		np = (ListMem *)0;
		s->val.v = 0; /* for null lists */

		while(lp != (ListMem *)0) {
			xp = (ListMem *)malloc(sizeof(ListMem));
			if(xp == (ListMem *)0)
				return(-1);
			if(memvalclonedata(&(xp->data),&(lp->data)))
				return(-1);
			if(np == (ListMem *)0) {
				s->val.v = (void *)xp;
				np = xp;
			} else {
				np->next = xp;
				np = xp;
			}
			xp->next = (ListMem *)0;
			lp = lp->next;
		}
		break;

	case NFR_TYPE_RECORDER:	/* XXX violates small index rule */
		recorderIncRef( (recorder_t *)v->val.v );
		s->val.v = v->val.v;
		break;

	case NFR_TYPE_SCHEMADATA:
	case NFR_TYPE_INDIRECT:
		s->val.v = v->val.v;
		break;

	case NFR_TYPE_IPV4HOST:
		s->val.h = v->val.h;
		break;

	case NFR_TYPE_IPV4NET:
		s->val.n = (ipv4net_t *)malloc( sizeof(ipv4net_t) );
		if (s->val.n == 0) {
			return(-1);
		}
		*s->val.n = *v->val.n; 
		break;

	case NFR_TYPE_BLOB:
		if (v->val.b && (v->val.b->b_flags & BLOB_FLAGS_STATIC)) {
		    /* make our own copy */
		    blob_t * ob = v->val.b;
		    if (ob) {
			s->val.b = blobNew( ob->b_data, ob->b_len, BLOB_FLAGS_MALLOC );
		    }
		} else {
		    s->val.b = v->val.b;
		    blobIncRef( s->val.b );
		}
		break;

	case NFR_TYPE_ETHMAC:
		s->val.e = v->val.e;
		break;

	case NFR_TYPE_PATTERN:
		/* we can't clone these directly BUT we can make a new one. */
		s->val.p = patternNew( v->val.p->p_str );
		break;

	case NFR_TYPE_TRIGGER:
		triggerIncRef(v->val.t);
		s->val.t = v->val.t;
		break;

	case NFR_TYPE_MAPBAG:
		mapBagIncRef(v->val.v);
		s->val.v = v->val.v;
		break;

	case NFR_TYPE_MAP:
		mapIncRef(v->val.v);
		s->val.v = v->val.v;
		break;

	default:
		panic("unknown data type in memvalclonedata()");
	}
	return(0);
}

/*
** ----------------------------------------
*/

int
memtmpvalclonedata(MemVal *s, MemVal *v)
{
	ListMem	*lp;
	ListMem	*np;
	ListMem	*xp;

	s->typ = v->typ;

	switch(v->typ) {
	case NFR_TYPE_ERR:
		s->val.i = v->val.i;
		break;
	case NFR_TYPE_INT:
		s->val.i = v->val.i;
		break;
	case NFR_TYPE_STR:
		s->val.c = (char *)tmpalloc(strlen(v->val.c) + 1);
		if(s->val.c == (char *)0)
			return(-1);
		strcpy(s->val.c,v->val.c);
		break;
	/* recursively clone a list */
	case NFR_TYPE_LIST:
		lp = (ListMem *)v->val.v;
		np = (ListMem *)0;
		s->val.v = 0;	/* for null lists */

		while(lp != (ListMem *)0) {
			xp = (ListMem *)tmpcalloc(sizeof(ListMem));
			if(xp == (ListMem *)0)
				return(-1);
			if(memtmpvalclonedata(&(xp->data),&(lp->data)))
				return(-1);
			if(np == (ListMem *)0) {
				s->val.v = (void *)xp;
				np = xp;
			} else {
				np->next = xp;
				np = xp;
			}
			xp->next = (ListMem *)0;
			lp = lp->next;
		}
		break;

	case NFR_TYPE_RECORDER:	/* XXX violates small index rule */
		s->val.v = v->val.v;
		tmpCallOnFree( recorderDecRef, v->val.t );
		recorderIncRef( (recorder_t *)s->val.v );
		break;

	case NFR_TYPE_SCHEMADATA:
	case NFR_TYPE_INDIRECT:
		s->val.v = v->val.v;
		break;

	case NFR_TYPE_IPV4HOST:
		s->val.h = v->val.h;
		break;

	case NFR_TYPE_IPV4NET:
		s->val.n = (ipv4net_t *)tmpcalloc( sizeof(ipv4net_t) );
		if (s->val.n == 0) {
			return(-1);
		}
		*s->val.n = *v->val.n; 
		break;

	case NFR_TYPE_BLOB:
		if (v->val.b->b_flags & BLOB_FLAGS_STATIC) {
		    /* we need to make our own */
		    blob_t * ob = v->val.b;
		    s->val.b = blobNew( ob->b_data, ob->b_len, BLOB_FLAGS_MALLOC );
		} else {
		    blobIncRef( v->val.b );
		    s->val.b = v->val.b;
		}
		tmpCallOnFree( blobDecRef, v->val.b );
		break;

	case NFR_TYPE_ETHMAC:
		s->val.e = v->val.e;
		break;

	case NFR_TYPE_PATTERN:
		/* we can't clone these directly BUT we can make a new one. */
		s->val.p = patternNew( v->val.p->p_str );
		break;

	case NFR_TYPE_TRIGGER:
		/*
		** When we hafta do something when we release the storage,
		** now is the time to tell the list that..
		*/
		tmpCallOnFree( triggerDecRef, v->val.t );
		s->val.t = v->val.t;
		triggerIncRef(v->val.t);
		break;

	case NFR_TYPE_MAPBAG:
		tmpCallOnFree( mapBagDecRef, v->val.t );
		mapBagIncRef(v->val.v);
		s->val.v = v->val.v;
		break;

	case NFR_TYPE_MAP:
		tmpCallOnFree( mapDecRef, v->val.t );
		mapIncRef(v->val.v);
		s->val.v = v->val.v;
		break;

	default:
		panic("unknown data type in memtmpvalclonedata()");
	}
	return(0);
}

/*
** ----------------------------------------
** memTmpNew.
** tmpalloc'd MemVal.... This is just glue.
** Make a MemVal to just hold one...
*/

MemVal *
memTmpNew( type, value )
int type;
void * value;
{
    MemVal * m = tmpcalloc( sizeof(*m) );
    m->typ = type;
    switch (type) {
	case NFR_TYPE_ERR:
	    m->val.i = (int)value;
	    break;

	case NFR_TYPE_INT:
	    m->val.i = (int)value;
	    break;

	case NFR_TYPE_STR:
	    m->val.c = (char *)value;
	    break;

	case NFR_TYPE_LIST:
	case NFR_TYPE_RECORDER:
	case NFR_TYPE_SCHEMADATA:
	    m->val.v = value;
	    break;

	case NFR_TYPE_IPV4HOST:
	    m->val.h = *(struct in_addr *)value;
	    break;

	case NFR_TYPE_IPV4NET:
	    m->val.n = (ipv4net_t *)value;
	    break;

	case NFR_TYPE_BLOB:
	    m->val.b = (blob_t *)value;
	    break;

	case NFR_TYPE_ETHMAC:
	    m->val.e = *(ethMac_t *)value;
	    break;

	case NFR_TYPE_PATTERN:
	    m->val.p = patternNew( (char *)value );
	    break;

	case NFR_TYPE_TRIGGER:
	    m->val.t = value;
	    break;

	default:
	    /* need smart witty error message here */
	    break;
    }
    return(m);
}

/*
** ----------------------------------------
** memvalSize
**
*/

int
memvalSize( MemVal * m )
{
    switch(m->typ) {
	case NFR_TYPE_STR:
	    return( sizeof(MemVal)+strlen(m->val.c) );
	case NFR_TYPE_IPV4NET:
	    return( sizeof(MemVal)+sizeof(ipv4net_t) );
	case NFR_TYPE_BLOB:
	    return( sizeof(MemVal)+m->val.b->b_len );
	case NFR_TYPE_PATTERN:
	    return( sizeof(MemVal)+sizeof(pattern_t)+strlen(m->val.p->p_str) );
	case NFR_TYPE_LIST:
	    {
		ListMem	*lp = (ListMem *)m->val.v;
		int total = 0;

		while(lp != (ListMem *)0) {
		    total += memvalSize( &lp->data ) +
			    (sizeof(ListMem) - sizeof(MemVal));
		    lp = lp->next;
		}
		return( total );

	    }
	case NFR_TYPE_MAP:
	    {
		int total = 0;
		Map * map = m->val.v;
		while(map) {
		    Sym * sym = map->m_sym;
		    while(sym) {
			total += memvalSize( &sym->s_data ) + sizeof(Sym);
			if (SYM_FLAG_TYPE(sym) == SYM_FLAG_TYPE_NAME) {
			    total += sym->s_length;
			} else
			if (SYM_FLAG_TYPE(sym) == SYM_FLAG_TYPE_MEMVAL) {
			    total += memvalSize( sym->s_memVal );
			}
		    }
		    total += sizeof(*map);
		    map = map->m_next;
		}
		return(total);
	    }
	case NFR_TYPE_MAPBAG:
	    {
		int total = 0;
		MapBag * mb = (MapBag *)m->val.v;

		if (mb) {
		    int i;
		    total += sizeof(*mb);
		    for (i=0; i<mb->mb_width; ++i) {
			Map * map = mb->mb_maps[i];
			while(map) {
			    Sym * sym = map->m_sym;
			    while(sym) {
				total += memvalSize( &sym->s_data ) + sizeof(Sym);
				if (SYM_FLAG_TYPE(sym) == SYM_FLAG_TYPE_NAME) {
				    total += sym->s_length;
				} else
				if (SYM_FLAG_TYPE(sym) == SYM_FLAG_TYPE_MEMVAL) {
				    total += memvalSize( sym->s_memVal );
				}
			    }
			    total += sizeof(*map);
			    map = map->m_next;
			}
		    }
		}
		return(total);
	    }
	    break;
    }
    return( sizeof(MemVal) );

}

/*
** ----------------------------------------
** tmpstrdup.
** like strdup, but allocs into tmpalloc.
*/
char *
tmpstrdup( s )
char * s;
{
    if (s == 0) {
	return(0);
    } else {
	int len = strlen(s);
	char * str = (char *)tmpalloc( len+1 );
	strcpy( str, s );
	return(str);
    }
}

/*
** ----------------------------------------
** listfree.
** 
*/

void
listfree(ListMem *l)
{
    ListMem	*p;

    while(l != (ListMem *)0) {
	p = l->next;
	memvalfreedata(&(l->data));
	free(l);
	l = p;
    }
}



/*
** ----------------------------------------
** just like malloc() but keep track of the storage
*/

static tmpalloc_alwayszeroInit = 1;
static int tmpalloc_alwayszero = 0;

void	*
tmpalloc(int siz)
{
	struct	tmplist	*lp;

	if(freetmpl != (struct tmplist *)0) {
		lp = freetmpl;
		freetmpl = freetmpl->next;
	} else {
		if((lp = (struct tmplist *)malloc(sizeof(struct tmplist))) == (struct tmplist *)0)
			return(0);
	}
	lp->tmpFree = 0;
	lp->dat = malloc((unsigned)siz);
	
	if (tmpalloc_alwayszeroInit == 1) {
	    char * env;
	    tmpalloc_alwayszeroInit = 0;
	    /* perform the getenv here for the value of the variable */
	    {
		extern int tmpalloc_alwayszero;
		char *s = getenv("NFR_TMPZERO");
		if (s) {
		    tmpalloc_alwayszero = atoi(s);
		}
	    }

	}
	if (tmpalloc_alwayszero) {
		memset(lp->dat, 0, siz);
	}

	/* enList it */
	lp->next = tmpl;
	tmpl = lp;
	return(lp->dat);
}


/*
** ----------------------------------------
** like tmpalloc, but clears the storage.
*/

void	*
tmpcalloc(int siz)
{
	struct	tmplist	*lp;

	if(freetmpl != (struct tmplist *)0) {
		lp = freetmpl;
		freetmpl = freetmpl->next;
	} else {
		if((lp = (struct tmplist *)malloc(sizeof(struct tmplist))) == (struct tmplist *)0)
			return(0);
	}
	lp->tmpFree = 0;
	lp->dat = calloc((unsigned)siz, sizeof(char) );

	/* enList it */
	lp->next = tmpl;
	tmpl = lp;
	return(lp->dat);
}

/*
** ----------------------------------------
** call when we're freeing storage.  Usually used to decRef
** structures which were incRef's by memtmpvalclonedata
*/

void	*
tmpCallOnFree( free, dat )
void	(*free)(void *);
void *	dat;
{
	struct	tmplist	*lp;

	if(freetmpl != (struct tmplist *)0) {
		lp = freetmpl;
		freetmpl = freetmpl->next;
	} else {
		if((lp = (struct tmplist *)malloc(sizeof(struct tmplist))) == (struct tmplist *)0)
			return(0);
	}
	lp->dat = dat;
	lp->tmpFree = free;

	/* enList them */
	lp->next = tmpl;
	tmpl = lp;
	return(dat);
}


/*
** ----------------------------------------
** put something on the temp list to free later.
** this results in what is basically a deferred free() - executed
** at the end of each machine run.
*/
void
tmpputonfree(void *p)
{
	struct	tmplist	*lp;

	if(freetmpl != (struct tmplist *)0) {
		lp = freetmpl;
		freetmpl = freetmpl->next;
	} else {
		if((lp = (struct tmplist *)malloc(sizeof(struct tmplist))) == (struct tmplist *)0)
			return;
	}
	lp->dat = p;
	lp->next = tmpl;
	tmpl = lp;
}



/*
** ----------------------------------------
** free all the temporaries
*/

void
tmpfree()
{
	struct	tmplist	*lp = tmpl;
	struct	tmplist	*np;

	if(tmpl == (struct tmplist *)0)
		return;

	while(lp != (struct tmplist *)0) {
		if(lp->dat != 0) {
			if (lp->tmpFree) {
			    /* not our job */
			    (*lp->tmpFree)(lp->dat);
			} else {
			    (void)free(lp->dat);
			}
		}
		np = lp->next;
		lp->next = freetmpl;
		freetmpl = lp;
		lp = np;
	}
	tmpl = (struct tmplist *)0;
}


