/*
** Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
** All rights reserved.
**
** Use and distribution of this software and its source code
** are governed by the terms and conditions of the
** Network Flight Recorder Software License ("LICENSE.TXT")
*/

/*
** ------------------------------
**
** nfrd.
**
** This is the packet reading engine.
**
** Manage:
**
** -becomming a daemon (signals, process group, etc)
**
** -multiple input streams (several different promiscuous listeners)
**
** -command input stream (hmm... socket open to everywhere XXX?)
**
** -reading new functions
**
** -maintaining global state (scheduling global state rewrites when
**  the state changes).
**
**
*/

#include "nfrd.h"

#include "pcap.h"

char * argv0;
  

/*
** global variables.
** --------------------
**
** restart - set to one, when the poll loop needs to be exited...
** logFile - describes how to log.  (this isn't alerts)
** timeOut - how long to wait between polls...
**
*/

int restart = 0;

#ifndef DEFAULT_LOG_FILE
#define DEFAULT_LOG_FILE "nfrd.log"
#endif
char *  logFile = DEFAULT_LOG_FILE;

#ifndef	DEFAULT_CONF_FILE
#define	DEFAULT_CONF_FILE "~/etc/nfrd.cfg"
#endif
char * confFile = DEFAULT_CONF_FILE;

#ifndef	DEFAULT_VAR_DIR
#define DEFAULT_VAR_DIR	"~/etc/vars/"
#endif
char * varDir = DEFAULT_VAR_DIR;

#ifndef DEFAULT_NFRD_FIFO
#define	DEFAULT_NFRD_FIFO "~/etc/run/nfrd.fifo"
#endif
char * defaultNfrdFifo = DEFAULT_NFRD_FIFO;

#ifndef	DEFAULT_NFRD_PID
#define DEFAULT_NFRD_PID "~/etc/run/nfrd.pid"
#endif

char *	nfrHost = 0;		/* set to defaultNfrdFifo as first line in main */
int	nfrPort = 0;

int timeOut = 1000;		/* maybe every minute would make sense */
int recorderTimeOut = 1;	/* seconds to sync recorders */

int bFlag;			/* reincarnate variable during _b_oot  */
int dFlag;			/* debugging flag */
int vFlag;			/* verbose flag */

char * 	argv0;


/*
** ----------------------------------------
** nfrCf.
**
** Provide an external interface for other
** nfrd components to review the contents of
** the nfrd configuration file.
*/

TCf * nfrCf;

nfrCfInit( cfName )
char * cfName;
{
    nfrCf = Cf_New( 32 );

    /* tell the alertd we've the config path we're reading */
    Alert_Args_Generate( NFR_SRC_NFRD,
	NFR_NFRD_CONFIGPATH, cfName, 0 );

    if ((Cf_Read( cfName, nfrCf )) == LIBNFR_ERR_UNDEFINED) {
	/* troubles, we can't tell if its parsing or an error. */
	if (errno) {
	    Alert_Args_Generate( NFR_SRC_NFRD,
		NFR_NFRD_CONFIGERROR, cfName, strerror(errno), 0 );
	} else {
	    Alert_Args_Generate( NFR_SRC_NFRD,
		NFR_NFRD_CONFIGERROR, cfName, "parsing error", 0 );
	}
	return(-1);
    }
    return(0);
}

/*
** nfrCfGetVal.
** nfrd cover to get the value out of the nfrd config file.
**
*/

char * 
nfrCfGetVal( char * s )
{
    if (nfrCf) {
	return Cf_GetVal(s, nfrCf);
    } else {
	return(0);
    }
}

/*
** ----------------------------------------
** nfrdSync.
**
** Send a sync to each of the active recorders...
** (ie: ones with an active connection)
*/
static Machine * recvMac = 0;

void
nfrdSync( void * v ) {

    if (recvMac == 0) {
	recvMac = machinenew();
	if (recvMac == 0) {
	    /* XXX didn't get one.  */
	    Alert_Generate( NFR_SRC_NFRD, NFR_NFRD_SYNCNOMAC );
	    return;
	}
    }

    /*
    ** If we call recorderSync with a machine to use,
    ** it will use that machine to send timestamp
    ** packets.  If we call with zero, we'll just
    ** flush any outbound data.
    */
    recorderSync( recvMac );
    /* recorderSync( 0 );  */
}

/*
** nfrPid.
**
** Write the pid file...  leave the name we wrote into lying around.
**
*/
char * pidFileName;

nfrPid()
{
    pidFileName = nfrPath( DEFAULT_NFRD_PID, 0);

    /* write the pid file. */
    {
	FILE * f;
	if ((f = fopen( pidFileName, "w") ) == 0) {
	    Alert_Args_Generate( NFR_SRC_NFRD, NFR_ERR_CANTWRITE, pidFileName, 0 );
	    return;
	}
	fprintf( f, "%d\n", getpid() );
	fclose(f);
    }
}

/*
** nfrdUninit.
*/

nfrdUninit()
{
    unlink(pidFileName);
    /*
    ** remove fifo.
    */
    if(nfrHost && (nfrPort == 0)) {
	unlink(nfrHost);
    }
}

/*
**
*/
nfrdHalt()
{
    nfrdSync( 0 );
    /*
    ** clean up any memory we've got lying around :^)
    */
    tmpfree();

    restart = 1;
}

/*
** nfrdload.
**
** used to set a scope and a piece of .nfr code
** into the engine.  This allows quick early loading
** of packages, its done before interfaces are
** made avilable for input.
**
*/

nfrdLoad( scope, path )
char * scope;
char * path;
{
    char *argv[8];

    /*
    ** if the scope exists, set the scope.
    */
    if (scope) {
	argv[0] = scope;
	argv[1] = 0;
	comandScopeCmd( -1, argv );
    }

    /*
    ** set up to perform the read...
    */
    if (path) {
	argv[0] = path;
	argv[1] = 0;
	commandReadCmd( -1, argv );
    }
}


/*
** keep a vector of the directory names
** which we'll do package loads from.
*/

#define	PKGCNT	8
char	**pkgs;
int	pkgI = 0;
int	pkgMax = 0;



/*
** --------------------
**
** main.
**
*/




main( argc, argv )
int argc;
char * argv[];
{
    argv0 = argv[0];
    nfrHost = defaultNfrdFifo;

    scopeInit();
    /*
    ** we create a global filter.  Its not really a filter,
    ** but a sheel to hold some info, which we use to find
    ** symbols.
    */
    globalFilterInit();

    /*
    ** daemon housekeeping.
    ** set up signal's processing routines...
    */
    sigInit();
    recorderInit( );
    ipInit();
    symInit();
    intfInit();

    Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_START, "$Header: /home/master/nfrbase/nfr/nfrd/nfrd.c,v 1.18 1997/12/14 20:56:07 mts Exp $", argv0, 0 );

    {
	int i;
	for (i=1; i<argc; ++i) {
	    if (argv[i][0] == '-') switch(argv[i][1]) {
		case 'b':	/* reincarnate variables during boot */
		    bFlag++;
		    break;

		case 'f':	/* name of the fifo */
		    if (argv[i][2]) {
			nfrHost = &argv[i][2];
		    } else {
			nfrHost = argv[++i];
		    }
		    break;

		case 'l':	/* how to log. */
		    if (argv[i][2]) {
			logFile = &argv[i][2];
		    } else {
			logFile = argv[++i];
		    }
		    break;

		case 't':	/* timeout */
		    if (argv[i][2]) {
			timeOut = atoi( &argv[i][2] );
		    } else {
			timeOut = atoi( argv[++i] );
		    }
		    break;

		case 's':	/* recorder sync timeout */
		    if (argv[i][2]) {
			recorderTimeOut = atoi( &argv[i][2] );
		    } else {
			recorderTimeOut = atoi( argv[++i] );
		    }
		    if (recorderTimeOut < 1) {
			/*
			** silently enforce the lowest limit
			*/
			recorderTimeOut = 1;
		    }
		    break;

		case 'c':	/* conf */
		    if (argv[i][2]) {
			confFile = &argv[i][2];
		    } else {
			confFile = argv[++i];
		    }
		    break;

		case 'p':	/* port */
		    if (argv[i][2]) {
			nfrPort = atoi(&argv[i][2]);
		    } else {
			nfrHost = "localhost";
			nfrPort = atoi(argv[++i]);
		    }
		    /*
		    ** note that we reset the nfrHost only
		    ** if we set the port..
		    */
		    if (nfrHost == defaultNfrdFifo) {
			nfrHost = "localhost";
		    }
		    break;

		case 'v':	/* varDir */
		    if (argv[i][2]) {
			varDir = &argv[i][2];
		    } else {
			varDir = argv[++i];
		    }
		    break;
			
		case 'd':	/* debug */
		    dFlag++;
		    break;
		
		case 'r':	/* run mode tracing */
		    runSetDebug();
		    break;

		case 'V':	/* verbose */
		    vFlag++;
		    break;

		default:
		usage:
		    fprintf( stderr, "%s: usage nfrd [ -l logFileName ] "
			"[ -t timeOut ] [ -c confFile ] [ -v varDir ]\n\t"
			"[ -d ] [ -r ] [ -V ] [ -f fifoName ] "
			"[ -p port ] [ packageDir ... ] \n",
			argv0 );
		    exit(1);
	    } else {
		/*
		** Directories to be loaded...
		*/
		if (pkgI >= pkgMax) {
		    pkgMax += PKGCNT;
		    if (pkgs == 0) {
			pkgs = (char **)calloc( pkgMax, sizeof(*pkgs) );
		    } else {
			pkgs = (char **)realloc( pkgs, pkgMax*sizeof(*pkgs) );
		    }
		}
		pkgs[pkgI++] = argv[i];
	    }
	}
    }

    logOpen( logFile );

    pollInit();
    pollOpen();

    if (nfrHost == 0) {
	nfrHost = DEFAULT_NFRD_FIFO;
    }

    /*
    ** Now try to read alert resource information.
    */
    nfr_blt_alertInit();


    /*
    ** The configuration file contains some info which can change the
    ** way we startup.  Currently we only shovel the names of some
    ** interfaces into it.   The interfaces are named using the 
    ** nfr_intf variable, this is a space separaed string with 
    ** the names of the default interfaces.
    */
    {
	char * cfName = nfrPath( confFile, 0 );
	if (nfrCfInit( cfName ) == 0) {
	    /*
	    ** See if we can find the nfr_intf in the config file...
	    ** use it as the name of all interfaces to init.
	    */
	    {
		char * intfs = nfrCfGetVal( "nfr_intf" );
		/*
		** for each of the named intf's, init the intf.
		*/
		if (intfs) {
		    char ** intfList = crack( intfs );
		    {
			char ** intf = intfList;
			while( intf[0] ) {
			    if (intfOpen( intf[0] )) {
				/*
				** only tell alertd the interfaces
				** we've opened.  intfOpen will
				** complain about interfaces we
				** can't open.
				*/
				Alert_Args_Generate( NFR_SRC_NFRD,
				    NFR_NFRD_NEWINTF, 
				    format( "nfrd.cfg: %s", intf[0] ), 0 );
			    }
			    intf++;
			}
			
		    }
		    free(intfList);
		}
	    }
	}

	if (cfName) {
	    free(cfName);
	}
    }
    /*
    ** if the environment variable NFRUID is set we'll suid there.
    */
    {
	char * nfruid;
	if ((nfruid = getenv( "NFRUID" )) != 0) {
	    /* don't worry about failures, we may not be root */
	    if (nfrSetuid( atoi(nfruid) ) > -1) {
		Alert_Args_Generate( NFR_SRC_NFRD,
		    NFR_NFRD_UIDSET, 
		    format( "%d", atoi(nfruid) ), 0 );
	    }
	}
    }
    /*
    ** if the environment variable NFRGID is set we'll set gid there.
    */
    {
	char * nfrgid;
	if ((nfrgid = getenv( "NFRGID" )) != 0) {
	    /* don't worry about failures, we may not be root */
	    if (nfrSetgid( atoi(nfrgid) ) > -1) {
		Alert_Args_Generate( NFR_SRC_NFRD,
		    NFR_NFRD_GIDSET, 
		    format( "%d", atoi(nfrgid) ), 0 );
	    }
	}
    }

    /*
    ** Create the command/control processing stream.
    ** this is done after reading the 
    */
    if (nfrHost && nfrPort) {
	commandInit();
	if (commandSockOpen( nfrHost, nfrPort ) == 0) {
	    Alert_Args_Generate( NFR_SRC_NFRD, NFR_ERR_CANTOPENIPHOST, nfrHost,
		format( "%d", nfrPort ), 0 );
	    exit(1);
	}
	{
	    Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_READFROMSOCK, nfrHost,
		format( "%d", nfrPort ), 0 );
	}
    } else if (nfrHost) {
	char * name;
	commandInit();
	if (commandFifoOpen( name = nfrPath( nfrHost, 0 ) ) == 0) {
	    Alert_Args_Generate( NFR_SRC_NFRD, NFR_ERR_CANTOPENFIFO, nfrHost, 0 );
	    exit(1);
	}
	Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_READFROMFIFO, nfrHost, 0 );
	free(name);
    }


    /*
    ** Get some basic directory names...
    */

    /* varDir -- where we write varibles */
    varDir = nfrPath( varDir, 0 );

    /*
    ** Start the boot process.
    ** hide variable reincarnation behind an option.
    */
    if (bFlag) {
	bootGlobals( varDir , (char *)0 );
    }

    /*
    ** ready to work.
    */
    nfrPid();

    /*
    ** load any of the packages we found.
    */
    {
	int i;
	for (i=0; i<pkgI; ++i) {
	    pkgLoad( /* starting scope */0, pkgs[i], nfrdLoad );
	}
	free(pkgs);
	pkgs = 0;

	/*
	** set the scope back to the global scope...
	*/
	{
	    char * args[2];
	    args[0] = "global";
	    args[1] = 0;
	    comandScopeCmd( -1, args );
	}

    }

    /*
    ** set up to reclaim memory every second.
    ** Since this is called from the main, and since
    ** no .nfr code can suspend execution, this ought
    ** to be safe
    */
    {
	 timeval_t when = util_makeTv( /* sec */ 1, /* usec */ 0 );
	 to_t * to = toNew( &when, TO_FLAGS_SPECIAL | TO_FLAGS_REPEAT, 0, tmpfree );
    }

    /*
    ** set up to sync any nfrd every recorderTimeOut seconds.
    */
    {
	timeval_t when = util_makeTv( /* sec */ recorderTimeOut, /* usec */ 0 );
	to_t * to = toNew( &when, TO_FLAGS_SPECIAL | TO_FLAGS_REPEAT, 0, nfrdSync );      
    }

    {
	timeval_t localTO, now;
	
	localTO = util_makeTv( /* sec */ 0, /* usec */ 0 );
	gettimeofday( &now, 0 );
	
	while (restart == 0) {
	    timeval_t * tv;

	    gettimeofday( &now, 0 );
	    if (tv = toTop() ) {
		while (toCmp( tv, &now ) <= 0) {
		    /*
		    ** note: all of the timely activies are done
		    ** in timeout.  These include (but aren't
		    ** limited to): reassembleGC, intfSample,
		    ** nfrSync, intfSample, etc.
		    */
		    toDispatch();
		    gettimeofday( &now, 0 );
		    tv = toTop();
		}
	    }

	    if (tv) {
		localTO = util_diffTv( tv, &now );
		if (localTO.tv_sec < 0) {
		    /* turn this into an alert in the future */
		    log( FL, "neg timeout %d.%6.6d tv %d.%6.6d diff %d.%6.6d\n",
			    now.tv_sec, now.tv_usec,
			    tv->tv_sec, tv->tv_usec,
			    localTO.tv_sec, localTO.tv_usec);
		}
	    } else {
	       localTO.tv_sec = timeOut / 1000;
	       localTO.tv_usec = (timeOut % 1000) * 1000;
	       printf( "No timeouts.\n" );
	    }

	    pollPerform( util_tv2msec( &localTO ) );

	}
	    /*
	    ** This would be a good place to clear out the tmpmalloc
	    ** of the engine ... there's nothing that can be running
	    ** at this layer.
	    ** XXX
	    */
    }


    pollClose();
    if (nfrHost && nfrPort) {
	commandClose();
    }

    /*
    ** We ought to save our global state once more, just before we exit...
    */

    /*
    ** normal exit.
    */
    Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_STOP, "$header$", argv0, 0 );
    exit(0);
}
