/*
** Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
** All rights reserved.
**
** Use and distribution of this software and its source code
** are governed by the terms and conditions of the
** Network Flight Recorder Software License ("LICENSE.TXT")
*/

#ifndef	_INCL_NFRD_H

#include "sys-include.h"
#include "log.h"
#include "counter.h"
#include "libnfr_err.h"

/*
**	Top-level master definitions file for nfrd. This mostly
** contains definitions of data structures and opcodes for the virtual
** machine interpreter.
**
**	Marcus J. Ranum, Network Flight Recorder, Inc. 1996
*/


/* return codes from compiler */
#define	NFR_COMPILE_EOF		0	/* EOF */
#define	NFR_COMPILE_FILT	1	/* compiled a filter OK */
#define	NFR_COMPILE_FUNC	2	/* compiled a function OK */
#define	NFR_COMPILE_BAD		3	/* syntax error */
#define	NFR_COMPILE_SYSERR	4	/* system error */

#define	NFR_OP_MASK		0xfff	/* provides for 4096 ops. */
#define	NFR_OP(x)		((x)&NFR_OP_MASK)

					/* hide the line number of the .nfr
					** in the op */
#define	NFR_OP_LINE_MASK	0xffff0000
#define	NFR_OP_LINE(x)		(((unsigned)(((x) & NFR_OP_LINE_MASK))) >> 16)

/* pseudo-machine instructions - these are VERY CISC. ;) */
#define	NFR_OP_STOP		0	/* cease execution */
#define	NFR_OP_POP		1	/* discard value from stack */
#define	NFR_OP_NPUSH		2	/* push a number on the stack */
#define	NFR_OP_SPUSH		3	/* push a string on the stack */
#define	NFR_OP_CPUSH		4	/* push parameter count onto stack */
#define	NFR_OP_PPUSH		5	/* push numbered parameter onto stack */
#define	NFR_OP_LPUSH		6	/* push list onto stack */
#define	NFR_OP_NULPUSH		7	/* push a NULL on the stack */
#define	NFR_OP_LASGN		8	/* assign a filter-local value */
#define	NFR_OP_CALL		9	/* call to user-defined function */
#define	NFR_OP_RETNV		10	/* return NO value */
#define	NFR_OP_RETV		11	/* return with value */
#define	NFR_OP_ADD		12	/* add */
#define	NFR_OP_SUB		13	/* subtract */
#define	NFR_OP_MUL		14	/* multiply */
#define	NFR_OP_DIV		15	/* divide */
#define	NFR_OP_NEG		16	/* negate */
#define	NFR_OP_EQ		17	/* negate */
#define	NFR_OP_NE		18	/* negate */
#define	NFR_OP_LT		19	/* less than */
#define	NFR_OP_LTE		20	/* less than/equal to */
#define	NFR_OP_GT		21	/* greater than */
#define	NFR_OP_GTE		22	/* greater than/equal to */
#define	NFR_OP_NOT		23	/* truth inversion (election year) */
#define	NFR_OP_AND		24	/* and */
#define	NFR_OP_OR		25	/* or */
#define	NFR_OP_IF		26	/* if statement */
#define	NFR_OP_IFCOND		27	/* if condition terminator */
#define	NFR_OP_LEVAL		28	/* evaluate a local var */
#define	NFR_OP_PACKET		29	/* look up a part of a packet */
#define	NFR_OP_RECORD		30	/* send a schema to a recorder */ 
#define	NFR_OP_GASGN		31	/* assign a global value XXX */
#define	NFR_OP_GEVAL		32	/* evaluate a global var */
#define NFR_OP_IPV4HOSTPUSH	33	/* push an ip address */
#define NFR_OP_IPV4NETPUSH	34	/* push a ip-net */
#define	NFR_OP_ON		35	/* on condition */
#define	NFR_OP_FUNCON		36	/* on condition */
#define	NFR_OP_OFF		37	/* off the on condition */
#define	NFR_OP_DECL_INSIDE	38	/* declare/inside */
#define	NFR_OP_DECL_GLOBAL	39	/* delcare global */
#define	NFR_OP_FOREACH		40	/* foreach iterator */
#define	NFR_OP_WHILE		41	/* while iterator */
#define	NFR_OP_BREAK		42	/* break iterator exit */
#define	NFR_OP_INSIDE		43	/* binary op INSIDE */
#define	NFR_OP_ARRAYGREF	44	/* array reference, global array */
#define	NFR_OP_ARRAYLREF	45	/* array reference, local array */
#define	NFR_OP_ARRAYEREF	46	/* array reference, from expression */
#define	NFR_OP_ARRAYGASGN	47	/* array assign, global array */
#define	NFR_OP_ARRAYLASGN	48	/* array assign, local array */
#define	NFR_OP_ARRAYEASGN	49	/* array assign, local array */
#define	NFR_OP_ETHMACPUSH	50	/* push a mac address. */
#define	NFR_OP_ECALL		51	/* call a function defined by an expr */
#define	NFR_OP_BOR		52	/* bitwise or */
#define	NFR_OP_BAND		53	/* bitwise and */
#define	NFR_OP_LSHIFT		54	/* left shift */
#define	NFR_OP_RSHIFT		55	/* right shift */
#define	NFR_OP_XOR		56	/* exclusive or */
#define	NFR_OP_INV		57	/* invert bits */


/* run-states - used to halt interpreter in case of error */
#define	NFR_RUN_OK		0	/* alles klar */
#define	NFR_RUN_STACK		1	/* stack overflow, underflow, full */
#define	NFR_RUN_FRAME		2	/* frame overflow, underflow, full */
#define	NFR_RUN_CPU		3	/* cpu ticks exceeded */
#define	NFR_RUN_ILLEGAL		4	/* illegal/unknown instruction */
#define	NFR_RUN_BREAK		5	/* exit foreach/while loop */
#define	NFR_RUN_RETURN		6	/* for returns inside loops */


/* data types: currently stored as a CHAR - limit 254 */
#define	NFR_TYPE_ERR		0
#define	NFR_TYPE_INT		1
#define	NFR_TYPE_STR		2
#define	NFR_TYPE_LIST		3	/* bucket class */
#define NFR_TYPE_RECORDER	4	/* recorder type */
#define NFR_TYPE_SCHEMADATA	5	/* pass around schema data */
#define	NFR_TYPE_IPV4HOST	6	/* one specific host number */
#define	NFR_TYPE_IPV4NET	7	/* a net description */
#define	NFR_TYPE_BLOB		8	/* blob type */
#define	NFR_TYPE_ETHMAC		9	/* ethernet mac type */
#define	NFR_TYPE_MAPBAG		10	/* map bag. doesn't get passed out of engine */
#define	NFR_TYPE_MAP		11	/* map. doesn't get passed out of engine */
#define	NFR_TYPE_INDIRECT	12	/* val.v points to another Sym * */
#define	NFR_TYPE_PATTERN	13	/* returns from regcomp */
#define	NFR_TYPE_TRIGGER	14	/* trigger values (for off statement) */


/* error codes */
#define	NFR_ERR_NONE		0	/* unspecified error */
#define	NFR_ERR_USER		1	/* user specified error */
#define	NFR_ERR_STACK		2	/* stack over/underflow */
#define	NFR_ERR_FRAME		3	/* frame shortage */
#define	NFR_ERR_ZDIV		4	/* division by zero */
#define	NFR_ERR_NONUM		5	/* numeric operation on non-number */
#define	NFR_ERR_OOM		6	/* lack of memory */
#define	NFR_ERR_NOVAR		7	/* undefined variable */
#define	NFR_ERR_NOFUNC		8	/* no such function or builtin */
#define	NFR_ERR_BADVAL		9	/* cannot extract packet value */
#define	NFR_ERR_ARGUMENT	10	/* bad argument(s) to function/builtin */
#define	NFR_ERR_PROTOCOL	11	/* not the correct protocol */
#define	NFR_ERR_NULL		12	/* null */

/* common #defines */

#define	min(a,b)	( ((a) < (b)) ? (a) : (b) )
#define	max(a,b)	( ((a) > (b)) ? (a) : (b) )

typedef struct timeval timeval_t;


/* representation for a net description. */

typedef struct {
	struct	in_addr	ip_addr;
	int		ip_mask;
} ipv4net_t;
/* a virtual stack memory data representation */


typedef struct blob {
	short		b_refCnt;
	short		b_flags;
	char *		b_data;
	int		b_len;
} blob_t;

#define	BLOB_FLAGS_MALLOC	1	/* b_data was alloc'd */
#define	BLOB_FLAGS_STATIC	2	/* don't free this thing. */

typedef struct ethMac {
	char 		e_addr[6];
} ethMac_t;

typedef	struct config {
	char *		c_name;
	char *		c_value;
} config_t;


typedef struct recorder {
    struct recorder *	r_next;		/* list of them */
    int			r_refCnt;	/* reference count */
    int			r_fd;		/* socket interface */
    int			r_pid;		/* process id */
    FILE *		r_file;		/* stream associated with r_fd */
    int			r_dirtyCnt;	/* number of schemas written */
    int			r_cnt;		/* total number of schemas written */
    counter		r_bytes;	/* total number of bytes written */
    int			r_errCnt;	/* total times we didn't write due to an err */
    char *		r_schema;	/* if the schema for this recorder is named */
    config_t **		r_configs;	/* recorder configuration */
    time_t		r_epoch;	/* when it was last fork/exec'd */
    char **		r_argv;		/* arguments to start recorder */
    int			r_restarts;	/* # of times the recorder started */
} recorder_t;


typedef struct pattern {
    char *		p_str;		/* the actual pattern */
    regex_t		p_regex;	/* the compiled pattern */
} pattern_t;

/*
** Some time later, we'll flesh this type out...
*/
typedef struct schemaData {
	int dummy;	/* have to have something for some compilers */
} schemaData_t;



/*
** --------------------
*/

typedef	struct {
	union	{
		char		*c;
		int		i;
		long		l;
		ipv4net_t	*n;	/* net */
		struct in_addr	h;	/* host */
		ethMac_t	e;	/* eth address */
		blob_t 		*b;	/* blob */
		pattern_t	*p;	/* pattern */
		struct trigger	*t;	/* trigger */
		void		*v;
	} val;
	char	typ;	/* beware the max ! */
			/* cause of structure alignment, 
			** making the typ a short ought to
			** leave the sizeof(MemVal) intact */
} MemVal;

/*
** Sym
** symbols -- lists of name/value pairs.
**
** The values of symbols are always MemVals.  The names
** can take on a variety of styles.  In the simplest case,
** the name can be a char*, with malloc'd data.  If the
** symbol's name length can fit directly into the same
** pointer space, we place the name directly into the 
** structure (instead of the pointer).  For more complicated
** situations (arrays), we use other kinds of names.  The
** array index becomes the name of the symbol, and the MemVal
** represents the value.  The index'es of arrays can be
** any MemVal, and we try to optimize a few of the NFR
** types.  For example, strings are treated like symbol
** names, and use the same length optimzation.  Ints are 
** stored directly into the same place as the pointer,
** and as a last resort, MemVal references are also
** stored as indicies.
*/

#define	SYM_NAME_UMAX	(sizeof(int))	/* max length of the local symbol */

typedef struct sym {
	union {
	    char		*cp;	/* symbol name */
	    MemVal		*m;	/* memval */
	    void		*v;	/* blob */
	    struct in_addr	h;	/* host */
	    int			i;	/* integer */
	    char		c[SYM_NAME_UMAX];
	} s_Name;
	MemVal		s_data;		/* associated value */
	short		s_flags;	/* see below */
	u_short		s_length;	/* if nonzero, length of the name */
	struct	sym *	s_next;
} Sym;

#define SYM_FLAG_DIRTY		0x100
#define SYM_FLAG_PICKLEERROR	0x200

#define	SYM_FLAG_TYPE(x)	(((x)->s_flags) & 0xff)
#define	SYM_FLAG_TYPE_SET(x,t)	((x)->s_flags = ((x)->s_flags & 0xff00) | (t & 0xff))

#define	SYM_FLAG_TYPE_NAME	0
#define	SYM_FLAG_TYPE_INT	1
#define	SYM_FLAG_TYPE_MEMVAL	2
#define	SYM_FLAG_TYPE_HOST	3
#define	SYM_FLAG_TYPE_VOID	4

/*
** solaris like using s_host for something.
*/
#ifdef	s_host
#undef	s_host
#endif

#define	s_name		s_Name.cp
#define	s_memVal	s_Name.m
#define	s_void		s_Name.v
#define s_int		s_Name.i
#define s_host		s_Name.h
#define s_chars		s_Name.c

#define	SYM_LENGTH_MAX	((unsigned)0xffff)

/*
** map.
** We collect together lists of sym's based on some 'index'
** currently, the index is just a small positive integer.
*/

typedef struct map {
	struct map *	m_next;	
	int		m_index;	/* later, maybe a MemVal. */
	char *		m_name;		/* could be null. */
	Sym *		m_sym;	
	int		m_refCnt;
} Map;

typedef struct mapBag {
	int	mb_refCnt;	/* reference counting */
	int	mb_width;	/* width of the hash */
	Map **	mb_maps;	/* reference to lists */
} MapBag;

/*
** trigger.
** we reuse the sym structure cause it looks like that...
*/

typedef struct trigger {
    char *	t_func;		/* name of function to call */
    char *	t_scope;	/* scope of the functin to call */
    int		t_ndx;		/* current index (used while building) */
    int	 	t_max;		/* max index of the t_whom/t_nv arrays */
    char **	t_whom;		/* array of names to associate with the n-v data */
    Sym **	t_nv;		/* name value pairs. */
    int		t_flags;	/* flags associated with the trigger */

    void	(*t_free)(struct trigger *);	/* routine to call to remove the trigger */
    void *	t_rock;		/* keep track of stack's info */
    int		t_refCnt;	/* keep track of how many are point'n at it */
} trigger_t;

#define	T_FLAGS_FUNC	1	/* trigger created inside filter trigger clause */


typedef (*mapCallout_t)( Map *, void *, char * );

/* list elements (for sets) */
typedef	struct	listmem {
	MemVal		data;
	struct	listmem	*next;
} ListMem;

/* userIntf -- connection to active telnet session */

typedef struct userintf {
	int	ui_trace;	/* the user wants to trace instructions */
	int	ui_refCnt;	/* keep track of how many things know about us */
	int	ui_fd;		/* output file descriptor */
	FILE *	ui_file;	/* output stream associated with fd */
} Userintf_t;


/* an assembled filter or function (many per nfrd) */
typedef	struct	{
	int		*f_mem;		/* compiled instructions */
	int		f_memsiz;	/* memory size */
	ListMem		**f_list;	/* list and external object table */
	int		f_listsiz;	/* list table size */
	Sym		*f_syms;	/* filter symbol table (unhashed) */
	int		f_indSymCnt;	/* number of INDIRECTs on the SYM list */
	char		*f_name;	/* filter name (in string table) */
	trigger_t	*f_trigger;/* reference to trigger associated with this theng */
	Userintf_t	*f_ui;	/* pointer to user interface */
	char 		*f_scope;	/* scope related to the func/filter */
	int		f_scopeNdx;	/* scope index number */
	int		f_useCount;	/* # of times called */
	timeval_t	f_runTime;	/* run time. */
	double		f_sumSq;	/* sum/sq of run time */
	double		f_stdDev;	/* stdDev of run time */
	char *		f_fromFile;	/* name of the file this was read from */
} Filter;


/* a stack frame for user-defined function calls. */
typedef	struct	{
	int	ret_sp;		/* where stack pointer should wind up */
	int	ret_ac;		/* # of args */
	int	ret_break;	/* frame allows break */
} Frame;


/* main runtime environment (one per nfrd) */
typedef	struct	{
	Frame	*m_fram;	/* funct call frames */
	MemVal	*m_mem;		/* machine memory */
	int	m_fp;		/* frame pointer */
	int	m_maxfp;	/* frame size */
	int	m_sp;		/* stack pointer */
	int	m_maxsp;	/* stack size */
} Machine;


/* builtin function table entries (see bltin.c and bltin directory) */
typedef struct  {
        char    *nam;
        void     (*func)();
} Bltin;


/* filter ring */
typedef	struct	filterlist {
	Filter			*f;
	struct	filterlist	*next;
} FilterList;

/* interface for poll routines */
typedef (*pollCall_t)( int fd, void * rock ); /* what gets called when we have input */

/*
** --------------------
** timeout structure.
** it appears here so that routines
** can create timly ops.
*/

typedef	void (*to_special_t)( void * );	/* what gets called when we dispatch a timeout */

typedef struct to {
    struct to *		to_next;
    timeval_t		to_tv;		/* value we're counting down */
    timeval_t		to_tvOrig;	/* original time offset */
    int			to_flags;	/* bits, and special ops */
    union to_trigSpecial {		/* either a trigger or a func call. */
	trigger_t *	to_Trigger;	/* if related to a trigger */
	to_special_t	to_Special;	/* if a function call. */
    } to_trigSpecial;
} to_t;

#define TO_FLAGS_SPECIAL	0x1	/* not a trigger, but instead, call the special */
#define TO_FLAGS_ONCE		0x2	/* only once, not over and over */
#define TO_FLAGS_REPEAT		0x4	/* over and over and over and over */

#define	to_trigger	to_trigSpecial.to_Trigger
#define	to_special	to_trigSpecial.to_Special

/*
** internal structure used to help in hashing array indexes.
**
*/
typedef struct arrayRef {
    char *		ar_base;	/* char * of where to start hash */
    int			ar_len;		/* #char's to hash */
    int			ar_hash;	/* hash value */
} arrayRef_t;

/* soft limit on runtime stack growth */
#ifndef	NFR_LIMIT_STACK
#define	NFR_LIMIT_STACK 4096
#endif


/* soft limit on runtime CPU ticks for a given filter call */
#ifndef	NFR_LIMIT_CPUTICK
#define	NFR_LIMIT_CPUTICK 10000
#endif

/* handy */
#define	MILLION 	1000000
#define	HIGH_UINT	0xffffffff


/* soft limit on runtime call stack frames for a given filter call */
#ifndef	NFR_LIMIT_FRAME
#define	NFR_LIMIT_FRAME 4096
#endif

/* our trivial buffering scheme, so we don't get stuck on read(1).  */
typedef struct buff {
    char *      b_data;			/* buffer location */
    int         b_len;			/* how much data in the buffer */
    int         b_loc;			/* where data starts in the buffer */
    int         b_size;			/* malloc'd size of b_data */
} buff_t;

/* que -- simple doubly linked list */
/* i'd prefer to use queue_t, but streams already use that name :^( */

typedef struct que {
    struct que *  q_forw;
    struct que *  q_back;
    void *          q_data;
} que_t;


/*
** Any include files which need some of the nfrd delcarations...
*/
#include "intf.h"
#include "pcap.h"
#include "alert.h"
#include "cf.h"
#include "nfr_res.h"

extern double sqrt(double);


extern	void	panic(char *, ...);			/* util.c */
extern	timeval_t util_diffTv( timeval_t *, timeval_t *);/* util.c */
extern	timeval_t util_makeTv( int, int);		/* util.c */
extern	void	util_incrTv( timeval_t *, timeval_t *);	/* util.c */
extern	double	util_tv2usec( timeval_t * );		/* util.c */
extern	char *	format(char * fmt, ... );		/* util.c */
extern	double	util_stdDev(double sum, double sumSq, int cnt );/* util.c */
extern	char *	util_str( char * s );			/* util.c */
extern	char *	util_err2str( int i );			/* util.c */
extern	char *	util_ntoa( int );			/* util.c */

extern	Sym	*symlook(char *,Sym *);			/* sym.c */
extern	Sym	*symnew(char *,MemVal *);		/* sym.c */
extern	Sym	*symadd(Sym *,Sym *);			/* sym.c */
extern	Sym	*symremove(Sym *,Sym *);		/* sym.c */
extern	char	*symName(Sym *);			/* sym.c */
extern	int	symCmp( Sym *, char * );		/* sym.c */
extern	void	symfreelist(Sym *);			/* sym.c */
extern	MapBag	*globalMapBag;				/* sym.c */

extern	MapBag	*arraySet( MemVal *, MemVal *, MemVal * ); /* sym.c */
extern	Sym	*arrayGet(  MemVal *, MemVal *, arrayRef_t * );/* sym.c */

extern	Filter	*funcfind(char *name, char *scope);	/* sym.c */
extern	int	funcadd(Filter *);			/* sym.c */
extern	int	funcdelete(Filter *);			/* sym.c */
extern	void	functable_init();			/* sym.c */

extern	Map *	mapNew(Map * n, int i, char * ni,Sym * s); /* sym.c */
extern	void	mapDecRef( void * );			/* sym.c */
extern	void	mapFree(Map * n);			/* sym.c */

extern	int	mapBagBigger( MapBag * );		/* sym.c */
extern	MapBag*	mapBagNew( int width );			/* sym.c */
extern	int	mapBagHash( MapBag *, int index );	/* sym.c */
extern	void	mapBagDecRef( void * );			/* sym.c */
extern	void	mapBagFree( MapBag * );			/* sym.c */
extern	void	mapBagAdd( MapBag *, Map * );		/* sym.c */
extern	MapBag *mapBagRenew( MapBag *, int width );	/* sym.c */
extern	Map *	mapBagFind( MapBag *, int ndx, char * name );/* sym.c */
extern 	int	strHash( char * s );			/* sym.c */

extern	void	setyyinput(FILE *);			/* parse.y */
extern	int	yyparse();				/* parse.y */
extern	Filter	*get_compiled_filter();			/* parse.y */
extern	void	clear_compiler_error();			/* parse.y */
extern	int	perform( char * );			/* parse.y */
extern	trigger_t *triggerNew();			/* parse.y */
extern	char *	scope;					/* parse.y */
extern	int 	scopeCnt;				/* parse.y */
extern	char *	globalScope;				/* parse.y */
extern	int 	globalScopeNdx;				/* parse.y */
extern	Filter *globalFilter;				/* parse.y */

extern	Machine	*machinenew();				/* mem.c */
extern	void	machinefree( Machine * );		/* mem.c */
extern	void	filtfree(Filter *);			/* mem.c */
extern	void	listfree(ListMem *);			/* mem.c */
extern	void	memvalfreedata(MemVal *);		/* mem.c */
extern	int	memvalclonedata(MemVal *, MemVal *);	/* mem.c */
extern	MemVal *memTmpNew( int type, void * value );	/* mem.c */
extern	char *  tmpstrdup( char * );			/* mem.c */
extern	void    *tmpalloc(int);				/* mem.c */
extern	void    *tmpcalloc(int);			/* mem.c */
extern	void	tmpputonfree(void *);			/* mem.c */
extern	void	tmpfree();				/* mem.c */
extern	blob_t*	blobNew( char *, int, int );		/* mem.c */
extern	blob_t*	blobSet( blob_t *, char *, int, int, int );/* mem.c */
extern	void	blobFree( void * );			/* mem.c */
extern	blob_t*	blobIncRef( blob_t * );			/* mem.c */
extern	void	blobDecRef( void * );			/* mem.c */

extern	char *	scopeSet(char * s); 			/* run.c */
extern	char *	scopeGet( );				/* run.c */
extern	int	machinerun(Machine *,Filter *,int);	/* run.c */
extern	void	machinereset(Machine *);		/* run.c */
extern	Filter *runCurrentGet();			/* run.c */
extern	Filter *runCurrentSet( Filter * f );		/* run.c */
extern	int	runCurrentOp;				/* run.c */
extern 	Sym*	run_globalLookup( char * name );	/* run.c */
extern 	Sym*	run_lookup( Filter * f, char * s, int op ); /* run.c */
extern	int	run_globalAssign( char * name, MemVal * val);/* run.c */

extern	Bltin	*bltlookup(char *);			/* bltin.c */

extern	int	packet_var(char *, char *);		/* packet.c */
extern	int	packet_get(int, MemVal *);		/* packet.c */

extern	Userintf_t * uiIncr( Userintf_t * ui );		/* user.c */
extern	Userintf_t * uiSetCurr( Userintf_t * ui );	/* user.c */
extern	Userintf_t * uiCurr( ); 			/* user.c */
extern	Userintf_t * uiFind( int ); 			/* user.c */
extern	int	 uiCurrPrintf( char *, ... );		/* user.c */

extern  char **crack( char * );                         /* recorder.c */

extern  buff_t * buffNew( char * data, int len );       /* buffer.c */
extern  void  buffFree( buff_t * );                     /* buffer.c */
extern  char * getsBuff( buff_t * );                    /* buffer.c */
extern  int getcBuff( buff_t * );                       /* buffer.c */

extern	MemVal *	unpickle( FILE * );		/* unpickle.c */

extern int	packetSec(MemVal *);			/* packet.c */
extern int	packetUsec(MemVal *);			/* packet.c */
extern int	packetLen(MemVal *);			/* packet.c */
extern int	packetNonce(MemVal *);			/* packet.c */
extern int	packetBlob(MemVal *);			/* packet.c */
extern int	ethType(MemVal *);			/* packet.c */
extern int	eth8023(MemVal *);			/* packet.c */
extern int	ethLen(MemVal *);			/* packet.c */
extern int	ethSrc(MemVal *);			/* packet.c */
extern int	ethDst(MemVal *);			/* packet.c */
 
extern int	ipSrc(MemVal *);			/* packet.c */
extern int	ipDst(MemVal *);			/* packet.c */
extern int	ipProto(MemVal *);			/* packet.c */
extern int	ipIsIp(MemVal *);			/* packet.c */
extern int	ipLen(MemVal *);			/* packet.c */
extern int	ipLenError(MemVal *);			/* packet.c */
extern char *	ipEnd();				/* packet.c */
extern struct ip * ipHdr();				/* packet.c */

extern int	scopeName(MemVal *);			/* packet.c */
extern int	scopeIndex(MemVal *);			/* packet.c */

extern int     icmpIsIcmp(MemVal *);			/* packet.c */
extern int     icmpIsHdr(MemVal *);			/* packet.c */
extern int     icmpType(MemVal *);			/* packet.c */
extern int     icmpCode(MemVal *);			/* packet.c */
extern int     icmpCksum(MemVal *);			/* packet.c */
extern int     icmpBlob(MemVal *);			/* packet.c */

extern int     udpIsUdp(MemVal *);			/* packet.c */
extern int     udpIsHdr(MemVal *);			/* packet.c */
extern int     udpSport(MemVal *);			/* packet.c */
extern int     udpDport(MemVal *);			/* packet.c */
extern int     udpBlob(MemVal *);			/* packet.c */

extern int     tcpIsTcp(MemVal *);			/* packet.c */
extern int     tcpIsHdr(MemVal *);			/* packet.c */
extern int     tcpSport(MemVal *);			/* packet.c */
extern int     tcpDport(MemVal *);			/* packet.c */
extern char *	tcpEnd();				/* packet.c */
extern struct tcphdr * tcpHdr();			/* packet.c */
extern int	tcpStream(MemVal *);			/* tcpip.c */
extern int	tcpBlob(MemVal *);			/* tcpip.c */
extern int	tcpBytes(MemVal *);			/* tcpip.c */
extern int	tcpLength(MemVal *);			/* tcpip.c */

extern int	tcpConn(MemVal *);			/* tcpip.c */
extern int	tcpConnEpoch(MemVal *);			/* tcpip.c */
extern int	tcpConnPktSec(MemVal *);		/* tcpip.c */
extern int	tcpConnHash(MemVal *);			/* tcpip.c */
extern int	tcpConnSrc(MemVal *);			/* tcpip.c */
extern int	tcpConnDst(MemVal *);			/* tcpip.c */
extern int	tcpConnSport(MemVal *);			/* tcpip.c */
extern int	tcpConnDport(MemVal *);			/* tcpip.c */
extern int	tcpConnForwBytes(MemVal *);		/* tcpip.c */
extern int	tcpConnBackBytes(MemVal *);		/* tcpip.c */
extern int	tcpConnForwLost(MemVal *);		/* tcpip.c */
extern int	tcpConnBackLost(MemVal *);		/* tcpip.c */
extern int	tcpConnSave(MemVal *);			/* tcpip.c */
extern int	tcpConnSyn(MemVal *);			/* tcpip.c */
extern int	tcpConnSynAck(MemVal *);		/* tcpip.c */
extern int	tcpConnRst(MemVal *);			/* tcpip.c */
extern int	tcpConnFin(MemVal *);			/* tcpip.c */

extern int	tcpConnMap(MemVal *);			/* tcpip.c */
extern int	tcpConnSym(MemVal *);			/* tcpip.c */

extern que_t * queNew( void * d );			/* queue.c */
extern que_t * queAdd( que_t * q, que_t * n );		/* queue.c */
extern que_t * queDel( que_t * d );			/* queue.c */
extern void queFree( que_t * f );			/* queue.c */
extern que_t * queForw( que_t * q );			/* queue.c */
extern que_t * queBack( que_t * q );			/* queue.c */

extern char * nfrHome(char *);				/* home.c */
extern char * nfrPath(char *, ... );			/* home.c */

extern	int	listlen( ListMem *);			/* bltin-list.c */
extern	char *	liststrcat( ListMem * list, char * str );/* bltin-list.c */
extern	int	liststrlen( ListMem * list );		/* bltin-list.c */
extern	char *	listblobcat( ListMem * list, char * str );/* bltin-list.c */
extern	int	listbloblen( ListMem * list );		/* bltin-list.c */
extern	char *	listbytescat( ListMem * list, char * str );/* bltin-list.c */
extern	int	listbyteslen( ListMem * list );		/* bltin-list.c */

extern  pattern_t * patternNew( char * str );		/* bltin-regex.c */

extern	void 	triggerDecRef( void * /* liar! */ );	/* trigger.c */

extern	timeval_t * toTop();				/* timeout.c */
extern	int toCmp( timeval_t *, timeval_t *);		/* timeout.c */
extern	to_t * toNew( timeval_t *, int, trigger_t *, to_special_t );

extern	void	ipInit();				/* tcpip.c */
extern	void	intfInit();				/* intf.c */

extern	void	symInit();				/* sym.c */
extern	void	mapBagIterate( MapBag *, mapCallout_t, void *, char * ); /* sym.c */

extern	void	recorderInit( );			/* recorder.c */
extern	int	recorderUidSet( int );			/* recorder.c */
extern	void	recorderDecRef( void * );		/* recorder.c */

extern int nfrAlphasort( const void * d1, const void *  );/* scandir.c */
extern int nfrScandir( char * , struct dirent ***,
	int (*select)(struct dirent *),
	int (*compare)( const void *, const void * ) );

extern int pkgLoad( char *, char *, int (*perform)( char *, char * ) );/* pkgLoad.c */
extern int cfgSelect( struct dirent * );

extern int fileType( char * path );			/* file.c */

extern char * nfrCfGetVal( char * );			/* nfrd.c */

#define	_INCL_NFRD_H
#endif
