/*
** Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
** All rights reserved.
**
** Use and distribution of this software and its source code
** are governed by the terms and conditions of the
** Network Flight Recorder Software License ("LICENSE.TXT")
*/

/*
** pace pcap.
**
** Read a libpcap file, provide a fifo, which can be read...
** which has the pcap data correctly paced put into the fifo.
**
*/

#include "nfrd.h"

#ifndef	DEFAULT_PACE_FIFO
#define	DEFAULT_PACE_FIFO	"~/etc/run/pace.fifo"
#endif

char * argv0;

int dFlag;
int vFlag;
int tFlag;

/*
** ----------------------------------------
**
*/

int	fifoFd = -1;			/* service port */
int	fifoOutFd = -1;			/* accept fd */

fifoOpen( name )
char * name;
{
    struct sockaddr_un un;

    /* just in case its been there before */
    if (unlink(name) == 0) {
	log( FL, "fifoOpen: old <%s> removed\n", name );
    }

    fifoFd = socket( AF_UNIX, SOCK_STREAM, 0 );
    if (fifoFd < 0) {
	/* error */
	log( FL, "fifoOpen: can't create socket %s\n", name );
	return(0);
    }
    if (strlen(name)+1 > sizeof(un.sun_path)) {
	/* error */
	log( FL, "fifoOpen: name too long %s\n", name );
	return(0);
    }

    un.sun_family = AF_UNIX;
    strcpy(  un.sun_path, name );

    if (bind( fifoFd, (struct sockaddr *)&un, sizeof(un) ) < 0) {
	log( FL, "fifoOpen: can't bind to %s\n", name );
	perror("bind");
	return(0);
    }

    if ((listen( fifoFd, 255 )) < 0) {
	log( FL, "fifoOpen: can't listen(255) %s\n", name );
	perror("listen");
	return(0);
    }

    {
	int len = sizeof(un);
	if ((fifoOutFd = accept( fifoFd, (struct sockaddr *)&un, &len )) < 0) {
	    log( FL, "fifoOpen: accept failed\n" );
	    perror( "accept" );
	    return(0);
	}
    }

    log( FL, "fifoOpen: fd %d\n", fifoFd );
    return(1);
}

/*
** --------------------
**
*/

fifoClose( name )
char * name;
{
    if (name) {
	unlink(name);
    }
    close(fifoFd);
    close(fifoOutFd);
}

/*
** --------------------
**
** packetProcess.
*/

struct timeval oldTimer;		/* last time we saw */

void
packetProcess( v, pkt, pk )
u_char * v;
struct pcap_pkthdr * pkt;
const u_char *pk;
{
    /*
    ** view the timeout.
    */
    if (vFlag) {
	printf( "%d.%06d len %d (%x %x %x)\n", 
	    pkt->ts.tv_sec, pkt->ts.tv_usec, pkt->len,
	    pkt, ((char *)pkt)+pkt->len, pk );
    }

    if (oldTimer.tv_sec == 0) {
	/* don't wait. */
    } else {
	/*
	** compute the differenence ... in seconds, and in usec...
	*/
	struct timeval wait;

	wait.tv_sec = pkt->ts.tv_sec - oldTimer.tv_sec;
	wait.tv_usec = pkt->ts.tv_usec - oldTimer.tv_usec;
	if (wait.tv_usec < 0) {
	    wait.tv_usec += 1000000;
	    wait.tv_sec--;
	}
	if (vFlag) {
	    printf( "Wait... %d.%6d\n", wait.tv_sec, wait.tv_usec );
	}
	/*
	** Use a select to wait the difference.
	*/
	{
	    int err = select( /*n*/ 0, /*rFds*/0, /*wFds*/0, /*xFds*/0, &wait );
	    if (err < 0) {
		perror("select");
	    }
	}
	if (vFlag) {
	    printf( "Wait... done\n" );
	}
    }
    oldTimer = pkt->ts;

    if (vFlag) {
	printf( "sending packet.\n" );
    }

    /*
    ** modify the sec/usec values to reflect more
    ** realistic values.... times which are closer to `now'.
    ** This isn't perfect, cause the pace differences aren't
    ** really maintained, but to get that, you don't wanna use -t
    */
    if (tFlag) {
	gettimeofday( &pkt->ts, 0 );
    }

    if (fifoOutFd) {
	write( fifoOutFd, pkt, sizeof(*pkt) );
	write( fifoOutFd, pk, pkt->len );
    }
}

/*
** --------------------
**
** main.
**
*/

char *	fifoName = 0;
char *	freeName = 0;
char *	logFile = 0;
char *	pcapFile;

pcap_t * pcap;


main( argc, argv )
int argc;
char * argv[];
{
    argv0 = argv[0];

    {
	int i;
	for (i=1; i<argc; ++i) {
	    if (argv[i][0] == '-') switch(argv[i][1]) {
		case 'l':	/* how to log. */
		    if (argv[i][2]) {
			logFile = &argv[i][2];
		    } else {
			logFile = argv[++i];
		    }
		    break;

		case 'd':	/* debug */
		    dFlag++;
		    break;
		
		case 'V':	/* verbose */
		    vFlag++;
		    break;

		case 't':	/* update time on each packet 'for ya write it */
		    tFlag++;
		    break;

		default:
		usage:
		    Usage:
		    fprintf( stderr, "usage: pace-pcap file fifo-path\n" ); 
		    exit(1);
	    } else {
		if (pcapFile == 0) {
		    pcapFile = argv[i];
		} else if (fifoName == 0) {
		    fifoName = argv[i];
		}

	    }
	}
    }

    if (pcapFile == 0) {
	goto Usage;
    }

    logOpen( logFile );

    log( FL, "pace-pcap. $header$\n" );

    /*
    ** open the saved pcap file... then schedule the writer...
    ** provide the fifo... wait for the connect...
    */

    {
	char error[256];
	pcap = pcap_open_offline( pcapFile, error );
	if (pcap == 0) {
	    log( FL, "pace-pcap: can't open pcapFile %s\n", pcapFile );
	    exit(1);
	}
    }

    if (fifoName == 0) {
	fifoName = nfrPath( DEFAULT_PACE_FIFO, 0 );
	log( FL, "pace-pcap: using default <%s>\n", fifoName );
	freeName = fifoName;
    }

    if (fifoName && (fifoOpen( fifoName ) == 0)) {
	log( FL, "pace-pcap: can't create fifo\n" );
	exit(1);
    }

    /*
    ** Accept is now completed... now start up the pcap reading...
    */
    pcap_dispatch( pcap, -1, packetProcess, (u_char *)fifoOutFd );


    /*
    ** normal exit.
    */

    fifoClose( fifoName );
    if (freeName) {
	free(freeName);
    }
    exit(0);
}

mark() {}
