/*
** Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
** All rights reserved.
**
** Use and distribution of this software and its source code
** are governed by the terms and conditions of the
** Network Flight Recorder Software License ("LICENSE.TXT")
*/


/*
** pace.c
**
** Support to read a paced fifo file... from pace-pcap
** Allows timely recorded data to get pushed into the
** engine.  Good for debugging the engine or .nfr's or
** for demo-ing particular features.
*/

#include "nfrd.h"

typedef struct pace {
	char *		p_name;		/* name of the thing opened. */
	int		p_fd;		/* file descriptor of fifo */
	buff_t *	p_buff;		/* where we keep partial data. */
	int		p_pkts;		/* number of packets. */
    	counter		p_bytes;	/* number o bytes */
} pace_t;

que_t * paceList;

int	paceTotalPackets = 0;

/*
** ------------------------------------------------------------
** paceNew
**
*/

pace_t *
paceNew( name )
char * name;
{
    pace_t * new = (pace_t *)calloc( sizeof(*new), 1 );
    new->p_name = strdup(name);
    {
	que_t * q = queNew(new);
	if (queAdd( paceList, q ) == 0) {
	    paceList = q;
	}
    }
    return(new);
}


/*
** ------------------------------------------------------------
** paceDel
*/

void
paceDel( q )
que_t * q;
{
    pace_t * ip = (pace_t *)q->q_data;
    if (paceList == q) {
	paceList = queForw(q);
    }
    queFree(q);
    free(ip->p_name);
    free(ip);
}

/*
** ------------------------------------------------------------
**
** paceInit
**
**
*/

int
paceInit( s )
char * s;
{
    return(0);
}

/*
** ------------------------------------------------------------
** paceRead.
** read from the fifo... should be at least one packet in the buffer.
** Use the buffer mechanism
*/

paceRead( fd, rock )
int fd;
void * rock;
{
    pace_t * p = (pace_t *)rock;
    buff_t * b = p->p_buff;
    int err = fillBuff( p->p_buff, fd );

    if (err < 0) {
	Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_PACEDONE, p->p_name, 0 );
	close(fd);
	pollUnCall(fd);
	return;
    }

    /*
    ** Look for as many packets as we can find in the
    ** buffer, then yank them out... 
    */
    {
	int d = (b->b_len - b->b_loc);
	while (d > sizeof(struct pcap_pkthdr)) {
	    struct pcap_pkthdr pkh[1];
	    memcpy( pkh, &b->b_data[b->b_loc], sizeof(pkh) );
	    if (d >= pkh->len) {
		/*
		** deliver a packet, move the buff loc...
		*/
		p->p_pkts++;
		p->p_bytes = counter_add( p->p_bytes,
			counter_from_int(pkh->len) );
		packetRecv( pkh, &b->b_data[b->b_loc+sizeof(pkh)] );
		b->b_loc += pkh->len + sizeof(pkh);
		paceTotalPackets++;
	    } else {
		/*
		** fillBuff will deal with aligning the newly read data from
		** the inbound fifo.
		*/
		return;
	    }
	    d = (b->b_len - b->b_loc);
	}
    }
}

/*
** ------------------------------------------------------------
**
**
*/

void *
paceOpen( fifoName )
char * fifoName;
{
    struct sockaddr_un un;
    int fd;

    fd = socket( AF_UNIX, SOCK_STREAM, 0 );
    if (fd < 0) {
	/* error */
	Alert_Args_Generate( NFR_SRC_NFRD, NFR_ERR_SOCKET, 0 );
	return(0);
    }
    if (strlen(fifoName)+1 > sizeof(un.sun_path)) {
	Alert_Args_Generate( NFR_SRC_NFRD,  NFR_ERR_FIFONAMETOOLONG, fifoName, 0 );
	return(0);
    }

    un.sun_family = AF_UNIX;
    strcpy(  un.sun_path, fifoName );

    if (connect( fd, (struct sockaddr *)&un, sizeof(un) ) < 0) {
	Alert_Args_Generate( NFR_SRC_NFRD, NFR_ERR_CONNECT, fifoName, 0 );
	return(0);
    }

    {
    pace_t * p = paceNew( fifoName );
    p->p_fd = fd;
    p->p_buff = buffNew( 0, 32*1023 );	/* enough for a few big packets */

    Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_NEWPACE, fifoName, 0 );
    pollCall( p->p_fd, paceRead, (void *)p );
    return(p);
    }

}

/*
** --------------------
**
*/

int
paceClose( v )
void * v;
{
    pace_t * p = (pace_t *)v;
    pollUnCall( p->p_fd );


    if (p->p_fd) {
	close(p->p_fd);
	p->p_fd = 0;
	free(p);
	return(0);
    } else {
	return(1);
    }
}

/*
** ------------------------------
**
*/

void
paceStats()
{
    uiCurrPrintf( "pace: totalPackets %d\n", paceTotalPackets );
    if (paceList) {
	que_t * q = paceList;
	do {
	    pace_t * p = (pace_t *)q->q_data;
	    {
		char cbuf[COUNTER_TEXT_SIZE];
		counter_itoa( cbuf, p->p_bytes );
		uiCurrPrintf( "pace: %s: packets %d\n", p->p_name, p->p_pkts);
		uiCurrPrintf( "pace: %s: bytes %s\n", p->p_name, cbuf );
	    }
	    q = q->q_forw;
	} while(q != paceList);
    }
}
