/*
** Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
** All rights reserved.
**
** Use and distribution of this software and its source code
** are governed by the terms and conditions of the
** Network Flight Recorder Software License ("LICENSE.TXT")
*/


#include "nfrd.h"
/*
** Packet value lookup routines. We check here for the type
** and value to be looked for. This assumes packet pre-processing
** as necessary has already been performed.
**
** 	Marcus J. Ranum, Network Flight Recorder, Inc. 1996
*/


/*
** global variables... describing the current packet...
** note: we use the nonce to figure out when to recreate
** state for some of the packet components. 
*/

char *			packet_data;
int    			packet_len;
struct pcap_pkthdr *	packet_hdr;
int			packet_nonce;

timeval_t		packet_arr;		/* last packet arrival */
timeval_t		packet_arrDiff;		/* last packet arrival diff */
double			packet_arrSumSq;	/* sum of square's of rate */
double			packet_arrStdDev;	/* stdDev of arrival rate */

/*
**
** --------------------------------------
**
**
*/

static	int	systime(MemVal *);

typedef	struct	{
	char	*proto;
	char	*val;
	int	(*func)();
} Ptab;


/*
** Handy, but will be removed as we understand all the alert problems...
*/
static int alerts = 1;

packetAlertsOff()
{
    alerts = 0;
}

packetAlertsOn()
{
    alerts = 1;
}

alertsSet() {
    return(alerts == 1);
}

/*
** --------------------------------------
**
** packet details which can be looked up.
*/

/* packet_var returns the offset into this table to put on stack */
static	Ptab	ops[] = {
	"packet",	"sec",		packetSec,
	"packet",	"usec",		packetUsec,
	"packet",	"len",		packetLen,
	"packet",	"nonce",	packetNonce,
	"packet",	"blob",		packetBlob,
	"eth",		"type",		ethType,
	"eth",		"8023",		eth8023,
	"eth",		"src",		ethSrc,
	"eth",		"dst",		ethDst,
	"eth",		"len",		ethLen,
	"ip",		"source",	ipSrc,
	"ip",		"src",		ipSrc,
	"ip",		"dest",		ipDst,
	"ip",		"dst",		ipDst,
	"ip",		"protocol",	ipProto,
	"ip",		"proto",	ipProto,
	"ip",		"len",		ipLen,
	"ip",		"lenError",	ipLenError,
	"ip",		"is",		ipIsIp,
	"icmp",		"is",		icmpIsIcmp,
	"icmp",		"hdr",		icmpIsHdr,
	"icmp",		"type",		icmpType,
	"icmp",		"code",		icmpCode,
	"icmp",		"cksum",	icmpCksum,
	"icmp",		"blob",		icmpBlob,
	"udp",		"is",		udpIsUdp,
	"udp",		"hdr",		udpIsHdr,
	"udp",		"sourceport",	udpSport,
	"udp",		"sport",	udpSport,
	"udp",		"destport",	udpDport,
	"udp",		"dport",	udpDport,
	"udp",		"blob",		udpBlob,
	"tcp",		"is",		tcpIsTcp,
	"tcp",		"hdr",		tcpIsHdr,
	"tcp",		"sourceport",	tcpSport,
	"tcp",		"sport",	tcpSport,
	"tcp",		"destport",	tcpDport,
	"tcp",		"dport",	tcpDport,
	"tcp",		"blob",		tcpBlob,
	"tcp",		"stream",	tcpStream,
	"tcp",		"bytes",	tcpBytes,
	"tcp",		"length",	tcpLength,
	"tcp",		"conn",		tcpConn,
	"tcp",		"connEpoch",	tcpConnEpoch,
	"tcp",		"connPktSec",	tcpConnPktSec,
	"tcp",		"connHash",	tcpConnHash,
	"tcp",		"connSrc",	tcpConnSrc,
	"tcp",		"connDst",	tcpConnDst,
	"tcp",		"connSport",	tcpConnSport,
	"tcp",		"connDport",	tcpConnDport,
	"tcp",		"connForwBytes",tcpConnForwBytes,
	"tcp",		"connBackBytes",tcpConnBackBytes,
	"tcp",		"connForwLost", tcpConnForwLost,
	"tcp",		"connBackLost",	tcpConnBackLost,
	"tcp",		"connMap",	tcpConnMap,
	"tcp",		"connSym",	tcpConnSym,
	"tcp",		"connSave",	tcpConnSave,
	"tcp",		"connSyn",	tcpConnSyn,
	"tcp",		"connSynAck",	tcpConnSynAck,
	"tcp",		"connRst",	tcpConnRst,
	"tcp",		"connFin",	tcpConnFin,
	"system",	"time",		systime,
	"ident",	"name",		scopeName,
	"ident",	"index",	scopeIndex,
	0,		0,		0
};



/*
** --------------------------------------
**
** Support routines.
**
*/

/*
**
*/

none(MemVal *m) {
    m->typ = NFR_TYPE_INT;
    m->val.i = 0;
    return(0);
}

/*
**
*/

null(MemVal *m) {
    m->typ = NFR_TYPE_ERR;
    m->val.i = NFR_ERR_NOFUNC;
    return(0);
}

/*
** protocol error.
** Trying to ask for protocol details when its the
** wrong protocol (like asking for tcp.sequence when
** its not a tcp packet)
**
*/

protocol(MemVal *m) {
    m->typ = NFR_TYPE_ERR;
    m->val.i = NFR_ERR_PROTOCOL;
    return(0);
}



/*
** this is *slightly* subtle. if we have no match, we return
** the function as -1. later on, if the function is called, we
** detect the -1 and return an error value. appropriately.
*/
int
packet_var(char *proto,char *val)
{
	Ptab	*p;
	int	ix;

	ix = 0;
	for(p = ops; p->proto != (char *)0; p++) {
		if(!strcasecmp(p->proto,proto) && !strcasecmp(p->val,val)) {
			return(ix);
		}
		ix++;
	}
	Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_PACKETVAR, proto, val, 0 );
	return(-1);
}



/*
** pull a value from a packet by calling the appropriate
** function. the routine is passed the correct stack
** address from the runtime. the right value is either
** put in here or by the function we call.
*/
int
packet_get(int val, MemVal *m)
{
	/* out of range */
	if(val < 0 || val > sizeof(ops) / sizeof(Ptab)) {
		m->typ = NFR_TYPE_ERR;
		m->val.i = NFR_ERR_NOFUNC;
		return( 0 );
	}
	return((*ops[val].func)(m));
}



static	int
systime(MemVal *m)
{
	long	now;

	time(&now);
	m->typ = NFR_TYPE_INT;
	m->val.i = now;
	return(0);
}

/*
** packetTriggerOff.
**
** Turn off a trigger.
**
*/

int packetTriggerReap = 0;
que_t * packetTriggers = 0;	/* list of triggers's */

void
packetTriggerOff( t )
trigger_t * t;
{
    que_t * q = packetTriggers;
    if (q) {
	do {
	    if (((void *)q->q_data == (void *)t->t_rock) && (q->q_data)) {
		/* XXX need to remove the packet filter */
		triggerDecRef(q->q_data);
		q->q_data = 0;
		packetTriggerReap++;

		if (q == packetTriggers) {
		    packetTriggers = queForw(q);
		}
		queFree(q);
		return;
	    }
	    q = q->q_forw;
	} while (q != packetTriggers);
    }
}

/*
** --------------------------------------
**
** packet Trigger.
** We currently have no parameters, we add this function to call when
** something interesting happens.
** 
** kinds of things we could add: packet interface types, possibly
** even mac addresses, packet lengths, some sort of bad packet
** test...
*/


int packetTotalPackets = 0;
int packetTotalTriggers = 0;

int
packetTriggerSet( t )
trigger_t * t;
{
    int i;
    for (i = 0; i<t->t_ndx; ++i) {
	if (strcmp( t->t_whom[i], "packet" ) == 0) {
	    Sym * l = t->t_nv[i];
	    que_t * n = queNew(t);
	    if (queAdd( packetTriggers, n ) == 0) {
		packetTriggers = n;
	    }

	    t->t_free = packetTriggerOff;
	    t->t_rock = (void *)n;

	    return( (int)n );
	} 
    }
}

/*
** deliver each packet.. 
** Take care, since we may modify the list of triggers which
** get applied to each packet when we perofrm the interpreted function.
**
** This is the lowest level packet arrival routine.  We call machine
** cleanup/garbage collection routines from here.
*/
packetTriggerDispatch()
{
    que_t * q = packetTriggers;
    int any = 0;
    if (q) {
	do {
	    que_t * qn = q->q_forw;
	    trigger_t * t = (trigger_t *)q->q_data;
	    if (t) {
		performArgs( t->t_func, t->t_scope, memTmpNew( NFR_TYPE_TRIGGER, t ), 0 );
		packetTotalTriggers++;
		any++;
	    }
	    q = qn;
	} while (packetTriggers && (q != packetTriggers) );
    }
    if (any) {
	tmpfree();
    }
    /*
    ** we reap any of the old packetTriggers now...
    */
    reapTriggers:

    q = packetTriggers;
    if (packetTriggerReap && q) {
	do  {
	    que_t * qn = q->q_forw;
	    if (q->q_data == 0) {
		if (q == packetTriggers) {
		    packetTriggers = queForw(q);
		}
		queFree(q);
	    }
	    q = qn;
	} while (packetTriggers && (q != packetTriggers));
	packetTriggerReap = 0;
    }

}

/*
** packetRecv.
**
** Send all packets here.  Every interface sends the packets
** to THIS routine.
**
*/

int             perform_cnt;
timeval_t       perform_runTime;
double          perform_sumSq;
double          perform_stdDev;

packetRecv( pcapHdr, pkt )
struct pcap_pkthdr * pcapHdr;
char * pkt;
{
    static Sym * apply = 0;

    packet_data = pkt;
    packet_len = pcapHdr->len;
    packet_hdr = pcapHdr;
    packet_nonce++;

    /* construct some stats */
    if (packetTotalPackets++ > 0) {
	double sq;
	timeval_t d = util_diffTv( &packet_hdr->ts, &packet_arr );
	double diff = (util_tv2usec(&d)/MILLION);
	util_incrTv( &packet_arrDiff, &d );
	packet_arrSumSq += (diff*diff);
	if (packetTotalPackets > 1) {
	    packet_arrStdDev = util_stdDev( 
		util_tv2usec(&packet_arrDiff)/packetTotalPackets/MILLION,
		packet_arrSumSq,
		packetTotalPackets );
	}

    }
    packet_arr = packet_hdr->ts;

    {
	timeval_t run_start, run_done, run_time;
	gettimeofday( &run_start, 0 );

	packetTriggerDispatch();
	reassemble();

	gettimeofday( &run_done, 0 );
	run_time = util_diffTv( &run_done, &run_start );

	if (perform_cnt++ > 0) {
	    double diff = (util_tv2usec(&run_time)/MILLION);
	    util_incrTv( &perform_runTime, &run_time );
	    perform_sumSq += diff;
	    if (perform_cnt > 1) {
		perform_stdDev = util_stdDev(
		    util_tv2usec(&perform_runTime)/perform_cnt/MILLION,
		    perform_sumSq, perform_cnt );
	    }  
	}
    }

    packet_data = 0;
    packet_len = 0;
    packet_hdr = 0;
	
}

packetStats()
{
    uiCurrPrintf( "packets: total %d triggers %d\n",
	packetTotalPackets, packetTotalTriggers );
    if (packetTotalPackets) {
	char tbuf[64];
	strftime( tbuf, sizeof(tbuf), "[%D %T]", localtime( &packet_arr.tv_sec ) );
	uiCurrPrintf( "packets: Arriv %s secTotal %g secTotal/cnt %g stdDev %g\n",
	    tbuf,  (util_tv2usec( &packet_arrDiff)/MILLION),
	    (util_tv2usec( &packet_arrDiff)/MILLION)/packetTotalPackets, packet_arrStdDev );
    }
}

 
performStats()
{
    if (perform_cnt > 0) {
	uiCurrPrintf( "perform: cnt %d sumSeconds %g sec/cnt %g sumSq %g stdDev %g\n",
	    perform_cnt,
	    util_tv2usec(&perform_runTime)/MILLION,
	    util_tv2usec(&perform_runTime)/MILLION/perform_cnt,
	    perform_sumSq, perform_stdDev ); 
    } else {
	uiCurrPrintf( "perform: cnt %d \n", perform_cnt );
    }
}


/*
** --------------------------------------
**
** Routines to collect values.
**
** The use of nonce in the routines below allow recomputation
** of the results only when necessary.
**
*/

/*
** return the seconds of the current packet.
*/
int
packet_sec()
{
    if (packet_hdr) {
	return(packet_hdr->ts.tv_sec);
    } else {
	return(0);
    }
}

int
packetSec( MemVal * m )
{
    if (packet_hdr) {
	m->typ = NFR_TYPE_INT;
	m->val.i = packet_hdr->ts.tv_sec;
	return(0);
    } else {
	null(m);
	return(0);
    }
}

/*
** return the microseconds of the current packet.
*/

int
packet_usec()
{
    if (packet_hdr) {
	return(packet_hdr->ts.tv_usec);
    } else {
	return(0);
    }
}

int
packetUsec( MemVal * m )
{
    if (packet_hdr) {
	m->typ = NFR_TYPE_INT;
	m->val.i = packet_hdr->ts.tv_usec;
	return(0);
    } else {
	null(m);
	return(0);
    }
}

int
packetLen( MemVal * m )
{
    if (packet_hdr) {
	m->typ = NFR_TYPE_INT;
	m->val.i = packet_len;
	return(0);
    } else {
	null(m);
	return(0);
    }
}

int
packetNonce( MemVal * m )
{
    if (packet_data) {
	m->typ = NFR_TYPE_INT;
	m->val.i = packet_nonce;
	return(0);
    } else {
	null(m);
	return(0);
    }
}

blob_t packetBlobVar[1];

int
packetBlob( MemVal * m )
{
    if (packet_data) {
	m->typ = NFR_TYPE_BLOB;
	m->val.b = blobSet( packetBlobVar,
		packet_data, packet_len, 1, BLOB_FLAGS_STATIC );
	return(0);
    } else {
	null(m);
	return(0);
    }
}

/*
** print some stats about the packet level interfaces...
**
*/

void
packetDisplay()
{
    if (packet_data) {
	uiCurrPrintf( "packet: nonce %d\n", packet_nonce );
	if (packet_hdr) {
	    char tbuf[64];
	    strftime( tbuf, sizeof(tbuf), "%D %T",
		localtime( &(packet_hdr->ts.tv_sec) ) );
	    uiCurrPrintf( "packet: time %s.%8.8d\n",
		tbuf, packet_hdr->ts.tv_usec );
	    uiCurrPrintf( "packet: caplen %d len %d\n",
		packet_hdr->caplen, packet_hdr->len );
	}
    }
}


/* return the end of the ethernet header */

    /*
    ** here, we're concerned with the type of 
    ** ethernet packet... whether 802.3, or
    ** dix style headers.
    */
    static int eth_nonce = 0;

    struct ethDetails {
	int	eth_type;
	char *	eth_src;
	char *	eth_dst;
	char *	eth_end;
	short	eth_8023;
	short	eth_length;
    } eth;

    struct e_header {
	char	eDst[6];
	char 	eSrc[6];
	char	eType[2];
    };

    struct e_header_8023 {
	char	eDst[6];
	char 	eSrc[6];
	char	eLen[2];
	char	eType[2];
    };

ethDetails() {
    struct e_header * e = (struct e_header *)packet_data;
    int type = ((e->eType[0]<<8) | (e->eType[1]));

    eth_nonce = packet_nonce;

    eth.eth_src = e->eSrc;
    eth.eth_dst = e->eDst;

    if (type < 1500) {
	/* seems this ought to be the length... its an 802.3 packet. */
	/* XXX compare this length to the pcal length */
	struct e_header_8023 * e = (struct e_header_8023 *)packet_data;
	eth.eth_length = type;
	eth.eth_type = (e->eType[0]<<8) | (e->eType[1]);
	eth.eth_8023 = 1;
	eth.eth_end = &e->eType[2];
    } else {
	eth.eth_type = type;
	eth.eth_length = packet_hdr->len;
	eth.eth_8023 = 0;
	eth.eth_end = &e->eType[2];
    }

}

/* return the end of the ethernet header. */
char *
ethTail()
{

    if (eth_nonce != packet_nonce) {
	ethDetails();
    }
    return eth.eth_end;
}

int
ethType(MemVal *m)
{
    if (eth_nonce != packet_nonce) {
	ethDetails();
    }
    m->typ = NFR_TYPE_INT;
    m->val.i = eth.eth_type;
    return(0);
}

int
eth8023(MemVal *m)
{
    if (eth_nonce != packet_nonce) {
	ethDetails();
    }
    m->typ = NFR_TYPE_INT;
    m->val.i = eth.eth_8023;
    return(0);
}

int
ethLen(MemVal *m)
{
    if (eth_nonce != packet_nonce) {
	ethDetails();
    }
    m->typ = NFR_TYPE_INT;
    m->val.i = eth.eth_length;
    return(0);
}

int
ethSrc(MemVal *m)
{
    if (eth_nonce != packet_nonce) {
	ethDetails();
    }
    m->typ = NFR_TYPE_ETHMAC;
    m->val.e = *(ethMac_t *)eth.eth_src;
    return(0);
}

int
ethDst(MemVal *m)
{
    if (eth_nonce != packet_nonce) {
	ethDetails();
    }
    m->typ = NFR_TYPE_ETHMAC;
    m->val.e = *(ethMac_t *)eth.eth_dst;
    return(0);
}

/*
** Show details about what is currently in the ethDetails structure.
*/
ethDisplay()
{
    uiCurrPrintf( "ethDisplay: nonce %d type %d src", eth_nonce, eth.eth_type );
    displayEnetMac( eth.eth_src );
    uiCurrPrintf( " dst " );
    displayEnetMac( eth.eth_dst );
    uiCurrPrintf(" end(index) %d length %d\n",
	eth.eth_end - packet_data, eth.eth_length );
}


    /*
    ** ip cracking.
    ** Since we could have ipOptions, we'll compute the end of the
    ** ip packet once...
    */
    static ip_nonce = 0;

    struct ipDetails {
	struct ip * 	ip;
	char *		ipEnd;
	char *		ipOpts;
	short		ipHLen;		/* header length */
	short		ipLen;		/* packet length from the header */
	short		ipLenError;	/* did this packet fit inside? */
	short		ipHasOpts;
	short		isIp;		/* is the current packet an ip packet */
    } ipd;


ipDetails()
{
    ip_nonce = packet_nonce;

    if (eth_nonce != packet_nonce) {
	ethDetails();
    }
    if (eth.eth_type != ETHERTYPE_IP) {
	ipd.ip = 0;
	ipd.ipEnd = 0;
	ipd.ipOpts = 0;
	ipd.ipHLen = 0;
	ipd.ipLen = 0;
	ipd.ipLenError = 0;
	ipd.ipHasOpts = 0;
	ipd.isIp = 0;
    } else {
	struct ip * i = (struct ip *)ethTail();

	ipd.ipHLen = i->ip_hl * 4;
	ipd.ipLen = int16( &(i->ip_len) );
	ipd.ip = i;
	ipd.ipEnd = &(((char *)i)[ipd.ipHLen]);
	ipd.ipHasOpts = 0;
	ipd.isIp = 1;
	if (ipd.ipHLen > 20) {
	    ipd.ipHasOpts = 1;
	}
	if (packet_len < (ipd.ipLen + (ethTail() - packet_data)) ) {
	    uiCurrPrintf( "ipDetails: short packet! packet %d ip_len %d packet_hdr %d difference %d\n",
		packet_len, ipd.ipLen, (ethTail() - packet_data),
		(ipd.ipLen + (ethTail() - packet_data)) - packet_len );
	    Display( 1, "ipDetails: short packet!" );
	}
    }
}

/* return the base of the ip header */
struct ip *
ipHdr()
{
    if (ip_nonce != packet_nonce) {
	ipDetails();
    }
    if (ipd.isIp) {
	return ipd.ip;
    }
    return(0);
}

/* return the end of the ip header */
char *
ipEnd()
{
    if (ip_nonce != packet_nonce) {
	ipDetails();
    }
    if (ipd.isIp) {
	return ipd.ipEnd;
    }
    return(0);
}


/* return the length of the header */
int
ipHLen()
{
    if (ip_nonce != packet_nonce) {
	ipDetails();
    }
    if (ipd.isIp) {
	return ipd.ipHLen;
    }
    return(0);
}

/*
** ipIsIp
*/

int
ipIsIp(MemVal *m)
{
    if (packet_data) {
	struct ip * ip = ipHdr();
	m->typ = NFR_TYPE_INT;
	if (ip) {
	    m->val.i = ipd.isIp;
	    return(0);
	} else {
	    m->val.i = 0; /* false, but ok result */
	    return(0);
	}
    } else {
	if (alertsSet()) {
	    Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_NOPACKET, "ipIsIp", 0 );
	}
	return(none(m));
    }
}

/*
** ipSrc
*/

int
ipSrc(MemVal *m)
{
    if (packet_data) {
	struct ip * ip = ipHdr();
	if (ip) {
	    m->typ = NFR_TYPE_IPV4HOST;
	    { /* XXX faster than calling memcpy */
		unsigned char * f = (unsigned char *)&ip->ip_src.s_addr;
		m->val.i = ntohl( (f[0]<<24) | (f[1]<<16) | (f[2]<<8) | (f[3]) );
	    }
	    return(0);
	} else {
	    return(protocol(m));
	}
    } else {
	if (alertsSet()) {
	    Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_NOPACKET, "ipSrc", 0 );
	}
	return(none(m));
    }
}

/*
** ipDst
*/

int
ipDst(MemVal *m)
{
    if (packet_data) {
	struct ip * ip = ipHdr();
	if (ip) {
	    m->typ = NFR_TYPE_IPV4HOST;
	    { /* XXX faster than calling memcpy */
		unsigned char * f = (unsigned char *)&ip->ip_dst.s_addr;
		m->val.i = ntohl( (f[0]<<24) | (f[1]<<16) | (f[2]<<8) | (f[3]) );
	    }
	    return(0);
	} else {
	    return(protocol(m));
	}
    } else {
	if (alertsSet()) {
	    Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_NOPACKET, "ipDst", 0 );
	}
	return(none(m));
    }
}

/*
** ipProto
*/

int
ipProto(MemVal *m)
{
    if (packet_data) {
	struct ip * ip = ipHdr();
	if (ip) {
	    m->typ = NFR_TYPE_INT;
	    m->val.i = ip->ip_p;
	    return(0);
	} else {
	    return(protocol(m));
	}
    } else {
	if (alertsSet()) {
	    Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_NOPACKET, "ipProto", 0 );
	}
	return(none(m));
    }
}

/*
** ipLen
*/
int
ipLen(MemVal *m)
{
    if (packet_data) {
	struct ip * ip = ipHdr();
	if (ip) {
	    m->typ = NFR_TYPE_INT;
	    m->val.i = ipd.ipLen;
	    return(0);
	} else {
	    return(protocol(m));
	}
    } else {
	if (alertsSet()) {
	    Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_NOPACKET, "ipLen", 0 );
	}
	return(none(m));
    }
}

/* return the value of the ip length error ... computed in ipDetails */
int
ipLenError(MemVal *m)
{
    if (packet_data) {
	struct ip * ip = ipHdr();
	if (ip) {
	    m->typ = NFR_TYPE_INT;
	    m->val.i = ipd.ipLenError;
	    return(0);
	} else {
	    return(protocol(m));
	}
    } else {
	if (alertsSet()) {
	    Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_NOPACKET, "ipLenError", 0 );
	}
	return(none(m));
    }
}

/*
**
*/

ipDisplay() 
{
    uiCurrPrintf( "ipDisplay: nonce %d ip (index) %d end (index) %d ipOpts (index) %d ",
	ip_nonce, ((char *)ipd.ip)-packet_data,
	ipd.ipEnd-packet_data, ipd.ipOpts-packet_data );
    uiCurrPrintf( "ipHasOpts %d isIp %d\n", ipd.ipHasOpts, ipd.isIp );
}


/*
** ----------------------------------------
** icmp.
**
**
*/

static icmp_nonce;
struct icmpDetails {
    struct icmp *	icmpHdr;		/* icmp header address */
    short		isIcmp;			/* is this a icmp packet */
    short		isHdr;			/* does this have a udp header */
} icmpd;

/*
**
*/
struct udphdr *
icmpDetails() {
    struct ip * ip = ipHdr();

    icmp_nonce = packet_nonce;

    if (ip) {
	icmpd.icmpHdr = (struct icmp *)ipEnd();
	icmpd.isIcmp = (ip->ip_p == IPPROTO_ICMP);
	icmpd.isHdr = ((int16(&ip->ip_off) & 0x1fff) == 0);
	return;
    } else {
	icmpd.icmpHdr = 0;
	icmpd.isIcmp = 0;
	icmpd.isHdr = 0;
	return(0);
    }
	
}

struct icmp *
icmpHdr()
{
    if (icmp_nonce != packet_nonce) {
	icmpDetails();
    }
    if (icmpd.isIcmp) {
	return(icmpd.icmpHdr);
    }
    return(0);
}


int
icmpIsIcmp(MemVal *m)
{
    if (packet_data) {
	struct icmp * up = icmpHdr();
	m->typ = NFR_TYPE_INT;
	if (up) {
	    m->val.i = icmpd.isIcmp;
	    return(0);
	} else {
	    m->val.i = 0; /* false, but not an error */
	    return(0);
	}
    } else {
	if (alertsSet()) {
	    Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_NOPACKET, "icmpIsIcmp", 0 );
	}
	return(none(m));
    }
}

int
icmpIsHdr(MemVal *m)
{
    if (packet_data) {
	struct icmp * icmpp = icmpHdr();
	if (icmpp) {
	    m->typ = NFR_TYPE_INT;
	    m->val.i = icmpd.isHdr;
	    return(0);
	} else {
	    return(protocol(m));
	}
    } else {
	if (alertsSet()) {
	    Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_NOPACKET, "icmpIsIcmp", 0 );
	}
	return(none(m));
    }
}

int
icmpType(MemVal *m)
{
    if (packet_data) {
	struct icmp * icmpp = icmpHdr();
	if (icmpp) {
	    if (icmpd.isHdr) {
		m->val.i = icmpp->icmp_type;
		m->typ = NFR_TYPE_INT;
		return(0);
	    }
	}
	return(protocol(m));
    } else  {
	if (alertsSet()) {
	    Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_NOPACKET, "icmpType", 0 );
	}
	return(none(m));
    }
}

int
icmpCode(MemVal *m)
{
    if (packet_data) {
	struct icmp * icmpp = icmpHdr();
	if (icmpp) {
	    if (icmpd.isHdr) {
		m->val.i = icmpp->icmp_code;
		m->typ = NFR_TYPE_INT;
		return(0);
	    }
	}
	return(protocol(m));
    } else {
	if (alertsSet()) {
	    Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_NOPACKET, "icmpCode", 0 );
	}
	return(none(m));
    }
}

int
icmpCksum(MemVal *m)
{
    if (packet_data) {
	struct icmp * icmpp = icmpHdr();
	if (icmpp) {
	    if (icmpd.isHdr) {
		unsigned char * v = (unsigned char *)&icmpp->icmp_cksum;
		m->val.i = ( (v[0] << 8) | (v[1]) );
		m->typ = NFR_TYPE_INT;
		return(0);
	    }
	}
	return(protocol(m));
    } else {
	if (alertsSet()) {
	    Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_NOPACKET, "icmpCksum", 0 );
	}
	return(none(m));
    }
}

blob_t icmpBlobVar[1];

int
icmpBlob(MemVal *m)
{
    if (packet_data) {
	struct icmp * icmpp = icmpHdr();
	if (icmpp) {
	    if (icmpd.isHdr) {
		char * v = (char *)(&(icmpd.icmpHdr->icmp_cksum)+1);
		m->typ = NFR_TYPE_BLOB;
		m->val.b = blobSet( icmpBlobVar, v,
			packet_len - (v - packet_data), 1, BLOB_FLAGS_STATIC );
		return(0);
	    }
	}
	return(protocol(m));
    } else {
	if (alertsSet()) {
	    Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_NOPACKET, "icmpBlob", 0 );
	}
	return(none(m));
    }
}
/*
** --------------------------------------
**
** icmp Trigger.
** We currently only support the port parameters.
*/

typedef struct icmpTrigger {
    trigger_t *	ict_trigger;
    MemVal * 	ict_src;
    MemVal *	ict_dst;
    MemVal * 	ict_type;
    MemVal *	ict_code;
    int		ict_flags;
} icmpTrigger_t;

#define	ICT_FLAGS_SRC	0x1
#define	ICT_FLAGS_DST	0x2
#define	ICT_FLAGS_HOST	0x4

#define	ICT_FLAGS_TYPE	0x10
#define	ICT_FLAGS_CODE	0x20



icmpTrigger_t *
icmpTriggerNew( src, dst, code, type, flags, trigger )
MemVal * src;
MemVal * dst;
MemVal * code;
MemVal * type;
int	flags;
trigger_t * trigger;
{
    icmpTrigger_t * new = (icmpTrigger_t *)malloc( sizeof(*new) );
    /* assert new? */
    new->ict_src = src;
    new->ict_dst = dst;
    new->ict_type = type;
    new->ict_code = code;
    new->ict_flags = flags;
    new->ict_trigger = trigger;
    return(new);
}

void
icmpTriggerFree(n)
icmpTrigger_t * n;
{
    if (n->ict_trigger) {
	trigger_t * t = n->ict_trigger;
	n->ict_trigger = 0;
	triggerDecRef(t);
    }
    free(n);
}

que_t * icmpTriggers = 0;	/* list of triggers's */

/*
** icmpTriggerOff.
**
** Turn off a trigger.
**
*/

int icmpTriggerReap = 0;

void
icmpTriggerOff( t )
trigger_t * t;
{
    que_t * q = icmpTriggers;
    if (q) {
	do {
	    if (((void *)q->q_data == (void *)t->t_rock) && (q->q_data)) {
		icmpTriggerFree(q->q_data);
		q->q_data = 0;
		icmpTriggerReap++;

		if (q == icmpTriggers) {
		    icmpTriggers = queForw(q);
		}
		queFree(q);

		return;
	    }
	    q = q->q_forw;
	} while (q != icmpTriggers);
    }
}

int
icmpTriggerSet( t )
trigger_t * t;
{
    int i;
    int err = 0;
    int flags = 0;
    MemVal * src = 0, * dst = 0, * code = 0, * type = 0;

    for (i = 0; i<t->t_ndx; ++i) {
	if (strcmp( t->t_whom[i], "icmp" ) == 0) {
	    Sym * l = t->t_nv[i];
	    {
		for( l = t->t_nv[i]; l ; l = l->s_next ) {
		    MemVal * m = &l->s_data;

		    if (symCmp( l, "src" ) == 0) {
			if (flags & (ICT_FLAGS_SRC|ICT_FLAGS_HOST)) {
			    uiCurrPrintf( "icmpTriggerSet: can't set src/host twice\n" );
			    err = 1;
			}
			flags |= ICT_FLAGS_SRC;
			src = m;
		    } else 
		    if (symCmp( l, "dst" ) == 0) {
			if (flags & (ICT_FLAGS_DST|ICT_FLAGS_HOST)) {
			    uiCurrPrintf( "icmpTriggerSet: can't set dst/host twice\n" );
			    err = 1;
			}
			flags |= ICT_FLAGS_DST;
			dst = m;
		    } else 
		    if (symCmp( l, "host" ) == 0) {
			if (flags & (ICT_FLAGS_HOST|ICT_FLAGS_SRC|ICT_FLAGS_DST)) {
			    uiCurrPrintf( "icmpTriggerSet: can't set code twice\n" );
			    err = 1;
			}
			flags |= ICT_FLAGS_HOST;
			src = m;
		    } else 
		    if (symCmp( l, "code" ) == 0) {
			if (flags & ICT_FLAGS_CODE) {
			    uiCurrPrintf( "icmpTriggerSet: can't set code twice\n" );
			    err = 1;
			}
			flags |= ICT_FLAGS_CODE;
			code = m;
		    } else 
		    if (symCmp( l, "type" ) == 0) {
			if (flags & ICT_FLAGS_TYPE) {
			    uiCurrPrintf( "icmpTriggerSet: can't set type twice\n" );
			    err = 1;
			}
			flags |= ICT_FLAGS_TYPE;
			type = m;
		    } else {
			/* XXX error */
			err = 1;
		    }
		}
	    }
	} 
    }
    /*
    ** Make one, if it seems ok;
    */
    if (err == 0) {
	icmpTrigger_t * ict = icmpTriggerNew( src, dst, code, type, flags, t );
	que_t * n = queNew(ict);

	t->t_free = icmpTriggerOff;
	t->t_rock = (void *)ict;

	if (queAdd( icmpTriggers, n ) == 0) {
	    icmpTriggers = n;
	} 
	return( (int)n );
    } else {
	uiCurrPrintf( "icmpTriggerNew: error in trigger.\n" );
    }
    return(0);
}


int icmpTotalTriggers = 0;

/*
** deliver each icmp packet.. 
** Take care, since we may modify the list of triggers which
** get applied to each packet when we perofrm the interpreted function.
**
** This is the lowest level packet arrival routine.  We call machine
** cleanup/garbage collection routines from here.
*/
icmpTriggerDispatch( )
{
    que_t * q = icmpTriggers;
    int any = 0;
    if (q) {
	do {
	    que_t * qn = q->q_forw;
	    MemVal m[1];
	    icmpTrigger_t * ict = (icmpTrigger_t *)q->q_data;

	    if ( (ict->ict_flags == 0) /* no flags */
	    || ( ( ((ict->ict_flags & (ICT_FLAGS_HOST|ICT_FLAGS_SRC|ICT_FLAGS_DST)) == 0)
		  ||((ict->ict_flags & ICT_FLAGS_HOST)
			&& (tEqual(ict->ict_src, (ipSrc(m), m))
				|| tEqual( ict->ict_src, (ipDst(m), m) )))
		  ||(((ict->ict_flags & (ICT_FLAGS_SRC|ICT_FLAGS_DST)) == ICT_FLAGS_SRC)
			&& tEqual(ict->ict_src, (ipSrc(m), m)) )
		  ||(((ict->ict_flags & (ICT_FLAGS_DST|ICT_FLAGS_DST)) == ICT_FLAGS_DST)
			&& tEqual(ict->ict_dst, (ipDst(m), m)) )
		  ||(((ict->ict_flags & (ICT_FLAGS_DST|ICT_FLAGS_DST)) == (ICT_FLAGS_SRC|ICT_FLAGS_DST))
			&& tEqual(ict->ict_dst, (ipDst(m), m))
			&& tEqual(ict->ict_src, (ipSrc(m), m)) ) )
		&& (((ict->ict_flags & (ICT_FLAGS_CODE)) == 0)
		     || (tEqual(ict->ict_code, (icmpCode(m), m))) )
		&& (((ict->ict_flags & (ICT_FLAGS_TYPE)) == 0)
		     || (tEqual(ict->ict_type, (icmpType(m), m))) ) )
	    ) {
		/*
		** perform the triger.
		*/
		performArgs( ict->ict_trigger->t_func, ict->ict_trigger->t_scope,
			memTmpNew( NFR_TYPE_TRIGGER, ict->ict_trigger ), 0 );
		icmpTotalTriggers++;
		any++;
	    }
	    q = qn;
	} while (icmpTriggers && (q != icmpTriggers) );
    }
    /*
    ** we reap any of the old packetTriggers now...
    */
    reapTriggers:

    q = icmpTriggers;
    if (icmpTriggerReap && q) {
	do  {
	    que_t * qn = q->q_forw;
	    if (q->q_data == 0) {
		if (q == icmpTriggers) {
		    icmpTriggers = queForw(q);
		}
		queFree(q);
	    }
	    q = qn;
	} while (icmpTriggers && (q != icmpTriggers));
	icmpTriggerReap = 0;
    }

}

/*
** icmpRecv.
**
** Routine which is called on reception of icmp packets, to see
** if there's anything to do
*/

icmpRecv() {
    if (icmpTriggers) {
	icmpTriggerDispatch( );
    }
}

/*
**
*/

icmpDisplay()
{
    uiCurrPrintf( "icmpDisplay: nonce %d header (index) %d isIcmp %d isHdr %d\n",
	icmp_nonce,  ((char *)icmpd.icmpHdr) - packet_data,
	icmpd.isIcmp, icmpd.isHdr);
}


/*
** ----------------------------------------
** udp.
**
**
*/

    static udp_nonce;
    struct udpDetails {
	struct udphdr *	udpHdr;		/* udp header address */
	char *		udpEnd;		/* end of udp packet */
	short		isUdp;		/* is this a udp packet */
	short		isHdr;		/* does this have a udp header */
    } udpd;

/*
**
*/
struct udphdr *
udpDetails() {
    struct ip * ip = ipHdr();

    udp_nonce = packet_nonce;

    if (ip) {
	udpd.udpHdr = (struct udphdr *)ipEnd();
	udpd.udpEnd = (char *)&udpd.udpHdr[1];
	udpd.isUdp = (ip->ip_p == IPPROTO_UDP);
	udpd.isHdr = ((int16(&ip->ip_off) & 0x1fff) == 0);
	return;
    } else {
	udpd.udpHdr = 0;
	udpd.udpEnd = 0;
	udpd.isUdp = 0;
	udpd.isHdr = 0;
	return(0);
    }
	
}

struct udphdr *
udpHdr()
{
    if (udp_nonce != packet_nonce) {
	udpDetails();
    }
    if (udpd.isUdp) {
	return(udpd.udpHdr);
    }
    return(0);
}


int
udpIsUdp(MemVal *m)
{
    if (packet_data) {
	struct udphdr * up = udpHdr();
	m->typ = NFR_TYPE_INT;
	if (up) {
	    m->val.i = udpd.isUdp;
	    return(0);
	} else {
	    m->val.i = 0; /* false, but not an error */
	    return(0);
	}
    } else {
	if (alertsSet()) {
	    Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_NOPACKET, "udpIsUdp", 0 );
	}
	return(none(m));
    }
}

int
udpIsHdr(MemVal *m)
{
    if (packet_data) {
	struct udphdr * up = udpHdr();
	if (up) {
	    m->typ = NFR_TYPE_INT;
	    m->val.i = udpd.isHdr;
	    return(0);
	} else {
	    return(protocol(m));
	}
    } else {
	if (alertsSet()) {
	    Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_NOPACKET, "udpIsUdp", 0 );
	}
	return(none(m));
    }
}

int
udpSport(MemVal *m)
{
    if (packet_data) {
	struct udphdr * up = udpHdr();
	if (up) {
	    if (udpd.isHdr) {
		unsigned char * v = (unsigned char *)&up->uh_sport;
		m->val.i = ( (v[0] << 8) | (v[1]) );
		m->typ = NFR_TYPE_INT;
		return(0);
	    }
	}
	return(protocol(m));
    } else  {
	if (alertsSet()) {
	    Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_NOPACKET, "udpSport", 0 );
	}
	return(none(m));
    }
}

int
udpDport(MemVal *m)
{
    if (packet_data) {
	struct udphdr * up = udpHdr();
	if (up) {
	    if (udpd.isHdr) {
		unsigned char * v = (unsigned char *)&up->uh_dport;
		m->val.i = ( (v[0] << 8) | (v[1]) );
		m->typ = NFR_TYPE_INT;
		return(0);
	    }
	}
	return(protocol(m));
    } else {
	if (alertsSet()) {
	    Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_NOPACKET, "udpDport", 0 );
	}
	return(none(m));
    }
}

blob_t udpBlobVar[1];

int
udpBlob(MemVal *m)
{
    if (packet_data) {
	struct udphdr * up = udpHdr();
	if (up) {
	    if (udpd.isHdr) {
		char * v = (char *)udpd.udpEnd;
		m->typ = NFR_TYPE_BLOB;
		m->val.b = blobSet( udpBlobVar, v,
			packet_len - (v - packet_data), 1, BLOB_FLAGS_STATIC );
		return(0);
	    }
	}
	return(protocol(m));
    } else {
	if (alertsSet()) {
	    Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_NOPACKET, "udpBlob", 0 );
	}
	return(none(m));
    }
}
/*
** --------------------------------------
**
** udp Trigger.
** We currently only support the port parameters.
*/

typedef struct udpTrigger {
    trigger_t *	ut_trigger;
    MemVal * 	ut_src;
    MemVal *	ut_dst;
    MemVal * 	ut_sport;
    MemVal *	ut_dport;
    int		ut_flags;
} udpTrigger_t;

#define	UT_FLAGS_SPORT	0x1
#define	UT_FLAGS_DPORT	0x2
#define	UT_FLAGS_PORT	0x4	/* use sport for PORT */

#define UT_FLAGS_SRC	0x10
#define UT_FLAGS_DST	0x20
#define UT_FLAGS_HOST	0x40	/* use src for HOST */

#define	UT_FLAGS_MATCH	(UT_FLAGS_SPORT|UT_FLAGS_DPORT|UT_FLAGS_PORT|UT_FLAGS_SRC|UT_FLAGS_DST|UT_FLAGS_HOST)

udpTrigger_t *
udpTriggerNew( src, dst, sport, dport, flags, trigger )
MemVal * src;
MemVal * dst;
MemVal * sport;
MemVal * dport;
int	flags;
trigger_t * trigger;
{
    udpTrigger_t * new = (udpTrigger_t *)malloc( sizeof(*new) );
    /* assert new? */
    new->ut_src = src;
    new->ut_dst = dst;
    new->ut_sport = sport;
    new->ut_dport = dport;
    new->ut_flags = flags;
    new->ut_trigger = trigger;
    return(new);
}

void
udpTriggerFree(n)
udpTrigger_t * n;
{
    if (n->ut_trigger) {
	trigger_t * t = n->ut_trigger;
	n->ut_trigger = 0;
	triggerDecRef(t);
    }
    free(n);
}

que_t * udpTriggers = 0;	/* list of triggers's */
/*
** udpTriggerOff.
**
** Turn off a trigger.
**
*/

int udpTriggerReap = 0;

void
udpTriggerOff( t )
trigger_t * t;
{
    que_t * q = udpTriggers;
    if (q) {
	do {
	    if (((void *)q->q_data == (void *)t->t_rock) && (q->q_data)) {
		udpTrigger_t * ut = q->q_data;
		q->q_data = 0;
		udpTriggerFree(ut);
		udpTriggerReap++;

		if (q == udpTriggers) {
		    udpTriggers = queForw(q);
		}
		queFree(q);
		return;
	    }
	    q = q->q_forw;
	} while (q != udpTriggers);
    }
}

int
udpTriggerSet( t )
trigger_t * t;
{
    int i;
    int err = 0;
    int flags = 0;
    MemVal * sport = 0, * dport = 0, * src = 0, * dst = 0;

    for (i = 0; i<t->t_ndx; ++i) {
	if (strcmp( t->t_whom[i], "udp" ) == 0) {
	    Sym * l = t->t_nv[i];
	    {
		for( l = t->t_nv[i]; l ; l = l->s_next ) {
		    MemVal * m = &l->s_data;

		    if (symCmp( l, "sport" ) == 0) {
			if (flags & (UT_FLAGS_SPORT|UT_FLAGS_PORT)) {
			    /* can't do this */
			    uiCurrPrintf( "udpTriggerSet: can't set PORT/SPORT twice\n" );
			    err = 1;
			}
			flags |= UT_FLAGS_SPORT;
			sport = m;
		    } else
		    if (symCmp( l, "dport" ) == 0) {
			if (flags & (UT_FLAGS_DPORT|UT_FLAGS_PORT)) {
			    uiCurrPrintf( "udpTriggerSet: can't set PORT/DPORT twice\n" );
			    err = 1;
			}
			flags |= UT_FLAGS_DPORT;
			dport = m;
		    } else 
		    if (symCmp( l, "port" ) == 0) {
			if (flags & (UT_FLAGS_DPORT|UT_FLAGS_SPORT)) {
			    uiCurrPrintf( "udpTriggerSet: can't set (SPORT/DPORT) and PORT\n" );
			    err = 1;
			}
			flags |= UT_FLAGS_PORT;
			sport = m;
		    } else 
		    if (symCmp( l, "src" ) == 0) {
			if (flags & (UT_FLAGS_SRC|UT_FLAGS_HOST)) {
			    uiCurrPrintf( "udpTriggerSet: can't set (SRC/HOST) and SRC\n" );
			    err = 1;
			}
			flags |= UT_FLAGS_SRC;
			src = m;
		    } else 
		    if (symCmp( l, "dst" ) == 0) {
			if (flags & (UT_FLAGS_DST|UT_FLAGS_HOST)) {
			    uiCurrPrintf( "udpTriggerSet: can't set (SRC/HOST) and DST\n" );
			    err = 1;
			}
			flags |= UT_FLAGS_DST;
			dst = m;
		    } else 
		    if (symCmp( l, "host" ) == 0) {
			if (flags & (UT_FLAGS_DST|UT_FLAGS_SRC)) {
			    uiCurrPrintf("udpTriggerSet: can't set (SRC/DST) and HOST\n" );
			    err = 1;
			}
			flags |= UT_FLAGS_HOST;
			src = m;
		    } else {
			uiCurrPrintf( "udpTriggerSet: unknown trigger name %s\n",
				symName(l) );
			err = 1;
		    }
		}
	    }
	} 
    }
    /*
    ** Make one, if it seems ok;
    */
    if (err == 0) {
	udpTrigger_t * ut = udpTriggerNew( src, dst, sport, dport, flags, t );
	que_t * n = queNew(ut);

	t->t_free = udpTriggerOff;
	t->t_rock = (void *)ut;

	if (queAdd( udpTriggers, n ) == 0) {
	    udpTriggers = n;
	}
	return( (int)n );
    }
    return(0);
}


int udpTotalTriggers = 0;

/*
** deliver each udp packet.. 
** Take care, since we may modify the list of triggers which
** get applied to each packet when we perofrm the interpreted function.
**
** This is the lowest level packet arrival routine.  We call machine
** cleanup/garbage collection routines from here.
*/
udpTriggerDispatch( )
{
    que_t * q = udpTriggers;
    int any = 0;
    if (q) {
	do {
	    que_t * qn = q->q_forw;
	    MemVal m[1];
	    udpTrigger_t * ut = (udpTrigger_t *)q->q_data;

	    if ( (ut->ut_flags == 0) || /* no flags */
	    ( /* otherwise theres some flags set.. first try ports */
	     ( ((ut->ut_flags & (UT_FLAGS_SPORT|UT_FLAGS_DPORT|UT_FLAGS_PORT)) == 0)
	     || ((( (ut->ut_flags & UT_FLAGS_PORT)
		&& (tEqual(ut->ut_sport, (udpSport(m), m)) || tEqual(ut->ut_sport, (udpDport(m), m))))
	       || (((ut->ut_flags & (UT_FLAGS_SPORT|UT_FLAGS_DPORT)) == UT_FLAGS_SPORT)
		&& tEqual(ut->ut_sport, (udpSport(m), m) ))
	       || (((ut->ut_flags & (UT_FLAGS_SPORT|UT_FLAGS_DPORT)) == UT_FLAGS_DPORT)
		&& tEqual(ut->ut_dport, (udpDport(m), m) ))
	       || (((ut->ut_flags & (UT_FLAGS_SPORT|UT_FLAGS_DPORT)) == (UT_FLAGS_SPORT|UT_FLAGS_DPORT))
		&& (tEqual(ut->ut_sport, (udpSport(m), m)) && tEqual(ut->ut_dport, (udpDport(m), m) ))) ) ) )
	    && ( ((ut->ut_flags & (UT_FLAGS_SRC|UT_FLAGS_DST|UT_FLAGS_HOST)) == 0)
	     || (((ut->ut_flags & UT_FLAGS_HOST)
		&& (tEqual(ut->ut_src, (ipSrc(m), m) ) || tEqual(ut->ut_src, (ipDst(m), m) )))
	       || (((ut->ut_flags & (UT_FLAGS_SRC|UT_FLAGS_DST)) == UT_FLAGS_SRC)
		&& tEqual(ut->ut_src, (ipSrc(m), m) ))
	       || (((ut->ut_flags & (UT_FLAGS_SRC|UT_FLAGS_DST)) == UT_FLAGS_DST)
		&& tEqual(ut->ut_dst, (ipDst(m), m) ))
	       || (((ut->ut_flags & (UT_FLAGS_SRC|UT_FLAGS_DST)) == (UT_FLAGS_SRC|UT_FLAGS_DST))
		&& (tEqual(ut->ut_src, (ipSrc(m), m) ) && tEqual(ut->ut_dst, (ipDst(m), m) ))) ) ) ) )

		    {
		    /*
		    ** perform the triger.
		    */
		    performArgs( ut->ut_trigger->t_func, ut->ut_trigger->t_scope,
			    memTmpNew( NFR_TYPE_TRIGGER, ut->ut_trigger ), 0 );
		    udpTotalTriggers++;
		    any++;
	    }
	    q = qn;
	} while (udpTriggers && (q != udpTriggers) );
    }
    /*
    ** we reap any of the old packetTriggers now...
    */
    reapTriggers:

    q = udpTriggers;
    if (udpTriggerReap && q) {
	do  {
	    que_t * qn = q->q_forw;
	    if (q->q_data == 0) {
		if (q == udpTriggers) {
		    udpTriggers = queForw(q);
		}
		queFree(q);
	    }
	    q = qn;
	} while (udpTriggers && (q != udpTriggers));
	udpTriggerReap = 0;
    }

}

/*
** udpRecv.
**
** Routine which is called on reception of udp packets, to see
** if there's anything to do
*/

udpRecv() {
    if (udpTriggers) {
	if (udp_nonce != packet_nonce) {
	    udpDetails();
	}
	udpTriggerDispatch( );
    }
}

/*
**
*/

udpDisplay()
{
    uiCurrPrintf( "udpDisplay: nonce %d header (index) %d end (index) %d isUdp %d isHdr %d\n",
	udp_nonce,  ((char *)udpd.udpHdr) - packet_data,
	udpd.udpEnd - packet_data,
	udpd.isUdp, udpd.isHdr);
}



/*
** ----------------------------------------
** tcp.
*/

    static tcp_nonce;
    struct tcpDetails {
	struct tcphdr	tcpInt;		/* internal header, used in some situations */
	struct tcphdr *	tcpHdr;		/* tcp header address */
	char *		tcpEnd;		/* end of tcp packet */
	short		tcpHasOptions;	/* do we have any options? */
	short		isTcp;		/* is this a tcp packet */
	short		isHdr;		/* does this have a tcp header */
    } tcpd;

/*
**
*/
struct tcphdr *
tcpDetails() {
    struct ip * ip = ipHdr();

    tcp_nonce = packet_nonce;

    if (ip) {
	struct tcphdr * tcphdr = (struct tcphdr *)ipEnd();
	tcpd.tcpHdr = tcphdr;
	if ((int)(tcpd.tcpHdr) & 0x2) {
	    /* on some machines, this is very bad. */
	    memcpy( &tcpd.tcpInt, tcpd.tcpHdr, sizeof(tcpd.tcpInt) );
	    tcpd.tcpHdr = &tcpd.tcpInt;
	}
	tcpd.tcpEnd = (char *)tcphdr + (tcpd.tcpHdr->th_off * 4);
	tcpd.tcpHasOptions = (tcpd.tcpHdr->th_off > 5);
	tcpd.isTcp = (ip->ip_p == IPPROTO_TCP);
	tcpd.isHdr = ((int16(&ip->ip_off) & 0x1fff) == 0);
	return;
    } else {
	tcpd.tcpHdr = 0;
	tcpd.tcpEnd = 0;
	tcpd.tcpHasOptions = 0;
	tcpd.isTcp = 0;
	tcpd.isHdr = 0;
	return(0);
    }
	
}

struct tcphdr *
tcpHdr()
{
    if (tcp_nonce != packet_nonce) {
	tcpDetails();
    }
    if (tcpd.isTcp) {
	return(tcpd.tcpHdr);
    }
    return(0);
}

char *
tcpEnd()
{
    if (tcp_nonce != packet_nonce) {
	tcpDetails();
    }
    if (tcpd.isTcp) {
	return(tcpd.tcpEnd);
    }
    return(0);
}


int
tcpIsTcp(MemVal *m)
{
    if (packet_data) {
	struct tcphdr * tp = tcpHdr();
	m->typ = NFR_TYPE_INT;
	if (tp) {
	    m->val.i = tcpd.isTcp;
	    return(0);
	} else {
	    m->val.i = 0; /* false, but not an error */
	    return(0);
	}
    } else {
	if (alertsSet()) {
	    Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_NOPACKET, "tcpIsTcp", 0 );
	}
	return(none(m));
    }
}

int
tcpIsHdr(MemVal *m)
{
    if (packet_data) {
	struct tcphdr * tp = tcpHdr();
	if (tp) {
	    m->typ = NFR_TYPE_INT;
	    m->val.i = tcpd.isHdr;
	    return(0);
	} else {
	    return(protocol(m));
	}
    } else {
	if (alertsSet()) {
	    Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_NOPACKET, "tcpIsTcp", 0 );
	}
	return(none(m));
    }
}

int
tcpSport(MemVal *m)
{
    if (packet_data) {
	struct tcphdr * tp = tcpHdr();
	if (tp) {
	    if (tcpd.isHdr) {
		unsigned char * v = (unsigned char *)&tp->th_sport;
		m->val.i = ( (v[0] << 8) | (v[1]) );
		m->typ = NFR_TYPE_INT;
		return(0);
	    }
	}
	return(protocol(m));
    } else  {
	if (alertsSet()) {
	    Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_NOPACKET, "tcpSport", 0 );
	}
	return(none(m));
    }
}

int
tcpDport(MemVal *m)
{
    if (packet_data) {
	struct tcphdr * tp = tcpHdr();
	if (tp) {
	    if (tcpd.isHdr) {
		unsigned char * v = (unsigned char *)&tp->th_dport;
		m->val.i = ( (v[0] << 8) | (v[1]) );
		m->typ = NFR_TYPE_INT;
		return(0);
	    }
	}
	return(protocol(m));
    } else {
	if (alertsSet()) {
	    Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_NOPACKET, "tcpDport", 0 );
	}
	return(none(m));
    }
}

/*
**
*/

tcpDisplay()
{
    uiCurrPrintf( "tcpDisplay: nonce %d header (index) %d end (index) %d isTcp %d isHdr %d\n",
	tcp_nonce, ((char *)tcpd.tcpHdr)-packet_data,
	tcpd.tcpEnd-packet_data,
	tcpd.isTcp, tcpd.isHdr );
}

/*
**
*/

int
tcpPacketTriggerSet( t )
trigger_t * t;
{
    return(0);
}

/*
** scope.
**
** These don't really belong here, they're more suited to be inside
** run.c or sym.c... but ...
*/

scopeName( MemVal * m )
{
    if (runCurrentGet()) {
	m->typ = NFR_TYPE_STR;
	if (runCurrentGet()->f_scope) {
	    m->val.c = runCurrentGet()->f_scope;
	} else {
	    m->val.c = "";
	}
	return(0);
    }
    return(1);
}

scopeIndex( MemVal * m )
{
    if (runCurrentGet()) {
	m->typ = NFR_TYPE_INT;
	m->val.i = runCurrentGet()->f_scopeNdx;
	return(0);
    }
    return(1);
}
