/*
** Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
** All rights reserved.
**
** Use and distribution of this software and its source code
** are governed by the terms and conditions of the
** Network Flight Recorder Software License ("LICENSE.TXT")
*/



#include "nfrd.h"



/*
**	Main virtual machine interpreter. Includes inline
**  functions for stack operations, stack trace debugging, and
**  call execution.
**
**	Marcus J. Ranum, Network Flight Recorder, Inc. 1996
*/



static	int	returning = 0;				/* break out of function */
static	int	runstate = NFR_RUN_OK;			/* flag runtime problems */
static	int	cputicklimit = NFR_LIMIT_CPUTICK;	/* max cpu ticks/filter */
static	int	cputicks = 0;				/* current cpu ticks */
static	int	stackdepthlimit = NFR_LIMIT_STACK;	/* max stack depth */
static	int	stackframelimit = NFR_LIMIT_FRAME;	/* max call depth */


/* for runtime debugging ... #define NFR_RUNDEBUG */

/*
** tracing instructions...
*/
int	rFlag;

#ifdef	NFR_RUNDEBUG

static  void	run_printtyp(MemVal *);
static	void	run_printop(Machine *, Filter *, int, int, int);
#define	RUNDEBUG(x)	if (rFlag) x
runSetDebug()
{
    rFlag = 1;
}
#else

/*
** not tracing instructions.
*/
#define	RUNDEBUG(x)	;
runSetDebug()
{
    printf( "Run level tracing not compiled into this version of nfrd\n" );
}
#endif


/* push a value on the machine stack */
void
stack_push(Machine *m,MemVal *val)
{
	if(m->m_sp + 1 >= stackdepthlimit) {
		m->m_sp++;
		runstate = NFR_RUN_STACK;
		return;
	}

	/* machine stack too small: grow it! */
	if(m->m_sp >= m->m_maxsp) {
		m->m_maxsp = m->m_maxsp * 2;
		m->m_mem = (MemVal *)realloc(m->m_mem,sizeof(MemVal) * m->m_maxsp);
		if(m->m_mem == (MemVal *)0) {
			m->m_sp++;
			runstate = NFR_RUN_STACK;
			return;
		}
		RUNDEBUG( printf("machine stack grows: size=%d\n",m->m_maxsp) );
	}

	m->m_mem[m->m_sp] = *val;
	m->m_sp++;
}




/* take a value off the machine stack */
void
stack_pop(Machine *m,MemVal *val)
{
	m->m_sp = m->m_sp - 1;

	/* this should never happen but it's cheap to check */
	if(m->m_sp < 0) {
		val->typ = NFR_TYPE_ERR;
		val->val.i = NFR_ERR_STACK;
		runstate = NFR_RUN_STACK;
		RUNDEBUG( printf("machine stack pointer trashed (less than zero!)\n"));
		return;
	}

	/* this should never happen, either */
	if(m->m_sp >= m->m_maxsp) {
		val->typ = NFR_TYPE_ERR;
		val->val.i = NFR_ERR_STACK;
		runstate = NFR_RUN_STACK;
		RUNDEBUG(
		    printf("machine stack pointer trashed (greater than %d!)\n",m->m_maxsp) );
		return;
	}

	*val = m->m_mem[m->m_sp];
}



/* push a new frame on the stack */
int
frame_push(Machine *m,int ret_sp,int ret_ac, int ret_break)
{
	/* too many dwell levels stop processing */
	if(m->m_fp + 1 >= stackframelimit) {
		runstate = NFR_RUN_FRAME;
		RUNDEBUG( printf("machine frame limit (%d) exceeded\n",m->m_fp) );
		return(-1);
	}

	/* frame overflow */
	if(m->m_fp >= m->m_maxfp) {
		m->m_maxfp = m->m_maxfp * 2;
		m->m_fram = (Frame *)realloc(m->m_fram,sizeof(Frame) * m->m_maxfp);
		if(m->m_fram == (Frame *)0) {
			runstate = NFR_RUN_FRAME;
			return(-1);
		}
		RUNDEBUG( printf("machine frames grow: size=%d\n",m->m_maxfp) );
	}

	m->m_fram[m->m_fp].ret_sp = ret_sp;
	m->m_fram[m->m_fp].ret_ac = ret_ac;
	m->m_fram[m->m_fp].ret_break = ret_break;

	RUNDEBUG( printf("\nFRAME %d pushed, ret_sp=%d ret_ac=%d ret_break %d\n",m->m_fp,ret_sp,ret_ac, ret_break) );
	m->m_fp++;
	return(0);
}




Frame	*
frame_pop(Machine *m)
{
	if(m->m_fp >= m->m_maxfp) {
		RUNDEBUG(
		printf("machine frame pointer trashed (greater than %d!)\n",m->m_maxfp) );
		runstate = NFR_RUN_FRAME;
		return((Frame *)0);
	}

	m->m_fp = m->m_fp - 1;

	/* this shouldn't happen but why not be prepared */
	if(m->m_fp < 0) {
		RUNDEBUG( printf("machine frame pointer trashed (less than zero!)\n") );
		runstate = NFR_RUN_FRAME;
		return((Frame *)0);
	}
	RUNDEBUG( printf("\nFRAME %d pop, ret_sp=%d ret_ac=%d ret_break=%d\n", m->m_fp,
		m->m_fram[m->m_fp].ret_sp,
		m->m_fram[m->m_fp].ret_ac,
		m->m_fram[m->m_fp].ret_break) );
	return(&m->m_fram[m->m_fp]);
}




/*
** is_equal.
** handle comparision of equality or non-equality.
*/
int
is_equal(MemVal *lhs,MemVal *rhs)
{
	/* both lists - UCK! */
	if(lhs->typ == NFR_TYPE_LIST && rhs->typ == NFR_TYPE_LIST) {
		ListMem	*lp1 = (ListMem *)lhs->val.v;
		ListMem *lp2 = (ListMem *)rhs->val.v;

		while((lp1 != (ListMem *)0) && (lp2 != (ListMem *)0)) {
			if(is_equal(&(lp1->data),&(lp2->data)) == 0) {
				return(0);
			}
			lp1 = lp1->next;
			lp2 = lp2->next;
		}
		if ((lp1 == (ListMem *)0) && (lp2 == (ListMem *)0)) {
		    return(1);
		}
		return(0);
	}
#ifdef	OLD_LIST_COMPARE
	/* if comparing with a list, compare all elements recursively */
	if(rhs->typ == NFR_TYPE_LIST) {
		ListMem	*lp;

		lp = (ListMem *)rhs->val.v;
		while(lp != (ListMem *)0) {
			if(is_equal(&(lp->data),lhs))
				return(1);
			lp = lp->next;
		}
		return(0);
	}
	if(lhs->typ == NFR_TYPE_LIST) {
		ListMem	*lp;

		lp = (ListMem *)lhs->val.v;
		while(lp != (ListMem *)0) {
			if(is_equal(&(lp->data),rhs))
				return(1);
			lp = lp->next;
		}
		return(0);
	}
#endif	OLD_LIST_COMPARE


	/* compare strings */
	if(lhs->typ == NFR_TYPE_STR && rhs->typ == NFR_TYPE_STR) {
		if ((lhs->val.c == 0) || (rhs->val.c == 0)) {
		    if ((lhs->val.c == 0) && (rhs->val.c == 0)) {
			return(1);
		    } else {
			return(0);
		    }
		}
		if(!strcmp(lhs->val.c,rhs->val.c))
			return(1);
	}
	/* strings and patterns */
	if ((lhs->typ == NFR_TYPE_STR && rhs->typ == NFR_TYPE_PATTERN)
	|| (lhs->typ == NFR_TYPE_PATTERN && rhs->typ == NFR_TYPE_STR)) {
		if (patternMatch( lhs, rhs )) {
			return(1);
		}
	}

	/* compare errors */
	if(lhs->typ == NFR_TYPE_ERR && rhs->typ == NFR_TYPE_ERR)
		return(1);

	/* compare ints */
	if(lhs->typ == NFR_TYPE_INT && rhs->typ == NFR_TYPE_INT)
		return(rhs->val.i == lhs->val.i);

	if(lhs->typ == NFR_TYPE_IPV4HOST && rhs->typ == NFR_TYPE_IPV4HOST) {
		return(lhs->val.h.s_addr == rhs->val.h.s_addr);
	}
	if(lhs->typ == NFR_TYPE_IPV4HOST && rhs->typ == NFR_TYPE_IPV4NET) {
		return( (lhs->val.h.s_addr & rhs->val.n->ip_mask)
			== rhs->val.n->ip_addr.s_addr );
	}
	if(rhs->typ == NFR_TYPE_IPV4HOST && lhs->typ == NFR_TYPE_IPV4NET) {
		return( (rhs->val.h.s_addr & lhs->val.n->ip_mask)
			== lhs->val.n->ip_addr.s_addr );
	}
	if(rhs->typ == NFR_TYPE_IPV4NET && lhs->typ == NFR_TYPE_IPV4NET) {
		return( memcmp( rhs->val.n, lhs->val.n, sizeof( *rhs->val.n ) ) == 0 );
	}

	if (lhs->typ == NFR_TYPE_ETHMAC && rhs->typ == NFR_TYPE_ETHMAC) {
		return(memcmp( &rhs->val.e, &lhs->val.e, sizeof(lhs->val.e.e_addr) ) == 0);
	}
	if (lhs->typ == NFR_TYPE_PATTERN && rhs->typ == NFR_TYPE_PATTERN) {
		return(strcmp( lhs->val.p->p_str, rhs->val.p->p_str ) == 0 );
	}
	if (rhs->typ == NFR_TYPE_TRIGGER && rhs->typ == NFR_TYPE_TRIGGER) {
		return( lhs->val.t == rhs->val.t );
	}
	/* anything else can't be equal */
	return(0);
}

/*
** isElementOf
**
** binary operations against lists...
**
** inside does not commutive.
** 
*/

int
isElementOf( MemVal * lhs, MemVal * rhs ) {

	if (rhs->typ != NFR_TYPE_LIST) {
	    return(0);
	}
	{
	    /*
	    ** We're trying to find a list inside the other list.
	    ** open up the other lists, and compare entries...
	    */
	    ListMem	*rp;

	    rp = (ListMem *)rhs->val.v;
	    while(rp != (ListMem *)0) {
		if (is_equal( lhs, &(rp->data) )) {
		    return(1);
		}
		rp = rp->next;
	    }
	}
	return(0);
}

/*
** tEqual
** There are some times where we want to test for list inclusion
** along with normal equal tests. This happens mainly in trigger
** comparison routines, so that you can not only test against
** values, but lists of values.
*/

int
tEqual(MemVal *lhs,MemVal *rhs)
{
    int eq;

    if ((eq = is_equal( lhs, rhs )) == 0) {
	if ((lhs->typ == NFR_TYPE_LIST) && (rhs->typ != NFR_TYPE_LIST)) {
	    return( isElementOf( rhs, lhs ) );
	} else
	if ((lhs->typ != NFR_TYPE_LIST) && (rhs->typ == NFR_TYPE_LIST)) {
	    return( isElementOf( lhs, rhs ) );
	}
	return(0);
    } else {
	return(eq);
    }
}


/*
** test_bool
** return a meaningful boolean value from an object
*/
static	int
test_bool(MemVal *d)
{
	if(d->typ == NFR_TYPE_INT && d->val.i == 0) {
		return(0);
	}
	if(d->typ == NFR_TYPE_STR && *(d->val.c) == '\0') {
		return(0);
	}
	if(d->typ == NFR_TYPE_ERR) {
		return(0);	/* errors should always be false */
	}
	return(1);
}



/*
** numeric_op
** handle basic binary numeric operations. returning nonzero indicates an
** error condition (currently relevant only in division by zero).
** the "answer" is left in d1
*/
static	void
numeric_op(int op,MemVal *d1,MemVal *d2)
{
	switch(NFR_OP(op)) {
	case NFR_OP_ADD:
		d1->val.i = d2->val.i + d1->val.i;
		break;
	case NFR_OP_SUB:
		d1->val.i = d2->val.i - d1->val.i;
		break;
	case NFR_OP_DIV:
		if(d1->val.i == 0) {
			d1->typ = NFR_TYPE_ERR;
			d1->val.i = NFR_ERR_ZDIV;
		} else {
			d1->val.i = d2->val.i / d1->val.i;
		}
		break;
	case NFR_OP_MUL:
		d1->val.i = d2->val.i * d1->val.i;
		break;
	case NFR_OP_LT:
		d1->val.i = (d2->val.i < d1->val.i);
		break;
	case NFR_OP_LTE:
		d1->val.i = (d2->val.i <= d1->val.i);
		break;
	case NFR_OP_GT:
		d1->val.i = (d2->val.i > d1->val.i);
		break;
	case NFR_OP_GTE:
		d1->val.i = (d2->val.i >= d1->val.i);
		break;
	case NFR_OP_BOR:
		d1->val.i = (d2->val.i | d1->val.i);
		break;
	case NFR_OP_XOR:
		d1->val.i = (d2->val.i ^ d1->val.i);
		break;
	case NFR_OP_BAND:
		d1->val.i = (d2->val.i & d1->val.i);
		break;
	case NFR_OP_LSHIFT:
		d1->val.i = (d2->val.i << d1->val.i);
		break;
	case NFR_OP_RSHIFT:
		d1->val.i = (((unsigned int)d2->val.i) >> d1->val.i);
		break;
	}
}

/*
** run_assign.
**
** pass in the scope index, a reference to a Sym * (usually
** the local symbols of the function), the name of the symbol, whether its 
** local or global (via the op), and the value.
**
** if the value is null, then we remove the symbol from existance.
**
** Return 1 on success, 0 on failure.
*/

int
run_assign( Filter * f, char * s, int op, MemVal * val )
{
    Sym ** ref;
    Sym * sp = 0;
    Map * m;

    if (NFR_OP(op) == NFR_OP_GASGN) {
	/*
	** If the global symbol isn't previously
	** defined, then we make it in a map
	** indexed by the function... if its not
	** there, make one.
	*/
	if (globalMapBag == 0) {
	    globalMapBag = mapBagNew( 0 );
	}
	m = mapBagFind( globalMapBag, f->f_scopeNdx, f->f_scope );
	if ( m == 0 ) {
	    m = mapNew( 0, f->f_scopeNdx, f->f_scope, 0 );
	    mapBagAdd( globalMapBag, m );
	}
	ref = &m->m_sym;

    } else if (NFR_OP(op) == NFR_OP_LASGN) {
	if (f == 0) {
	    /* XXX log something */
	    return(0);
	} else {
	    ref = &f->f_syms;
	}
    } else {
	Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_SYMANTICERR,
	    "run_assign: invalid op !(NFR_OP_GASGN|NFR_OP_LASGN)", 0 );
	return(0);
    }

    if((sp = symlook(s,*ref)) == (Sym *)0) {
	if (memvalIsNull(val)) {
	    /* sure we can remove a symbol which isn't there  */
	    return(1);
	}
	sp = symnew(s, val);
	if(sp == (Sym *)0) {
	    /*
	    ** failure, can't make the new symbol.
	    */
	    return(0);
	}
	*ref = symadd(sp,*ref);
    } else {
	/* if it exists, it could be an indirect store. */
	if (sp->s_data.typ == NFR_TYPE_INDIRECT) {
	    /*
	    ** look up the symbol on the OTHER list.
	    */
	    Map * map;
	    while (sp && sp->s_data.typ == NFR_TYPE_INDIRECT) {
		map = sp->s_data.val.v;
		sp = symlook( s, map->m_sym );
	    }

	    if ((sp = symlook( s, map->m_sym )) == (Sym *)0) {
		if (memvalIsNull(val)) {
		    /* sure we can remove a symbol that isn't there */
		    return(1);
		}
		/* add the new symbol to the map */
		sp = symnew( s, val );
		if (sp == (Sym *)0) {
		    return(0);
		}
		map->m_sym = symadd( sp, map->m_sym );
	    } else {
		memvalfreedata(&(sp->s_data));
		if (memvalIsNull(val)) {
		    /* remove the symbol */
		    *ref = symremove( sp, *ref );
		} else {
		    memvalclonedata(&(sp->s_data), val );
		}
	    }
	} else {
	    memvalfreedata(&(sp->s_data));
	    if (memvalIsNull(val)) {
		/* remove the symbol */
		*ref = symremove( sp, *ref );
	    } else {
		memvalclonedata(&(sp->s_data), val);
	    }
	}
    }
    symDirty(sp);
    return(1);
}

/*
** Convience routine.
*/
run_globalAssign( char * name, MemVal * val )
{
    return(run_assign( globalFilter, name, NFR_OP_GASGN, val ));
}

/*
** run_lookup
**
** Look up a variable in either a local or a global context.
** Return 0 for failure, the sym * for success.
**
** We also look up symbols names with the ':' separator.
** we assume that prefix means global scope.
** This was orignally used to acquire values of alert names.
*/
Filter run_lookup_filt[1];

Sym *
run_lookup( Filter * f, char * s, int op )
{
    Sym * sp = 0;
    Sym * sym = 0;

    if (NFR_OP(op) == NFR_OP_GEVAL) {
	/*
	** if the globalMapBag isn't there yet, then
	** no variables are ready
	*/
	if (globalMapBag != 0) {
	    Map * m;
	    m = mapBagFind( globalMapBag, f->f_scopeNdx, f->f_scope );
	    if (m) {
		sym = m->m_sym;
	    }
	}
    } else if (NFR_OP(op) == NFR_OP_LEVAL) {
	sym = f->f_syms;
    } else {
	Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_SYMANTICERR,
		"run_lookup: bad value of op !(NFR_OP_GEVAL/NFR_OP_LEVAL)", 0 );
	return(0);
    }

    if (sym) {
	
	sp = symlook( s, sym );
	if(sp == (Sym *)0) {
	    /* see if there's something  special about the name. */
	    char * ndx;
	    globalLookup:
	    if ((ndx = strchr( s, ':' )) != 0) {
		Sym * val;
		/*
		** i don't like clobbering storage which isn't our
		** property to touch.  This is quick, tho, and we
		** return the data to its original value.
		*/
		ndx[0] = 0;
		run_lookup_filt->f_scope = s;
		run_lookup_filt->f_scopeNdx = strHash(s);
		val = run_lookup( run_lookup_filt, ndx+1, NFR_OP_GEVAL );
		ndx[0] = ':';
		return(val);
	    } else {
		return(0);
	    }
	} else {
	    if (sp->s_data.typ == NFR_TYPE_INDIRECT) {
		Map * map;
		/*
		** need a recursion sanity check.. how many times?
		*/
		while (sp && sp->s_data.typ == NFR_TYPE_INDIRECT) {
		     map = (Map *)sp->s_data.val.v;
		     sp = symlook(s, map->m_sym );
		}

		if (sp == (Sym *)0) {
		    return(0);
		} else {
		    return(sp);
		}
	    } else {
		return(sp);
	    }
	}
    } else {
	goto globalLookup;
    }
    return(0);
}

/*
**
*/

Sym *
run_globalLookup( char * name )
{
    return( run_lookup( globalFilter, name, NFR_OP_GEVAL ) );
}

/*
** a little machinery to keep track of the current filter.
*/

Filter * runCurrent;		/* current filter we're running. */

Filter *
runCurrentSet( Filter *f )
{
    Filter * old = runCurrent;
    runCurrent = f;
    return(old);
}

Filter *
runCurrentGet()
{
    return(runCurrent);
}

/*
** a little machinery to keep track of the current op.
*/

int	runCurrentOp;


/*
** There are several OK status from an interpetive exit.
** The NFR_RUN_RETURN can be returned from a call to the
** interpreter iff we returned from the subroutine.  That
** return is used by the engine to help deal with breaks inside
** while or foreach loops.
*/
#define	RUN_OK(x)	((x == NFR_RUN_OK) || (x == NFR_RUN_RETURN))

/*
** loop across a list of machine instructions, obeying them dutifully.
** we *COULD* get some extra speed out of stuff like ASGN, EVAL, and
** so on, by avoiding calls to stack_pop and stack_push, but that
** kind of hackery can be done if it proves necessary.
*/
int
machinerun(Machine *m,Filter *f,int pc)
{
	MemVal	d1;		/* stack data */
	MemVal	d2;
	Frame	*framep;	/* misc. frame pointer */
	Filter	*funcp;		/* used for function CALL */
	Bltin	*bltnp;		/* used for builtin CALL */
	Sym	*sp;		/* var ptr for locals. */
	int	op;		/* saved operator */
	int	tmp1;
	int	tmp2;
	int	tmp3;
	int	runval;
	Filter *oldRunCurrent = runCurrentSet(f);

	RUNDEBUG( (void)printf("Entering machinerun(pc=%d)\n",pc) );

	/* unset returning */
	returning = 0;


	/* main loop */
	while(NFR_OP(op = f->f_mem[pc]) != NFR_OP_STOP) {
		if(++cputicks >= cputicklimit)
			runstate = NFR_RUN_CPU;

		RUNDEBUG( (void)run_printop(m,f,pc,m->m_sp,op) );

		if(!RUN_OK(runstate)) {
			runCurrent = oldRunCurrent;
			return(runstate);
		}


		switch(NFR_OP(runCurrentOp = op)) {

		/* handle binary numeric operations */
		case NFR_OP_ADD:
		case NFR_OP_SUB:
		case NFR_OP_DIV:
		case NFR_OP_MUL:
		case NFR_OP_LT:
		case NFR_OP_LTE:
		case NFR_OP_GT:
		case NFR_OP_GTE:
		case NFR_OP_BOR:
		case NFR_OP_XOR:
		case NFR_OP_BAND:
		case NFR_OP_LSHIFT:
		case NFR_OP_RSHIFT:
		stack_pop(m,&d1);
		stack_pop(m,&d2);
		if(d1.typ != NFR_TYPE_INT || d2.typ != NFR_TYPE_INT) {
			d2.val.i = NFR_ERR_NONUM;
			d2.typ = NFR_TYPE_ERR;
			stack_push(m,&d2);
			break;
		}
		numeric_op(op,&d1,&d2);
		stack_push(m,&d1);
		break;



		/*
		** call a function or builtin by name.
		** The parameter must be something we can
		** use to look up a name.  Currently, the
		** only thing that count that way is 
		** a string...
		*/
		case NFR_OP_ECALL:
		    pc++;
		    tmp2 = f->f_mem[pc]; /* arg count */
		    d1 = (m->m_mem[m->m_sp-tmp2-1]);
		    if (d1.typ == NFR_TYPE_STR) {
			tmp3 = (m->m_sp - tmp2); /* caller's stack addr */
			tmp1 = tmp3-1; /* where to unroll our stack to. */
			goto nfr_op_ecall;
		    } else {
			d1.typ = NFR_TYPE_ERR;
			d1.val.i = NFR_ERR_USER;
			m->m_sp -= tmp2;
			m->m_sp --;
		    }
		break;

		/*
		** call a function or builtin. builtins will be
		** overridden by defined functions with the same
		** name, UNLESS builtins are specified with a
		** leading '_'.
		*/
		case NFR_OP_CALL:

		/* name of function/builtin is in next pc */
		pc++;
		d1.val.c = f->f_list[f->f_mem[pc]]->data.val.c;
		pc++;
		tmp2 = f->f_mem[pc]; /* arg count */

		funcBuiltIn:
		tmp3 = (m->m_sp - tmp2); /* return stack addr */
		tmp1 = tmp3;

		nfr_op_ecall:
		/* if name doesn't start with '_' look for function */
		if(d1.val.c[0] != '_') {
			funcp = funcfind(d1.val.c, f->f_scope);

			/* found a function: call it */
			if(funcp != (Filter *)0) {
				returning = 0;

				if (funcp->f_indSymCnt) {
				    symFixIndirect(funcp->f_syms);
				    funcp->f_indSymCnt = 0;
				}
				/*
				** it is our responsibility to repair
				** the stack when it's done with it.
				*/
				if(frame_push(m,tmp3,tmp2,0)) {
					d2.typ = NFR_TYPE_ERR;
					d2.val.i = NFR_ERR_FRAME;
					m->m_sp = tmp1; /* unroll the stack manually */
					stack_push(m,&d2);
					break;
				}
				funcp->f_useCount++;
				runval = machinerun(m,funcp,0);
				RUNDEBUG( log(FL,"CALL: returned from %s %d\n",
					d1.val.c, runval) );
				if (!RUN_OK(runval)) {
					runCurrent = oldRunCurrent;
					RUNDEBUG( printf("Exiting machinerun %d: %d\n",
						__LINE__, runval) );
					return(runval);
				}
				runval = NFR_RUN_OK;

				returning = 0;
				break;
			}
		}

		/* look for a builtin to execute */
		if((bltnp = bltlookup(d1.val.c)) == (Bltin *)0) {
			/* not found: do some cleanup */
			d2.typ = NFR_TYPE_ERR;
			d2.val.i = NFR_ERR_NOFUNC;
			m->m_sp = tmp3; /* unroll the stack manually */
			RUNDEBUG( log( FL, "can't find func %s\n", d1.val.c ) );
			stack_push(m,&d2);
			break;
		}

		/* call it */
		(*bltnp->func)(m,tmp2,tmp3,tmp3);

		/* new stack pointer is unrolled */
		m->m_sp = tmp3 + 1;
		returning = 0;
		break;

		/*
		** assign to a local/global variable. 
		*/
		case NFR_OP_LASGN:
		case NFR_OP_GASGN:
		{
		    /* next instruction is name of variable */
		    pc++;
		    /* and take the value to assign to off the stack. */
		    stack_pop(m,&d2);

		    if (run_assign( f, f->f_list[f->f_mem[pc]]->data.val.c, op, &d2 ) == 0) {
			d2.val.i = NFR_ERR_OOM;
			d2.typ = NFR_TYPE_ERR;
			stack_push(m, &d2);
		    } else {
			stack_push(m,&d2);
		    }
		}
		break;

		/*
		** array assignment.
		*/

		case NFR_OP_ARRAYGASGN:
		case NFR_OP_ARRAYLASGN:
		{
		    char * name;
		    int lop = NFR_OP_GEVAL;
		    int aop = NFR_OP_GASGN;
		    /* next instruction is name of variable */
		    pc++;
		    /* and take the value to assign to off the stack. */
		    stack_pop(m,&d2);
		    stack_pop(m,&d1);

		    if (NFR_OP(op) == NFR_OP_ARRAYLASGN) {
			lop = NFR_OP_LEVAL;
			aop = NFR_OP_LASGN;
		    }

		    name = f->f_list[f->f_mem[pc]]->data.val.c;
		    if (((sp = run_lookup( f, name, lop )) == 0)
		    ||  (sp->s_data.typ != NFR_TYPE_MAPBAG)) {
			MemVal new[1];
			new->val.v = 0;
			new->typ = NFR_TYPE_MAPBAG;

			run_assign( f, name, aop, new );
			sp = run_lookup( f, name, lop );
		    }
		    arraySet( &(sp->s_data), &d1, &d2 );
		    stack_push(m,&d2);
		}
		break;


		/*
		** array assignment, but without a direct reference.
		** instead, we check the first thing on the stack
		** to make sure its a MAPBAG.
		*/
		case NFR_OP_ARRAYEASGN:
		{
		    MemVal mb[1];
		    stack_pop(m,&d2);
		    stack_pop(m,&d1);
		    stack_pop(m,mb);
		    if (mb->typ != NFR_TYPE_MAPBAG) {
			MemVal err[1];
			err->typ = NFR_TYPE_ERR;
			err->val.i = NFR_ERR_USER;
			printf( "NFR_OP_ARRAYEASGN: not mapbag: %d\n", mb->typ ); 
		    } else {
			arraySet( mb, &d1, &d2 );
		    }
		}
		break;

		/*
		** push the parameter count on the stack. sneak a look at the
		** stack frame, and push the arg count.
		*/
		case NFR_OP_CPUSH:
		if(m->m_fp == 0)
			d2.val.i = 0;
		else
			d2.val.i = m->m_fram[m->m_fp - 1].ret_ac;
		d2.typ = NFR_TYPE_INT;
		stack_push(m,&d2);
		break;




		case NFR_OP_EQ:
		stack_pop(m,&d1);
		stack_pop(m,&d2);
		d1.val.i = is_equal(&d2,&d1);
		d1.typ = NFR_TYPE_INT;
		stack_push(m,&d1);
		break;

		case NFR_OP_INSIDE:
		stack_pop(m,&d1);
		stack_pop(m,&d2);
		d1.val.i = isElementOf(&d2,&d1);
		d1.typ = NFR_TYPE_INT;
		stack_push(m,&d1);
		break;


		case NFR_OP_NE:
		stack_pop(m,&d1);
		stack_pop(m,&d2);
		d1.val.i = (!is_equal(&d2,&d1));
		d1.typ = NFR_TYPE_INT;
		stack_push(m,&d1);
		break;


		/* push a NULL on the stack */
		case NFR_OP_NULPUSH:
		d2.val.i = NFR_ERR_NULL;
		d2.typ = NFR_TYPE_ERR;
		stack_push(m,&d2);
		break;


		/* push a local/global variable on the stack */
		case NFR_OP_LEVAL:
		case NFR_OP_GEVAL:
		{
		    Sym * sym;
		    MemVal * mem = &d2;

		    pc++;
		    if ((sp = run_lookup( f, f->f_list[f->f_mem[pc]]->data.val.c, op )) == NULL) {
			d2.val.i = NFR_ERR_NOVAR;
			d2.typ = NFR_TYPE_ERR;
			mem = &d2; /* sanity */
		    } else {
			mem = &sp->s_data;
		    }
		    stack_push( m, mem );
		}
		break;

		/*
		** array references.
		** Two values are needed, the expression to use as the
		** array reference, and the array name.  For right now
		** we'll use singly subscripted arrays...
		*/
		case NFR_OP_ARRAYLREF:
		case NFR_OP_ARRAYGREF:
		{
		    MemVal *  mem;
		    int lop = NFR_OP_GEVAL;
		    char * name;
		    stack_pop(m,&d1);
		    pc++;

		    if (NFR_OP(op) == NFR_OP_ARRAYLREF) {
			lop = NFR_OP_LEVAL;
		    }
		    name = f->f_list[f->f_mem[pc]]->data.val.c;
		    if ((sp = run_lookup( f, name, lop )) == NULL) {
			d2.val.i = NFR_ERR_NOVAR;
			d2.typ = NFR_TYPE_ERR;
			mem = &d2; /* sanity */
		    } else {
			mem = &sp->s_data;
			mark();
		    }
		    /* array reference. */
		    {
			Sym * s = arrayGet( mem, &d1, 0 );
			if (s) {
			    stack_push(m,&s->s_data);
			} else {
			    d2.val.i = NFR_ERR_NOVAR;
			    d2.typ = NFR_TYPE_ERR;
			    stack_push(m, &d2);
			}
		    }
		}
		break;


		/*
		** In this case, we aren't looking up the symbol,
		** but just checking to see if the expression
		** at the top of the stack evaluates to 
		** a mapBag.  If it does, then we just get the
		** thing.
		*/
		case NFR_OP_ARRAYEREF:
		{
		    stack_pop(m,&d2);	/* the index */
		    stack_pop(m,&d1);	/* the mapbag */

		    if (d1.typ == NFR_TYPE_MAPBAG) {
			Sym * s = arrayGet( &d1, &d2, 0 );
			if (s) {
			    stack_push(m,&s->s_data);
			} else {
			    d2.val.i = NFR_ERR_NOVAR;
			    d2.typ = NFR_TYPE_ERR;
			    stack_push(m, &d2);
			    printf( "NFR_OP_ARRAYEREF: no element: %d\n", d1.typ ); 
			}
		    } else {
			d2.val.i = NFR_ERR_USER;
			d2.typ = NFR_TYPE_ERR;
			stack_push(m, &d2);
			printf( "NFR_OP_ARRAYEREF: not mapbag: %d\n", d1.typ ); 
		    }
		}
		break;



		/* pull a coded value from the current packet */
		case NFR_OP_PACKET:
		/* next instruction is number of variable */
		pc++;
		tmp1 = f->f_mem[pc];
		if(packet_get(tmp1, &d2)) {
			d2.val.i = NFR_ERR_BADVAL;
			d2.typ = NFR_TYPE_ERR;
			stack_push(m,&d2);
		} else
			stack_push(m,&d2);
		break;

		/* connect to the recorder's outbound port */
		case NFR_OP_RECORD:
		    pc++;
		    tmp2 = f->f_mem[pc];	/* saved number of expressions */
		    d1.val.c = "send";
		    goto funcBuiltIn;
		break;




		/*
		** cope with an if statement by unrolling the address
		** in the program to jump to, making a recursive
		** call to run() with that program counter, and returning
		** to execute. we do some weird stack-fiddling but it works.
		*/
		case NFR_OP_IF:
		mark();
		pc++;
		tmp1 = f->f_mem[pc];	/* if/true part */
		pc++;
		tmp2 = f->f_mem[pc];	/* if/else part */
		pc++;
		tmp3 = f->f_mem[pc];	/* continuation */

		/*
		** this is subtle. incrementing the program counter points it
		** to the assembled code for the condition. run the condition
		** and resume operation at the continuation branch.
		*/
		pc++;
		runval = machinerun(m,f,pc);
		if(!RUN_OK(runval)) {
			runCurrent = oldRunCurrent;
			RUNDEBUG( printf("Exiting machinerun %d: %d\n",
				__LINE__, runval) );
			return(runval);
		}
		runval = NFR_RUN_OK;

		/* the results of the "if" condition are on the stack */
		stack_pop(m,&d1);

		/* test the results of the "if" condition */
		if(test_bool(&d1)) {
			if(tmp1 != 0) {
				runval = machinerun(m,f,tmp1);
				if (runval != NFR_RUN_OK) {
					runCurrent = oldRunCurrent;
					RUNDEBUG( printf("Exiting machinerun %d: %d\n",
						__LINE__, runval) );
					return(runval);
				}
			}
		} else {
			if(tmp2 != 0) {
				runval = machinerun(m,f,tmp2);
				if (runval != NFR_RUN_OK) {
					runCurrent = oldRunCurrent;
					RUNDEBUG( printf("Exiting machinerun %d: %d\n",
						__LINE__, runval) );
					return(runval);
				}
			}
		}

		/* the "if" or "else" side may have a return in it */
		if(returning) {
			runCurrent = oldRunCurrent;
			RUNDEBUG( printf("Exiting machinerun %d: %d\n",
				__LINE__, tmp1) );
			return(NFR_RUN_OK);
		}

		/* cheat. the program counter will be post-incremented */
		pc = tmp3 - 1;
		break;



		/* push a numeric value from program space onto the stack. */
		case NFR_OP_NPUSH:
		pc++;
		d2.val.i = f->f_mem[pc];
		d2.typ = NFR_TYPE_INT;
		stack_push(m,&d2);
		break;

		case NFR_OP_IPV4HOSTPUSH:
		pc++;
		d2.val.h.s_addr = f->f_mem[pc];
		d2.typ = NFR_TYPE_IPV4HOST;
		stack_push(m,&d2);
		break;

		case NFR_OP_IPV4NETPUSH:
		{
		    ipv4net_t * n = (ipv4net_t *)tmpcalloc( sizeof(*n) );
		    pc++;
		    n->ip_addr.s_addr = f->f_mem[pc];
		    pc++;
		    n->ip_mask = f->f_mem[pc];
		    d2.val.n = n;
		}
		d2.typ = NFR_TYPE_IPV4NET;
		stack_push(m,&d2);
		break;


		/* pop (and throw away) a value off the stack */
		case NFR_OP_POP:
		stack_pop(m,&d1);
		break;



		/*
		** return from a function call or primary machine run.
		** this is done by popping a stack frame, using it
		** to reset the stack, and pushing a fake value
		** onto the stack, or a real one, depending.
		*/
		case NFR_OP_RETNV:
		case NFR_OP_RETV:
		returning++;
		tmp1 = NFR_RUN_OK;

		/* if returning a value, acquire a value to return. */
		if(NFR_OP(op) == NFR_OP_RETV)
			stack_pop(m,&d1);

		/*
		** if we were inside a break scope, we need to
		** return to the upper level, after popping 
		** all the break frames...
		*/
		if (m->m_fp > 0 && m->m_fram[m->m_fp-1].ret_break) {
		    tmp1 = NFR_RUN_RETURN;
		    while (m->m_fram[m->m_fp-1].ret_break) {
			frame_pop(m);
		    }
		}

		/* we are leaving run() entirely */
		if(m->m_fp == 0) {
			runCurrent = oldRunCurrent;
			RUNDEBUG( printf("Exiting machinerun: %d\n", tmp1) );
			return(tmp1);
		}

		if((framep = frame_pop(m)) == 0) {
			runCurrent = oldRunCurrent;
			RUNDEBUG( printf("Exiting machinerun %d: %d\n",
				__LINE__, -1) );
			return(-1);
		}
		m->m_sp = framep->ret_sp;

		if(NFR_OP(op) == NFR_OP_RETNV) {
			d2.val.i = NFR_ERR_NOVAR;
			d2.typ = NFR_TYPE_ERR;
			stack_push(m,&d2);
		} else {
			stack_push(m,&d1);
		}
		RUNDEBUG( printf("RETNV/RETV resets stack pointer to %d (after push %d)\n",
			framep->ret_sp, m->m_sp) );
		RUNDEBUG({
			blt_echo(&m->m_mem[m->m_sp - 1]);
		})

		runCurrent = oldRunCurrent;
		RUNDEBUG( printf("Exiting machinerun %d: %d\n", __LINE__, tmp1) );
		return(tmp1);



		/*
		** push a parameter value onto the stack.
		** sneak a look at the stack frame, find the offset
		** to the value in question, and push it again -
		** unless it is invalid. grotty, huh?
		*/
		case NFR_OP_PPUSH:
		pc++;

		/* don't let anyone look at values we don't have or $0 or less */
		if(m->m_fp <= 0 || f->f_mem[pc] > m->m_fram[m->m_fp - 1].ret_ac || f->f_mem[pc] <= 0) {
			d2.typ = NFR_TYPE_ERR;
			d2.val.i = NFR_ERR_NOVAR;
			stack_push(m,&d2);
		} else {
			int	sa;

			sa = m->m_fram[m->m_fp - 1].ret_sp + f->f_mem[pc] - 1;
			stack_push(m,&(m->m_mem[sa]));
		}
		break;



		/*
		** push a string onto the stack from program space.
		** use the index into the program string space
		** (the next value in the program space) to get
		** a pointer to the string, which is pushed onto
		** the stack.
		*/
		case NFR_OP_SPUSH:
		d2.typ = NFR_TYPE_STR;
		pc++;
		d2.val.c = f->f_list[f->f_mem[pc]]->data.val.c;
		stack_push(m,&d2);
		break;



		/*
		** push a list onto the stack. to simplify
		** recursive data structures we stash it as a
		** correctly typed void *
		*/
		case NFR_OP_LPUSH:
		d2.typ = NFR_TYPE_LIST;
		pc++;
		d2.val.v = f->f_list[f->f_mem[pc]];
		stack_push(m,&d2);
		break;

		/*
		** push an ethernet mac address onto the stack.
		*/
		case NFR_OP_ETHMACPUSH:
		pc++;
		d2 = f->f_list[f->f_mem[pc]]->data;
		stack_push(m,&d2);
		break;



		/* logical joins */
		case NFR_OP_AND:
		case NFR_OP_OR:
		stack_pop(m,&d1);
		stack_pop(m,&d2);

		if(NFR_OP(op) == NFR_OP_AND)
			d1.val.i = (test_bool(&d1) && test_bool(&d2));
		else
			d1.val.i = (test_bool(&d1) || test_bool(&d2));
		d1.typ = NFR_TYPE_INT;
		stack_push(m,&d1);
		break;



		/*
		** negate/invert by popping top value, neg/inverting it, and
		** pushing the result.
		*/
		case NFR_OP_NEG:
		stack_pop(m,&d1);
		if(d1.typ != NFR_TYPE_INT) {
			d2.val.i = NFR_ERR_NONUM;
			d2.typ = NFR_TYPE_ERR;
			stack_push(m,&d2);
			break;
		}
		d1.val.i = - (d1.val.i);
		stack_push(m,&d1);
		break;



		/*
		** invert by popping top value, inverting it, and
		** pushing the result.
		*/
		case NFR_OP_NOT:
		stack_pop(m,&d1);
		d2.val.i = !(test_bool(&d1));
		d2.typ = NFR_TYPE_INT;
		stack_push(m,&d2);
		break;

		case NFR_OP_INV:
		stack_pop(m,&d1);
		if (d1.typ != NFR_TYPE_INT) {
		    d1.typ = NFR_TYPE_ERR;
		    d1.val.i = NFR_ERR_USER;
		} else {
		    d1.val.i = ~d1.val.i;
		}
		stack_push(m,&d1);
		break;



		/* IF conditions have a magic end tag */
		case NFR_OP_IFCOND:
		runCurrent = oldRunCurrent;
		RUNDEBUG( printf("Exiting machinerun %d : %d\n",
			__LINE__, NFR_RUN_OK) );
		return(NFR_RUN_OK);



		/* function trigger evaluation. */
		case NFR_OP_FUNCON:
		run_perform_on(m, f, 1);
		break;

		/* ON trigger evaluation. */
		case NFR_OP_ON:
		run_perform_on(m, f, 0);
		break;


		/* OFF trigger evaluation. */
		case NFR_OP_OFF:
		run_perform_off(m, f);
		break;


		/* DECLARE/INSIDE */
		case NFR_OP_DECL_INSIDE:
		pc++;
		d1.val.c = f->f_list[f->f_mem[pc]]->data.val.c;
		stack_pop(m,&d2);
		run_perform_decl_inside( f, d1.val.c, &d2 );
		break;

		/* DECLARE/GLOBAL */
		case NFR_OP_DECL_GLOBAL:
		pc++;
		d1.val.c = f->f_list[f->f_mem[pc]]->data.val.c;
		run_perform_decl_global( f, d1.val.c );
		break;

		/* FOREACH */
		case NFR_OP_FOREACH:
		{
		    /* the first thing is what gets assigned. 
		    ** the thing on the stack better be a list.
		    ** the fourth thing is the location to continue after we're done.
		    */
		    char * iterator;
		    int next;
		    pc++;
		    iterator = f->f_list[f->f_mem[pc]]->data.val.c;
		    pc++;
		    next = f->f_mem[pc];
		    pc++;

		    stack_pop(m,&d2);

		    /* we want a way to tell break its ok... to break */
		    if (frame_push(m,m->m_fram[m->m_fp-1].ret_sp,f->f_mem[pc], 1)) {
			d2.typ = NFR_TYPE_ERR;
			d2.val.i = NFR_ERR_FRAME;
			stack_push(m,&d2);
			break;
		    }

		    if (d2.typ == NFR_TYPE_LIST) {
			ListMem * mem = d2.val.v;
			while (mem) {
			    /* run the interpreter starting at the next instruction till
			    ** it stops.
			    */

			    RUNDEBUG( printf("FOREACH, next iteration: $%s\n", iterator) );
			    if (run_assign( f, iterator, NFR_OP_LASGN, &mem->data )) {
				if ((runval = machinerun( m, f, pc )) == NFR_RUN_BREAK) {
				    runval = NFR_RUN_OK;
				    break;
				} else  if (runval == NFR_RUN_RETURN) {
				    if(m->m_fp == 0) {
					RUNDEBUG( printf("Exiting machinerun %d: %d\n",
						__LINE__, NFR_RUN_OK) );
					return(NFR_RUN_OK);
				    } else {
					RUNDEBUG( printf("Exiting machinerun %d: %d\n",
						__LINE__, NFR_RUN_RETURN) );
					return(NFR_RUN_RETURN);
				    }
				} else {
				    /* test errors? */
				}
			    }
			    mem = mem->next;
			}
		    } else if (d2.typ == NFR_TYPE_MAPBAG) {
			MapBag * mb = (MapBag *)d2.val.v;
			int mbi;

			if (mb) {
			    for ( mbi =0; mbi < mb->mb_width; ++mbi ) {
				Map * map = mb->mb_maps[mbi];
				while (map) {
				    Sym * s = map->m_sym;
				    while (s) {
					/*
					** assign the iterator to the
					** value of the sym.
					*/
					MemVal mv[1], *mp = mv;

					switch(SYM_FLAG_TYPE(s)) {
					    case SYM_FLAG_TYPE_NAME:
						mv->typ = NFR_TYPE_STR;
						mv->val.c = symName(s);
						mp = mv;
						break;

					    case SYM_FLAG_TYPE_INT:
						mv->typ = NFR_TYPE_INT;
						mv->val.i = s->s_int;
						mp = mv;
						break;

					    case SYM_FLAG_TYPE_HOST:
						mv->typ = NFR_TYPE_IPV4HOST;
						mv->val.h = s->s_host;
						mp = mv;
						break;

					    case SYM_FLAG_TYPE_MEMVAL:
						mp = s->s_memVal;
						break;

					    default:
						log( FL, "foreach/array, unknown type\n" );
					}
					if (run_assign( f, iterator,
						NFR_OP_LASGN, mp )) {
					    if ((runval = machinerun( m, f, pc )) == NFR_RUN_BREAK) {
						runval = NFR_RUN_OK;
						break;
					    } else  if (runval == NFR_RUN_RETURN) {
						if(m->m_fp == 0) {
						    RUNDEBUG( printf("Exiting machinerun %d: %d\n",
							    __LINE__, NFR_RUN_OK) );
						    return(NFR_RUN_OK);
						} else {
						    RUNDEBUG( printf("Exiting machinerun %d: %d\n",
							    __LINE__, NFR_RUN_RETURN) );
						    return(NFR_RUN_RETURN);
						}
					    } else {
						/* test errors? */
					    }
					}
					s = s->s_next;
				    }
				map = map->m_next;
				}
			    }
			}
		    }
		    frame_pop(m);
		    pc = next-1;  /* continue here (note: pc gets incrd) */
		}
		break;

		/* WHILE */
		case NFR_OP_WHILE:
		{
		    int next, statement;
		    pc++;
		    statement = f->f_mem[pc];
		    pc++;
		    next = f->f_mem[pc];
		    pc++;
		    /* we want a way to tell break its ok... to break */
		    if (frame_push(m,m->m_fram[m->m_fp-1].ret_sp,f->f_mem[pc], 1)) {
			d2.typ = NFR_TYPE_ERR;
			d2.val.i = NFR_ERR_FRAME;
			stack_push(m,&d2);
			break;
		    }

		    /* run the machine at the next instrction
		    ** till we stop... review the condition.
		    */
		    while ((runval = machinerun(m,f,pc)) == NFR_RUN_OK) {
			stack_pop(m,&d1);
			RUNDEBUG( printf("WHILE, iteration:\n") );
			if (test_bool(&d1)) {
			    if ((runval = machinerun( m, f, statement )) == NFR_RUN_BREAK) {
				runval = NFR_RUN_OK;
				break;
			    } else if (runval == NFR_RUN_RETURN) {
				if (m->m_fp == 0) {
				    RUNDEBUG( printf("Exiting machinerun %d: %d\n",
					__LINE__, NFR_RUN_OK) );
				    return(NFR_RUN_OK);
				} else {
				    RUNDEBUG( printf("Exiting machinerun %d: %d\n",
					__LINE__, NFR_RUN_RETURN) );
				    return(NFR_RUN_RETURN);
				}
			    }
			} else {
			    break;
			}
		    }
		    frame_pop(m);
		    pc = next-1;
		}
		break;

		/* BREAK */
		case NFR_OP_BREAK:
		    if ((m->m_fp > 0) && m->m_fram[m->m_fp-1].ret_break) {
			RUNDEBUG( printf("Exiting machinerun %d: %d\n",
				__LINE__, NFR_RUN_BREAK ) );
			return(NFR_RUN_BREAK);
		    } else {
			/* XXX log? */
			Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_SYMANTICERR,
			    "run: break not in loop", 0 );
		    }
		break;

		default:
		runstate = NFR_RUN_ILLEGAL;
		panic("illegal instruction");
		}
		pc++;
	}

	/* when we hit a STOP we are done. */
	if(m->m_sp != 0 && m->m_fp == 0) {
	    int i;
	    printf("WARNING stack not drained!! ptr=%d\n",m->m_sp);
	    for (i=0; i<m->m_sp; ++i) {
		printf( "Stack:%d: ", i );
		blt_echo( &m->m_mem[i] );
		printf( "\n" );
	    }
	}

	returning = 0;
	runCurrentSet(oldRunCurrent);
	RUNDEBUG( printf("Exiting machinerun %d: %d\n", __LINE__, NFR_RUN_OK ) );
	return(NFR_RUN_OK);
}

/*
** Perform the on condition.
** Review the triggers, and call appropriate routines to service them.
**
** Save info about whether the trigger is begin evaluated inside
** a function, or in a ON clause.  Tag the trigger_t with a flag,
** when we're in a function.  Some triggerSet's may dis-allow
** some options if part of a function invocation.
*/


run_perform_on( Machine *m, Filter * f, int insideFunc )
{
    /*
    ** lets see what the stack looks like.
    */
    MemVal func[1], ident[1], name[1], num[1];
    char * identName;
    
    stack_pop(m,func);
    stack_pop(m,num);
    stack_pop(m,ident);
    if (ident->typ != NFR_TYPE_STR) {
	identName = 0;
    } else {
	identName = ident->val.c;
    }
    stack_pop(m,name);
    /*
    ** make sure both type's match.
    */

    if ((func->typ == NFR_TYPE_STR)
     && (name->typ == NFR_TYPE_STR)
     && (num->typ == NFR_TYPE_INT)) {
	/*
	** See if we can find who to send it to...
	*/
	int ti = triggerFind( name->val.c );
	if (ti > -1 /* -1 == not found */ ) {
	    trigger_t * new = triggerNew( func->val.c, f->f_scope );
	    Sym * s = 0;
	    int n = num->val.i;
	    int i;

	    if (insideFunc) {
		/*
		** If a filter is free'd, then we also hafta
		** free the trigger associated with the filter.
		** If the user issues an OFF for the filter
		** inside the filter, then we hafta clear
		** the trigger for the filter.
		*/
		new->t_flags |= T_FLAGS_FUNC;
		f->f_trigger = new;
		triggerIncRef(f->f_trigger);
	    }

	    /*
	    ** Find all the name value pairs, then make the trigger_t
	    */ 
	    for (i=0; i<n; ++i) {
		Sym * nv;
		MemVal nv_name[1];
		MemVal nv_value[1];
		stack_pop( m, nv_name );
		stack_pop( m, nv_value );
		if (nv_name->typ != NFR_TYPE_STR) {
		    Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_INTERNALERROR,
			"run_perform_on: internal format error" );
		} else {

		    nv = symnew( nv_name->val.c, nv_value );
		    if (nv == 0) {

			Alert_Args_Generate( NFR_SRC_NFRD, NFR_ERR_OUTOMEMORY,
			    "malloc", "run_perform_on: symnew", 0 );
		    }
		    s = symadd( nv, s );
		}
	    }
	    triggerAdd( new, name->val.c, s );
	    {
		int cookie;
		/*
		** Perform a global assignement of the values of the
		** cookie (to use in the off values), to the name
		** of the symbol described in the statement (identName).
		*/
		cookie = triggerFunc( name->val.c, new );
		if (cookie && identName) {
		    MemVal mv[1];
		    mv->typ = NFR_TYPE_TRIGGER;
		    mv->val.i = cookie; /* well, its really a trigger_t */

		    run_globalAssign( identName, mv );
		}
	    }
	}
	else {
	    uiCurrPrintf( "Trigger ignored: no such trigger type `%s'\n",
		name->val.c );
	}
    } else {
	/* error, bad symantics. */
	uiCurrPrintf( "Error in trigger: internal format error\n" );
    }
}

/*
** off.
**
** See who knows something about this cookie.
*/

int
run_perform_off( Machine *m, Filter *f )
{
    MemVal mv[1];
    int cookie;

    stack_pop( m, mv );
    cookie = mv->val.i;

    if (mv->typ != NFR_TYPE_TRIGGER) {
	/* XXX bad cookie */
    } else {
	return(triggerOff( mv->val.t ));
    }
}

/*
** run_perform_decl_inside.
**
** Create the local symbol, if one isn't there.  Set up the
** symbol so that references to it will extract the symbol
** withing the SYM list.
**
*/

int
run_perform_decl_inside( Filter * f, char * s, MemVal * m )
{
    Sym * sp = symlook( s, f->f_syms );
    if (m->typ == NFR_TYPE_MAP) {
	if (sp == 0)  {
	    MemVal n;
	    n.typ = NFR_TYPE_INDIRECT;
	    mapIncRef(m->val.v);
	    n.val.v = m->val.v;

	    sp = symnew( s, &n );
	    f->f_syms = symadd( sp, f->f_syms );
	} else {
	    sp->s_data.typ = NFR_TYPE_INDIRECT;
	    mapIncRef(m->val.v);
	    sp->s_data.val.v = m->val.v;
	}
	f->f_indSymCnt++;
	/*
	** We also need to create the symbol on the Sym *
	*/
    } else {
	/* if the symbol exists on the list, change its type... */
	if (sp) {
	    sp->s_data.typ = NFR_TYPE_INT;
	    sp->s_data.val.i = 0;
	}

    }
}

/*
** run_perform_decl_global
**
** Create a global variable.
** All externally named global variables are actually local to
** a 'ident'... that way we don't get name clashes.  If we
** get a normal global assignment, we actually create a new
** name prefixed by the ident name. The only way to get a name
** which is a real global is by using the DECLARE name GLOBAL.
*/

run_perform_decl_global( Filter *f, char * s )
{
    /*
    ** First see if the globalMapBag is in place.
    */
    if (globalMapBag == 0) {
	globalMapBag = mapBagNew( 0 );
    }
    /*
    ** Now see if the global map is in place...
    */
    {
	Map * globalMap = mapBagFind( globalMapBag, globalScopeNdx, globalScope );
	if (globalMap == 0) {
	    /*
	    ** Make the global map.
	    */
	    globalMap =  mapNew( 0, globalScopeNdx, globalScope, 0 );
	    globalMap->m_name = strdup( globalScope );
	    mapBagAdd( globalMapBag, globalMap );
	}
	/*
	** now make a declaration for the symbol in the globalMapBag
	** and make a local symbol... but make it indirect... and 
	** have it refer to the global map's map.
	*/
	{
	    Sym * sp = symlook( s, f->f_syms );
	    if (sp == 0) {
		/* make one */
		MemVal n;
		n.typ = NFR_TYPE_INDIRECT;
		n.val.v = globalMap;
		mapIncRef(globalMap);
		sp = symnew( s, &n );
		f->f_syms = symadd( sp, f->f_syms );
	    } else {
		sp->s_data.typ = NFR_TYPE_INDIRECT;
		sp->s_data.val.v = globalMap;
		mapIncRef(globalMap);
	    }
	    f->f_indSymCnt++;
	}
    }
}

void
machinereset(Machine *m)
{
	m->m_sp = 0;
	m->m_fp = 0;
	cputicks = 0;
	runstate = NFR_RUN_OK;
}


#ifdef	NFR_RUNDEBUG
void
run_printlist(MemVal *mp)
{
	ListMem	*p;

	for(p = (ListMem *)mp->val.v; p != (ListMem *)0; p = p->next) {
		run_printtyp(&(p->data));
		if(p->next != (ListMem *)0)
			printf(", ");
	}
}



static void
run_printtyp(MemVal *p)
{
	switch(p->typ) {
	case NFR_TYPE_ERR:	printf(" ERR=%d",p->val.i); return;
	case NFR_TYPE_INT:	printf(" INT=%d",p->val.i); return;
	case NFR_TYPE_STR:	printf(" STR=\"%s\"",p->val.c); return;
	case NFR_TYPE_LIST:	printf(" LIST="); run_printlist(p); return;
	case NFR_TYPE_RECORDER:	printf(" RECORDER=%x", p->val.v ); return;
	case NFR_TYPE_SCHEMADATA: printf(" SCHEMADATA=%x", p->val.v ); return;
	case NFR_TYPE_IPV4HOST:	printf( " IPV4HOST=%x", p->val.h.s_addr );
				return;
	case NFR_TYPE_IPV4NET:  {
					struct in_addr mask;
					mask.s_addr = p->val.n->ip_mask;
					printf( " IPV4NET=%s",
						inet_ntoa( p->val.n->ip_addr ) );
					printf( "%s", inet_ntoa( mask ) );
				}
				return;
	case NFR_TYPE_BLOB:	if (p->val.b) {
				    printf( " BLOB=[%d,\"", p->val.b->b_len );
				    dump( 0, p->val.b->b_data, p->val.b->b_len );
				    printf( "\"]" );
				} else {
				    printf( " BLOB=<null>" );
				}
				return;
	case NFR_TYPE_ETHMAC:	printf( " ETHMAC=" );
				displayEnetMac( (char *)&p->val.e );
				return;
	case NFR_TYPE_MAP:	printf( " MAC=%x", p->val.v ); 
				return;
	case NFR_TYPE_MAPBAG:	printf( " MAPBAG=%x", p->val.v );
				return;
	default:
				printf(" Unknown!(%d)",p->typ);
	}
}



void
run_printst(Machine *m,int sp)
{
	run_printtyp(&(m->m_mem[sp]));
}



/*
this must essentially reflect the operations in run() which should
act as a sanity check. if life gets WEIRD make sure that the operations
performed in run() match the diagnostics printed here!!!!
*/
static	void
run_printop(Machine *m, Filter *f, int pc, int sp, int op)
{
	printf("pc=%d, sp=%d, op=",pc,sp);

	switch(NFR_OP(op)) {
	case NFR_OP_STOP:	printf("STOP");
				break;
	case NFR_OP_POP:	printf("POP"); run_printst(m,sp - 1);
				break;
	case NFR_OP_NPUSH:	printf("NPUSH %d",f->f_mem[pc + 1]);
				break;
	case NFR_OP_SPUSH:	printf("SPUSH \"%s\"",f->f_list[f->f_mem[pc + 1]]->data.val.c);
				break;
	case NFR_OP_PPUSH:	printf("PPUSH $%d",f->f_mem[pc + 1]);
				break;
	case NFR_OP_LPUSH:	printf("LPUSH <list>");
				break;
	case NFR_OP_ETHMACPUSH:	printf("ETHMACPUSH ");
				break;
	case NFR_OP_CPUSH:	if(m->m_fp != 0)
					printf("CPUSH %d",m->m_fram[m->m_fp - 1].ret_ac);
				else
					printf("CPUSH 0");
				break;
	case NFR_OP_LASGN:	printf("LASGN %s=(",f->f_list[f->f_mem[pc + 1]]->data.val.c);
				run_printst(m,sp - 1);
				printf(")");
				break;
	case NFR_OP_CALL:	printf("CALL %s:%s",
					f->f_scope,
					f->f_list[f->f_mem[pc + 1]]->data.val.c);
				break;
	case NFR_OP_RETNV:	printf("RETNV");
				break;
	case NFR_OP_RETV:	printf("RETV");
				run_printst(m,sp - 1);
				break;
	case NFR_OP_ADD:	printf("ADD");
				run_printst(m,sp - 1);
				run_printst(m,sp - 2);
				break;
	case NFR_OP_SUB:	printf("SUB");
				run_printst(m,sp - 1);
				run_printst(m,sp - 2);
				break;
	case NFR_OP_MUL:	printf("MUL");
				run_printst(m,sp - 1);
				run_printst(m,sp - 2);
				break;
	case NFR_OP_DIV:	printf("DIV");
				run_printst(m,sp - 1);
				run_printst(m,sp - 2);
				break;
	case NFR_OP_NEG:	printf("NEG");
				run_printst(m,sp - 1);
				break;
	case NFR_OP_EQ:		printf("EQ");
				run_printst(m,sp - 1);
				run_printst(m,sp - 2);
				break;
	case NFR_OP_NE:		printf("NE");
				run_printst(m,sp - 1);
				run_printst(m,sp - 2);
				break;
	case NFR_OP_LT:		printf("LT");
				run_printst(m,sp - 1);
				run_printst(m,sp - 2);
				break;
	case NFR_OP_LTE:	printf("LTE");
				run_printst(m,sp - 1);
				run_printst(m,sp - 2);
				break;
	case NFR_OP_GT:		printf("GT");
				run_printst(m,sp - 1);
				run_printst(m,sp - 2);
				break;
	case NFR_OP_GTE:	printf("GTE");
				run_printst(m,sp - 1);
				run_printst(m,sp - 2);
				break;
	case NFR_OP_NOT:	printf("NOT");
				run_printst(m,sp - 1);
				break;
	case NFR_OP_AND:	printf("AND");
				run_printst(m,sp - 1);
				run_printst(m,sp - 2);
				break;
	case NFR_OP_OR:		printf("OR");
				run_printst(m,sp - 1);
				run_printst(m,sp - 2);
				break;
	case NFR_OP_BAND:	printf("BAND");
				run_printst(m,sp - 1);
				run_printst(m,sp - 2);
				break;
	case NFR_OP_BOR:	printf("BOR");
				run_printst(m,sp - 1);
				run_printst(m,sp - 2);
				break;
	case NFR_OP_LSHIFT:	printf("LSHIFT");
				run_printst(m,sp - 1);
				run_printst(m,sp - 2);
				break;
	case NFR_OP_RSHIFT:	printf("RSHIFT");
				run_printst(m,sp - 1);
				run_printst(m,sp - 2);
				break;
	case NFR_OP_IF:		printf("IF");
				break;
	case NFR_OP_IFCOND:	printf("IFCOND");
				break;
	case NFR_OP_LEVAL:	printf("LEVAL %s",f->f_list[f->f_mem[pc + 1]]->data.val.c);
				break;
	case NFR_OP_PACKET:	printf("PACKET %d",f->f_mem[pc + 1]);
				break;
	
	case NFR_OP_GEVAL:	printf("GEVAL %s",f->f_list[f->f_mem[pc + 1]]->data.val.c);
				break;
	case NFR_OP_GASGN:	printf("GASGN %s=(",f->f_list[f->f_mem[pc + 1]]->data.val.c);
				run_printst(m,sp - 1);
				printf(")");
				break;
	case NFR_OP_RECORD:	printf("RECORD <more later>" );
				break;
	case NFR_OP_DECL_INSIDE:printf("DECLARE %s INSIDE",
					f->f_list[f->f_mem[pc]]->data.val.c );
				break;
	case NFR_OP_DECL_GLOBAL:printf("DECLARE %s GLOBAL",	
					f->f_list[f->f_mem[pc]]->data.val.c );
				break;

	case NFR_OP_FOREACH:	printf("FOREACH into %s (end at %d) ",
					f->f_list[f->f_mem[pc+1]]->data.val.c,
					f->f_mem[pc+2]);
				break;

	case NFR_OP_WHILE:	printf("WHILE (do %d) (end %d)",
					f->f_mem[pc+1], f->f_mem[pc+2] );
				break;
	case NFR_OP_ARRAYGREF:	printf("ARRAYGREF: %d ", f->f_mem[pc+1] );
				break;
	case NFR_OP_ARRAYLREF:	printf("ARRAYLREF: %d", f->f_mem[pc+1] );
				break;
	case NFR_OP_ARRAYGASGN:	printf("ARRAYGASGN: %d", f->f_mem[pc+1] );
				break;
	case NFR_OP_ARRAYLASGN:	printf("ARRAYLASGN: %d", f->f_mem[pc+1] );
				break;
	default:
		printf("Unknown!(%d)", NFR_OP(op) );
	}
	printf(" cptick=%d, runstate=%d\n",cputicks,runstate);
}
#endif
