/*
** Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
** All rights reserved.
**
** Use and distribution of this software and its source code
** are governed by the terms and conditions of the
** Network Flight Recorder Software License ("LICENSE.TXT")
*/

#include <stdio.h>
#include <stddef.h>
#include <string.h>

#include "nfrd.h"


/*
**	 Symbol table and function table management routines.
** These routines manage local and function tables, as well as
** interfaces to global tables.
**
**	Marcus J. Ranum, Network Flight Recorder, Inc. 1996
*/


/* free entries are stacked, not freed */
static	Sym		*freel = (Sym *)0;

/* global variables */
MapBag			*globalMapBag = (MapBag *)0;

#ifndef	NFR_FUNCTABLEWIDTH
#define	FUNCTABLEWIDTH	31
#endif
static	FilterList	*functable[FUNCTABLEWIDTH];


/*
** sym Routines.
**
** symnew, symadd, symremove, symlook, symfreelist
** symFixIndirect, symDirty, symSync.
**
*/

/*
** ------------------------------
** symnew
**
** allocate a new (local) symbol table entry
*/

Sym	*
symnew(char *name, MemVal *op)
{
    Sym	*ret;

    if(freel != (Sym *)0) {
	ret = freel;
	freel = freel->s_next;
    } else {
	if((ret = (Sym *)malloc(sizeof(Sym))) == 0)
	    return((Sym *)0);
    }
    ret->s_flags = 0;
    ret->s_length = 0;
    ret->s_next = 0;

    if (name) {
	unsigned len = strlen(name);
	ret->s_length = len;
	if (len > SYM_LENGTH_MAX) {
	    ret->s_length = SYM_LENGTH_MAX;
	}
	if (len <= sizeof(char *)) {
	    memcpy( &ret->s_name, name, len );
	} else {
	    if((ret->s_name = (char *)malloc(strlen(name) + 1)) == (char *)0) {
		free(ret);
		return((Sym *)0);
	    }
	    strcpy(ret->s_name,name);
	}
    }
    if(memvalclonedata(&(ret->s_data),op)) {
	free(ret->s_name);
	free(ret);
	return((Sym *)0);
    }
    return(ret);
}


/*
** ------------------------------
** symadd 
**
** add a new local symbol table entry
*/

Sym	*
symadd( Sym * sym, Sym * list )
{
    sym->s_next = list;
    return(sym);
}


/*
** ------------------------------
** symremove
**
** find the entry in the list, remove
** it from the list.  Return the new
** list.  (The new list needs to be
** returned in case the removed element
** is the First element in the list)
*/

Sym	*
symremove( Sym *sym, Sym *list )
{
    Sym * l = list;

    if ((sym == 0) || (l == 0)) {
	return( 0 );
    }
    if (l == sym) {
	return(l->s_next);
    }
    while( l->s_next ) {
	if (l->s_next == sym) {
	    l->s_next = l->s_next->s_next;
	    return(list);
	}
	l = l->s_next;
    }
    /*
    ** didn't find it on the list
    */
    return( (Sym *)-1);
}

/*
** ------------------------------
** symlook
**
** search a local symbol table
*/

Sym	*
symlook(char *str,Sym *table)
{
    Sym	*ret;
    if (str != 0) {
	for(ret = table; ret != (Sym *)0; ret = ret->s_next) {
	    if (ret->s_length <= SYM_NAME_UMAX) {
		if ((memcmp( str, &ret->s_name, ret->s_length) == 0)
		&&  (str[ret->s_length] == 0)) {
		    return(ret);
		}
	    } else if(strcmp(str,ret->s_name) == 0) {
		    return(ret);
	    }
	}
    }
    return(0);
}

/*
** ------------------------------
** symCmp
** compare the string passed in, to the symbol,
** see if they match.
*/

int
symCmp( sym, s )
Sym * sym;
char * s;
{
    if (sym) {
	if (sym->s_length <= SYM_NAME_UMAX) {
	    return( memcmp( &sym->s_name, s, sym->s_length ) );
	} else {
	    return( strcmp( sym->s_name, s ) );
	}
    } 
    return 0;
}

/*
** ------------------------------
**
*/

char symNameVar[SYM_NAME_UMAX+1];

char *
symName( sym )
Sym * sym;
{
    if (sym) {
	if (sym->s_length <= SYM_NAME_UMAX) {
	    memcpy( symNameVar, &sym->s_name, sym->s_length );
	    symNameVar[sym->s_length] = 0;
	    return(symNameVar);
	} else {
	    return(sym->s_name);
	}
    }
    return "";
}

/*
** ------------------------------
** symfreelist
**
** free a Sym *, and all its next references.
*/


/* free a local symbol table */
void
symfreelist(Sym *l)
{
    Sym	*p;
	
    p = l;
    while(l != 0) {
	memvalfreedata(&(l->s_data));
	if (l->s_length && (l->s_length <= SYM_NAME_UMAX) ) {
	    /* don't free anything.  name is in there... */
	} else {
	    if (l->s_name) {
		char * name = l->s_name;
		l->s_name = 0;
		free(name);
	    }
	}
	l = l->s_next;
	p->s_next = freel;
	freel = p;
	p = l;
    }
}

/*
** ------------------------------
** symDirty
**
** mark a symbol as dirty.
*/

Sym	*
symDirty( Sym * sym )
{
	sym->s_flags |= SYM_FLAG_DIRTY;
	return(sym);
}

/*
** ------------------------------
** symFixIndirect
**
** When there are INDIRECT syms associated with the filter,
** we need to walk that list, and change them to something else..
**
*/

int
symFixIndirect( Sym * s )
{
    int cnt = 0;
    
    while(s) {
	MemVal * m = &s->s_data;
	if (m->typ == NFR_TYPE_INDIRECT) {
	    m->typ = NFR_TYPE_INT;
	    cnt++;
	}
	s = s->s_next;
    }
    return(cnt);
}


/*
** ------------------------------
** symSync
**
** sync variables to some file name space. 
** use the variable names
*/

void
symSync(Sym *l, char * dir)
{
    Sym *p = l;
    while(p) {
	if (p->s_flags & SYM_FLAG_DIRTY) {
	    FILE * pickled;
	    char buffer[1024];

	    /*
	    ** maybe we should just mkdir without stating.
	    */
	    struct stat s[1];
	    if (stat( dir, s ) < 0) {
		if (errno == ENOENT) {
		    /* create it */
		    mkdir( dir, 0755 );
		}
	    }


	    {
		sprintf( buffer, "%s/%s", dir, symName(p) );

		if (vFlag) {
		    log( FL, "symSync: syncing %s into %s\n", symName(p), buffer );
		}

		if (pickled = fopen( buffer, "w" )) {
		    pickle( pickled, &p->s_data );
		    fclose(pickled);
		    p->s_flags &= ~SYM_FLAG_PICKLEERROR;
		} else {
		    /*
		    ** Complain once...
		    */
		    if ((p->s_flags & SYM_FLAG_PICKLEERROR) == 0) {
		       p->s_flags |= SYM_FLAG_PICKLEERROR;
			   perror( format( "%s into %s:%s", symName(p), buffer, "pickled-fopen" ) );
		    }
		}
		p->s_flags &= ~(SYM_FLAG_DIRTY);
	    }
	}
	p = p->s_next;
    }
}

/*
** ----------------------------------------
** symPrint.
** Print the value of any symbols which match a specific pattern.
**
*/

typedef struct symPrint {
	char *		sp_scope;
	regex_t		sp_scopeReg;
	char *		sp_name;
	regex_t		sp_nameReg;
} symPrint_t;

/*
** ----------------------------------------
**
** Called for each scope which matches.
*/
symPrintScopeDetail( m, rock, info )
Map * m;
void * rock;
char * info;
{
    symPrint_t * sp = (symPrint_t *)rock;

    while(m) {
	Sym * s = m->m_sym;
	while (s) {
	    if ((regexec( &sp->sp_scopeReg, m->m_name, 0, 0, 0 ) == 0)
	    &&  (regexec( &sp->sp_nameReg, symName(s), 0, 0, 0 ) == 0)) {
		uiCurrPrintf( "%s %s:%s ", info, m->m_name, symName(s) );
		blt_echo( &s->s_data );
		uiCurrPrintf( "\n" );
	    }
	    s = s->s_next;
	}
	m = m->m_next;
    }
}

void
symPrint( variable )
char * variable;
{
    symPrint_t sp[1];
    /*
    ** break the name into a scope and a name.
    ** if no scope exists, then assume the scope is global.
    */
    char * scope;
    char * name;
    char * colon;

    if ((colon = strchr( variable, ':' )) == 0) {
	name = variable;
	scope = "global";
    } else {
	name = colon+1;
	scope = variable;
	colon[0] = 0; /* i don't like this... its not ours to touch */
    }


    /*
    ** compile both the scope and the name into regular
    ** expression matches.  Then iterate through the
    ** maps, first we'll do the global symbols, then
    ** we'll use the tcp connections...
    */
    sp->sp_scope = scope;
    if (regcomp( &sp->sp_scopeReg, scope, REG_EXTENDED|REG_NOSUB )) {
	return;
    }

    sp->sp_name = name;
    if (regcomp( &sp->sp_nameReg, name, REG_EXTENDED|REG_NOSUB )) {
	return;
    }
    if (globalMapBag) {
	mapBagIterate( globalMapBag, symPrintScopeDetail, sp, "global" );
    }
    tcpAssemblePrint( symPrintScopeDetail, sp );

    if (colon) {
	colon[0] = 0;
    }

}



static	int
cornedbeefhash(s,siz)
char	*s;
int	siz;
{
	register unsigned int	c = 0;
	register unsigned int	g;
	while(*s) {
		c = (c << 1) + *s++;
		if(g = (c & 0xF0000000)) {
			c = c ^ (g >> 24);
			c = c ^ g;
		}
	}
	c = c % 211;
	return(c % siz);
}

extern	int	funcdelete(Filter *);			/* sym.c */

void
functable_init()
{
	int	x;

	for(x = 0; x < FUNCTABLEWIDTH; x++)
		functable[x] = (FilterList *)0;
}


/*
** ----------------------------------------
** filtfree
** delete the interrior portion of a filter,
** then reset the state of the filter...
*/

void
filtfree(Filter *f)
{
    int	x;

    /* if there's a trigger assciated with the filter,
    ** free the filter...
    */
    if (f->f_trigger) {
	trigger_t * t = f->f_trigger;
	/*
	** yes, we know its a filter function.
	** we're taking care of it...
	*/
	t->t_flags &= ~T_FLAGS_FUNC;
	f->f_trigger = 0;
	triggerOff(t);
	triggerDecRef(t);
    }

    /* free program text */
    if (f->f_mem) {
	free(f->f_mem);
	f->f_mem = 0;
    }

    if (f->f_list) {
	/* free external/list memory */
	for(x = 0; x < f->f_listsiz; x++) {
	    if (f->f_list[x]) {
		ListMem * l = f->f_list[x];
		f->f_list[x] = 0;
		listfree(l);
	    }
	}
	f->f_listsiz = 0;
	free(f->f_list);
    }

    /* free local symbol table */
    if (f->f_syms) {
	Sym * s = f->f_syms;
	f->f_syms = 0;
	symfreelist(s);
    }

    if (f->f_fromFile) {
	free(f->f_fromFile);
	f->f_fromFile = 0;
    }

    if (f->f_scope) {
	free(f->f_scope);
	f->f_scope = 0;
    }
    free(f);
}

/*
** ----------------------------------------
** funcfind.
** find a function
**
** First try to find a function with local scope to match the
** procedure...  then try to match a global one... notice that
** we don't hash on anything but the function name... that will 
** keep all the functions with `similar' names on the same list...
*/
Filter	*
funcfind(char *s, char * scope)
{
	FilterList	*fp = functable[cornedbeefhash(s,FUNCTABLEWIDTH)];


	if (scope) {
	    FilterList * lp = fp;
	    while(lp != (FilterList *)0) {
		if(!strcmp(s,lp->f->f_name)
		&& !strcmp(lp->f->f_scope, scope )) {
		    return(lp->f);
		}
		lp = lp->next;
	    }
	}
	/*
	** Now see if the symbol we've got has a colon (":") in it.
	** if so, use is as a `scope:name' call, and find the
	** function.
	*/
	{
	    char * colon;
	    if ((colon = strchr( s, ':' )) != 0) {
		char * nscope = s;
		int nscopeLen = colon - s;
		char * ns = &colon[1];
		fp = functable[cornedbeefhash(ns,FUNCTABLEWIDTH)];

		while(fp != (FilterList *)0) {
		    if (!strcmp(ns,fp->f->f_name)
		    && ((strncmp(fp->f->f_scope, nscope, nscopeLen) == 0)
			&&(fp->f->f_scope[nscopeLen] == 0)) ) {
			return(fp->f);
		    }
		    fp = fp->next;
		}
	    }
	}
	{
	/*
	** now try for globals. (maybe we should do that?)
	*/
	FilterList * lp = fp;
	while(lp != (FilterList *)0) {
	    if(!strcmp(s,lp->f->f_name) && !strcmp(lp->f->f_scope, "global")) {
		return(lp->f);
	    }
	    lp = lp->next;
	}
	return((Filter *)0);
	}
}

/*
** ----------------------------------------
** funcUnload.
** interface to unload all filters with a specific
** scope name.
** note: This isn't sufficient to destroy all data about
** a scope.  We also need to flush any non-local variables
** associated with this scope.
**
*/

funcUnload( scope )
char * scope;
{
	FilterList	*lp;
	int x;

	for(x = 0; x < FUNCTABLEWIDTH; x++) {
	    lp = functable[x];

	    while(lp->next != (FilterList *)0) {
		if (lp->next->f) {
		    if(strcmp(scope, lp->next->f->f_scope) == 0) {
			FilterList * f = lp->next;
			printf( "funcUnload: function %s\n", f->f->f_name );
			lp->next = f->next;

			filtfree(f->f);
			f->next = 0;
			free(f);
			/* notice we don't move lp down the list */

		    } else {
			lp = lp->next;
		    }
		} else {
		    lp = lp->next;
		}
	    }
	}
}



/* add a function - replace if already extant */
int
funcadd(Filter *f)
{
	FilterList	*lp;
	FilterList	*np;
	int		x;

	x = cornedbeefhash(f->f_name,FUNCTABLEWIDTH);
	lp = functable[x];
	while(lp != (FilterList *)0) {

	    /* it exists, replace it after freeing the old one */
	    if(!strcmp(f->f_name,lp->f->f_name)
	    && !strcmp(f->f_scope,lp->f->f_scope)) {
		filtfree(lp->f);
		lp->f = f;
		return(0);
	    }
	    lp = lp->next;
	}

	if((np = (FilterList *)malloc(sizeof(FilterList))) == (FilterList *)0)
		return(-1);
	np->f = f;
	np->next = functable[x];
	functable[x] = np;
	return(0);
}

/* stats - return some basic stats about functions */

funcStat(Filter *f)
{
	/*
	** We may wann keep some start/stop elapsed time info
	** But not now...
	*/
	uiCurrPrintf( "func: %s:%s (ndx %d): Use %d\n",
		f->f_scope, f->f_name, f->f_scopeNdx, f->f_useCount );
	if (f->f_useCount) {
	    uiCurrPrintf( "func: %s:%s totalTime %g AvgTime %g stdDev %g\n",
		f->f_scope, f->f_name, util_tv2usec(&f->f_runTime)/MILLION,
		util_tv2usec(&f->f_runTime)/MILLION/f->f_useCount, f->f_stdDev );
		
	}
}

/* all functions ... */
funcStats()
{
	FilterList	*lp;
	int x;

	for(x = 0; x < FUNCTABLEWIDTH; x++) {
	    lp = functable[x];
	    while(lp != (FilterList *)0) {
		if (lp->f) {
			funcStat(lp->f);
		}
		lp = lp->next;
	    }
	}
}


/*
** -----------------
** maps and mapMod.
**
**
**
*/

Map *
mapNew( m, i, ni, s )
Map * m;
int i;
char * ni;
Sym * s;
{
    Map * n = (Map *)calloc( 1, sizeof(*n) );
    n->m_next = m;
    n->m_index = i;
    n->m_name = strdup(ni);
    n->m_sym = s;
    n->m_refCnt = 1;
    return(n);
}

void
mapFree( m )
Map * m;
{
    if (m->m_name) {
	char * name = m->m_name;
	m->m_name = 0;
	free( name );
    }
    if (m->m_sym) {
	Sym * s = m->m_sym;
	m->m_sym = 0;
	symfreelist( s );
    }
    free(m);
}

Map *
mapIncRef(m)
Map * m;
{
    if (m) {
	m->m_refCnt++;
    } else {
	mark();
    }
    return(m);
}

void
mapDecRef( v )
void * v;
{
    Map * m = (Map *)v;
    if (m) {
	if (--m->m_refCnt == 0) {
	    mapFree(m);
	}
    }
}

/*
** Return an eligible next size for the mapBag.
*/

	/*
	** this is just a list of primes, each one being about
	** twice as large as the previous.
	*/
static int primeMapBag[] = {
	29, 61, 127, 257, 521, 1049, 2099, 4201,
	0
};

int
mapBagBigger( mp )
MapBag * mp;
{
    if (mp == 0) {
	return(primeMapBag[0]);
    }

    {
	int i;
	for (i = 0; primeMapBag[i]; ++i ) {
	    if (mp->mb_width < primeMapBag[i]) {
		if (primeMapBag[i]) {
		    return(primeMapBag[i]);
		}
	    }
	}
    }
}


/*
** mapBagNew
**
** Since the hash is a simple mod, using prime numbers for
** the width is a wise choice.  Using binary width's could
** help, but only if the hash width was compiled in...
**
** If zero is passed in as a width, we use the first prime
** in biggerMapBag.
**
*/

MapBag *
mapBagNew( width )
int width;
{
    MapBag * mb = (MapBag *)calloc( 1, sizeof(*mb) );

    if ((mb->mb_width = width) == 0) {
	mb->mb_width = mapBagBigger(0);
    }
    mb->mb_maps = (Map **)calloc( mb->mb_width, sizeof(*mb->mb_maps) );
    mb->mb_refCnt = 1;
    return(mb);
}

/*
** mapBagIncRef
*/

MapBag *
mapBagIncRef( mb )
MapBag * mb;
{
    if (mb) {
	mb->mb_refCnt++;
    }
    return(mb);
}

/*
**
** mapBagDecRef.
** we return a null to indicate the entry no longer
** exists.  We return the mb to indiate its still there.
*/


void
mapBagDecRef( v )
void *v;
{
    MapBag * mb = (MapBag *)v;
    if (mb) {
	if (--mb->mb_refCnt == 0) {
	    mapBagFree(mb);
	}
    }
}

/*
** strHash.
** create an index for a string.
*/
int
strHash( char * s )
{
    int v = 0;
    int c = 0;
    
    while (*s) {
	c++;
	v += *s;
	v ^= (*s << 5) + (*s << c);
	s++;
    }
    return(v);
}

/*
** memHash
** just like strHash, but we use a passed in
** length.
*/

int
memHash( char * s, int l )
{
    int v = 0;
    int c = 0;
    
    while (l--) {
	c++;
	v += *s;
	v ^= (*s << 5) + (*s << c);
	s++;
    }
    return(v);
}

/*
** mapBagMod
** We could do something fancier,
** but this is ok for a first cut.
*/

int
mapBagMod( mb, i )
MapBag * mb;
int i;
{
    int mod = i % mb->mb_width;
    if (mod >= 0) {
	return( mod );
    }
    return( - mod );
}

/*
** mapBagFree
** even tho some of the mapDecRef's may not zero,
** that doesn't matter.  We're losin all our references.
**
*/

void
mapBagFree( mb )
MapBag * mb;
{
    if (mb) {
	{
	    int i;
	    for (i=0; i < mb->mb_width; ++i) {
		Map * map = mb->mb_maps[i];
		while (map) {
		    mapDecRef( map );
		    map = map->m_next;
		}
	    }
	}
	free(mb->mb_maps);
	free(mb);
    }
}

/*
** mapBagAdd
** add the map into the bag.
*/

void
mapBagAdd( mb, m )
MapBag * mb;
Map *m;
{   /* add `m' to the new map. */
    int hash = mapBagMod( mb, m->m_index );
    m->m_next = mb->mb_maps[ hash ];
    mb->mb_maps[hash] = m;
}

/*
** mapBagRenew.
**
** Change the size of the map.
** to expand, or shrink the map. 
** Maybe we ought to call it when the length
** of any 'find' turns up a length of greater
** then 2 * the current width.
*/

MapBag *
MapBagRenew( mb, width )
MapBag * mb;
int width;
{
    if (mb == 0) {
	return mapBagNew( width );
    }
    if (width && (mb->mb_width == width)) {
	return( mb );
    } else {
	if (width == 0) {
	    width = mapBagBigger( mb );
	}

	{
	    MapBag * new = mapBagNew( width );
	    /*
	    ** iterate over each of the old members, and add it
	    ** to the new members.
	    */
	    int i;
	    for (i=0; i<mb->mb_width; ++i ) {
		Map * m;
		while ( m = mb->mb_maps[i] ) {
		    mb->mb_maps[i] = m->m_next; /* take m off the old list */

		    mapBagAdd( new, m );
		}
	    }
	}
    }
}

/*
** findMapBag.
** look up a Map, based on the index 
**
** note: we may want to change the index i in the map
** structure into a MemVal, and use engine equal tests
** to see when we match.
*/

Map *
mapBagFind( mb, i, name )
MapBag * mb;
int i;
char * name;
{
    int hash = mapBagMod( mb, i );
    Map * m = mb->mb_maps[hash];

    for( ; m ; m = m->m_next ) {
	if ((m->m_index == i)
	&& ((name == 0) || (strcmp( name, m->m_name ) == 0)))  {
	    return(m);
	}
    }
    return(0);
}

/*
** mapBagIterate
**
** call out for each map we have in the mapBag.
** rock belongs to the caller, i
*/

void
mapBagIterate( mb, callout, rock, info )
MapBag * mb;
mapCallout_t callout;
void * rock;
char * info;
{
    int width = mb->mb_width;
    int i;

    for ( i = 0; i < width; ++i ) {
	Map * m;
	for ( m = mb->mb_maps[i]; m; m = m->m_next ) {
	    callout( m, rock, info );
	}
    }
}

/*
**
** Sync the variables in the mapBag
*/

mapBagSync( MapBag * mb, char * dir )
{
    if (mb) {
	int i;
	for (i = 0; i < mb->mb_width; i++) {
	    Map * m = mb->mb_maps[i];
	    while (m) {
		/*
		** Create a new directory for the variable to be sync'ed into.
		** Prefix the directory with the name of the ident.
		*/
		char * prefix = m->m_name ? m->m_name : "none";
		char * newDir = nfrPath( dir, prefix, 0 );
		symSync( m->m_sym, newDir );
		free(newDir);

		m = m->m_next;
	    }
	}
    }
}

/*
** mapBagMem.
**
** print some details about memory usage of the mapBag passed in.
*/

mapBagMem( MapBag * mb )
{
    if (mb) {
	int total = sizeof(*mb);
	uiCurrPrintf( "memory: mb %x sizeof %d witdh %d\n", mb, total, mb->mb_width );
	{
	    int i;
	    for (i = 0; i < mb->mb_width; i++) {
		Map * m = mb->mb_maps[i];
		while (m) {
		    total += sizeof(*m);
		    if (m->m_sym) {
			Sym * s = m->m_sym;
			while (s) {
			    total += memvalSize( &s->s_data ) +
				(sizeof(Sym) - sizeof(MemVal)); 
			    s = s->s_next;
			}
		    m = m->m_next;
		    }
		}
	    }
	}
	return(total);
    }
    return(0);
}

/*
** arrays.
** ------------------------------
** arrays are implemented by using
** mapbag's and their Sym *'s.
** Instead of hiding names in the
** Sym's we hide the index values.
**
** There is an optomization tho,
** if the length of the data we're storing
** is less than (or equal) to the size
** of a pointer, we don't allocation the
** memory associated with the data...
**
** arrayHash arrayGet, arraySet
*/

/* typedef struct arrayRef {
**	char *	ar_base;
**	int	ar_len;
**	int	ar_hash;
** } arrayRef_t;
*/

arrayRef_t *
arrayHash( MemVal * ndx, arrayRef_t * ar )
{
    switch(ndx->typ) {
	case NFR_TYPE_INT:
	    ar->ar_base = (char *)&ndx->val.i; /* questionable */
	    ar->ar_len = sizeof(ndx->val.i);
	    break;
	case NFR_TYPE_STR:
	    ar->ar_base = ndx->val.c;
	    ar->ar_len = strlen(ar->ar_base);
	    break;
	case NFR_TYPE_IPV4HOST:
	    ar->ar_base = (char *)&ndx->val.h.s_addr;
	    ar->ar_len = sizeof(ndx->val.h.s_addr);
	    break;
	case NFR_TYPE_ETHMAC:
	    ar->ar_base = (char *)&ndx->val.e.e_addr;
	    ar->ar_len = sizeof(ndx->val.e.e_addr);
	    break;
	case NFR_TYPE_PATTERN:
	    ar->ar_base = (char *)&ndx->val.p->p_str;
	    ar->ar_len = strlen(ar->ar_base);
	    break;
	case NFR_TYPE_BLOB:
	    ar->ar_base = (char *)&ndx->val.b->b_data;
	    ar->ar_len = ndx->val.b->b_len;
	    break;
	case NFR_TYPE_IPV4NET:
	    ar->ar_base = (char *)ndx->val.n;
	    ar->ar_len = sizeof(ipv4net_t);
	    break;
	default:
	    /* these aren't supported... yet */
	    /* we just need a consistent way to make a hash
	    ** for each type.
	    */
	    return(0);
    }
    ar->ar_hash = memHash( ar->ar_base, ar->ar_len );
    return(ar);
}

/*
** ------------------------------
** arrayGet
**
*/

Sym *
arrayGet( MemVal * mem, MemVal * ndx, arrayRef_t * ar )
{
    if (mem->typ != NFR_TYPE_MAPBAG) {
	/* error, not an array */
	return(0);
    }
    if (mem->val.v == 0) {
	mem->val.v = mapBagNew( 0 );
    }
    /*
    ** hash the ndx, then run down the list looking for
    ** matches.  We stop on the first match, cause 
    ** there's only supposed to be one...
    */

    {
    arrayRef_t arrayRef[1];
    if (ar) {
	ar = arrayHash( ndx, ar );
    } else {
	ar = arrayHash( ndx, arrayRef );
    }

	if (ar) {
	    MapBag * mb = (MapBag *)mem->val.v;
	    int i = mapBagMod( mb, ar->ar_hash );
	    Map * m = mb->mb_maps[i];

	    for( ; m ; m = m->m_next ) {
		if (m->m_index == ar->ar_hash) {
		    /* walk this symlist. */
		    Sym * s = m->m_sym;
		    while (s) {
			/* most common case */
			if ((ndx->typ == NFR_TYPE_INT)
			&& (SYM_FLAG_TYPE(s) == SYM_FLAG_TYPE_INT)) {
			    if (ndx->val.i == s->s_int) {
				return(s);
			    }
			} else
			if ((ndx->typ == NFR_TYPE_IPV4HOST)
			&& (SYM_FLAG_TYPE(s) == SYM_FLAG_TYPE_HOST)) {
			    if (ndx->val.h.s_addr == s->s_host.s_addr) {
				return(s);
			    }
			} else 
			if ((ndx->typ == NFR_TYPE_STR)
			&& (SYM_FLAG_TYPE(s) == SYM_FLAG_TYPE_NAME)) {
			    if (s->s_length <= SYM_NAME_UMAX) {
				if (memcmp( s->s_chars, ar->ar_base, ar->ar_len ) == 0) {
				    if (ar->ar_base[s->s_length] == 0) {
					return(s);
				    }
				}
			    } else {
				if (strcmp( s->s_name, ar->ar_base ) == 0) {
				    return(s);
				}
			    }
			} else {
			    /* all cases. */
			    switch(SYM_FLAG_TYPE(s)) {
				case SYM_FLAG_TYPE_NAME:
				case SYM_FLAG_TYPE_INT:
				case SYM_FLAG_TYPE_VOID:
				    log( FL, "arrayGet: can't compare elements\n" );
				    break;

				case SYM_FLAG_TYPE_MEMVAL:
				    if (is_equal( ndx, s->s_memVal )) {
					return(s);
				    }
				    break;

			    }
			}
			s = s->s_next;
		    }
		}
	    }
	}
    }
    /* didn't find an entry. */
    return(0);
}

/*
** ------------------------------
** arraySet
**
*/

MapBag *
arraySet( MemVal * mem, MemVal * ndx, MemVal * value )
{
    arrayRef_t ar[1];
    Sym * s = arrayGet( mem, ndx, ar );
    MapBag * mb = (MapBag *)mem->val.v;

    if (s == 0) {
	/* need to add another entry. */
	if (mem->typ != NFR_TYPE_MAPBAG) {
	    /* error, not an array */
	    return(0);
	}
	if (memvalIsNull(value)) {
	    /*  don't do nuthin else, */
	    return(mb);
	}
	if (mb == 0) {
	    mb = mapBagNew( 0 );
	}
	{
	    int i = mapBagMod( mb, ar->ar_hash );

	    s = symnew( 0, value );
	    s->s_length = ar->ar_len;

	    /*
	    ** We try to optimize how we store the index'es
	    ** into the array.
	    */
	    switch (ndx->typ) {
		case NFR_TYPE_INT:
		    SYM_FLAG_TYPE_SET(s, SYM_FLAG_TYPE_INT);
		    s->s_int = ndx->val.i;
		    break;

		case NFR_TYPE_STR:
		    SYM_FLAG_TYPE_SET(s, SYM_FLAG_TYPE_NAME);
		    s->s_length = ar->ar_len;
		    if (ar->ar_len > SYM_LENGTH_MAX) {
			s->s_length = SYM_LENGTH_MAX;
		    }
		    if (ar->ar_len == 0) {
			s->s_chars[0] = 0;
		    } else if (ar->ar_len <= SYM_NAME_UMAX) {
			memcpy( s->s_chars, ar->ar_base, ar->ar_len );
		    } else {
			s->s_name = (char *)malloc( ar->ar_len+1 );
			strcpy( s->s_name, ar->ar_base );
		    }
		    break;

		case NFR_TYPE_IPV4HOST:
		    SYM_FLAG_TYPE_SET(s, SYM_FLAG_TYPE_HOST);
		    s->s_host = ndx->val.h;
		    break;
			
		default:
		    {
			s->s_memVal = (MemVal *)malloc( sizeof(*(s->s_memVal)) );
			SYM_FLAG_TYPE_SET(s, SYM_FLAG_TYPE_MEMVAL);
			memvalclonedata( s->s_memVal, ndx );
		    }
		    break;
	    }

	    {
		Map *  map = mb->mb_maps[i];
		if (map == 0) {
		    mb->mb_maps[i] = mapNew( mb->mb_maps[i], ar->ar_hash, "array", s );
		} else {
		    if (ar->ar_hash == map->m_index) {
			/* same map */
			map->m_sym = symadd( s, map->m_sym );
		    } else {
			/* new map */
			Map * new = mapNew( mb->mb_maps[i], ar->ar_hash, "array", s );
			new->m_next = map;
			mb->mb_maps[i] = new;
		    }

		}
	    }
	}

    } else {
	memvalfreedata( &(s->s_data) );
	if (memvalIsNull(value)) {
	    /* remove the index */
	    int cnt = 0; /* debugging */
	    Map * list = 0; /* belongs lower XXX debuggin. */
	    int i = mapBagMod( mb, ar->ar_hash );
	    Map * map = mb->mb_maps[ i ];

	    while(map) {
		if (map->m_index == ar->ar_hash) {
		    /*
		    ** free any resources acquired for the index.
		    */
		    switch (ndx->typ) {
			case NFR_TYPE_INT:
			case NFR_TYPE_IPV4HOST:
			    /* no additional resources */
			    break;

			case NFR_TYPE_STR:
			    if ((ar->ar_len > SYM_NAME_UMAX) && s->s_name) {
				free( s->s_name );
				s->s_name = 0;
			    }
			    break;

			default:
			    {
				memvalfreedata( s->s_memVal );
			    }
			    break;
		    }
		    /* this is the correct map */
		    map->m_sym = symremove( s, map->m_sym );
		    if (map->m_sym == (Sym *)-1) {
			panic( "arraySet: while removing index, couldn't find sym in list\n" );
		    }
		    if (map->m_sym == 0) {
			/*
			** remove this map from the list of maps
			*/
			/* Map * */ list = mb->mb_maps[ i ];
			if (list == map) {
			    mb->mb_maps[i] = map->m_next;
			    mapDecRef(map);
			} else {
			    while (list->m_next) {
				if (list->m_next == map) {
				    list->m_next = map->m_next;
				    mapDecRef(map);
				    break;
				}
				list = list->m_next;
			    }
			    /*
			    ** XXX bad assumption. we ought to check
			    ** to make sure we SAW that thing on the list.
			    */
			}
		    }
		    break;
		}
		cnt++;
		map = map->m_next;
	    }

	} else {
	    /* set the index's value */
	    memvalclonedata( &(s->s_data), value );
	}
    }
    return(mb);
}

/*
** symSyncIntf.
** interface between the timout stuff and the syncing
** routines for variables.
*/

void
symSyncIntf( void * v )
{
    extern char * varDir;
    mapBagSync( globalMapBag, varDir );
}

/*
** symInit.
** set up the timeout interval for sym's value syncing.
*/

void
symInit()
{
    timeval_t when = util_makeTv( /* sec */ 60, /* usec */ 0 );
    to_t * to = toNew( &when, TO_FLAGS_SPECIAL | TO_FLAGS_REPEAT, 0, symSyncIntf );
}
