/*
** Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
** All rights reserved.
**
** Use and distribution of this software and its source code
** are governed by the terms and conditions of the
** Network Flight Recorder Software License ("LICENSE.TXT")
*/

#include 	"nfrd.h"

/*
** triggers.
*/

/*
** triggerFunctions.
** ------------------------------
** Collect all the names of the trigger
** types into one place; easier to find.
**
*/

extern int packetTriggerSet( trigger_t * );
extern int ipPacketTriggerSet( trigger_t * );
extern int ipTriggerSet( trigger_t * );
extern int icmpTriggerSet( trigger_t * );
extern int udpTriggerSet( trigger_t * );
extern int tcpPacketTriggerSet( trigger_t * );
extern int tcpTriggerSet( trigger_t * );
extern int toTriggerSet( trigger_t * );         /* time-out's */

struct triggerFuncs {
    char *      	tf_name;
    int 		(*tf_func)( trigger_t * );
} triggerFuncs[] = {
    "packet",       packetTriggerSet,
    "ip",           ipTriggerSet,
    "icmp",	    icmpTriggerSet,
    "udp",          udpTriggerSet,
    "tcpPacket",    tcpPacketTriggerSet,
    "tcp",          tcpTriggerSet,
    "timeout",      toTriggerSet,
};


/*
** triggerFind
** ------------------------------
** validate the trigger type.
*/

int
triggerFind( name )
char * name;
{
    int i;
    for( i=0; i< (sizeof(triggerFuncs)/sizeof(triggerFuncs[0])); ++i ) {
	if (strcmp( triggerFuncs[i].tf_name, name ) == 0) {
		return(i);	
	}
    }
    return(-1);
}

/*
** triggerFunc
** ------------------------------
** call the trigger function, passing the trigger, and returning
** the cookie associated with the trigger.
**
*/
int
triggerFunc( name, t )
char * name;
trigger_t * t;
{
    int i = triggerFind(name);
    if (i > -1) {
	return( (triggerFuncs[i].tf_func)(t) );
    }
    return(0);
}



/*
** triggerNew
** ------------------------------
**
** make a trigger structure out of nothing
*/

trigger_t *
triggerNew( func, scope )
char * func;
char * scope;
{
    trigger_t * new = (trigger_t *)calloc( sizeof(*new), 1 );
    new->t_func = strdup(func);
    new->t_scope = strdup(scope);
    new->t_refCnt = 1;
    return new;
}

/*
** ----------------------------------------
** triggerFree.
**
** We don't easily create, nor destroy one of
** these, when they're copied to another variable
** or into an expression, we just make references
** to it.
**
*/

void
triggerFree(t)
trigger_t * t;
{
    int i;

    if (t->t_refCnt != 0) {
	mark();
    }
    /*
    ** If this trigger hasn't been turned off yet,
    ** now since we're freeing it, we'd better do that.
    */
    if (t->t_free) {
	triggerOff(t);
    }

    for (i=0; i<t->t_ndx; ++i) {
	if (t->t_whom && t->t_whom[i]) {
	    free(t->t_whom[i]);
	    t->t_whom[i] = 0;
	}
	if (t->t_nv && t->t_nv[i]) {
	    symfreelist(t->t_nv[i]);
	    t->t_nv[i] = 0;
	}
    }

    if (t->t_whom) {
	free(t->t_whom);
	t->t_whom = 0;
    }
    if (t->t_nv) {
	free(t->t_nv);
	t->t_nv = 0;
    }

    if (t->t_func) {
	free(t->t_func);
	t->t_func = 0;
    }

    if (t->t_scope) {
	free(t->t_scope);
	t->t_scope = 0;
    }
    free(t);
}

/*
** ----------------------------------------
** triggerIncRef.
*/

trigger_t *
triggerIncRef( t )
trigger_t * t;
{
    if (t) {
	t->t_refCnt++;
    }
    return(t);
}

/*
** ----------------------------------------
** triggerDecRef.
** Odd declaration since its used in some
** generic free situations.
*/

void 
triggerDecRef( p )
void * p;
{
    trigger_t * t = (trigger_t *)p;
    if (t) {
	if (--t->t_refCnt == 0) {
		triggerFree(t);
	}
    }
}

/*
** triggerAdd
** ------------------------------
** add another new sym list to the trigger, along with its 'direct' value.
*/
#define TRIG_INCR	10

void
triggerAdd( t, whom, nv )
trigger_t * t;
char * whom;
Sym * nv;
{
    if (t) {
	if (t->t_ndx == t->t_max) {
	    if (t->t_max == 0) {
		t->t_whom = (char **)calloc( sizeof(t->t_whom[0]), t->t_max+TRIG_INCR );
		t->t_nv = (Sym **)calloc( sizeof(t->t_nv[0]), t->t_max+TRIG_INCR );
	    } else {
		t->t_whom = (char **)realloc( t->t_whom,
			sizeof(t->t_whom[0])*(t->t_max+TRIG_INCR) );
		t->t_nv = (Sym **)realloc( t->t_nv, 
			sizeof(t->t_nv[0])*(t->t_max+TRIG_INCR) );
	    }
	    t->t_max += TRIG_INCR;
	}
	t->t_whom[t->t_ndx] = strdup(whom);
	t->t_nv[t->t_ndx] = nv;
	t->t_ndx++;
    } else {
	Alert_Args_Generate( NFR_SRC_NFRD, NFR_NFRD_INTERNALERROR,
		"triggerAdd: passed zero `t'" );
    }
}


/*
** triggerStat
** ------------------------------
** Display detail about the trigger.
*/

triggerStat( t )
trigger_t * t;
{
    if (t) {
	uiCurrPrintf( "t: %x func %s:%s ndx %d max %d rock %d flags %x\n",
		t, t->t_scope, t->t_func, t->t_ndx, t->t_max, t->t_rock, t->t_flags );
	{
	    int i;
	    for (i=0; i<t->t_ndx; ++i) {
		Sym * s = t->t_nv[i];
		uiCurrPrintf( "t: %s %x:\n",
		    t->t_whom[i] ? t->t_whom[i] : "<none>", s );
		while (s) {
		    uiCurrPrintf( "t:	%s flags %x:<", symName(s), s->s_flags );
		    blt_echo( &s->s_data );
		    uiCurrPrintf( ">\n" );
		    s = s->s_next;
		}
	    }
	}
    }
}

/*
** triggerOff
** ----------------------------------------
**
** The call here free's any rock associated with
** the trigger.  Its an indirect call, and 
** suffers from the inability to determine
** where the next line of code gets executed.
** SO, i'd like to list the possible targets
** of the call...
**
**   packetTriggerOff
**   tcpTriggerOff
**   udpTriggerOff
**   toTriggerOff (timeouts)
**
*/

triggerOff( t )
trigger_t * t;
{
    if (t->t_free) {
	/*
	** if we have a function which is the target
	** of a trigger (cause of its filter clause),
	** we need to release any reference to the
	** trigger we scarf away... cause we're responsible
	** for it.
	*/
	if (t->t_flags & T_FLAGS_FUNC) {
	    /* see if the trigger is still hooked up here */
	    Filter * f = funcfind(t->t_func, t->t_scope);
	    if (f) {
		if (f->f_trigger) {
		    if (f->f_trigger != t) {
			/* hey, this is wierd ? XXX */
		    } else {
			/* we're freeing it right now */
			printf( "triggerOff: %x zero\n", f->f_trigger );
			f->f_trigger = 0;
			triggerDecRef(t);
		    }
		}
	    }
	}
	t->t_free(t);
	t->t_free = 0;

	triggerDecRef(t); /* we may need a reference count here */

	return(1);
    } else {
	/* XXX couldn't reset the ON condition */
	return(0); 
    }
}
