/*
** Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
** All rights reserved.
**
** Use and distribution of this software and its source code
** are governed by the terms and conditions of the
** Network Flight Recorder Software License ("LICENSE.TXT")
*/

#include "autounpickle.c"

/*
** This is a sample of how to use the unpickling interfaces.
**
** This code is appended to automatically generated code to
** construct the object file.  To see how its used, review the
** makefile... it should look something like this:
**
** pickle.o:       pickle.c schema-pickle
**        ${TCLSH} ./schema-pickle Schemas/basic.sch -c pickle.c
**
**
** This example converts pickled data to MemVals.
**
** There are three sections to this file.
**
** The first is a collection of routines to convert the
** pickled schema's into interpreter internal representations (MemVals)
**
** The second is a cover routine to manage that conversion process.
**
** The third section calls the unpickling routines, then calls
** the cover routine to manage the conversion.
**
**
*/

/*
** Each of the specific conversions has an associated routine.
** We could place each of these converstion routines into a different
** file, allowing different conversion for different situations.
*/

/*
** --------------------
*/

MemVal *
unpickle2Int( MemVal * m, void * d )
{
    basicIntSchema_t * b = (basicIntSchema_t *)d;
    m->val.i = b->basicInt;
    free(d);
    return(m);
}

/*
** --------------------
*/

MemVal * 
unpickle2Str( MemVal * m, void * d )
{
    basicStrSchema_t * b = (basicStrSchema_t *)d;
    m->val.c = b->basicStr;
    free(d);
    return(m);
}

/*
** --------------------
*/

MemVal * 
unpickle2List( MemVal * m, void * d )
{
    basicListSchema_t * b = (basicListSchema_t *)d;
    m->val.v = (void *)b->basicList;
    free(d);
    return(m);
}

/*
** --------------------
*/

MemVal * 
unpickle2SchemaData( MemVal * m, void * d )
{
    basicSchemaDataSchema_t * b = (basicSchemaDataSchema_t *)d;
    m->val.v = (void *)b->basicSchemaData;  /* more later */
    free(d);
    return(m);
}

/*
** --------------------
*/

MemVal * 
unpickle2Recorder( MemVal * m, void * d )
{
    basicRecorderSchema_t * b = (basicRecorderSchema_t *)d;
    m->val.v = (void *)b->basicRecorder;  /* we may wanna hava type inda MemVal */
    free(d);
    return(m);
}

/*
** --------------------
*/

MemVal * 
unpickle2Ipv4host( MemVal * m, void * d )
{
    basicIpv4HostSchema_t * b = (basicIpv4HostSchema_t *)d;
    m->val.h = b->basicIpv4Host; 
    free(d);
    return(m);
}

/*
** --------------------
*/

MemVal * 
unpickle2Ipv4net( MemVal * m, void * d )
{
    basicIpv4NetSchema_t * b = (basicIpv4NetSchema_t *)d;
    m->val.n = b->basicIpv4Net;  
    free(d);
    return(m);
}

/*
** --------------------
*/

MemVal * 
unpickle2Blob( MemVal * m, void * d )
{
    basicBlobSchema_t * b = (basicBlobSchema_t *)d;
    m->val.b = b->basicBlob;
    free(d);
    return(m);
}

/*
** ------------------------------------------------------------
*/

/*
**
** unpickle2MemVal.
**
** Conversions we know about.... how to get from some pickled value
** the the nfr MemVal we know and love.
*/

struct unpickle2type {
	int		up_schema;		/* schema type number */
	int		up_nfrd;		/* nfrd.h type value */
	MemVal *	(*up_cvt)(MemVal * m, void * d);
} unpickle2type[] = {
	
    SCHEMA_TYPE_basicInt,		NFR_TYPE_INT,		unpickle2Int,
    SCHEMA_TYPE_basicStr,		NFR_TYPE_STR, 		unpickle2Str,
    SCHEMA_TYPE_basicList,		NFR_TYPE_LIST,		unpickle2List,
    SCHEMA_TYPE_basicRecorder,		NFR_TYPE_RECORDER,	unpickle2Recorder,
    SCHEMA_TYPE_basicSchemaData,	NFR_TYPE_SCHEMADATA,	unpickle2SchemaData,
    SCHEMA_TYPE_basicIpv4Host,	 	NFR_TYPE_IPV4HOST,	unpickle2Ipv4host,
    SCHEMA_TYPE_basicIpv4Net,	 	NFR_TYPE_IPV4NET,	unpickle2Ipv4net,
    SCHEMA_TYPE_basicBlob,		NFR_TYPE_BLOB,		unpickle2Blob,

};

MemVal *
unpickle2MemVal( void * d, int schemaType )
{
    int i;
    int e = (sizeof(unpickle2type)/sizeof(unpickle2type[0]));

    for (i = 0; i<e; ++i) {
	if (schemaType == unpickle2type[i].up_schema) {
	    MemVal * m = (MemVal *)malloc( sizeof(*m) );

	    m->typ = unpickle2type[i].up_nfrd;
	    return (*unpickle2type[i].up_cvt)( m, d );
	}
    }
    return(0);
}

/*
** ------------------------------------------------------------
*/

/*
** unpickle.
**
** From an inbound stream, identify the schema's we know about,
** and then get them into a memory object identifying the thing.
** Then convert it into a value we can use ...
**
*/

MemVal *
unpickle( FILE * f )
{
    char word[128];
    if (fscanf( f, "%s", word ) == 1) {
	if (strcmp( word, "type:" ) == 0) {
	    /* get the type */
	    if (fscanf( f, "%s", word ) == 1 ) {
		if (isdigit(word[0])) {
		    /* scan the numbers. */
		    int i;
		    int e = sizeof(unpickleTypeInts)/sizeof(unpickleTypeInts[0]);
		    int v = atoi( word );

		    for (i=0; i<e; ++i) {
			if (v == unpickleTypeInts[i].up_type) {
			    /* found it */
			    void * d = (*unpickleTypeInts[i].up_unp)(f);
			    if (d == 0) {
				return 0;
			    }
			    return( unpickle2MemVal( d, unpickleTypeStrs[i].up_hostType ) );
			}
		    }
		    /*
		    ** can't find the type.
		    */
		    return(0);
		} else {
		    /* scan the strings. */
		    int i;
		    int e = sizeof(unpickleTypeStrs)/sizeof(unpickleTypeStrs[0]);

		    for (i=0; i<e; ++i) {
			if (strcmp( word, unpickleTypeStrs[i].up_type ) == 0) {
			    /* found it */
			    void * d = (*unpickleTypeStrs[i].up_unp)(f);
			    if (d == 0) {
				return 0;
			    }
			    return( unpickle2MemVal( d, unpickleTypeStrs[i].up_hostType ) );
			}
		    }
		}
	    }
	} else {
	    return(0);
	}
	
    } else {
	return(0);
    }
}
