/*
** Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
** All rights reserved.
**
** Use and distribution of this software and its source code
** are governed by the terms and conditions of the
** Network Flight Recorder Software License ("LICENSE.TXT")
*/


/*
** user.c
**
** Routines to support writing info back to a telnetted
** session.  Allows debugging of components...
** info is dropped when the connetion is torn down.
*/

#include	"nfrd.h"

/*
**
** user session issues.
**
** we'd like to be able to associate logging output,
** parse errors, etc with a telnet session... that way
** we can forward info to that session as long as its
** connected.   When it does die, we don't have to
** relinquish resources
**
** To meet those goals, we have a Userintf_t structure,
** which describes all the info to get some data to
** a user.  We keep track of the fd's and map them to
** the Userintf_t's in a simple way...
*/

/*
** ----------------------------------------
**
** External state... describes current user interface...
*/

static Userintf_t * uiCurrent;

Userintf_t *
uiCurr() {
    return(uiCurrent);
}

Userintf_t *
uiSetCurr( Userintf_t * ui )
{
    /* XXX check to see if its on the list? */
    Userintf_t * old = uiCurrent;
    uiCurrent = ui;

    return( old );
}

Userintf_t *
uiSetTrace( Userintf_t * ui, int flag )
{
    ui->ui_trace = flag;
    return(ui);
}
 
/*
** ----------------------------------------
*/

  
typedef struct	uiFdMap {
	int		uf_fd;
	Userintf_t *	uf_ui;
} uiFdMap_t;

uiFdMap_t * uiFdMap;

int	uiFdCnt = 0;
int	uiFdMax = 0;


/*
** send output to some interface.
**
** if the file descriptor in the ui is -1, then
** we (silently) return.
**
*/

uiPrintf( Userintf_t * ui, char * fmt, ... )
{
    if (ui->ui_fd < 0) {
	/* sorry, channel is shut down. */
	return;
    }
    if (ui->ui_file == 0) {
	/* only an output stream. */
	ui->ui_file = fdopen( ui->ui_fd, "w" );
    }

    {
    int rtn;
    va_list ap;
    va_start(ap, fmt);
    rtn = vfprintf( ui->ui_file, fmt, ap );
    {
	/* if the connection is closed, we MAY find out here. */
    	int err = fflush( ui->ui_file );
	if (err != 0) {
	    if (errno == EPIPE) {
		/* other end has shut down */
		pollUnCall( fileno( ui->ui_file ) );
		uiFdClose( fileno( ui->ui_file ) );
		return(-1);
	    }
	}
    }
    return(rtn);
    /* XXX flush? */
    }

}

/*
** same as uiPrintf, but uses current user...
*/
uiCurrPrintf( char * fmt, ... )
{
    Userintf_t * ui;
    FILE * o;

    if ((ui = uiCurr()) == 0) {
	extern FILE * logFILE; /* XXX this will be bad someday */
	if ((o = logFILE) == 0) {
	    o = stdout;
	}
    } else {
	if (ui->ui_fd < 0) {
	    /* sorry, channel is shut down. */
	    return;
	}
	if (ui->ui_file == 0) {
	    /* only an output stream. */
	    ui->ui_file = fdopen( ui->ui_fd, "w" );
	}
	o = ui->ui_file;
    }

    {
    int rtn;
    va_list ap;
    va_start(ap, fmt);

    rtn = vfprintf( o, fmt, ap );
    {	/* if the connection is closed, we MAY find out here. */
    	int err = fflush( o );
	if (err != 0) {
	    if (errno == EPIPE) {
		/* other end has shut down. */
		pollUnCall( fileno(o) );
		uiFdClose( fileno(o) );
		return(-1);
	    }
	}
    }
    return(rtn);
    /* XXX flush? */
    }
}


int
uiFindI( fd )
{
    int i;
    for (i=0; i<uiFdCnt; ++i) {
	if (fd == uiFdMap[i].uf_fd) {
	    return(i);
	}
    }
    return(-1);
}

Userintf_t *
uiFind( fd )
{
    int i = uiFindI(fd);
    if (i == -1) {
	return(0);
    } else {
	return(uiFdMap[i].uf_ui);
    }
}

/*
** When a file descriptor shuts down, we just
** clear the fd, and if there's a stream, we
** shut it down too.
** note we don't free any storage, since its
** the reference count which determines when
** to free them, not the activity of the fd's
*/

int
uiFdClose( fd )
{
    int i = uiFindI( fd );
    if (i < 0) {
	return(1);
    } else {
	Userintf_t * ui = uiFdMap[i].uf_ui;

	if (ui) {
	    if (ui->ui_file) {
		fclose( ui->ui_file );
		ui->ui_file = 0;
	    }
	    ui->ui_fd = -1;
	}
	uiFdMap[i].uf_fd = -1;
    }
    return(0);
}

/*
**
** 
*/

Userintf_t *
uiNew( int fd )
{
    Userintf_t * ui = (Userintf_t *)calloc( sizeof(*ui), 1 );
    /* XXX assert? */

    ui->ui_fd = fd;
    ui->ui_file = 0;

    /* need room for more? */
    if (uiFdCnt >= uiFdMax){
	uiFdMax += 10;
	if (uiFdMap) {
	    uiFdMap = (uiFdMap_t *)calloc( sizeof(*uiFdMap), uiFdMax );
	} else {
	    uiFdMap = (uiFdMap_t *)realloc( uiFdMap, (sizeof(*uiFdMap)*uiFdMax ) );
	}
    }
    /* keep track of all the ui's */
    {
	int i = uiFdCnt++;

	uiFdMap[i].uf_fd = fd;
	uiFdMap[i].uf_ui = ui;
    }
    return(ui);
}

/*
**
*/

int
uiFree( Userintf_t * ui )
{
    /* find the stucture in the list. */
    {
	int i;
	for (i=0; i<uiFdCnt; ++i) {
	    if (ui == uiFdMap[i].uf_ui) {
		break;
	    }
	}
	/* something went wrong. */
	if (i = uiFdCnt) {
	    return(1);
	}

	{
	    int j;
	    for (j=i+1; j<uiFdCnt; ++j, ++i) {
		uiFdMap[i] = uiFdMap[j];
	    }
	}
	uiFdCnt--;
    }
    free(ui);
    return(0);
}

/*
**
*/

Userintf_t * 
uiIncr( Userintf_t * ui )
{
    if (ui) {
	ui->ui_refCnt++;
    }
    return(ui);
}

/*
** uiDecr.
**
** We don't have any references to this routine yet...
** i guess that's cause we don't really know what to do
** about realoading procedures, and deleting them...
** That's when the decr's ought to be done...
**
**
*/

Userintf_t * 
uiDecr( Userintf_t * ui )
{
    if (--ui->ui_refCnt == 0) {
	uiFree(ui);
    }
}
