/******************************************************************** 
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT") 

  File:   main.c 
  Author: Kent Landfield <kent@nfr.net>
 
********************************************************************/ 
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
** getnfrpid(char *pidfile)
**
** Returns:
**    Success:   pid > 0  
**    Errors:
**           Pidfile missing:       0
**           Pidfile open failure: -1
**           Pidfile format error: -2
*/

pid_t getnfrpid(char *pidfile)
{
    struct stat stbuf;
    FILE *fp;
    pid_t pid = 0;

    if (stat(pidfile,&stbuf) == 0) {
        if ((fp = fopen(pidfile,"r")) == NULL)  /* Open the pid file */
            return(-1);                         /* Troubles - bye    */

        if (fscanf(fp,"%d", &pid) != 1) {       /* Read the PID      */
            fclose(fp);                         /* Troubles, cleanup */
            return(-2);                         /* Troubles - bye    */
        }
        fclose(fp);                             /* Done so close it  */
    }
    return(pid);
}

/*
** testnfrpid(char *pidfile)
**
** Returns:
**    < 0 - error occurred
**      0 - process id in file not currently running
**    > 0 - valid running process id
*/

pid_t testnfrpid(char *pidfile)
{
    pid_t pid;

    if ((pid = getnfrpid(pidfile)) < 0)  /* get the pid in the pidfile */
        return(pid);                     /* : Error or none there...   */

    if (kill(pid,0) == 0)            /* Check if proc is running */
        return(pid);                 /* Yep  */
    return(0);                       /* Nope */
}

/*
** setnfrpid(char *pidfile)
**
** Returns: 
**    Success: 0 -  pidfile created and this is the only version running.
**    Errors:
**             1 -  Another process already running.
**            -1 -  Pidfile open read failed
**            -2 -  Pidfile format error encountered.
**            -3 -  Pidfile open write failed
*/
 
int setnfrpid(char *pidfile)
{
    int pid;
    FILE *fp;
  
    if ((fp = fopen(pidfile,"r")) != NULL) {  /* Open the pid file */
        if (fscanf(fp,"%d", &pid) != 1) {       /* Read the PID      */
            fclose(fp);                         /* Troubles, cleanup */
            return(-2);                         /* Troubles - bye    */
        }
        fclose(fp);                             /* Done so close it  */

        if (pid > 0) {                       /* If I found a valid pid     */ 
            if (kill(pid,0) == 0)            /* : Check if proc is running */
                return(1);                   /* :: Yep. Not the one && only */
        }
    }
    if ((fp = fopen(pidfile,"w")) == NULL)   /* Create the new pidfile */
         return(-3);                         /* : Whoa, troubles - bye */
  
    fprintf(fp,"%d\n",getpid());        /* Put current process info in */
    fclose(fp);                         /* Housekeeping                */
    return(0);                          /* Return I'm the one and only */
}

#ifdef TESTNFRPID

/*
** usage()
**
** Print the command's usage message
*/

int usage(char *progname)
{
   fprintf(stderr,"usage: %s [-cgt] pidfile\n", progname);
   fprintf(stderr,"\t-c  : create pidfile - setnfrpid()\n");
   fprintf(stderr,"\t-g  : return pid from existing pidfile - getnfrpid()\n");
   fprintf(stderr,"\t-t  : test existing pidfile - testnfrpid()\n");
   return(1);
}


/*
** Ye Olde Main.
**
*/

int main(int argc, char **argv)
{
   int c;
   pid_t fpid;
   int testfor;
   int gettest;
   int createpidfile;

   gettest = testfor = createpidfile = 0;

   if (argc > 1) {
      while ((c = getopt(argc, argv, "gtc")) != EOF) {
         switch (c) {
             case 'g':   /* check for */
                 gettest++;
                 break;
             case 't':   /* check for */
                 testfor++;
                 break;
             case 'c':   /* create */
                 createpidfile++;
                 break;
             default: return(usage(argv[0]));
         }
      }
   }
   
   /* Did they supply a pidfile to create/read ? */

   if ((optind >= argc) || (argc == 1)) {
       fprintf(stderr,"Not! You MUST specify a pidfile.\n",argv[0]);
       return(usage(argv[0]));
   }

   if (gettest) {
      fpid = getnfrpid(argv[optind]);
      fprintf(stderr,"%d = getnfrpid(%s)\n", fpid, argv[optind]);
      switch(fpid) {
          case 0 : fprintf(stderr,"Missing pidfile: %s\n", argv[optind]); 
                   break;
          case -1: fprintf(stderr,"Pidfile open read failed: %s\n", 
                                   argv[optind]); 
                   break;
          case -2: fprintf(stderr,"Pidfile format error: %s\n", argv[optind]); 
                   break;
          default: fprintf(stderr,"%s contains PID %d\n", argv[optind], fpid); 
                   break;
      }
   }

   if (testfor) {
      fpid = testnfrpid(argv[optind]);
      fprintf(stderr,"%d = testnfrpid(%s)\n", fpid, argv[optind]);
      switch(fpid) {
          case 0 : fprintf(stderr,"Not running\n"); 
                   break;

          case -1: fprintf(stderr,"Pidfile open read failed: %s\n", 
                                   argv[optind]); 
                   break;
          case -2: fprintf(stderr,"Pidfile format error: %s\n", argv[optind]); 
                   break;
          default: fprintf(stderr,"%s contains valid PID %d: running\n", 
                           argv[optind], fpid); 
                   break;
      }
   }

   if (createpidfile) {
      fpid = setnfrpid(argv[optind]);
      fprintf(stderr,"%d = setnfrpid(%s)\n", fpid, argv[optind]);
      switch(fpid) {
          case 0 : fprintf(stderr,"%s created successfully\n", argv[optind]); 
                   break;
          case 1 : fprintf(stderr,"Another process already running: %s\n", 
                                   argv[optind]); 
                   break;
          case -1: fprintf(stderr,"Pidfile open read failed: %s\n", 
                                   argv[optind]); 
                   break;
          case -2: fprintf(stderr,"Pidfile format error: %s\n", argv[optind]); 
                   break;
          case -3: fprintf(stderr,"Pidfile open write failed: %s\n", 
                                   argv[optind]); 
                   break;
          default: fprintf(stderr,"%s contains PID %d\n", argv[optind], fpid); 
                   break;
      }
   }

   return(0);
}
#endif
