/******************************************************************** 
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT") 

  File:   main.c 
  Author: Kent Landfield <kent@nfr.net>
 
********************************************************************/ 

#include <sys/types.h>
#include <sys/stat.h>

#include <sys/wait.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <pwd.h>
#include <stdarg.h>
#include <unistd.h>
#include <setjmp.h>
#include <syslog.h>
#include "alert.h"
#include "nfr_res.h"

#define NFRWATCH_PIDFILE "etc/run/nfrwatch.pid"

/* 
** NFRHOME and NFRUSER must be declared in the config file.
*/

/*
** SHELL and IFS not changeable except here 
*/
#define SHELL "/bin/sh"
#define IFS   " \t\n"

/* 
** Simple defaults - changable in config file 
*/
#define PATH    "/bin:/usr/bin:"
#define TMPDIR  "/tmp"

/* 
** Forbidden to change the following.
**     Note: HOME is set to NFRHOME
**     Note: LOGNAME and USER are set to NFRUSER
**     Note: NFRUID and NFRGID are set to the numeric uid and gid internally
*/
char *disallowed[] = { "SHELL","IFS","HOME","LOGNAME","USER","NFRUID","NFRGID",NULL };
#define NUMDISALLOWED 7

#define INITIALIZED 100

#define NORESPAWN 0
#define RESPAWN   1

#define NOPRIV   0
#define PRIV     1

#ifndef MAXENV
#define MAXENV  64
#endif

struct wentry {
	char *section;       /* Program label                  */
	char *start_cmdline; /* Command line to execute        */
	int  priv;           /* Execute as root or as NFR user */ 
	int  niceval;        /* Execute with the nice value    */ 
	int  respawn;        /* Respawn flag.                  */ 
	char *pidfile;       /* path to the pid file           */
	int  termsig;        /* signal to send to terminate    */
	char *termcmd;       /* command to run to terminate    */
        struct wentry *prev; /* Pointer to previous in list    */
        struct wentry *next; /* Pointer to next in list        */
};
struct wentry *Phead[BUFSIZ]; 

struct variable {
    char    *name;
    char    *value;
};
struct variable *vars[MAXENV];
char *newenvp[MAXENV];

/*
** NFR User Information
*/
char *Nfrhomedir;
char *Nfruser;
uid_t Nfruid;
gid_t Nfrgid;

/*
** General counters and control variables
*/
int envcnt = 0;
int entcnt = 0;
int verbose = 0;
int debug = 0;
unsigned int sleeptime = 60;

jmp_buf env; 

extern char *optarg;
extern int optind, opterr, optopt;

int getopt();
int isspace();

char *spacestrip(char *);
char *itoa(int);
int blankline(char *);
pid_t getnfrpid(char *);
pid_t testnfrpid(char *);
int setnfrpid(char *);
char **mkargv(int *, char ***, char *);

/*
** usage()
**
** Print the command's usage message
*/

int usage(char *progname)
{
   fprintf(stderr,"usage: %s [-cdv] [-t seconds] config_file\n", progname);
   fprintf(stderr,"\t-c          : Check supplied config file for syntax\n");
   fprintf(stderr,"\t-d          : Turn on debugging\n");
   fprintf(stderr,"\t-v          : Turn on verbose output\n");
   fprintf(stderr,"\t-t seconds  : Recheck period - defaults to %d seconds\n", sleeptime);
   return(1);
}

/*
** forbidden()
**
** check and make sure the variable to be added 
** is not in the "disallowed" list.
*/

int forbidden(char *envvar)
{
    register int i;
    char *p;

    for(i = 0; i < NUMDISALLOWED; i++) {
        p = disallowed[i];
        if (strcasecmp(p,envvar) == 0)
            return(1);
    }
    return(0);
}

/*
** init_wentry()
**
** initialize an allocated struct wentry members
*/

void init_wentry(struct wentry *t)
{
    t->section = NULL;
    t->start_cmdline = NULL;
    t->priv = 0;
    t->niceval = INITIALIZED;
    t->respawn = 0;
    t->pidfile = NULL;
    t->termsig = 0;
    t->termcmd = NULL;
    t->next = NULL;
}

/*
** free_wentry()
**
** free the previously allocated memory in a struct wentry.
*/

void free_wentry(struct wentry *t)
{
    if (t->section)          free(t->section);
    if (t->start_cmdline)    free(t->start_cmdline);
    if (t->pidfile)          free(t->pidfile);
    if (t->termcmd)          free(t->termcmd);

    free((char *) t);

    return;
}

/*
** free_list()
**
** Free all memory in the watched entries list.
*/

void free_list(struct wentry **phead)
{
    struct wentry *t;

    while(*phead != NULL) {
        /* delete the first element in the list */
        t = *phead;
        *phead = (*phead)->next;
        free_wentry(t);
    }
    return;
}

/*
** print_entry()
**
** Used to print out a watched entry for debugging purposes.
*/

void print_entry(struct wentry *t)
{
    if (t->section) 
        fprintf(stderr,"\n\tSection:\t[%s]\n",t->section);

    if (t->start_cmdline) 
        fprintf(stderr,"\tStart Command:\t[%s]\n",t->start_cmdline);

    if (t->priv == PRIV)
        fprintf(stderr,"\tExecute as: \t[%s]\n","root");
    else
        fprintf(stderr,"\tExecute as: \t[%s]\n","NFR USER");

    if (t->niceval == INITIALIZED )
        fprintf(stderr,"\tNice Value: \t[%s]\n","default");
    else
        fprintf(stderr,"\tNice Value: \t[%d]\n",t->niceval);

    if (t->respawn == RESPAWN)
        fprintf(stderr,"\tRespawn:\t[%s]\n","Yes");
    else
        fprintf(stderr,"\tRespawn:\t[%s]\n","No");

    if (t->pidfile)
        fprintf(stderr,"\tPIDfile:\t[%s]\n",t->pidfile);

    if (t->termsig > 0) 
        fprintf(stderr,"\tTermination:\t[kill -%d pid]\n",t->termsig);
    else
        fprintf(stderr,"\tTermination:\t[%s]\n",t->termcmd);
    return;
}

/*
** print_list()
**
** Used to print out a list of watched entries for debugging purposes.
*/

void print_list(struct wentry **phead)
{
    struct wentry *t;
    
    if ((t = *phead) != NULL) {
        while(t != NULL) {
          print_entry(t);
          t = t->next;
        }
    }
    return;
}


/*
** expandvars()
**
** Expand Environment variables inline.
*/

char *expandvars(char *str)
{
    static char ebuf[BUFSIZ];
    char tmpbuf[BUFSIZ];
    char *rsp;
    register int i;
    register int varlen;
    struct variable *t;

    strcpy(ebuf, str);

    for (i = 0; i < BUFSIZ; i++)
        tmpbuf[i] = '\0';

    while ((rsp = strchr(ebuf,'$')) != NULL) {
            
         /* variable found, need to expand inline */

         (void) strncpy(tmpbuf,ebuf,(rsp-ebuf));
         tmpbuf[(rsp-ebuf)] = '\0';

         ++rsp;

         /* locate the variable to expand */

         for (i = 0; i < envcnt; i++) {
             t = vars[i];
             varlen = strlen(t->name);
             if (strncasecmp(rsp, t->name, varlen) == 0) {
                 /* 
                 ** Found a valid replacement variable.
                 ** Copy the variable's value onto the tmpbuf.
                 ** Increment the pointer past the variable.
                 ** Copy the remaining portion of the ebuf into the tmpbuf.
                 ** Copy the tmpbuf back to ebuf.
                 */
                 strcat(tmpbuf,t->value);
                 rsp += varlen;
                 if (*rsp != '\0') 
                     strcat(tmpbuf,rsp);
                 strcpy(ebuf,tmpbuf);
                 break;
             }
         }
    }
    return(ebuf);
}

/*
** expand_environment()
**
** Expand the entire list of environment variables inline
*/

void expand_environment()
{
   register int i;
   char *p;

   for (i = 0; i < envcnt; i++) {
       if (strchr(vars[i]->value, '$') != NULL) {
             p = expandvars(vars[i]->value);
             if (vars[i]->value)
                 free (vars[i]->value);
             vars[i]->value = (char *)malloc(strlen(p)+1);
             strcpy(vars[i]->value, p);
       }
   }
}

/*
** expand_watch_entries()
**
** Expand variables in the list of watched entries.
*/

void expand_watch_entries()
{
    char *p;
    struct wentry *t;

    t = *Phead;

    while (t != NULL) {
        if (t->start_cmdline && ((p = strchr(t->start_cmdline,'$')) != NULL)) {
             p = expandvars(t->start_cmdline);
             if (t->start_cmdline)
                 free (t->start_cmdline);
             t->start_cmdline = (char *)malloc(strlen(p)+1);
             strcpy(t->start_cmdline, p);
        }
        if (t->pidfile && ((p = strchr(t->pidfile,'$')) != NULL)) {
             p = expandvars(t->pidfile);
             if (t->pidfile)
                 free (t->pidfile);
             t->pidfile = (char *)malloc(strlen(p)+1);
             strcpy(t->pidfile, p);
        }
        if (t->termcmd && ((p = strchr(t->termcmd,'$')) != NULL)) {
             p = expandvars(t->termcmd);
             if (t->termcmd)
                 free (t->termcmd);
             t->termcmd = (char *)malloc(strlen(p)+1);
             strcpy(t->termcmd, p);
        }
        t = t->next;
    }
}


/*
** add_entry()
**
** Add a watch entry to the list of watched entries.
*/

int add_entry(struct wentry **phead, char *buffer, int num)
{
    int i;
    struct wentry *t, *prev;
    register char *rp;
    register char *cp;

    char *f1 = NULL;    /* section */
    char *f2 = NULL;    /* start_cmdline */
    char *f3 = NULL;    /* priv */
    char *f4 = NULL;    /* nice */
    char *f5 = NULL;    /* respawn */
    char *f6 = NULL;    /* pidfile */
    char *f7 = NULL;    /* termination */

    f1 = buffer;
    rp = f1;

    for (i = 0; i < 6;i++) {
        while(*rp && (*rp != ':'))
            ++rp;
        if (!*rp) {
            fprintf(stderr,"error: line %d: incorrect # of fields.\n",num);
            return(0);
        }

        *rp++ = '\0';

        switch(i) {
           case 0: f2 = rp; break;
           case 1: f3 = rp; break;
           case 2: f4 = rp; break;
           case 3: f5 = rp; break;
           case 4: f6 = rp; break;
           case 5: f7 = rp; break;
        }
    }

    /* list in order */

    prev = NULL;
    t = *phead;
    
    /* search for insertion position */
    while (t != NULL) {
        prev = t;
        t = t->next;
    }

    if ((t = (struct wentry *) malloc(sizeof(struct wentry))) == NULL) {
        fprintf(stderr,"nfrwatch: add_entry: out of memory\n");
        return(0);
    }

    t->prev = prev;

    init_wentry(t);  /* initialize the data structure */

    /* Save the section name */

    cp = spacestrip(f1);
    if (!*cp) {
       fprintf(stderr,"nfrwatch: Line %d: no section specified. Terminating.\n", num);
        return(0);
    }
    t->section = (char *)malloc(strlen(cp)+1);
    strcpy(t->section, cp);

    
    /* Save the start command line */

    cp = spacestrip(f2);
    if (!*cp) {
        fprintf(stderr,"nfrwatch: Line %d: start command missing. Terminating.\n", num);
        return(0);
    }
    t->start_cmdline = (char *)malloc(strlen(cp)+1);
    strcpy(t->start_cmdline, cp);

    /* TBD - Need to get anal here. Check if process specified exists */

    /* Determine root execution status */

    cp = spacestrip(f3);
    if (*cp == 'Y' || *cp == 'y')
        t->priv = PRIV;
    else 
        t->priv = NOPRIV;

    /* Determine nice status */

    cp = spacestrip(f4);
    if (*cp)
        t->niceval = atoi(cp);

    /* Check if the process nice value is between 0 and 39 */

    if (t->niceval != INITIALIZED && (t->niceval < 0 || t->niceval > 39)) {
        fprintf(stderr,"nfrwatch: Line %d: invalid nice value specified. Terminating.\n", num);
        return(0);
    }
    
    /* Determine respawn status */

    cp = spacestrip(f5);
    if (*cp == 'r' || *cp == 'R')
        t->respawn = RESPAWN;
    else 
        t->respawn = NORESPAWN;
    
    /* Save pidfile location */

    cp = spacestrip(f6);
    if (!*cp) {
        fprintf(stderr,"nfrwatch: Line %d: no pidfile specified. Terminating.\n", num);
        return(0);
    }
    t->pidfile = (char *)malloc(strlen(cp)+1);
    strcpy(t->pidfile, cp);


    /* Save termination signal mode */

    cp = spacestrip(f7);
    if (!*cp) {
        fprintf(stderr,"nfrwatch: Line %d: no termination means specified. Terminating.\n", num);
        return(0);
    }

    if (strcasecmp(cp, "SIGTERM") == 0)
        t->termsig = SIGTERM;
    else if (strcasecmp(cp, "SIGHUP") == 0)
        t->termsig = SIGHUP;
    else if (strcasecmp(cp, "SIGINT") == 0)
        t->termsig = SIGINT;
    else if (strcasecmp(cp, "SIGKILL") == 0)
        t->termsig = SIGKILL;
    else if (strcasecmp(cp, "SIGUSR1") == 0)
        t->termsig = SIGUSR1;
    else if (strcasecmp(cp, "SIGUSR2") == 0)
        t->termsig = SIGUSR2;
    else {
        t->termsig = -1;  /* must be a command */

        /* strip any "quote marks" */

        if (*cp == '"')
            cp++;
        if ((rp = strchr(cp,'"')) != NULL)
             *rp = '\0';

        t->termcmd = (char *)malloc(strlen(cp)+1);
        strcpy(t->termcmd, cp);

        /* TBD - Need to get anal here. Check if process specified exists */
    }

    entcnt++;  /* Increment the entry counter */

    if (prev == NULL) {
        t->next = NULL;
        *phead = t;
    }
    else {                       
        t->next = prev->next;
        prev->next = t;
    }
    return(1);
}

/*
** add_envvar()
**
** Add an environment pair to the vars list
*/

int add_envvar(char *vname, char *val)
{
    if ((vars[envcnt] = (struct variable *) 
                   malloc(sizeof(struct variable))) == NULL) {
        fprintf(stderr,"nfrwatch: add_envvar: out of memory\n");
        return(-1);
    }

    /* save the variable name */

    vars[envcnt]->name = (char *)malloc(strlen(vname)+1);
    strcpy(vars[envcnt]->name, vname);

    /* save the variable value */

    vars[envcnt]->value = (char *)malloc(strlen(val)+1);
    strcpy(vars[envcnt]->value, val);

    envcnt++;
    return(1);
}

/*
** add_env()
**
** check to see if the buffer passed in is in the right format
** (var=val), check to see if it's an allowed variable and the
** add it to the vars list if it is.
*/

int add_env(char *buffer, int num)
{
    char *val;

    /*
    ** Separate into Variable and Value 
    */

    if ((val = strchr(buffer, '=')) == NULL) {
        fprintf(stderr,"nfrwatch: %d: Env format error\n", num);
        return(-1);
    }

    *val++ = '\0';

    /*
    ** First need to check and see if the variable 
    ** is in the forbidden list. If so, print an error message
    ** and continue;
    */
    if (forbidden(buffer)) {
        fprintf(stderr,"nfrwatch: %d: You CANNOT reset %s, ignoring\n", num, buffer);
        return(0);
    }

    return(add_envvar(buffer, val));
}

/*
** setup_environment()
**
** Read the comment below...
*/

void setup_environment(void) 
{
    struct stat stbuf;
    struct passwd *pw;
    char tb[BUFSIZ];
    int c, fnd;

    /*
    ** NFR Required configuration variables:
    **
    ** NFRHOME - directory that nfr is to be run from
    ** NFRUSER - name of account that nfr should run as (same as LOGNAME)
    **
    ** Internally set by NFRWATCH:
    **
    ** SHELL   - default shell 
    ** IFS     - For safety sake reset for proper usage.
    ** HOME    - directory that nfr is to be run from
    ** LOGNAME - name of account that nfr should run as
    ** USER    - name of account that nfr should run as (same as LOGNAME)
    **
    ** Optionally set by NFRWATCH:
    **
    ** PATH    - Default execution path
    ** TMPDIR  - Path to temporary file creation directory 
    ** 
    ** Created and set by NFRWATCH:
    **
    ** NFRUID  - to be filled in in this function with the numeric uid
    ** NFRGID  - to be filled in in this function with the numeric gid
    **
    ** To setup the environment, 
    **    - Assure NFRHOME specified.
    **    - Assure NFRUSER specified.
    **    - get the uid and gid for the NFRUSER.
    **    - Set NFRUID to uid's value
    **    - Set NFRGID to gid's value
    **    - Set HOME to NFRHOME's value
    **    - Set LOGNAME to NFRUSER's value
    **    - Set USER to NFRUSER's value
    **    - Set SHELL to compiled value
    **    - Set IFS to compiled value
    **    - Assure PATH is set, else use default
    **    - Assure TMPDIR is set, else use default
    */

    /*
    ** Check and see if NFRHOME specified and the directory exists. 
    */

    for (c = 0; c < envcnt; c++) {
        if (strcasecmp(vars[c]->name, "NFRHOME") == 0) {
            /* Found it. Check and make sure it is not blank */
            if (!vars[c]->value) {
                fprintf(stderr,"nfrwatch: NFRHOME not set! Terminating.\n");
                exit(1);
            }

            /* Now check and see if it is a valid directory */
            if (stat(vars[c]->value, &stbuf) != 0) {
                fprintf(stderr,"nfrwatch: NFRHOME %s directory missing! Terminating.\n",
                         vars[c]->value);
                exit(1);
            }
            if ((stbuf.st_mode & S_IFMT) != S_IFDIR) {
                fprintf(stderr,"nfrwatch: NFRHOME %s is not a directory! Terminating.\n", vars[c]->value);
                exit(1);
            }

            Nfrhomedir = (char *)malloc(strlen(vars[c]->value)+1);
            strcpy(Nfrhomedir,vars[c]->value);
        }
    }
 
    /*
    ** Check and see if NFRUSER specified. 
    ** If found then verify the user is a valid system user.
    */

    for (c = 0; c < envcnt; c++) {
        if (strcasecmp(vars[c]->name, "NFRUSER") == 0) {
            /* Found it. Check and make sure it is not blank */
            if (!vars[c]->value) {
                fprintf(stderr,"nfrwatch: NFRUSER not set! Terminating.\n");
                exit(1);
            }
            /*
            ** Now check and see if it is a valid system user 
            */
            if ((pw = getpwnam(vars[c]->value)) == NULL) { 
                fprintf(stderr,"nfrwatch: NFRUSER %s invalid, Terminating\n",
                         vars[c]->value);
                exit(1);
            }
            /*
            ** CYA:  some getpwnam implementations leave the 
            **       passwd file open. Don't need that...
            */
            endpwent(); 

            /*
            ** It is so save it off.
            */
            Nfruser = (char *)malloc(strlen(vars[c]->value)+1);
            strcpy(Nfruser,vars[c]->value);
        }
    }

    /*
    ** Need to setup the NFRUID and NFRGID variables
    */

    Nfruid = pw->pw_uid;
    Nfrgid = pw->pw_gid;

    sprintf(tb,"%ld",Nfruid);
    add_envvar("NFRUID", tb);

    sprintf(tb,"%ld",Nfrgid);
    add_envvar("NFRGID", tb);

    /*
    ** Set HOME to NFRHOME's value
    */
    add_envvar("HOME", Nfrhomedir);

    /*
    ** Set LOGNAME and USER to NFRUSER's value
    */
    add_envvar("LOGNAME", Nfruser);
    add_envvar("USER", Nfruser);

    /*
    ** Set SHELL to compiled value
    */
    add_envvar("SHELL", SHELL);

    /*
    ** Assure PATH is set, else use default
    */
    for (fnd = 0, c = 0; c < envcnt; c++) {
        if (strcasecmp(vars[c]->name, "PATH") == 0) {
            fnd = 1;
            break;
        }
    }
    if (!fnd)
        add_envvar("PATH", PATH);

    /*
    ** Assure TMPDIR is set, else use default
    */
    for (fnd = 0, c = 0; c < envcnt; c++) {
        if (strcasecmp(vars[c]->name, "TMPDIR") == 0) {
            fnd = 1;
            break;
        }
    }
    if (!fnd)
        add_envvar("TMPDIR", TMPDIR);

    /*
    ** Set IFS to compiled value
    */
    add_envvar("IFS", IFS);

    /*
    ** DONE PUTTING THINGS INTO THE VARS LIST
    */

    if (verbose && debug) {
        fprintf(stderr,"\nRaw Variables:\n\n");
        for (c = 0; c < envcnt; c++)
            fprintf(stderr,"\t%s=%s\n", vars[c]->name, vars[c]->value);
    }

    /*
    ** Expand any variable usage in the environment 
    ** or in the individual entries 
    */

    expand_environment();


    /* Now need to construct the actual execution environment */

    
    for (c = 0; c < envcnt; c++) {
         sprintf(tb,"%s=%s", vars[c]->name, vars[c]->value);
         newenvp[c] = strdup(tb);
    }

    expand_watch_entries();

    if (verbose) {
        if (debug) {
            fprintf(stderr,"\nExpanded Variables:\n\n");
            for (c = 0; c < envcnt; c++)
                fprintf(stderr,"\t%s=%s\n",vars[c]->name, vars[c]->value);
        }

        fprintf(stderr,"\nExecution Environment:\n\n");
        for (c = 0; c < envcnt; c++)
            fprintf(stderr,"\tENV[%0.2d] == %s\n", c, newenvp[c]);

        fprintf(stderr,"\nCommand Entries:\n");
        print_list(Phead);
    }
}

/*
** read_config()
**
** Read the config file and setup the needed structures
**
** Format:
**  var=value
**    or
**  +Program:Start Command:Priv:Nice:Respawn:Pidfile:Termination Command
**    +nfrd:bin/start_nfrd:N:0:R:~/etc/run/nfrd.pid:"bin/ctlnfrd halt"
*/

int read_config(char *configfile)
{
    char *cp;
    char *buf;
    char *newbuf;
    char buf2[BUFSIZ];
    int lineno;
    int freebuf;
    int nbuflen;
   
    FILE *fp;

    if ((fp = fopen(configfile,"r")) == NULL) {
        fprintf(stderr,"Panic: Cant open %s. Bailing out!\n", configfile);
        return(0);
    }

    /* initialize things */

    buf = (char *)malloc(BUFSIZ);
    nbuflen = BUFSIZ;
    lineno = 0;
    freebuf = 0;

    *Phead = NULL;   /* initial setup of list */

    while (fgets(buf, BUFSIZ, fp) != NULL) {
          lineno++;
          if (*buf == '#' || blankline(buf))
               continue;

          /* 
          ** Need to see if there is a continuation. If so
          ** then we must combine the line into one line
          ** for easing parsing.
          */

          while ((cp = strchr(buf,'\134')) != NULL) {
               lineno++;
               *cp = '\0';

                if (fgets(buf2, BUFSIZ, fp) == NULL) {
                    fprintf(stderr,"Error: %d: EOF in continuation.\n",lineno);
                    fclose(fp);
                    return(0);
                }
                if (*buf2 == '#' || blankline(buf2)) {
                    fprintf(stderr,"Error: %d: bad continuation.\n",lineno);
                    fclose(fp);
                    return(0);
                }

                /* strip whitespace from buf2 */

                if (!isspace(*buf2)) {
                    fprintf(stderr,"Error: %d: leading whitespace missing.\n", 
                                       lineno);
                    fclose(fp);
                    return(0);
                }

                cp = buf2;
                while (*cp && (isspace(*cp))) 
                       cp++;

                nbuflen += BUFSIZ;
                newbuf = (char *)malloc(nbuflen);
                sprintf(newbuf,"%s%s",buf, cp); /* put two lines together */
                free(buf);                /* Free the existing buf        */
                buf = newbuf;             /* Point at the expanded buffer */
                freebuf = 1;
          }
         
          /*
          **  At this point, buf is the buffer with the complete line in it 
          **  Strip the newline
          */
          if ((cp = strchr(buf,'\n')) != NULL) 
              *cp ='\0';

          /*
          ** Add watch entry
          */
          if (*buf == '+') {
              if (!add_entry(Phead,buf+1,lineno)) {  /* add entry to table */
                  fclose(fp);
                  return(0);
              }
          }
          else { /* Add Environment variable */
              if (add_env(buf,lineno) == -1) {  /* add entry to table */
                  fclose(fp);
                  return(0);
              }
          }

          if (freebuf) {   /* Don't let the buffer grow out of control */
              free(buf);   /* A bit wasteful but only done at startup  */
              freebuf = 0;
              buf = (char *)malloc(BUFSIZ);
          }

    }
    if (*buf)
        free(buf);

    fclose(fp);
    return(1);
}


/*
** one_and_only(char *pidfile)
**
**    Create a pid file for this process and print error messages 
*/
  
int one_and_only(char *pidfl)
{
    int pid;
  
    pid = setnfrpid(pidfl);  /* Try and create the pid file */

    switch(pid) {
       case 0 : syslog(LOG_INFO, "nfrwatch: starting");
                break;
       case 1 : syslog(LOG_ERR, "Another nfrwatch running. Terminating...");
                break;
       case -1: syslog(LOG_ERR,"%s open read failed! Terminating: %m",pidfl);
                break;
       case -2: syslog(LOG_ERR,"%s format error! Terminating: %m",pidfl);
                break;
       case -3: syslog(LOG_ERR,"%s open write failed! Terminating: %m",pidfl);
                break;
       default: syslog(LOG_ERR, "nfrwatch software error. Huh? Terminating.");
                break;
    }
    return(pid);
}

/*
** start_entry()
**
** Startup an individual process.
*/
int start_entry(struct wentry *t)
{
   pid_t pid;
   char **eargv;
   int  eargc;

   if ((pid = fork()) == -1) {  /* error */
        perror(t->section);
        return(-127);
   }
   else if (pid > 0)            /* parent */
        return(0);
             
   /* Child from here on out */
   
   /* Set the nice value if specified */

   print_entry(t);

   if (t->niceval != INITIALIZED)
       (void) nice((t->niceval - 20));

   /* setuid/setgid to the nfr user listed in the config file */

   if ((t->priv != PRIV) && (getuid() == 0)) {

       /* Limit the group access for this user to the primary nfr group */

       if (setgroups(1, &Nfrgid) < 0)       /* non-fatal at present */
           syslog(LOG_ERR, "nfrwatch cant setgroups: %m");

       /* Set the User privileges to the specified nfr user */

       if (setgid(Nfrgid) != 0) {
           perror(Nfruser);
           exit(-127);
       }

       /* Set the Group privileges to the specified nfr group */

       if (setuid(Nfruid) != 0) {
           perror(Nfruser);
           exit(-127);
       }
   }

   /*
   ** need to setup the argv vector array 
   */

   (void) mkargv(&eargc, &eargv, t->start_cmdline);

   /* Now execute the program with the new env */

   if (debug) {
       fprintf(stderr,"\tStart Command:\t[%s]\n",t->start_cmdline);
       fprintf(stderr,"%s executing as UID %ld - GID %ld\n", 
               eargv[0],getuid(),getgid());
   }

   alarm(0);  /* turn off any pending alarm */

   /* Reset any signals set to their system defaults */

   (void) signal(SIGCHLD, SIG_DFL);
   (void) signal(SIGTERM, SIG_DFL);

   /* Go for the gold... */

   (void)execve(eargv[0], (char **)eargv, (char **)newenvp);

   if (debug) {
       perror(eargv[0]);
       fprintf(stderr,"nfrwatch start failed for [%s]\n",t->start_cmdline);
       fprintf(stderr,"%s executing as UID %ld - GID %ld\n", 
               eargv[0],getuid(),getgid());
   }
   syslog(LOG_CRIT, "nfrwatch cant exec %s %m", eargv[0]);
   _exit(1);

   /* NOTREACHED */
   return(1);
}

static void signal_handler(sig)
int sig;
{
     switch (sig)     {
          case SIGTERM:   longjmp(env,sig);    /* process for termination    */
                          break;               /* break never reached        */
          case SIGCHLD:   longjmp(env,sig);    /* process for child's death  */
                          break;               /* break never reached        */
          default:        exit(sig);           /* ???                        */
     }
}

/*
** start_nfr_procs()
**
** Does the initial startup of the nfrd processes
*/

void start_nfr_procs()
{
    struct wentry *t;

   /* setup signal handlers */
   (void) signal(SIGCHLD, SIG_IGN);

    t = *Phead;
    
    while(t != NULL) {
          start_entry(t);
          sleep(1);
          t = t->next;
    }
    return;
}

/*
** check_nfr_procs()
**
**   Check to see if the processes specified are running.  If one or more
**   are found to have died, start the missing processes up again if the 
**   respawn flag is specified.
*/

void check_nfr_procs()
{
    struct wentry *t;

    t = *Phead;

    /* setup signal handlers */
    (void) signal(SIGCHLD, SIG_IGN);

    if (debug && verbose)
       fprintf(stderr,"checking for NFR processes\n");
    
    if (t != NULL) {
        while(t != NULL) {
            if (testnfrpid(t->pidfile) <= 0) {
                start_entry(t);
                sleep(1);
            }
            t = t->next;
        }
    }

    return;
}

/*
** stop_nfr_procs()
**
** Send a termination request to the individual processes.
**
** The general flow: 
**       The termination method can consist of sending one of the 
**       supported signals to the process or there can be a termination 
**       command specified that is simply run.  If a termsig is used
**       then we open the pidfile, get the pid and send the signal.
**       If there is no termsig specified then the termination
**       command (termcmd) is executed directly.
*/

void stop_nfr_procs()
{
    struct wentry *t;
    struct wentry *prev;
    pid_t tpid;
    
    prev = NULL;
    t = *Phead;

    while(t != NULL) {
          prev = t;
          t = t->next;
    }

    t = prev;

    while(t != NULL) {
        if (t->termsig > 0) {   /* By signal */
            if ((tpid = getnfrpid(t->pidfile)) > 0) {
                if (debug) 
                    fprintf(stderr,"Terminating %s: [kill -%d %d]",
                                      t->section, t->termsig, tpid);
                syslog(LOG_INFO,"Terminating %s: [kill -%d %d]",
                                    t->section,t->termsig, tpid);
                (void) kill(tpid, t->termsig);
            }
        }
        else {                  /* By command */
            if (debug) 
                fprintf(stderr,"Terminating %s: [%s]",t->section,t->termcmd);
            syslog(LOG_INFO,"Terminating %s: [%s]",t->section,t->termcmd);
            (void) system(t->termcmd);
        }
        t = t->prev;
    }
    return;
}


void keep_em_goin(void)
{
    int status;

    for (;;) {
        (void) signal(SIGCHLD, signal_handler);

        sleep(sleeptime);

        do {
           check_nfr_procs();
           /*
           ** It then loops checking the pid files listed to assure the the
           ** process are still active. If a process terminates and the respawn
           ** flag indicates a "Respawn" is requested then the start command is
           ** re-executed.
           */
       } while (wait3(&status, WNOHANG, (struct rusage *) NULL) > 0);
    }
    /* NOTREACHED */
}



/*
** Ye Olde Main.
**
*/

int main(int argc, char **argv)
{
   int c;
   int termination = 0;
   int check_config = 0;
   pid_t fpid;

   if (argc > 1) {
      while ((c = getopt(argc, argv, "cdvt:")) != EOF) {
         switch (c) {
             case 'c':   /* just check the cofigurtion file */
                 check_config++;
                 verbose++;
                 break;
             case 'd':   /* debugging stuff */
                 debug++;
                 verbose++;
                 break;
             case 'v':   /* verbose */
                 verbose++;
                 break;
             case 't':   /* sleeptime */
                 sleeptime = atoi(optarg);
                 break;
             default:   
                 return(usage(argv[0]));
         }
      }
   }
   
   /* Did they supply a config file to read ? */

   if ((optind >= argc) || (argc == 1)) {
       fprintf(stderr,"%s: Error: You MUST specify a config_file.\n",argv[0]);
       usage(argv[0]);
       return(1);
   }

   if (check_config) {
       if (!read_config(argv[optind]))
           exit(-1);
       setup_environment();

       /* checking configuration only so bail out */
       return(0);
   }

   /* 
   ** The idea here is to start NFR with the proper uid and gid
   ** and to begin execution in the proper place.  The user will
   ** supply a path to the config file for a specific instantiation
   ** of NFR.  Multiple NFRs can be started by supplying the paths
   ** to different configuration files.
   **
   ** Initially nfrwatch is started as root from an rc.* file.
   ** Upon startup it reads the config file and sets up memory tables.
   **
   ** - cds to the NFR HOME directory listed in the config file.
   ** - sets up a new enviroment for the child to execute commands with.
   ** - setuid/setgids to the nfr user listed in the config file.
   ** - executes the programs listed in the config file with the new env.
   **
   ** It then loops checking the pid files listed to assure the the 
   ** process are still active. If a process terminates and the respawn 
   ** flag indicates a "Respawn" is requested then the start command is 
   ** re-executed.
   */

   /*
   ** Personal Thought: 
   **      I am using syslog to log messages. What I may want to do 
   **      is create a function that recognizes if alertd has been 
   **      started.  In otherwords, use syslogd exclusive until alertd 
   **      is started and then log to both.  That way the alert facilities
   **      can be used for most things but syslog has a complete copy
   **      of everything in the event alertd craters. Just a thought...
   */

   (void)openlog("nfrwatch", LOG_PID, LOG_DAEMON);

   if ((fpid = fork()) > 0) {
        /* parent */
        exit (0);    /* nothing to be done here */
   }
   else if (fpid == -1) {
        perror(argv[0]);
        syslog(LOG_CRIT, "terminating - initial spawn failure: %m");
        exit(-127);    
   }

   if (!read_config(argv[optind]))
       exit(-1);

   setup_environment();
   
   /* The rest here is in the child */

   /* cd to the NFR HOME directory listed in the config file  */

   if (chdir(Nfrhomedir) != 0) {
       syslog(LOG_CRIT, "Terminating - cant change to %s: %m", Nfrhomedir);
       perror(Nfrhomedir);
       exit(-1);
   }
 
   /* make sure we are the only copy running */

   if (one_and_only(NFRWATCH_PIDFILE) != 0)
       return(1);

   start_nfr_procs();

   for(;;) {
       switch (setjmp(env)) {
           case SIGTERM:
                   termination++;
                   syslog(LOG_INFO,"Termination request received. Beginning");
                   if (debug) 
                       fprintf(stderr,"nfrwatch: beginning termination.\n");
                   break;
           case SIGCHLD:
                   check_nfr_procs();   /* process for child's death */
                   break;
           case 0: break;
       }
    
       if (termination) 
           break;

       (void) signal(SIGTERM, signal_handler);

       keep_em_goin();
      /* NOTREACHED */
   }

   /*
   ** Need to send termination commands/signals to existing processes,
   ** do some housekeeping, tell someone we're leaving and leave.
   */

   (void) signal(SIGCHLD, SIG_IGN);  /* Don't whine while I'm working   */
   (void) signal(SIGTERM, SIG_IGN);  /* I already know it quitting time */

   stop_nfr_procs();                 /* Tell all to quit... */

   unlink(NFRWATCH_PIDFILE);  /* clean up the file droppings */
   free_list(Phead);          /* clean up your play room.    */
   syslog(LOG_INFO,"Termination complete.");
   closelog();                /* close the door behind you...*/
   if (debug) 
       fprintf(stderr,"nfrwatch: termination complete.\n");
   return(0);
}
