/*****************************************************************************
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT")


  File:   rcmain.c
  Author: Andrew Lambeth <andrew@nfr.net>

 *****************************************************************************/

#include "sys-include.h"

/* #define DEBUG_LEV 1  */
#include "debug.h"

#include "libnfr.h"
#include "../lib/rc/rc.h"

int Usage(void)
{
  fprintf(stderr, 
	  "Usage: rc [-vh] -o OutputDir SrcName [ SrcName ... ]\n"
	  "options:\n"
	  "  v - increment verbosity\n"
	  "  h - spit out a header file with #define statements for each id\n"
	  "\n");
  return 1;
}


#define MAX_SRCS 32

int main(int argc, char **argv)
{
  PResFile pResFile;
  char HeaderFileName[] = "res.h";
  FILE *OutFile = NULL;
  FILE *In = NULL;
  BOOL DestExpected = FALSE;
  BOOL HaveDest = FALSE;
  int i = 0;
  char *p = NULL;
  ULONG AlertCount = 0, AlertSrcCount = 0, StringCount = 0;
  ULONG ResCount = 0, SevCount = 0, FacCount = 0;
  int Result = 0;
  PSymbol pSymbol;
  time_t t;
  char Dest[512];
  int DestLen = 0;
  char *SrcNames[MAX_SRCS];
  int NumSrcs = 0;
  int TotalLines = 0;
  BOOL WantHeader = FALSE;

  DBG_OUT("Start rc\n");

  i = 1;			/* parse cmd line */
  while(i < argc)
    {
      p = argv[i];
      if (*p == '-')
	{
	  while (*++p != 0)
	    {
	      switch (*p)
		{
		case 'h':
		  WantHeader = TRUE;
		  break;

		case 'o':
		  DestExpected = TRUE;
		  break;

		case 'v':
		  gVerbosity++;
		  break;

		default:
		  SExit(Usage());
		}
	    }
	}
      else if (DestExpected)
	{
	  strncpy(Dest, argv[i], sizeof(Dest) - 2);
	  DestLen = strlen(Dest);
	  Dest[DestLen++] = '/';
	  Dest[DestLen++] = 0;
	  DestExpected = FALSE;
	  HaveDest = TRUE;
	}
      else
	{
	  if(NumSrcs >= MAX_SRCS)
	    {
	      fprintf(stderr, "too many src files\n");
	      SExit(1);
	    }
	  SrcNames[NumSrcs++] = argv[i];
	}
      i++;
    }

  if(!HaveDest)
    {
      fprintf(stderr, "Must specify a destination path\n");
      exit(1);
    }


  DBG_OUT("-> Rc_Init()\n");

  Rc_Init(); /* initialize global structures used by Rc_Parse */

  DBG_OUT("<- Rc_Init()\n");

  for (i = 0; i < NumSrcs; i++)
    {
      In = fopen(SrcNames[i], "r");
      if (In == NULL)
	{
	  perror(SrcNames[i]);
	  exit(1);
	}

      gpCurFilename = SrcNames[i];

      DBG_OUT("-> Rc_Parse()\n");

      Rc_Parse(RC_MODE_SCAN | RC_MODE_COMPILE, In, NULL, NULL);
      
      DBG_OUT("<- Rc_Parse()\n");

      fclose(In);
    }

  /* the callbacks used in these functions just call DBG_OUT(), so unless
   * DEBUG_LEV > 0 they just act as counters */
  AlertCount = SymTab_ForEach(RES_ALERT, PrintAlertResFields, 0, 0, 
			      gpResTab->SymTab);
  AlertSrcCount = SymTab_ForEach(RES_ALERT_SRC, PrintAlertSrcFields, 0, 0, 
				 gpResTab->SymTab);
  StringCount = SymTab_ForEach(RES_STRING, PrintStringResFields, 0, 0, 
			       gpResTab->SymTab);
  FacCount = SymTab_ForEach(ID_FAC, PrintFacFields, 0, 0, gpResTab->SymTab);
  SevCount = SymTab_ForEach(ID_SEV, PrintSevFields, 0, 0, gpResTab->SymTab);
  ResCount = AlertCount;

  if( WantHeader && (gErrors == 0) )
    {
      /* create a header with the new definitions */
      Dest[DestLen] = 0;
      strncat(Dest, HeaderFileName, sizeof(Dest) - DestLen);
      OutFile = fopen(Dest, "w");
      if (! OutFile )
	{
	  perror(Dest);
	  exit(1);
	}
      t = time(0);
      fprintf(OutFile, "/*\n * DO NOT EDIT THIS FILE MANUALLY\n "
	      "* automatically generated by rc \n */\n\n");

      /* bug: should prolly make this say something unique */
      fprintf(OutFile,"#ifndef NFR_RES_H_\n#define NFR_RES_H_\n\n");
      
      fprintf(OutFile,"#define %-30s0x%.8X\n",
	      "NFR_ERR_SUCCESS", NFR_ERR_SUCCESS);
      fprintf(OutFile,"\n/* output facilities */\n");
      SymTab_ForEach(ID_FAC, PrintGenDefine, 0, OutFile, gpResTab->SymTab);
      fprintf(OutFile,"\n/* alert severities */\n");
      SymTab_ForEach(ID_SEV, PrintGenDefine, 0, OutFile, gpResTab->SymTab);

      fprintf(OutFile,"\n\n");
      fprintf(OutFile,"\n/* alert source resource codes */\n");
      SymTab_ForEach(RES_ALERT_SRC, PrintResDefine, 0, OutFile, 
		     gpResTab->SymTab);
      fprintf(OutFile,"\n/* alert resource codes */\n");
      SymTab_ForEach(RES_ALERT, PrintResDefine, 0, OutFile, 
		     gpResTab->SymTab);
      fprintf(OutFile,"\n/* string resource codes */\n");
      SymTab_ForEach(RES_STRING, PrintResDefine, 0, OutFile, 
		     gpResTab->SymTab);

      fprintf(OutFile,"\n\n#endif /* NFR_RES_H_ */\n\n");

      fclose(OutFile);
    }

  if (gErrors == 0)
    {
      /* write the binary alert resource file */
      Dest[DestLen - 1] = 0;
      strncat(Dest, RESFILE_ALERTS, sizeof(Dest) - DestLen);
      pResFile = ResFile_Create(Dest, AlertCount);
      SymTab_ForEach(RES_ALERT, ResSym_Write, 0, pResFile, gpResTab->SymTab);
      ResFile_Close(pResFile);

      /* write the binary alert source resource file */
      Dest[DestLen - 1] = 0;
      strncat(Dest, RESFILE_ALERT_SRCS, sizeof(Dest) - DestLen);
      pResFile = ResFile_Create(Dest, AlertSrcCount);
      SymTab_ForEach(RES_ALERT_SRC, ResSym_Write, 0, pResFile, 
		     gpResTab->SymTab);
      ResFile_Close(pResFile);

      /* write the binary string resource file */
      Dest[DestLen - 1] = 0;
      strncat(Dest, RESFILE_STRINGS, sizeof(Dest) - DestLen);
      pResFile = ResFile_Create(Dest, StringCount);
      SymTab_ForEach(RES_STRING, ResSym_Write, 0, pResFile, gpResTab->SymTab);
      ResFile_Close(pResFile);

      /* write the text facilities.cf file */
      /* bug: don't hardcode filenames */
      OutFile = fopen("facilities.cf", "w");
      if (! OutFile )
	{
	  perror("facilities.cf");
	  exit(1);
	}
      /* bug: remove this FAC_NULL stuff */
      /* fprintf(OutFile, "NFR_FAC_NULL = 0\n"); */
      SymTab_ForEach(ID_FAC, PrintFacNVPair, 0, OutFile, gpResTab->SymTab);
      fclose(OutFile);


      /* write the map file */
      Dest[DestLen - 1] = 0;
      strncat(Dest, RES_MAPFILE, sizeof(Dest) - DestLen);
      OutFile = fopen(Dest, "w");
      if (! OutFile )
	{
	  perror(Dest);
	  exit(1);
	}
      SymTab_ForEach(RES_ALERT, PrintMapEntry, 0, OutFile, 
		     gpResTab->SymTab);
      SymTab_ForEach(RES_ALERT_SRC, PrintMapEntry, 0, OutFile, 
		     gpResTab->SymTab);
      SymTab_ForEach(RES_STRING, PrintMapEntry, 0, OutFile, 
		     gpResTab->SymTab);
      fclose(OutFile);
    }


  if(gVerbosity > 1)
    {
      fprintf(stderr,"\nRC\n-----------------------------------------------"
	      "\n\t%ld lines processed\n\t%ld Alert facilities defined\n"
	      "\t%ld Alerts defined\n\t%ld Alert Sources defined\n"
	      "\t%ld Strings defined\n\t%ld warnings\n\t%ld errors\n"
	      "-----------------------------------------------\n\n",
          TotalLines, FacCount, AlertCount, AlertSrcCount, StringCount, 
	  gWarnings, gErrors);
    }


  Rc_Cleanup(); /* free global structures used by Rc_Parse() */

  SExit(gErrors);
  return -1;			/* not reached, just shutup cc -Wall */
}
