/***************************************************************************** 
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT") 

  File:   config.c 
  Author: Kent Landfield <kent@nfr.net>
 
 *****************************************************************************/ 

#include "sys-include.h"

#include "spaceman.h"
#include "alert.h"
#include "nfr_res.h"

#define ONE_K    1024
#define ONE_MEG  1000000
#define ONE_GIG  1000000000

struct node *Phead[BUFSIZ]; 

int valid_format(char *fptr);

extern int debug;
extern char *Configpath; /* config file to use */

void free_node(struct node *t)
{
    if (t->backend)          free(t->backend);
    if (t->recorder)         free(t->recorder);
    if (t->command)          free(t->command);
    if (t->location)         free(t->location);
    if (t->dbinfo.indexpath) free(t->dbinfo.indexpath);

    free((char *) t);

    return;
}

void free_list(struct node **phead)
{
    struct node *t;

    while(*phead != NULL) {
        /* delete the first element in the list */
        t = *phead;
        *phead = (*phead)->next;
        free_node(t);
    }
    return;
}

int blankline(char *bl_line)   /* Return 1 if blank and 0 otherwise */
{
     register char *blcp;

     for (blcp = bl_line; *blcp; blcp++) {
        if (!isspace(*blcp))
            return (0);
     }   
     return(1);
}

void init_node(struct node *t)
{
    t->backend = NULL;
    t->recorder = NULL;
    t->quota = 0;
    t->age = 0;
    t->priority = 0;
    t->action = NOACTION;
    t->command = NULL;
    t->location = NULL;
    t->next = NULL;
    t->dbinfo.indexpath = NULL;
    t->dbinfo.dbsize = 0;
    t->dbinfo.starttime = 0;
    t->dbinfo.endtime = 0;
}

int add_entry(struct node **phead, char *buffer, int num)
{
    int i;
    long val;
    struct node *t, *prev;
    register char *rp;
    register char *cp;
    char *f1 = NULL;    /* line type */
    char *f2 = NULL;    /* recorder or location */
    char *f3 = NULL;    /* quota */
    char *f4 = NULL;    /* age */
    char *f5 = NULL;    /* priority */
    char *f6 = NULL;    /* action */
    char *f7 = NULL;    /* command */
    char *f8 = NULL;    /* location */

    f1 = buffer;
    rp = f1;

    for (i = 0; i < 7;i++) {
        while(*rp && (*rp != ':'))
            ++rp;
        if (!*rp) {
            if (debug)
                report("Format error: line %d: incorrect # of fields.\n",num);
            else 
               Alert_Args_Generate(NFR_SRC_SPACEMAND, NFR_SPACEMAND_BADNUMFIELDS,
                                  itoa(num), 0);
            return(0);
        }

        *rp++ = '\0';

        switch(i) {
           case 0: f2 = rp; break;
           case 1: f3 = rp; break;
           case 2: f4 = rp; break;
           case 3: f5 = rp; break;
           case 4: f6 = rp; break;
           case 5: f7 = rp; break;
           case 6: f8 = rp; break;
        }
    }

    /* list in order */

    prev = *phead;
    t = *phead;
    
    /* search for insertion position */
    while (t != NULL) {
        prev = t;
        t = t->next;
    }

    if ((t = (struct node *) malloc(sizeof(struct node))) == NULL) {
        if (debug)
            report("add_entry: %s: out of memory\n", "malloc");
        else
            Alert_Generate(NFR_SRC_SPACEMAND, NFR_SPACEMAND_MALLOCERR);
        return(0);
    }

    init_node(t);  /* initialize the data structure */

    /*
    ** Determine if this is an indexed archive or simply a directory
    ** that is to be read. If it is a directory then backend contains
    ** the disk location and recorder is set to NULL and not used.
    */

    if (*f1 == 'R' || *f1 == 'r') {
        t->node_type = RKIVEINDEX;
        /* Now separate the backend/recorder entries 
        **
        ** Put a '*' in the recorder field if no recorder has
        ** been specified and the entry it pertains to all 
        */

        if ((cp = strchr(f2,'/')) != NULL) {
            *cp++ = '\0';
            if (*cp) {
                t->recorder = (char *)malloc(strlen(cp)+1);
                strcpy(t->recorder,cp);
            }
            else {
                t->recorder = (char *)malloc(2);
                strcpy(t->recorder,"*");
            }
        }
        else {
            t->recorder = (char *)malloc(2);
            strcpy(t->recorder,"*");
        }

        t->backend = (char *)malloc(strlen(f2)+1);
        strcpy(t->backend, f2);
    }
    else if (*f1 == 'D' || *f1 == 'd') {
        /*
        ** This is a directory entry. Need to allocate space for maximum
        ** length of a path to a file since directory expiration routines 
        ** will also use this storage.
        */
        t->node_type = DIRONLY;
        t->backend = (char *)malloc(BUFSIZ);
        strcpy(t->backend, f2);
    }
    else {
        if (debug)
            report("Type Error: line %d: bad type specified.\n",num);
        else 
            Alert_Args_Generate(NFR_SRC_SPACEMAND, 
                                   NFR_SPACEMAND_TYPE_ERR, itoa(num), 0);
        return(0);
    }
 
    /*
    ** Now save the quota info 
    **
    ** It is possible to have no QUOTA information.  In that case
    ** You must have AGE information. Checked later.
    */
    
    cp = f3;
    while(*cp && isdigit(*cp))
        ++cp;

    /*
    ** If no special expansion character specified consider
    ** the value as absolute bytes.
    */

    if (!*cp)  {
        if (*f3)       /* Is there something there to save */
            t->quota = atol(f3);
        else 
            t->quota = -1;
    }
    else {    /* determine the expansion character's value */
        switch (tolower(*cp)) {
             case 'k': val = ONE_K;   break; 
             case 'm': val = ONE_MEG; break;
             case 'g': val = ONE_GIG; break; 
             default : if (debug)
                           report("Quota Error: line %d: bad format char.\n",num);
                       else 
                           Alert_Args_Generate(NFR_SRC_SPACEMAND, 
                                   NFR_SPACEMAND_QUOTA_ERR, itoa(num), 0);
                       return(0);
        }
        *cp = '\0';
        t->quota = atol(f3) * val;;
    }

    /* Now save the age info */
    
    cp = f4;
    while(*cp && isdigit(*cp))
        ++cp;

    /*
    ** If no special expansion character specified consider
    ** the value as absolute bytes.
    */

    if (!*cp) {
        if (*f4)
            t->age = atol(f4);
        else if (t->quota == 0) {
             if (debug)
                 report("Error: line %d: must specify AGE OR QUOTA.\n",num);
             else 
                 Alert_Args_Generate(NFR_SRC_SPACEMAND, 
                                   NFR_SPACEMAND_WHICH_ONE, itoa(num), 0);
             return(0);
        }
        else 
            t->age = -1;
    }
    else {    /* determine the expansion character's value */
        switch (tolower(*cp)) {
             case 'm': val = SECSPERMIN; break;
             case 'h': val = SECSPERHOUR; break; 
             case 'd': val = SECSPERDAY;break;   
             case 'w': val = SECSPERWEEK;break;   
             default : if (debug)
                           report("Age Error: line %d: bad format char.\n",num);
                       else
                           Alert_Args_Generate(NFR_SRC_SPACEMAND, 
                                   NFR_SPACEMAND_AGE_ERR, itoa(num), 0);
                       return(0);
        }
        *cp = '\0';
        t->age = atol(f4) * val;;
    }

    if (t->quota == -1 && t->age == -1) {
        if (debug)
            report("Error: line %d: must specify AGE OR QUOTA.\n",num);
        else 
            Alert_Args_Generate(NFR_SRC_SPACEMAND, 
                                   NFR_SPACEMAND_WHICH_ONE, itoa(num), 0);
        return(0);
    }

    /* Now save the priority info */
    
    cp = f5;
    while(*cp && isdigit(*cp))
        ++cp;

    if (!*cp) {
        if (!*f5) {
            if (debug)
                report("Error: line %d: must specify Priority level.\n",num);
            else 
                Alert_Args_Generate(NFR_SRC_SPACEMAND, 
                                NFR_SPACEMAND_PRIORITY_GONE, itoa(num), 0);
             return(0);
        }
        t->priority = atoi(f5);
    }
    else {
        if (debug)
            report("%s\n\tPriority Error: %d: 0-10 only.\n", buffer,num);
        else 
            Alert_Args_Generate(NFR_SRC_SPACEMAND, 
                                NFR_SPACEMAND_PRIORITY_ERR, itoa(num), 0);
        return(0);
    }
    
    /* Now save the action info */

    if (strcasecmp(f6,"None") == 0)
        t->action = NOACTION;
    else if (strcasecmp(f6,"Purge") == 0)
        t->action = PURGE;
    else if (strcasecmp(f6,"Save") == 0)
        t->action = SAVE;
    else if (strcasecmp(f6,"Process") == 0)
        t->action = PROCESS;
    else if (strcasecmp(f6,"Process+Save") == 0)
        t->action = PROCESSNSAVE;
    else if (strcasecmp(f6,"Process+Purge") == 0)
        t->action = PROCESSNPURGE;
    else {
        if (debug)
           report("%s\n\tAction Error: %d: Invalid action defined\n",buffer,num);
        else
            Alert_Args_Generate(NFR_SRC_SPACEMAND,
                                NFR_SPACEMAND_ACTION_ERR, itoa(num), 0);
        return(0);
    }
 
    /*
    ** Now save the command info 
    ** If a PROCESS* specified assure there is an associated command
    ** to execute.  Check and make sure the pattern is valid and then
    ** save the pattern here for later use.
    */
  
    if (!*f7) {
        if (t->action == PROCESSNSAVE || t->action == PROCESSNPURGE) {
            if (debug)
                report("Error: line %d: Required command missing.\n",num);
            else
                Alert_Args_Generate(NFR_SRC_SPACEMAND,
                                NFR_SPACEMAND_CMD_MISSING, itoa(num), 0);

            return(0);
        }
        else
            t->command = NULL;
    }
    else if (!valid_format(f7)) {
        if (debug)
            report("Command Error: line %d: command format [%s] invalid.\n",num,f7);
        else 
            Alert_Args_Generate(NFR_SRC_SPACEMAND,
                                NFR_SPACEMAND_CMD_ERR, itoa(num), f7, 0);

        return(0);
    }
    else
        t->command = strdup(f7);
     
    /*
    ** Now save the location info 
    ** If SAVE is specified assure there is an associated LOCATION
    ** to save it to. Check and make sure the pattern is valid and
    ** then save the pattern for later use.
    */
  
    if ((cp = strchr(f8,':')) != NULL)  /*Remove trailing ':'*/
        *cp = '\0';

    if (!*f8) {
        if (t->action == SAVE) {
            if (debug)
                report("Error: line %d: Save Location not specified.\n",num);
            else 
                Alert_Args_Generate(NFR_SRC_SPACEMAND,
                                NFR_SPACEMAND_SAVELOC_ERR, itoa(num), 0);
            return(0);
        }
        else
            t->location = NULL;
    }
    else if (!valid_format(f8)) {
        if (debug)
            report("Location Error: line %d: location format [%s] invalid.\n",num,f7);
        else 
            Alert_Args_Generate(NFR_SRC_SPACEMAND,
                                NFR_SPACEMAND_LOC_ERR, itoa(num), f8, 0);
        return(0);
    }
    else
        t->location = strdup(f8);

    /*
    ** At this point the structure is filled for the present.
    ** Time to add it to the list of structures.
    */
     
    if (prev == NULL) {
        t->next = NULL;
        *phead = t;
    }
    else {                       
        t->next = prev->next;
        prev->next = t;
    }
    return(1);
}

/*
**   package/recorder:QUOTA:AGE:PRIORITY:ACTION:COMMAND:LOCATION:
**
** webwatcher/urlhist:1m::7:Process+Purge:\
**      /bin/mail -s 'URL History for %M %y" %f::
*/


int read_config(void)
{
    char *cp;
    char *buf;
    char *newbuf;
    char buf2[BUFSIZ];
    int lineno;
    int freebuf;
    int nbuflen;
   
    FILE *fp;

    if ((fp = fopen(Configpath,"r")) == NULL) {
         if (debug)
             report("Panic: Cant open the %s config file. Bailing out!\n", 
                   Configpath);
         else 
             Alert_Args_Generate(NFR_SRC_SPACEMAND,
                             NFR_SPACEMAND_CONFIG_MISSING, Configpath, 0);
         exit(-1);
    }

    buf = (char *)malloc(BUFSIZ);
    nbuflen = BUFSIZ;
    lineno = 0;
    freebuf = 0;

    *Phead = NULL;   /* initial setup of list */

    while (fgets(buf, BUFSIZ, fp) != NULL) {
          lineno++;
          if (*buf == '#' || blankline(buf))
               continue;

          /* 
          ** Need to see if there is a continuation. If so
          ** then we must combine the line into one line
          ** for easing parsing.
          */

          while ((cp = strchr(buf,'\134')) != NULL) {
               lineno++;
               *cp = '\0';

                if (fgets(buf2, BUFSIZ, fp) == NULL) {
                    if (debug)
                        report("Format error: %d: EOF in continuation.\n",lineno);
                    else 
                        Alert_Args_Generate(NFR_SRC_SPACEMAND,
                                NFR_SPACEMAND_EOF_IN_CONT_ERR, itoa(lineno), 0);
                    fclose(fp);
                    return(0);
                }
                if (*buf2 == '#' || blankline(buf2)) {
                    if (debug)
                        report("Format error: %d: bad continuation.\n", lineno);
                    else 
                        Alert_Args_Generate(NFR_SRC_SPACEMAND,
                                NFR_SPACEMAND_BADCONT_ERR, itoa(lineno), 0);
                    fclose(fp);
                    return(0);
                }

                /* strip whitespace from buf2 */

                if (!isspace(*buf2)) {
                    if (debug)
                       report("Format error: %d: leading whitespace missing.\n", lineno);
                    else
                        Alert_Args_Generate(NFR_SRC_SPACEMAND,
                                NFR_SPACEMAND_WSPACE_ERR, itoa(lineno), 0);

                    fclose(fp);
                    return(0);
                }

                cp = buf2;
                while (*cp && (isspace(*cp))) 
                       cp++;

                nbuflen += BUFSIZ;
                newbuf = (char *)malloc(nbuflen);
                sprintf(newbuf,"%s%s",buf, cp); /* put two lines together */
                free(buf);                /* Free the existing buf        */
                buf = newbuf;             /* Point at the expanded buffer */
                freebuf = 1;
          }
         
          /*
          **  At this point, buf is the buffer with the complete line in it 
          **  Strip the newline
          */
          if ((cp = strchr(buf,'\n')) != NULL) 
              *cp ='\0';

          if (!add_entry(Phead,buf,lineno)) {  /* add the entry to the table */
              fclose(fp);
              return(0);
          }

          if (freebuf) {   /* Don't let the buffer grow out of control */
              free(buf);   /* A bit wasteful but only done at startup  */
              freebuf = 0;
              buf = (char *)malloc(BUFSIZ);
          }

    }
    if (*buf)
        free(buf);

    fclose(fp);
    return(1);
}


void print_node(struct node *t)
{
    if (t->node_type == RKIVEINDEX) {
        report("\nBackend:\t[%s]\n",t->backend);
        if (t->recorder) 
            report("Recorder:\t[%s]\n",t->recorder);
        report("Type:\t\t[Indexed]\n");
    }
    else {
        report("\nDirectory:\t[%s]\n",t->backend);
        report("Type:\t\t[Directory]\n");
    }

    if (t->quota == -1)
        report("Quota:\t\t[%s]\n","Not Checking");
    else
        report("Quota:\t\t[%ld]\n",t->quota);

    if (t->age == -1)
        report("Age:\t\t[%s]\n","Not Checking");
    else
        report("Age:\t\t[%ld]\n",t->age);

    report("Priority:\t[%d]\n",t->priority);
    switch(t->action) {
        case NOACTION:      report("Action:\t\t[%s]\n","NOACTION");
                                break;
        case PURGE:         report("Action:\t\t[%s]\n","PURGE");
                                break;
        case SAVE:          report("Action:\t\t[%s]\n","SAVE");
                                break;
        case PROCESS:       report("Action:\t\t[%s]\n","PROCESS");
                                break;
        case PROCESSNSAVE:  report("Action:\t\t[%s]\n","PROCESS && SAVE");
                                break;
        case PROCESSNPURGE: report("Action:\t\t[%s]\n","PROCESS && PURGE");
                                break;
        default:            report("Internal ERROR: [%d]\n",t->action);
                                break;
    }
    report("Command:\t[%s]\n", t->command != NULL ? t->command : "NONE");
    report("Location:\t[%s]\n",t->location != NULL ? t->location : "NONE");

#ifndef SPACECONFIG
    report("Indexpath:\t[%s]\n",
                   t->dbinfo.indexpath != NULL ? t->dbinfo.indexpath : "NONE");
    report("DB Size:\t[%ld]\n", t->dbinfo.dbsize);
    report("DB Starttime:\t[%ld]\n", t->dbinfo.starttime);
    report("DB Endtime:\t[%ld]\n", t->dbinfo.endtime);
#endif 
    return;
}

void print_list(struct node **phead)
{
    struct node *t;
    
    if ((t = *phead) != NULL) {
        while(t != NULL) {
          print_node(t);
          t = t->next;
        }
    }
    return;
}

#ifdef SPACECONFIG

int debug = 1;
int verbose = 0;

char *Configpath = SPACEMAN_CONFIG;  /* config file to use */

extern FILE *stream;

int getopt();

extern char *optarg;
extern int optind, opterr, optopt;

int main(int argc, char **argv)
{
   int c;

   stream = stderr;
   opterr = 0;

   if (argc > 1) {
      while ((c = getopt(argc, argv, "dv")) != EOF) {
         switch (c) {
             case 'd': /* debugging stuff */
                 debug = 0;
                 break;
             case 'v': /* verbose output */
                 verbose = 1;
                 break;
            default : 
		fprintf(stderr,"usage: %s [-dv]\n", argv[0]);
                return(1);
         }
      }
   }

   if ((c = read_config())) {
       if (verbose)
           print_list(Phead);
   }
   else
       report("%s: Please correct errors and re-run.\n", argv[0]);

   free_list(Phead);

   return(c == 1 ? 0 : 1);       /* terminate this process */
}
#endif 
