/*****************************************************************************
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT")

  File:   direxp.c
  Author: Kent Landfield <kent@nfr.net>

 *****************************************************************************/
#include "sys-include.h"
#include "spaceman.h"

/* Functions:
** dbdirinfo - fill the dbinfo in a struct node
** direxpirebytime - expire directory of files by time
** direxpirebysize - expire directory of files by size
*/

extern int debug;

struct fileinfo {
    char *name;     /* filename                  */
    time_t mtime;   /* last modification time    */
    off_t size;     /* size of the file in bytes */
};

int nfiles;             /* # of elements space has been allocated for.   */
int files_index;        /* index of first unused in `files'.             */
struct fileinfo *files; /* list of files to be examined for expiration.  */

#ifndef INT_MAX
#define INT_MAX 2147483647
#endif

#if (INT_MAX <= 65535)
#define longcmp(a, b) ((a) < (b) ? -1 : (a) > (b) ? 1 : 0)
#else
#define longcmp(a, b) ((a) - (b))
#endif

int cmptimes(struct fileinfo *file1, struct fileinfo *file2)
{
  return longcmp(file1->mtime, file2->mtime);
}


void add_to_list(char *name, time_t mtime, off_t size)
{
    if (files_index == nfiles) {
        nfiles *=2;
        files = (struct fileinfo *) realloc (files, sizeof (*files) * nfiles);
    }

    files[files_index].name = strdup(name);
    files[files_index].mtime = mtime;
    files[files_index].size = size;
    files_index++;

    return;
}

/* dbdirinfo
**
**  refill the dbinfo structure with information in the directory.
*/
 
int dbdirinfo(struct node *t, char *dname)
{
    register char *nbp;
    register char *nep;
    register int i;
    struct dirent *dp;
    struct stat sbuf;
    DIR *dstream;

    if (debug)
       report("dbdirinfo(t, %s)\n", dname);

    /* add a slash to the end of the directory name */
    nbp = dname + strlen(dname);
    *nbp++ = '/';
    *nbp = '\0';

    if ((dstream = opendir(dname)) == NULL) {
        (void) fprintf(stderr, "can't open %s\n", dname);
        return(-1);
    }   

    t->dbinfo.dbsize = 0;
    t->dbinfo.starttime = 0;
    t->dbinfo.endtime = 0;

    while ((dp = readdir(dstream)) != NULL) {
        if (strcmp(dp->d_name, ".") == 0
            || strcmp(dp->d_name, "..") == 0)
            continue;
 
        for (i = 0, nep = nbp; dp->d_name[i] && i < MAXNAMLEN; i++)
            *nep++ = dp->d_name[i];
        *nep++ = '\0';

        if (stat(dname, &sbuf) == -1) {
            fprintf(stderr, "can't stat %s\n", dname);
            continue;
        }

        if ((sbuf.st_mode & S_IFMT) != S_IFREG)
            continue;

        /* Newest file so far ? */
        if (sbuf.st_mtime < t->dbinfo.starttime || t->dbinfo.starttime == 0) 
            t->dbinfo.starttime = sbuf.st_mtime;

        /* Oldest file so far ? */
        if (sbuf.st_mtime > t->dbinfo.endtime) 
            t->dbinfo.endtime = sbuf.st_mtime;
        
        t->dbinfo.dbsize += sbuf.st_size;
    }
    (void) closedir(dstream);
    *--nbp = '\0';

    if (debug) {
        report("%s: Size:\t\t[%ld]\n", dname, t->dbinfo.dbsize);
        report("%s: Starttime:\t[%ld]\n", dname, t->dbinfo.starttime);
        report("%s: Endtime:\t[%ld]\n", dname, t->dbinfo.endtime);
    }
    return(0);
}


/* direxpirebysize
**
** - expire a directory full of files by size based thresholds
** 
**  This is a bit tricky.  What we want here is to expire files by size.
**  The database has exceeded the limits so remove files until you are
**  under that limit.  The problem here is that you want to remove the
**  oldest files first.  The index method of expirebysize is easy because
**  the files are ordered for you.  That is not the case here. This can
**  get a bit expensive performance-wise.  
**
**  Read the directory and store the name, size last modification time of
**  file in a list.  Sort the list and then start expiring files by size
**  until you are under the configured threshold. 
*/
 
void direxpirebysize(char *dname, struct node *t, struct tm *now)
{
    register char *nbp;
    register char *nep;
    register int i;
    DIR *dstream;
    struct dirent *dp;
    struct stat sbuf;

    if (debug)
       report("direxpirebysize(%s, t, now)\n", dname);

    /* add a slash to the end of the directory name */
    nbp = dname + strlen(dname);
    *nbp++ = '/';
    *nbp = '\0';

    if ((dstream = opendir(dname)) == NULL) {
        (void) fprintf(stderr, "can't open %s\n", dname);
        return;
    }   

    nfiles = 100;
    files = (struct fileinfo *) malloc (sizeof (struct fileinfo) * nfiles);
    files_index = 0;

    while ((dp = readdir(dstream)) != NULL) {
        if (strcmp(dp->d_name, ".") == 0
            || strcmp(dp->d_name, "..") == 0)
            continue;
 
        for (i = 0, nep = nbp; dp->d_name[i] && i < MAXNAMLEN; i++)
            *nep++ = dp->d_name[i];
        *nep++ = '\0';

        if (stat(dname, &sbuf) == -1) {
            (void) fprintf(stderr, "can't stat %s\n", dname);
            continue;
        }

        if ((sbuf.st_mode & S_IFMT) != S_IFREG) 
            continue;

        add_to_list(dname, sbuf.st_mtime, sbuf.st_size);
    }
    (void) closedir(dstream);
    *--nbp = '\0';

    /*
    ** At this point the directory has been read and the files stored in 
    ** the list.  Time to sort the list and begin expiring the files 
    */

    if (debug)
       report("before qsort - %d \n", files_index);

    qsort(files, files_index, sizeof(struct fileinfo), cmptimes);

    if (debug)
       report("after qsort - %d \n", files_index);
    /*
    ** Gentlepeople... start your expires...
    */

    for (i = 0; i < files_index; i++) {
         expire_file(t, files[i].name, now); /* expire the file */
         if ((t->dbinfo.dbsize -= files[i].size) < t->quota) 
              break;
    }

    /*
    ** Mom says: "Cleanup your room!"...
    */

    for (i = 0; i < files_index; i++) 
      free (files[i].name);
    
    /*
    ** At this point the old files have been archived as configured. It 
    ** is now time to rescan and reset the t->dbinfo variables to what 
    ** is still on disk .
    */

    dbdirinfo(t, dname);

    return;
}

/* direxpirebytime
**
**  expire a directory full of files by time based thresholds
*/
 
void direxpirebytime(char *dname,struct node *t,time_t crnt_time,struct tm *now)
{
    register char *nbp;
    register char *nep;
    register int i;
    DIR *dstream;
    struct dirent *dp;
    struct stat sbuf;

    if (debug)
       report("direxpirebysize(%s, t, now)\n", dname);

    /* add a slash to the end of the directory name */
    nbp = dname + strlen(dname);
    *nbp++ = '/';
    *nbp = '\0';

    if ((dstream = opendir(dname)) == NULL) {
        (void) fprintf(stderr, "can't open %s\n", dname);
        return;
    }   

    while ((dp = readdir(dstream)) != NULL) {
        if (strcmp(dp->d_name, ".") == 0
            || strcmp(dp->d_name, "..") == 0)
            continue;
 
        for (i = 0, nep = nbp; dp->d_name[i] && i < MAXNAMLEN; i++)
            *nep++ = dp->d_name[i];
        *nep++ = '\0';

        /*
        ** This uses stat, instead of lstat, to follow the symlink.
        ** If the file read is a symlink, the syslink is currently left.
        ** This will result in unresolved symlinks.  Should I remove the
        ** symlinks instead or just check if a symlink and ignore any
        ** unresolved link conditions ?  The curious progammer want to know.
        */

        if (stat(dname, &sbuf) == -1) {
            (void) fprintf(stderr, "can't stat %s\n",dname);
            continue;
        }

        if ((sbuf.st_mode & S_IFMT) != S_IFREG) 
            continue;

        if ((sbuf.st_mtime + t->age) <= crnt_time)
             expire_file(t, dname, now); /* expire the file */
    }
    (void) closedir(dstream);
    *--nbp = '\0';

    /*
    ** At this point the old files have been archived as configured. It 
    ** is now time to rescan and reset the t->dbinfo variables to what 
    ** is still on disk .
    */

    dbdirinfo(t, dname);

    return;
}

