/*****************************************************************************
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT")

  File:   exp.c 
  Author: Kent Landfield <kent@nfr.net>
 
 *****************************************************************************/
#include "sys-include.h"

#include "spaceman.h"
#include "archive.h"
#include "alert.h"  
#include "nfr_res.h" 

#define SEGMENTS    10

extern int debug;

int lockfile(char *filename);
int unlockfile(char *filename);
char *itoa(int i);

int dbinfo(struct node *t, char *indexfile)
{
    FILE *ifp;
    FILE *fp;
    int num;
    struct indx_rec Irec;
    struct header Head;

    if (debug)
       report("in dbinfo(t, %s)\n", indexfile);

    /*
    ** attempt to open the index file
    */
 
    if ((ifp = fopen(indexfile,"rb")) == NULL) 
        return(-1);

    /*
    ** An index file has been found and opened.
    ** Get the start time of the index file from the first record 
    */
        
    if ((num = fscanf(ifp,"%s %s %s %ld",
        Irec.backend,Irec.ctype,Irec.path,&Irec.starttime)) != 4) {
        fclose(ifp);
        if (num == EOF) {
            /*
            ** There is nothing left in the index file. Need to
            ** reset variables to refect the current situation.
            */
            t->dbinfo.starttime = 0;
            t->dbinfo.dbsize = 0;
            return(0);
        }
        if (debug)
            report("%s: indexfile problem - inactive\n", indexfile);
        else
             Alert_Args_Generate(NFR_SRC_SPACEMAND, NFR_SPACEMAND_BADNINDEX,
                                  indexfile, 0);
        return(-1);
    }
      
    t->dbinfo.starttime = Irec.starttime;  /* Index file has starttime */
    t->dbinfo.dbsize = 0;

    /*
    ** Now compute the sizes - This would do much better if
    ** the archive sizes were stored in the index file... Sigh...
    ** This is a definite TBD!!! 
    */
    do {
         if ((fp = fopen(Irec.path,"rb")) != NULL) {
              /*
              ** Open the file, read the header and add it to the size
              */
              if (fread(&Head, sizeof(struct header), 1, fp) == 0)  {
                  if (debug)
                      report("%s: read problem - skipping\n", Irec.path);
                  else
                      Alert_Args_Generate(NFR_SRC_SPACEMAND, 
                                    NFR_SPACEMAND_INDEX_READ_ERR,Irec.path,0);
              }
              else {
                   t->dbinfo.dbsize += Head.archive_size;
                   t->dbinfo.endtime = Head.endtime;  /* last holds endtime */
              }
              fclose(fp);
         }
         else {
              if (debug)
                  report("%s missing - skipping\n",Irec.path);
              else
                  Alert_Args_Generate(NFR_SRC_SPACEMAND, 
                                     NFR_SPACEMAND_FILE_MISSING,Irec.path,0);
         }
    } while(fscanf(ifp,"%s %s %s %ld", 
                 Irec.backend,Irec.ctype,Irec.path,&Irec.starttime) != EOF);

    fclose(ifp);                      /* close the index file pointer */
    return(0);
}

int expirebytime(struct node *t, time_t crnt_time, struct tm *now)
{
     FILE *ifp;
     FILE *ofp;
     struct indx_rec Irec;
     char *toexpire[SEGMENTS];
     char newindex[BUFSIZ];
     int expcnt;
     int rc;
     int i;
     int eof_found;

     if (debug)
         report ("Time Expiration: %s/%s \n", t->backend, 
                 t->recorder == NULL ? "" : t->recorder);
     /*
     ** Check for expirations in blocks of SEGMENTS
     ** to cutdown on index re-writing.
     */

     sprintf(newindex,"%s.idx_",t->dbinfo.indexpath);

     while ((t->dbinfo.starttime > 0 && 
            (t->dbinfo.starttime + t->age) < crnt_time)) {

     if (debug)
        report("expirebytime: t->dbinfo [%ld] + t->age [%ld] == [%ld] < crnt_time [%ld]\nindexfile\n", t->dbinfo.starttime,t->age, (t->dbinfo.starttime + t->age), crnt_time, t->dbinfo.indexpath);

         /* 
         ** Examine the existing index file and see if there are
         ** records to be expired.
         */ 

         if ((ifp = fopen(t->dbinfo.indexpath,"rb")) == NULL) {
             if (debug)
                 report("%s: can not open indexfile - skiping\n", 
                   t->dbinfo.indexpath);
                else
                    Alert_Args_Generate(NFR_SRC_SPACEMAND, 
                                          NFR_SPACEMAND_INDEXFILE_OPEN,
                                            t->dbinfo.indexpath,0);
             return(-1);
         }

         /*
         ** An index file has been found and opened.
         ** Now read up to SEGMENT records.  This is done because we
         ** don't want to keep the index files locked too long. We
         ** locate the records to expire and if some are found then
         ** grab them is groups of SEGMENTS count, lock the index,
         ** regenerate the index and then unlock the index.  The 
         ** files to be expired are done so out-of-band and not with
         ** the index files locked.
         */

         expcnt = 0;
         eof_found = 0;

         while (expcnt < SEGMENTS) {
             if ((rc = fscanf(ifp,"%s %s %s %ld",
                   Irec.backend,Irec.ctype,Irec.path,&Irec.starttime)) != 4) {
                 if (rc != EOF) {
                     if (debug)
                         report("%s: index format problem\n",t->dbinfo.indexpath);
                     else
                         Alert_Args_Generate(NFR_SRC_SPACEMAND, 
                                          NFR_SPACEMAND_INDEX_FORMAT,
                                            t->dbinfo.indexpath,0);
                     fclose(ifp);
                     return(-1);
                 }
                 eof_found++;
                 break;
             }

             /* Is this a record to expire ? */

             if ((Irec.starttime + t->age) <= crnt_time) 
                  toexpire[expcnt++] = strdup(Irec.path);
             else {
                  /*
                  ** We are fine now. Time to quit the gathering 
                  ** of records to expire
                  */
                  break;
             }
         }

         /*
         ** Check and see if we have found records to expire 
         ** If so, we have work to do.
         */

         if (expcnt > 0) {
             /*
             ** Lock the index file and  open the new one 
             */
  
             lockfile(t->dbinfo.indexpath);
             if ((ofp = fopen(newindex,"wb")) == NULL) {
                  if (debug)
                      report("%s: can not open new indexfile - skipping\n", 
                               t->dbinfo.indexpath);
                  else
                      Alert_Args_Generate(NFR_SRC_SPACEMAND, 
                                          NFR_SPACEMAND_NEWINDEX,
                                            t->dbinfo.indexpath,0);
                 fclose(ifp);
                 unlockfile(t->dbinfo.indexpath);
                 return(-1);
             }

             /* 
             ** Should the last record read be expired or re-written
             ** back into the index file ? Check to make sure there
             ** is something there and we are not here due to eof.
             */

             if (!eof_found && ((Irec.starttime + t->age) > crnt_time)) {
                 /* Put the non-matching record into the new indexfile */
                 fprintf(ofp,"%s %s %s %ld\n", Irec.backend, Irec.ctype,
                         Irec.path,Irec.starttime);
             }

             /* rewrite the new indexfile */

             while (fscanf(ifp,"%s %s %s %ld", Irec.backend,Irec.ctype,
                           Irec.path,&Irec.starttime) != EOF) {
                    fprintf(ofp,"%s %s %s %ld\n", Irec.backend, Irec.ctype,
                           Irec.path,Irec.starttime);
             }

             /*
             ** Now to put the new index file back in place and unlock
             ** the file for backends to start writing to again.
             */

             fclose(ifp);
             fclose(ofp);

             if (rename(newindex,t->dbinfo.indexpath) != 0) {
                 if (debug)
                     report("rename failed for %s to %s - errno %d\n", 
                               newindex,t->dbinfo.indexpath, errno);
                 else
                      Alert_Args_Generate(NFR_SRC_SPACEMAND, 
                                          NFR_SPACEMAND_INDEXRENAME,
                                          newindex,t->dbinfo.indexpath, 
                                            itoa(errno), 0);

                 unlockfile(t->dbinfo.indexpath);
                 return(-1);
             }

             unlockfile(t->dbinfo.indexpath);

             /* Time to EXPIRE the actual files */

             for (i = 0; i < expcnt; i++) {
/*
NEED TO PUT BACK INDEX FILES THAT
ARE ORPHANED DUE TO ERRORS!!!
*/
                  expire_file(t, toexpire[i], now); /* expire the file   */
                  free(toexpire[i]);                /* cleanup your room */
             }

             /*
             ** At this point the new index file has been created, the
             ** old files archived as configured. It is now time to rescan
             ** the index and reset the t->dbinfo variables to what is still
             ** on disk .
             */

             if (dbinfo(t, t->dbinfo.indexpath) != 0) {    /* database info */
                 if (debug)
                     report("reset dbinfo failed for %s \n", t->dbinfo.indexpath); 
                 else
                      Alert_Args_Generate(NFR_SRC_SPACEMAND, 
                                          NFR_SPACEMAND_RESETDB,
                                          t->dbinfo.indexpath, 0);
             }
         }
         else /* nothing to expire */
             break;
     }
     return(0);
}

int expirebysize(struct node *t, struct tm *now)
{
     FILE *ifp;
     FILE *ofp;
     char *toexpire[SEGMENTS];
     char newindex[BUFSIZ];
     int expcnt;
     int rc;
     int i;
     int eof_found;
     long quota;
     struct indx_rec Irec;
     struct stat stbuf;

     if (debug)
         report ("Size Expiration: %s/%s \n", t->backend, 
                 t->recorder == NULL ? "" : t->recorder);
     /*
     ** Check for expirations in blocks of SEGMENTS
     ** to cutdown on index re-writing.
     */

     sprintf(newindex,"%s.idx_",t->dbinfo.indexpath);

     while (t->dbinfo.dbsize > 0 && t->dbinfo.dbsize > t->quota) {

         if (debug)
             report("expirebysize: t->dbinfo.dbsize [%ld] > t->quota [%ld]\nIndexpath: %s\n", t->dbinfo.dbsize,t->quota,t->dbinfo.indexpath);

         /* 
         ** Examine the existing index file and see if there are
         ** records to be expired.
         */ 

         if ((ifp = fopen(t->dbinfo.indexpath,"rb")) == NULL) {
             if (debug)
                 report("%s: can not open indexfile - skiping\n",
                   t->dbinfo.indexpath);
                else
                    Alert_Args_Generate(NFR_SRC_SPACEMAND,
                                          NFR_SPACEMAND_INDEXFILE_OPEN,
                                            t->dbinfo.indexpath,0);
             return(-1);
         }

         /*
         ** An index file has been found and opened.
         ** Now read up to SEGMENT records.  This is done because we
         ** don't want to keep the index files locked too long. We
         ** locate the records to expire and if some are found then
         ** grab them is groups of SEGMENTS count, lock the index,
         ** regenerate the index and then unlock the index.  The 
         ** files to be expired are done so out-of-band and not with
         ** the index files locked.
         */

         expcnt = 0;
         eof_found = 0;
         quota = t->dbinfo.dbsize;

         while (expcnt < SEGMENTS) {
             if ((rc = fscanf(ifp,"%s %s %s %ld",
                   Irec.backend,Irec.ctype,Irec.path,&Irec.starttime)) != 4) {
                 if (rc != EOF) {
                     if (debug)
                         report("%s: index format problem\n",t->dbinfo.indexpath);
                     else
                         Alert_Args_Generate(NFR_SRC_SPACEMAND,
                                          NFR_SPACEMAND_INDEX_FORMAT,
                                            t->dbinfo.indexpath,0);
                     fclose(ifp);
                     return(-1);
                 }
                 eof_found++;
                 break;
             }

             /*
             ** Add the record found to the list ot be removed
             ** then check to see if that meets the criteria
             ** of when to terminate the expire.  Gotta put the
             ** sizes in the index files....
             */

             toexpire[expcnt++] = strdup(Irec.path);

             if (stat(Irec.path, &stbuf) == 0) {
                 quota -= stbuf.st_size;
                 if (quota < t->quota) {
                     /*
                     ** We are fine now. Time to quit the gathering 
                     ** of records to expire
                     */
                     break;
                 }
             }
             /* if stat fails then add to the list and it will 
             ** get removed from the index file.
             */
         }

         /*
         ** Check and see if we have found records to expire 
         ** If so, we have work to do.
         */

         if (expcnt > 0) {
             /*
             ** Lock the index file and  open the new one 
             */
  
             lockfile(t->dbinfo.indexpath);
             if ((ofp = fopen(newindex,"wb")) == NULL) {
                  if (debug)
                      report("%s: can not open new indexfile - skipping\n", 
                               t->dbinfo.indexpath);
                  else
                      Alert_Args_Generate(NFR_SRC_SPACEMAND, 
                                          NFR_SPACEMAND_NEWINDEX,
                                            t->dbinfo.indexpath,0);
                 fclose(ifp);
                 unlockfile(t->dbinfo.indexpath);
                 return(-1);
             }

             /* rewrite the new indexfile */

             while (fscanf(ifp,"%s %s %s %ld", Irec.backend,Irec.ctype,
                           Irec.path,&Irec.starttime) != EOF) {
                    fprintf(ofp,"%s %s %s %ld\n", Irec.backend, Irec.ctype,
                           Irec.path,Irec.starttime);
             }

             /*
             ** Now to put the new index file back in place and unlock
             ** the file for backends to start writing to again.
             */

             fclose(ifp);
             fclose(ofp);

             if (rename(newindex,t->dbinfo.indexpath) != 0) {
                 if (debug)
                     report("rename failed for %s to %s - errno %d\n",
                               newindex,t->dbinfo.indexpath, errno);
                 else
                      Alert_Args_Generate(NFR_SRC_SPACEMAND,
                                          NFR_SPACEMAND_INDEXRENAME,
                                          newindex,t->dbinfo.indexpath,
                                            itoa(errno), 0);
                 unlockfile(t->dbinfo.indexpath);
                 return(-1);
             }

             unlockfile(t->dbinfo.indexpath);

             /* Time to EXPIRE the actual files */

             for (i = 0; i < expcnt; i++) {
                  expire_file(t, toexpire[i], now); /* expire the file   */
                  free(toexpire[i]);                /* cleanup your room */
             }

             /*
             ** At this point the new index file has been created, the
             ** old files archived as configured. It is now time to rescan
             ** the index and reset the t->dbinfo variables to what is still
             ** on disk .
             */

             if (dbinfo(t, t->dbinfo.indexpath) != 0) {    /* database info */
                 report("reset dbinfo failed for %s \n", t->dbinfo.indexpath); 
             }
         }
         else /* nothing to expire */
             break;
     }
     return(0);
}
