/*****************************************************************************
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT")

  File:   runlist.c 
  Author: Kent Landfield <kent@nfr.net>
 
 *****************************************************************************/
#include "sys-include.h"

#include "spaceman.h"
#include "archive.h" 
#include "alert.h" 
#include "nfr_res.h" 

#define SEGMENTS    10
#define TIME_EXPIRE  1
#define SIZE_EXPIRE  2

extern int debug;
extern struct node *Phead[]; 
extern unsigned int alarmtime;

char *nfrHome( char *b );
void print_node(struct node *t);

int rmfile(char *dbname, char *filename);
int copyfile(char *src, char *dest);
int process_archive(char *filename, char *command);
char *xpand(char *fptr, struct node *t, char *fname, struct tm *now);

/* Directory expiration */
int dbdirinfo(struct node *t, char *dname);
void direxpirebysize(char *dname, struct node *t, struct tm *now);
void direxpirebytime(char *dname,struct node *t,time_t crnt_time,struct tm *now);

/* Indexed expiration */
int dbinfo(struct node *t, char *indexfile);
int expirebytime(struct node *t, time_t crnt_time, struct tm *now);
int expirebysize(struct node *t, struct tm *now);

time_t lowtime;

void setup_dbvals(void) 
{
    char path[MAXPATHLEN];
    struct node *t;

    if (debug)
        report("setup_dbvals %s\n","beginning");

    /*
    ** Configuration list is built.
    ** Now fill in the indexpath for the appropriate entries.
    */

    t = *Phead;
    while(t != NULL) {

        if ((t->action != NOACTION) && (t->node_type == RKIVEINDEX)) {
            /* Build path to index file */
            sprintf(path,"%s/%s/%s/index",
                    nfrHome(INDEXDIR),t->backend,t->recorder);
            t->dbinfo.indexpath = strdup(path); /* Save path to indexfile  */
        }
        t = t->next;                     /* go for the next record  */
    }
}

int expire_file(struct node *t, char *filename, struct tm *now)
{
    int rc;
    char *dbname;
    char *dest; 
    char *command; 

    if (debug)
       report("in expire_file(t, %s)\n", filename);

    /* 
    ** Create the complete database name.
    ** Save it and free later.
    */

    if (t->recorder != NULL) {
        dbname = (char *) malloc(strlen(t->backend)+strlen(t->recorder)+2);
        sprintf(dbname,"%s/%s",t->backend,t->recorder);
    }
    else 
        dbname = strdup(t->backend);

    rc = -1;

    switch(t->action) {
        case NOACTION:   /* Shouldn't get here but... */
                /* Nothing is suppose to happen for this entry */
                if (debug)
                    report ("%s - no action for %s\n", dbname,filename);
                else
                  Alert_Args_Generate(NFR_SRC_SPACEMAND, 
                                          NFR_SPACEMAND_NOACTION,dbname,filename,0);
                break;   
        case PURGE:
                /*
                ** Simply remove the file - Not much use but it may
                ** have been created as a runtime file who contains
                ** data of little long term use.
                */
                rc = rmfile(dbname,filename);
                break;   
        case SAVE:
                /*
                ** Save the file to the specified location.  When the
                ** save completes then purge the file from its archive
                ** location.
                */
                dest = xpand(t->location, t, filename, now);
                if ((rc = copyfile(filename, dest)) == 0)
                     rc = rmfile(dbname,filename);
                break;   
        case PROCESS:
                /*
                ** Send the data to an command external to the 
                ** space management system to process the data.
                ** It is up to the external command to remove the 
                ** archived file.  
                */
                command = xpand(t->command, t, filename, now);
                rc = process_archive(filename, command);
                break;   
        case PROCESSNSAVE:
                /*
                ** Send the data to an command external to the 
                ** space management system to process the data.
                ** Upon successful completion of the command, the
                ** archived file is saved to the location and 
                ** then purged.
                */
                command = xpand(t->command, t, filename, now);
                if ((rc = process_archive(filename, command)) == 0) {
                     dest = xpand(t->location, t, filename, now);
                     if ((rc = copyfile(filename, dest)) == 0)
                          rc = rmfile(dbname,filename);
                }
                break;   
        case PROCESSNPURGE:
                /*
                ** Send the data to an command external to the 
                ** space management system to process the data.
                ** Upon successful completion of the command, the
                ** archived file is removed from the archive.
                */
                command = xpand(t->command, t, filename, now);
                if ((rc = process_archive(filename, command)) == 0) 
                     rc = rmfile(dbname,filename);
                break;   
      default:
                if (debug)
                    report ("%s - invalid action specified %d\n", dbname, t->action);
                else
                    Alert_Args_Generate(NFR_SRC_SPACEMAND, 
                                          NFR_SPACEMAND_BAD_ACTION,
                                            dbname, itoa(t->action),0);
                break;   
    }
    free(dbname);
    return(rc);
}

time_t runnode(struct node *t,struct tm *now,time_t crnt_time)
{
    time_t dbtime;

    if (debug)
        report("In runnode loop: %s\n",t->backend);

    if (t->node_type == RKIVEINDEX) {
        /*
        ** If there is no index path specified then we need to check from
        ** scratch again.  It may be that they put the info in the spaceman 
        ** config before they start up nfrd and the associated recorders.  
        ** If so we need to check if the index file as appeared since the 
        ** last runlist.
        */

        if (dbinfo(t, t->dbinfo.indexpath) != 0) 
            return(0); /* Nothing here to process... Skip it. */
    }
    else {
        /*
        ** Need to check if the directory specified exists. It may not.
        ** Do not consider this an error. The directory may be created
        ** later in the course of system execution.
        */ 
        if (dbdirinfo(t, t->backend) != 0) 
            return(0); /* Nothing here to process... Skip it. */
    }

    if (debug)
        print_node(t);

    /*
    ** First check to see if the archive size has been exceeded 
    */

    if ((t->quota != -1) && (t->dbinfo.dbsize > 0)) {
        if (debug)
            report("Size check: t->quota (%ld) < t->dbinfo.dbsize (%ld)\n",
                   t->quota,t->dbinfo.dbsize);

        if (t->dbinfo.dbsize > t->quota) {
            if (t->node_type == RKIVEINDEX) 
                expirebysize(t, now);
            else
                direxpirebysize(t->backend, t, now);
        }
    }
    
    /*
    ** Next check to see if the age of archive members has been exceeded 
    */

    if ((t->age != -1) && (t->dbinfo.starttime > 0)) {
        /*
        ** Now check to see if the archive times have been exceeded 
        */
        dbtime = t->dbinfo.starttime + t->age;

        if (debug) {
            report("Time Check: t->dbinfo.starttime (%ld) + t->age (%ld)\n",
               t->dbinfo.starttime,t->age);
            report("dbtime (%ld) < crnt_time(%ld)\n", dbtime, crnt_time);
            report("lowtime (%ld)\n",lowtime);
        }

        if (dbtime < crnt_time) {
            if (t->node_type == RKIVEINDEX) 
                expirebytime(t, crnt_time, now);
            else
                direxpirebytime(t->backend,t,crnt_time,now);
        }
    
        else if (dbtime > crnt_time && dbtime < lowtime) 
           lowtime = dbtime; /* calculate the alarm time */
    }
    return(lowtime);
}

void Runlist(void)
{
    struct node *t;
    struct tm *now;
    time_t crnt_time;

    if (debug)
        report("Runlist %s\n","Running");

    /* 
    ** Run through the list and see if there are files that
    ** need to be expired.  If so do so.
    ** This is not a forced expire due to overall space. This
    ** is a standard expire.  In many cases the standard expire
    ** will push things below the overall level.
    */

    crnt_time = time((time_t *)0);  /* get current runtime           */ 
    now = localtime(&crnt_time);    /* Convert to a tm structure     */
    lowtime = 0;

    t = *Phead;
    while(t != NULL) {
        if (t->action != NOACTION) 
            runnode(t,now,crnt_time);
        t = t->next;
    }

    /*
    ** Now need to see if the overall space used exceeds the system-wide
    ** limits.  If it does then we go into a forced expiration mode.
    **
    ** There are two different types of forced expiration, 
    **   - Spiral through all levels from priority 10 to priority 1
    **     removing a single file from each.
    **   - Remove all files (except the last two archived files) in 
    **     each priority level from 10 to 1 until below the system 
    **     threshold.
    ** In both cases stop forced expiration once below the system threshold.
    ** 
    ** TBD!!!
    */
    return;
}
