/***************************************************************************** 
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT") 

  File:   sstart.c 
  Author: Kent Landfield <kent@nfr.net>
 
 *****************************************************************************/

#include "sys-include.h"

#include "spaceman.h"
#include "alert.h"
#include "nfr_res.h"

extern char *Fifopath;
extern int FifoFdW;
extern int FifoFdR;
extern PDaemon Daemon_This;
extern int debug;

extern struct node *Phead[]; 

void free_list(struct node **phead);
void print_list(struct node **phead);
void setup_dbvals(void);
void Runlist(void);
int read_config(void);
int kill(pid_t pid, int sig);

int smack_spacemand(void)
{
    int pid;
    FILE *fp;
    struct stat stbuf;

    if (stat(SPACEMAN_PIDFILE,&stbuf) == 0) {
        /* 
        ** need to open the file, get the pid and check to see if it
        ** is running.  
        */ 
        if ((fp = fopen(SPACEMAN_PIDFILE,"r")) == NULL) {
            fprintf(stderr,"Can not open %s\n",SPACEMAN_PIDFILE);
            return(-1);
        }  
        if (fscanf(fp,"%d", &pid) != 1) {
            fprintf(stderr,"%s read error\n", SPACEMAN_PIDFILE);
            fclose(fp);  /* Done with it so close it */
            return(-1);
        }  
        fclose(fp);  /* Done with it so close it */

        if (kill(pid,SIGHUP) != 0) {
             if (errno == ESRCH) {
                 fprintf(stderr,"Spacemand not running.\n");
                 return(0);
             }
             else {
                 perror("spacemand: ");
                 return(-1);
             }
        }  
        fprintf(stderr,"Spacemand notified.\n");
        return(0);
    }
    fprintf(stderr,"Spacemand not running.\n");
    return(0);
}

void Rereadconfig(int sig)
{
   if (debug)
       report("Rereadconfig %d\n", sig);

    free_list(Phead); /* free the list */

    if (!read_config()) { /* read it again */
        if (debug)
            report("%s: config read failed\n", "PANIC");
        else
            Alert_Generate(NFR_SRC_SPACEMAND, NFR_SPACEMAND_REREAD_ERR);
        exit(-1);
    }
    setup_dbvals();

    if (debug) {
        report("%s: succeeded\n", "Configuration reinitialization");
        print_list(Phead);
    }
    else
        Alert_Generate(NFR_SRC_SPACEMAND, NFR_SPACEMAND_REINIT);

    print_list(Phead);
}

int process_running(int pid)
{
   /* This will need to be adopted for MS environments */

   if (debug)
       report("process_running %d\n",pid);

   if (kill(pid,0) == 0)
       return(1);

   return(0);
}

/* 
** Startup:
**        1. check to see if this is the only spacemand process running.
**           if not, die
**           if so, create file with pid in it.
**        2. create input fifo
**        3. Setup signal handlers - timer alarms, reread config files
**        4. read configuration file and setup memory structures.
**        5. Run through list looking for expired items
*/

int spaceman_start()
{
  int pid;
  FILE *fp;
  struct stat stbuf;

  if (debug)
      report("spaceman_start %s\n","running");

  if (stat(SPACEMAN_PIDFILE,&stbuf) == 0) {
      /*
      ** need to open the file, get the pid and check to see if it
      ** is running.  If so, then I need to go away...
      */
      if ((fp = fopen(SPACEMAN_PIDFILE,"r")) == NULL) {
          if (debug)
              report("Pidfile %s open read failed. Terminating\n",SPACEMAN_PIDFILE);
          else
              Alert_Args_Generate(NFR_SRC_SPACEMAND, 
                                   NFR_SPACEMAND_PIDFILE_OPEN, SPACEMAN_PIDFILE, 0);
          return(NFR_SPACE_CANTOPENFILE);
      }
      if (fscanf(fp,"%d", &pid) != 1) {
          if (debug)
              report("%s pidfile format error. Terminating\n", SPACEMAN_PIDFILE);
          else
              Alert_Args_Generate(NFR_SRC_SPACEMAND, 
                                   NFR_SPACEMAND_PIDFILE_FORMAT, SPACEMAN_PIDFILE, 0);
          fclose(fp);  /* Done with it so close it */
          return(NFR_SPACE_PIDREAD);
      }

      fclose(fp);  /* Done with it so close it */

      if (process_running(pid)) {
          if (debug)
              report("Spacemand %d already running. Terminating\n",pid);
          else
              Alert_Args_Generate(NFR_SRC_SPACEMAND, 
                                   NFR_SPACEMAND_RUNNING, itoa(pid), 0);
          return(NFR_SPACE_RUNNING);
      }
  }

  /* Create the new PID file */

  if ((fp = fopen(SPACEMAN_PIDFILE,"w")) == NULL) {
       if (debug)
          report("Pidfile %s open write failed. Terminating\n", SPACEMAN_PIDFILE);
       else
          Alert_Args_Generate(NFR_SRC_SPACEMAND, 
                               NFR_SPACEMAND_PIDFILE_WRITE, SPACEMAN_PIDFILE, 0);
       return(NFR_SPACE_CANTCREATPID);
  }

  /* Put the current process info in it. */

  pid = getpid();
  fprintf(fp,"%d",pid);
  fclose(fp);

  /* Remove the existing fifo and create anew. */

  remove(Fifopath);

  /* Create the input fifo  */

  if (mkfifo(Fifopath, 0666) == -1) {
      if (debug)
          report("Fifo %s creation failed. errno == %d. Terminating\n", Fifopath, errno);
      else
          Alert_Args_Generate(NFR_SRC_SPACEMAND, 
                               NFR_SPACEMAND_MKFIFO_ERR, Fifopath, itoa(errno), 0);
      (void) unlink(SPACEMAN_PIDFILE);
      return(NFR_SPACE_FIFOCREAT);
  }

  /* Open the input fifo so it blocks waiting for data
  **
  ** OK... This takes a bit of discussion...  I want to open up a
  ** FIFO that will be forever open as long as this process is running.
  ** Due to the semantics of FIFOs and their use with reads and writes
  ** it is necessary to play games. Normally a server opening a fifo 
  ** readonly wants to block on the fifo waiting for client messages.  
  ** A client normally sends a msg to a server process by opening the 
  ** fifo for the duration of the send and then closing it again. When
  ** there is no writer process, the read in the server returns a 0 (EOF)
  ** if no other clients have the fifo open write.  The server process
  ** is then forced to reopen the fifo readonly and wait for a client
  ** to open it for write.  This is not acceptable.  One way is to avoid 
  ** this is to open the fifo twice in the server process, once for reading 
  ** and once for writing.  That way there is always a reader and a writer.
  ** active.
  **
  ** Now you'll notice that I opened it read, then open it write and
  ** then I close the read side and reopen it.  The key here is the
  ** O_NDELAY flag.  Initially it is set for non-blocking so the open 
  ** returns without an error.  But I need it blocking so once the write
  ** open succeeds, I reopen it without the O_NDELAY so that reads on the
  ** input fifo block waiting for data on the fifo.  Are we confused yet ?
  */ 

  /* Open the input fifos */

  if ((FifoFdR = open(Fifopath, O_RDONLY | O_NDELAY)) == -1) {
       if (debug)
           report("Fifo %s open read failed. Terminating\n", Fifopath);
       else
           Alert_Args_Generate(NFR_SRC_SPACEMAND, 
                               NFR_SPACEMAND_FIFO_OPENR, Fifopath, itoa(errno), 0);
       (void) unlink(Fifopath);
       (void) unlink(SPACEMAN_PIDFILE);
       return(NFR_SPACE_FIFOOPEN);
  }
   
  if ((FifoFdW = open(Fifopath, O_WRONLY | O_NDELAY)) == -1) {
       if (debug)
           report("Fifo %s open write failed. errno == %d Terminating\n", Fifopath, errno);
       else
           Alert_Args_Generate(NFR_SRC_SPACEMAND, 
                               NFR_SPACEMAND_FIFO_OPENW, Fifopath, itoa(errno), 0);
       (void) close(FifoFdR);
       (void) unlink(Fifopath);
       (void) unlink(SPACEMAN_PIDFILE);
       return(NFR_SPACE_FIFOOPEN);
  }

  (void) close(FifoFdR);

  if ((FifoFdR = open(Fifopath, O_RDONLY)) == -1) {
       if (debug)
           report("Fifo %s reopen read failed. Terminating\n", Fifopath);
       else
           Alert_Args_Generate(NFR_SRC_SPACEMAND, 
                               NFR_SPACEMAND_FIFO_REOPENR, Fifopath, itoa(errno), 0);
       (void) close(FifoFdW);
       (void) unlink(Fifopath);
       (void) unlink(SPACEMAN_PIDFILE);
       return(NFR_SPACE_FIFOOPEN);
  }
   
  /* 
  ** read configuration file and setup memory structures.
  */

  if (!read_config()) {
      (void) close(FifoFdR);
      (void) close(FifoFdW);
      (void) unlink(Fifopath);
      (void) unlink(SPACEMAN_PIDFILE);
      return(NFR_SPACE_CONFIGBAD);
  }
 
  setup_dbvals();

  /* 
  ** Run through list looking for expired items
  */
  Runlist();

  if (debug)
      report("spacemand: %d running\n",pid);
  else
      Alert_Args_Generate(NFR_SRC_SPACEMAND,NFR_SPACEMAND_STARTED,itoa(pid),0);
   
  return 0;
}
