/***************************************************************************** 
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT") 

  File:   xpand.c 
  Author: Kent Landfield <kent@nfr.net>
 
 *****************************************************************************/  

#include "sys-include.h"

#include "spaceman.h"
#include "home.h"

char *days[] = { "Sunday", "Monday", "Tuesday", "Wednesday", 
                 "Thursday", "Friday", "Saturday" };
                   
char *months[] = { "January", "February", "March", "April", "May", 
                   "June", "July", "August", "September", "October",
                   "November", "December" };
/*
*****************************************************************
**
** add_char - add supplied character to output string
**
** Returns:    pointer to next position to add to.
**
*****************************************************************
*/

static char *add_char(char *ptr, char member_char)
{
    *ptr++ = member_char;
    return(ptr);
}

/*
*****************************************************************
**
** add_string - add supplied string to output string
**
** Returns:    pointer to next position to add to.
**
*****************************************************************
*/

static char *add_string(char *ptr, char *member_str)
{
    while(*member_str)
        *ptr++ = *member_str++;
    return(ptr);
}

/*
*****************************************************************
**
** xpand - fill in values substituting for magic cookies
**
**      Substitution cookies supported
** LOCATION:         This field is used if the data is to be copied to another
**                   disk location. This can be an NFS mounted location. It is 
**                   up to the administrator to assure that appropriate space 
**                   is available and mounted at LOCATION. LOCATION fields
**                   can use substitution characters as listed below.
**
**                       %b - name of backend the file was archived with
**                       %d - two digit day of month (1-28/30/31)
**                       %F - path to file being expired 
**                       %f - basename of file being expired 
**                       %m - two digit month of year (1-12)
**                       %r - name of recorder the file was archived with
**                       %y - four digit year (1990,..2001)
**                       %D - Day of the week (Sunday, Monday, ..)
**                       %M - Month of year (January, February, ..)
**
**  LOCATION: ~/data/%p/%r/%y/%m/%d/
**
**             /cdrom/nfrdata/1997-08/webwatcher/urlhist/28/
**
** Parameters:
**             fptr  - format string to fill 
**             t     - struct node *
**             fname - name of file being expired
**             now   - filled tm structure
**
** Returns:   expanded path string
**
*****************************************************************
*/

char *xpand(char *fptr, struct node *t, char *fname ,struct tm *now)
{
    static char format[BUFSIZ];
    register char *aptr;
    register char *cp;
    char dtstr[32];
    char c;
    char *rp;

    for (cp = format; cp < format + sizeof(format); cp++)
         *cp = '\0';
 
    aptr = fptr;
    cp = format;

    while ((c = *aptr++)) {
        if (c == '%') {
           switch (*aptr++) {
               case '%':  /* Add the % character */
                   cp = add_char(cp,'%');
                   continue;
               case '~':  /* Convert to NFR Home Directory */
                   cp = add_string(cp, nfrHome("~"));
                   continue;
               case 'b':     /* backend name  */
                   if (*t->backend)
                       cp = add_string(cp, t->backend);
                   continue;
               case 'd':     /* two digit day of month (1-31) */
                   sprintf(dtstr,"%.2d",now->tm_mday);
                   cp = add_string(cp, dtstr);
                   continue;
               case 'F':     /* Full path to file name passed in  */
                   if (*fname)
                       cp = add_string(cp, fname);
                   continue;
               case 'f':     /* file name passed in  */
                   if (*fname) {
                       if ((rp = strrchr(fname,'/')) != NULL)
                            rp++;
                       cp = add_string(cp, rp);
                   }
                   continue;
               case 'm':     /* two digit month of year (1-12) */
                   sprintf(dtstr,"%.2d",now->tm_mon+1);
                   cp = add_string(cp, dtstr);
                   continue;
               case 'r':     /* recorder name  */
                   if (*t->recorder)
                       cp = add_string(cp, t->recorder);
                   continue;
               case 'y':     /* 4 digit year */
                   sprintf(dtstr,"%.4d",now->tm_year+1900);
                   cp = add_string(cp, dtstr);
                   continue;
               case 'D':     /* day of the week */
                   cp = add_string(cp, days[now->tm_wday]);
                   continue;
               case 'M':     /* Month of year */
                   cp = add_string(cp, months[now->tm_mon]);
                   continue;
               default:
                   return(NULL);
           }    /* end switch */
        }
        *cp++ = c;
     }  /* end while */
     return(format);
}

/*
*****************************************************************
**
** valid_format - determine the admin has specified a valid format string.
**
** Substitution cookies supported described in xpand() above.
**
** WARNING: KEEP THESE TWO FUNCTIONS IN SYNC FORMAT CHARACTER-WISE!!!
**
**
** Parameters:
**             fptr  - format string 
**             t     - struct node *
**
** Returns:   0 - not a valid format
**            1 - valid format
**
*****************************************************************
*/
int valid_format(char *fptr)
{
    register char *aptr;
    char c;

    aptr = fptr;

    while ((c = *aptr++)) {
        if (c == '%') {
           switch (*aptr++) {
               case '%':  /* Add the % character */
                   continue;
               case '~':  /* Convert to NFR Home Directory */
                   continue;
               case 'b':     /* backend name  */
                   continue;
               case 'd':     /* two digit day of month (1-31) */
                   continue;
               case 'F':     /* full path to the file name passed in  */
                   continue;
               case 'f':     /* basename of path to file name passed in  */
                   continue;
               case 'm':     /* two digit month of year (1-12) */
                   continue;
               case 'r':     /* recorder name  */
                   continue;
               case 'y':     /* 4 digit year */
                   continue;
               case 'D':     /* day of the week */
                   continue;
               case 'M':     /* Month of year */
                   continue;
               default:
                   return(0);
           }    /* end switch */
        }
     }  /* end while */
     return(1);
}
