/*
        Copyright(C) 1997 Network Flight Recorder, Inc.
        All rights reserved.

        Use and distribution of this software and its source code
        are governed by the terms and conditions of the
        Network Flight Recorder Software License ("LICENSE.TXT")

*/

package graph;

import java.io.*;
import java.awt.*;

// BGCanvas: a canvas holding a bar graph; the bargraph needs to
// be redrawn every time we repaint, so we use this canvas to trap
// redraw events

class BGCanvas extends Canvas {

  private Image i;
  private BarGraph bg;

  BGCanvas(BarGraph bg) {
    super();
    this.bg = bg;
  }

  public Dimension minimumSize() {
    // make the size proportional to at least 5/7 the enclosing frame
    // so we have some space for titles, labels, etc.
    Container parent = getParent();
    Dimension size = parent.size();

    return new Dimension(75, 75);
  }

  public Dimension preferredSize() {
    // we like the size proportional to 3/4 the enclosing frame
    // so we have some space for titles, labels, etc.
    Container parent = getParent();
    Dimension size = parent.size();

    return new Dimension(100, 100);
  }

  public void update(Graphics g) {
    // since we're updating the whole bitmap from our stored image, 
    // don't clear the background
    paint(g);
  }


  // paint: regenerate the bar graph if data has been updated,
  // if the size of the bitmap has changed, or if we don't have
  // a copy of the bitmap lying around
  
  // save the image you make to obviate the need to regenerate it later

  public void paint(Graphics g) {

    // check to see if we need to write new data in our image
    boolean regen = bg.updated();

    int w = size().width;
    int h = size().height;

    if (i == null || 
	(i.getWidth(null) != w || 
	 i.getHeight(null) != h)) {

      // the size has changed or we didn't have a chached image at all

      if (i != null)
	i.flush();

      // create a new cached image
      i = createImage(w, h);

      // force the label fonts and axis images to be recalculated
      bg.y_axis_image = null;

      // force the image to be regenerated
      regen = true;
    }

    if (regen) {

      // size has changed, or contents... regenerate our
      // cached bitmap
      Graphics ig = i.getGraphics();

      ig.clearRect(0, 0, w, h);
      bg.display(ig, this);

      // explicitly dispose in case bitmaps are scarce resources
      ig.dispose();

    }

    g.drawImage(i, 0, 0, null);
  }

  public void reshape (int x, int y, int w, int h )
	{
	super.reshape(x,y,w,h);
	System.out.println("BG CANVAS RESHAPE "+x+" "+y+" "+w+" "+h);
	}
}
