/*
        Copyright(C) 1997 Network Flight Recorder, Inc.
        All rights reserved.

        Use and distribution of this software and its source code
        are governed by the terms and conditions of the
        Network Flight Recorder Software License ("LICENSE.TXT")

*/

package graph;



import java.io.*;
import java.util.*;
import java.awt.*;

//  BGDataList:  encapsulation of the bar graph list that
//  maintains its high, low, count, start, and end for the convenience
//  of the bar graph client.  the constructor reads in the next chunk
//  of a file.  

//  two lists can be merged, so the client that's reading the
//  data in chunk-wise can display intermediate graphs

public class BGDataList {

  private BGData start, end;

  private long min_y;
  private long max_y;
  private long data_total;
  private int n_data;

  private static boolean lock = false;

  // initialize a list from the contents of a file
  public BGDataList(EOFInputStream file, int max_data) 
       throws java.io.EOFException, java.io.IOException
  {
    min_y = Long.MAX_VALUE;
    max_y = Long.MIN_VALUE;
    n_data = 0;
    data_total = 0;

    do {

      try {
	String label = file.nextLine();	// eof is valid here

	long value;
	Color color;

	try {
	  value = Long.parseLong(file.nextLine());	// data value
	  color = parseColor(file.nextLine());		// data color
	} catch (NumberFormatException e) {
	  // parse error
	  throw new java.io.IOException();
	} catch (java.io.EOFException e) {
	  throw new java.io.IOException(); // must be a number here 
	}

	// add this data to our list
	BGData cur = new BGData(label, value, color);
	if (start == null)
	  start = end = cur;
	else
	  {
	    cur.setPrev(end);
	    end.setNext(cur);
	    end = cur;
	  }

	// keep a running sum and note max and min values
	data_total += value;

	if (value < min_y) {
	  min_y = value;
	}

	if (value > max_y) {
	  max_y = value;
	}

	++n_data;

	if (n_data == max_data)
	  break;

      } catch (java.io.EOFException e) {
	if (n_data == 0)
	  throw e;
	break;
      }

    } while (true);
  }

  private Color parseColor(String color) {
    if (color.equals("blue"))
      return Color.blue;
    else if (color.equals("red"))
      return Color.red;
    else if (color.equals("yellow"))
      return Color.yellow;
    else if (color.equals("cyan"))
      return Color.cyan;
    else if (color.equals("green"))
      return Color.green;
    else if (color.equals("darkGray"))
      return Color.darkGray;
    else
      return Color.black;
  }

  public synchronized void lock() {
    if (lock) {
      try {
	wait();
      } catch (InterruptedException e) {}
    }
  }

  public synchronized void unlock() {
    if (lock) {
      lock = false;
      notify();
    }
  }

  // merge:  append the given list to this one.  
  // NOTE: the given list is NOT copied, so traversals must
  // respect start and end pointers

  public void merge (BGDataList data) {
    BGData start = data.getStart();
    start.setPrev(end);
    end.setNext(start);
    end = data.getEnd();

    long new_min_y = data.getMinY();
    if (new_min_y < min_y)
      min_y = new_min_y;

    long new_max_y = data.getMaxY();
    if (new_max_y > max_y)
      max_y = new_max_y;

    n_data += data.getCount();
    data_total += data.getTotal();
  }

  public BGData getStart() {
    return start;
  }

  public BGData getEnd() {
    return end;
  }

  public long getMaxY() {
    return max_y;
  }

  public long getMinY() {
    return min_y;
  }

  public int getCount() {
    return n_data;
  }

  public long getTotal() {
    return data_total;
  }
}
