/*
        Copyright(C) 1997 Network Flight Recorder, Inc.
        All rights reserved.

        Use and distribution of this software and its source code
        are governed by the terms and conditions of the
        Network Flight Recorder Software License ("LICENSE.TXT")

*/

package graph;


import java.io.*;
import java.util.*;
import java.awt.*;

import nfr.misc.*;

/*
* A COMMENT ON COMMENTS:
*
* This program was written by a Java consultant.  The following list
* of things didn't work well, so I'm fixing them.  Note that some
* comments in this file may be misleading because they refer to the
* old implementation, but I didn't understand them and know to delete
* them.
*
* - It used to pick a bar width for the graph, then rummage around
* looking for a font that is about that size.  This made it crash
* on Netscape.  I had the consultant fix that bug, and the problem
* came back.  Now it picks a font and makes the bars that wide.
* (Actually, you pass in a font and a fontmetric.)
*
* - It only understood a fixed number of pixels for the X axis labels.
* I'm working on fixing that now.
*
*/

//  BarGraph:  a class that implements the bar graph version of the
//  abstract graph class.  

//  the constructor extracts header information from the data file; a viewer
//  client takes responsibility for reading subsequent data in, in "chunks"
//  and merging it into the growing bar graph.  

//  the graph implements horizontal  scrolling methods 
//  to track scroll right/left events, and the viewer
//  client calls "display" in order to generate a graph on a
//  given graphic tableau.  currently, "display" also takes a component
//  argument because I couldn't figure out any better way to induce creation
//  of an off-screen bitmap that to use the component's getImage() method.

//  axis labels are scaled to fit within the given space.  axes are rescaled
//  when their cached images have been set to null by an client resize routine.

public class BarGraph extends Graph {

  // NOTE: GAP must be < MIN_WIDTH
  private static final int MIN_WIDTH=13;//  min size of a bar
  public static final int HASH_WID=8;

  protected Font label_font;
  int label_font_height;
  FontMetrics label_fm;

  Font axis_font;

  BGData display_first;		// first displayed node

  // total number of displayed columns
  int cols;

  // scroll values
  int cur;		// ordinal position in scrolling 

  // input data
  String version;

  Vector titles;
  String y_axis_label;
  String x_axis_label;
  Image  y_axis_image;

  BGDataList data;		// the actual bar graph data

  boolean new_image;		// data updated since last display


  public BarGraph (EOFInputStream file, Font default_font, 
		FontMetrics default_font_metrics) 
       throws java.io.IOException, java.io.EOFException
  {

    display_first = null;
    cols = 1;
    cur = 0;

    label_font = default_font;
    label_font_height = default_font_metrics.getMaxAscent() +
		default_font_metrics.getMaxDescent();
    label_fm = default_font_metrics;

    axis_font = new Font(default_font.getName(),
		default_font.getStyle(), default_font.getSize() + 8);

    // read initial boilerplate from the file, defining
    // titles, labels, etc.

    /* 
     * file should start "bar-graph", followed by version id,  #titles, the
     * titles, x and y axis labels, max y-value, min y-value,
     * and the data (read by BGData)
     */
    String line = file.nextLine();

    if (line.equals("Error:")) {
        line = file.nextLine();
        new popup_bummer(line, "Close" );
        throw new java.io.IOException();
    }

    if (! line.equals("bar-graph")) {
        new nfr.misc.popup_bummer("web server sent "+line
                +", expected 'bar-graph'", "Close" ); 
        throw new java.io.IOException();
    } 

    // read in the version, in case we have some data 
    // format change in the future
    version = file.nextLine();

    line = file.nextLine();

    int n_titles;
    try {
      n_titles = Integer.parseInt(line);
    } catch (NumberFormatException e) {
      System.err.println("Error in titles!");
      throw new java.io.IOException();
    }

    titles = new Vector();
    for (int i=0; i<n_titles; i++) {
      titles.addElement(file.nextLine());
    }

    y_axis_label = file.nextLine();
    x_axis_label = file.nextLine();

    new_image = true;
  }


  /****************************************

   abstract graph methods

   ***************************************/

  // getCanvas: return a canvas representing this data
  public Canvas getCanvas() {
    Canvas c = new BGCanvas(this);
    return c;
  }

  public boolean hasVscroll() {
    return false;
  }

  public boolean hasHscroll() {
    return true;
  }

  // scroll: scroll the list left or right
  public synchronized boolean hScroll (int value) {

    if (display_first == null)
      return (false);

    if (value == cur)
      return (false);

    if (value > cur) {
      while (value > cur && 
	     (display_first.getNext() != null)) {
	cur++;
	display_first = display_first.getNext();
      }
    } else {
      while (value < cur && 
	     (display_first.getPrev() != null)) {
	cur--;
	display_first = display_first.getPrev();
      }
    }

    new_image = true;

    return (true);
  }

  public boolean vScroll (int value) {
    return false;
  }

  public int getHCount() {
    return (data == null) ? 0 : data.getCount();
  }

  public int getVCount() {
    return 1;
  }

  public int getHVisible() {
    return cols;
  }

  public int getVVisible() {
    return 1;
  }

  public Vector getTitles() {
    return titles;
  }


  /********************************

   bar graph methods

   ********************************/


  // updated: return whether image needs to be recalculated because
  // of a change in nature (e.g., scrolling or addition of new data).

  public synchronized boolean updated() {
    boolean needs_new = new_image;
    new_image = false;

    return needs_new;
  }

  public synchronized void merge (BGDataList new_data) {

    // note that we must redisplay
    new_image = true;

    if (data == null) {
      data = new_data;

      // values that won't croak, but won't scroll until after
      // we have done a "display" 
      display_first = data.getStart();
      cur = 1;
    } else {
      data.lock();
      data.merge(new_data);
      data.unlock();
    }
  }

  // display:  display as a bar graph within the space allotted
  public void display(Graphics gdisp, Component c) {

    // left to right
    if (data == null) {
      return;
    }

    gdisp.setFont(label_font);

    int width = c.size().width;
    int height = c.size().height;

    // ensure we're not trying to update the data 
    // while display is going on

    data.lock();

    long min_y = 0;
    long max_y = data.getMaxY();

    // make yourself a new context inside the border
    width -= 2*BORDER; height -= (2*BORDER);

    Graphics g = gdisp.create(BORDER, BORDER, width, height);

    try {

      draw_y_axis(g, c, min_y, max_y);

      draw_bars(g, c, min_y, max_y);

    } finally {
      g.dispose();
      data.unlock();
    }
  }

	int local_x_axis_off = 0;


  public void draw_y_axis(Graphics g, Component c, long min_y, long max_y) {

    long range = max_y - min_y;

    int height = g.getClipRect().height - (X_AXIS_OFF+GAP);
    int width = Y_AXIS_OFF+1;

    Graphics gax = g.create(0, 0, width, height);

    if (y_axis_image == null) {
      // cache your y axis label images
      Font f_save = gax.getFont();
      gax.setFont(label_font);
      ScaledTextVector sv = new ScaledTextVector(y_axis_label, 
						 ScaledTextVector.R_90,
					label_font);
      y_axis_image = sv.getImage(c, gax, height, LAB_HEIGHT);
      gax.setFont(f_save);
    }

/*
* the Y axis line looks stupid and it doesn't have any useful
* information on it anyway.  You can turn it off by changing this
* if (true) to if (false)
*
* bug: The Y axis doesn't come out in the right place because I fixed
* the X axis to have enough space for all the text for the data labels.
* I'll come back and fix it later, but for now, it looks better to
* suppress the Y axis.
*/
if (false) {

    // figure out the order of magnitude for the axis
    int exp = (int)Math.floor(Math.log(range) / Math.log(10));
    long inc = (long)Math.pow(10.0, exp);

    // draw the axis
    int y_origin = height;
    gax.drawLine(Y_AXIS_OFF, 0, Y_AXIS_OFF, y_origin);
    
    // draw some hashes and coordinate labels

    // calculate the space between hashes
    double hashcnt = ((double)range / inc);
    double hash_delta = (double)height / hashcnt;
    
    long data_val = min_y + (inc/10);
    for (double y = y_origin-hash_delta, cnt = 1; 
	 y >= 0; 
	 y -= hash_delta, cnt++, data_val += inc) {

      // major lines
      gax.drawLine(Y_AXIS_OFF-HASH_WID, (int)y, 
		   Y_AXIS_OFF, (int)y);
    }
}

    // draw in the y axis label
    int iwidth = y_axis_image.getHeight(null);
    int iy = Math.max((height-iwidth)/2, 0);
    int ix = Y_AXIS_OFF-HASH_WID-y_axis_image.getWidth(null);
    gax.drawImage(y_axis_image, ix, iy, null);

    gax.dispose();
  }

  public void draw_bars(Graphics g, Component c, 
			long min_y, long max_y) {

    // draw the bars in the available space
    int width = g.getClipRect().width;
    int height = g.getClipRect().height;

	/*
	* find the longest X label
	*/
	for (BGData cur = display_first; cur != null ; 
		cur = cur.getNext()) 
	{
		int size = label_fm.stringWidth(cur.getLabel());
		if (size > local_x_axis_off)
			local_x_axis_off = size;
	}

	/*
	* if it consumes too much of the window, we just chop
	* it off.  This makes labels get truncated, but too bad...
	* the alternative is you end up looking at a blank window
	* wondering where your data is.
	*/
	if (local_x_axis_off > height / 3)
		local_x_axis_off = height / 3;

	/*
	* select the axis font and draw the X axis label at the bottom 
	* of the window, centered.
	*/
	g.setFont(axis_font);
	FontMetrics temp_fm = g.getFontMetrics(axis_font);
	int x_label_baseline = temp_fm.getMaxDescent();
	int x_label_width = temp_fm.stringWidth(x_axis_label);

	g.drawString( x_axis_label, (width - x_label_width)/2
		, height - x_label_baseline );

	/*
	* compute where the X axis should be by coming up from the
	* bottom by the sum of
	*	height of the longest value label
	*	height of the X axis label
	*	gap between the X axis label and the value labels
	*/
	local_x_axis_off += temp_fm.getMaxAscent()+x_label_baseline + 2;

    g.translate(Y_AXIS_OFF+GAP, 0);
    width -= Y_AXIS_OFF+GAP;

    // create a context for the x-axis labels
    Graphics glab = g.create(0, height-local_x_axis_off, width, local_x_axis_off);

    // draw the x axis
    glab.drawLine(0, 0, width, 0);

    // create a clipping rectangle for the bar area
    height -= local_x_axis_off+GAP;
    g.clipRect(0, 0, width, height);

    // if the window is too small to draw anything, give up
    if (height <= LAB_HEIGHT)
      return;

    int bar_width = label_font_height;
    cols = width / ( bar_width + GAP );

    double delta = (float) (max_y - min_y);
    double bar_unit_percent =  (max_y != min_y)
      ? ((double)((height-LAB_HEIGHT) / delta))
      : 0.0;

    int x = 0;
    int i = 0;
    int max_col = 0;
    // get the max value within this set of data
    long max = Long.MIN_VALUE;
    for (BGData cur = display_first;
       cur != null && i < cols; 
       cur = cur.getNext(), i++) {
      if (cur.getValue() > max) {
      max_col = i;
      max = cur.getValue();
      }
    }

    i=0; x=0;
    for (BGData cur = display_first;
	 cur != null && i < cols; 
	 cur = cur.getNext(), x += bar_width+GAP, i++) {

      long data_val = cur.getValue();
      int bar_height = (int)((data_val-min_y) * bar_unit_percent);

      // print a bar of "value" height with label "label"
      g.setColor(cur.getColor());
      g.fillRect(x, height - bar_height, bar_width, bar_height);

/*
* I don't like this feature.  It also behaves erraticly, so I'm turning
* it off for now.
*/
      if (false && (i == max_col)) {
	// add an underlined label at our maximum value
	ScaledTextVector vsv = new ScaledTextVector(Long.toString(data_val),
						    ScaledTextVector.R_0,
						    label_font);
	Image im = vsv.getImage(c, g, label_fm);
	int iwidth=im.getWidth(null);
	int iy=height-bar_height-im.getHeight(null);
	int ix = (x > iwidth+GAP) ? x-iwidth-GAP : x+bar_width+GAP;

	// draw the label to the left or right of the bar, in the
	// margin above it, with a little underline
	g.drawImage(im, ix, iy, null);
	im.flush();
	g.setColor(Color.black);
	g.drawLine(ix, height-bar_height-2, ix+8, height-bar_height-2);
      }

/*
* here is where the X axis labels come from
*/
      // draw our label rotated appropriately 
      ScaledTextVector sv = new ScaledTextVector(cur.getLabel(), 
						 ScaledTextVector.R_90,
						 label_font);
      Image li = sv.getImage(c, glab, label_fm);
      glab.drawImage(li, x, 5, null);
      li.flush();
    }

    glab.dispose();
    g.translate(-(Y_AXIS_OFF+GAP), 0);

  }

}
