/*
        Copyright(C) 1997 Network Flight Recorder, Inc.
        All rights reserved.

        Use and distribution of this software and its source code
        are governed by the terms and conditions of the
        Network Flight Recorder Software License ("LICENSE.TXT")

*/

package graph;


import java.io.*;
import java.util.*;
import java.awt.*;
import nfr.misc.*;

//  PieChart:  a class that implements the pie chart version of the
//  abstract graph class.  

//  the constructor extracts header information from the data file; a viewer
//  client takes responsibility for reading subsequent data in, in "chunks"
//  and merging it into the growing bar graph.  

//  the graph implements no scrolling methods 
//  to track scroll events, and the viewer
//  client calls "display" in order to generate a graph on a
//  given graphic tableau.  currently, "display" also takes a component
//  argument because I couldn't figure out any better way to induce creation
//  of an off-screen bitmap that to use the component's getImage() method.

//  each pie section gets a label printed off in some appropriate direction,
//  based on which half of the circle you're in.

public class PieChart extends Graph {

  private static final int BORDER = 20;	//  pixel border around chart

  private Font font;

  // input data
  String version;

  Vector titles;

  BGDataList data;		// the actual pie chart data
  boolean new_image;		// data updated since last display

  public PieChart (EOFInputStream file, int chunk_size) 
       throws java.io.IOException, java.io.EOFException
  {
    font = new Font("TimesRoman", Font.BOLD, 10);

    // read initial boilerplate from the file, defining
    // titles, labels, etc.

    /* 
     * file should start "pie-chart", followed by version id,  #titles, the
     * titles, x and y axis labels, max y-value, min y-value,
     * and the data (read by BGData)
     */
    String line = file.nextLine();

    if (line.equals("Error:")) {
        line = file.nextLine();
        new popup_bummer(line, "Close" );
        throw new java.io.IOException();
    }

    if (! line.equals("pie-chart")) {
        new nfr.misc.popup_bummer("web server sent "+line
		+", expected 'pie-chart'", "Close" );
        throw new java.io.IOException();
    }

    // read in the version, in case we have some data 
    // format change in the future
    version = file.nextLine();

    line = file.nextLine();

    int n_titles;
    try {
      n_titles = Integer.parseInt(line);
    } catch (NumberFormatException e) {
      System.err.println("Error in titles!");
      throw new java.io.IOException();
    }

    titles = new Vector();
    for (int i=0; i<n_titles; i++) {
      titles.addElement(file.nextLine());
    }

    new_image = true;

    data = new BGDataList(file, chunk_size);
  }


  /****************************************

   abstract graph methods

   ***************************************/

  // getCanvas: return a canvas representing this data
  public Canvas getCanvas() {
    return new PCCanvas(this);
  }

  public Vector getTitles() {
    return titles;
  }


  /********************************

   pie chart methods

   ********************************/


  // updated: return whether image needs to be recalculated because
  // of a change in nature (e.g., scrolling or addition of new data).

  public synchronized boolean updated() {
    boolean needs_new = new_image;
    new_image = false;

    return needs_new;
  }

  public synchronized void merge (BGDataList new_data) 
  {
    // note that we must redisplay
    new_image = true;

    if (data == null) {
      data = new_data;
    } else {
      data.lock();
      data.merge(new_data);
      data.unlock();
    }
  }

  // display:  display as a bar graph within the space allotted
  public void display(Graphics gdisp, Component c) {

    // left to right
    if (data == null)
      return;

    int width = c.size().width;
    int height = c.size().height;

    // ensure we're not trying to update the data 
    // while display is going on
    data.lock();

    // make yourself a new context inside the border
    width -= 2*BORDER; height -= (2*BORDER);
    Graphics g = gdisp.create(BORDER, BORDER, width, height);

    // diameter is the most you can fit in our smallest dimension, minus
    // something for the labels

    int diameter = Math.min(width, height) - BORDER*2;
    int radius = diameter/2;
    if (diameter <= 0)
      return;

    // center the pie
    int x = (width-diameter) / 2;
    int y = (height-diameter) / 2;

    // draw the pie (fill it in black so crappy [nonexistent] antialiasing
    // doesn't show through white
    g.setColor(Color.black);
    g.fillArc(x-1, y-1, diameter+2, diameter+2, 0, 360);

    g.setFont(font);
    FontMetrics fm = g.getFontMetrics();

    try {

      if (data == null || data.getCount() == 0)
	return;

      // what is the total "value" represented by our pie
      // chart; slices are percentages of that value
      long total = data.getTotal();
      if (total == 0)
	return;

      // iterate through the n-1 of the values, assigning subsequent wedges
      // values proportional to the percentages they represent; polish the
      // last wedge off exactly

      int angle = 0;
      int n_data = data.getCount();

      for (BGData cur = data.getStart(); 
	   cur != null; 
	   cur = cur.getNext()) {

	// figure out what our portion of the pie is
	long value = cur.getValue();
	double percent = (double)value / total;
	Color color = cur.getColor();

	// don't let the fill color 
	g.setColor( (color == Color.black) ? Color.white : color );
	
	// and, draw it
	--n_data;
	int arc_angle = (n_data != 0) 
	  ? (int)(360 * percent) 
	  : 360 - angle;

	g.fillArc(x, y, diameter, diameter, angle, arc_angle);

	// draw a separator line between slices
	double rad_angle = (2 * Math.PI) * ((double)angle / 360.0);
	int run = (int)(Math.cos(rad_angle) *  radius);
	int rise = (int)(Math.sin(rad_angle) *  radius);
	int x_org = x+radius;
	int y_org = y+radius;

	g.setColor(Color.black);
	g.drawLine(x_org, y_org, x_org+run, y_org-rise);

	// draw the label on an angle half way into this slice
	rad_angle += ((2 * Math.PI) * ((double)arc_angle / 360.0)) / 2.0;
	x_org += (int)(Math.cos(rad_angle) *  radius);
	y_org -= (int)(Math.sin(rad_angle) *  radius);
	run = (int)(Math.cos(rad_angle) *  GAP);
	rise = (int)(Math.sin(rad_angle) *  GAP);

	g.drawLine(x_org, y_org, x_org+run, y_org-rise);
	String label = cur.getLabel();

	// labels on the left get printed to the left
	if (rad_angle > Math.PI/2 && rad_angle < (3*Math.PI/2))
	  x_org -= fm.stringWidth(label);

	// labels underneath get printed underneath
	if (rad_angle > Math.PI && rad_angle < (2*Math.PI))
	  y_org += fm.getHeight();

	g.drawString(label, x_org+run, y_org-rise);

	angle += arc_angle;
      }

    } finally {
      g.dispose();
      data.unlock();
    }
  }

}
