/*
        Copyright(C) 1997 Network Flight Recorder, Inc.
        All rights reserved.

        Use and distribution of this software and its source code
        are governed by the terms and conditions of the
        Network Flight Recorder Software License ("LICENSE.TXT")

*/

package graph;



import java.io.*;
import java.util.*;
import java.awt.*;

//  SPDataMatrix:  a sparse matrix of scatter plot information.  the
//  data is stored in a hashed column of hash tables; y values are hashed
//  in and return a hash table that can be used to retrieve x values

public class SPDataMatrix {

  private int n_data;			// current # data points
  private int chunk_size;		// # data to read at a crack
  private int xidx, yidx;		// scroll origin

  private Hashtable data;		// y-value indexed rows of data
  private Hashtable x_values;		// list of unique x values

  private Vector y_labels;		// y values in order received
  private Vector x_labels;		// x values in order received

  private final static int AVG_ROWS = 25;	// initial size of row hash 
  private final static int AVG_COLS = 25;	// initial size of column hash 

  private boolean lock = false;


  // initialize a list from the contents of a file
  public SPDataMatrix(int chunk_size) 
  {
    n_data = 0;
    this.chunk_size = chunk_size;
    xidx = yidx = 0;

    data = new Hashtable(AVG_ROWS);
    x_values = new Hashtable(AVG_COLS);

    y_labels = new Vector(AVG_ROWS);
    x_labels = new Vector(AVG_COLS);

    // other slots default to null
  }

  private Color parseColor(String color) {
    if (color.equals("blue"))
      return Color.blue;
    else if (color.equals("red"))
      return Color.red;
    else if (color.equals("yellow"))
      return Color.yellow;
    else if (color.equals("cyan"))
      return Color.cyan;
    else if (color.equals("green"))
      return Color.green;
    else if (color.equals("darkGray"))
      return Color.darkGray;
    else
      return Color.black;
  }

  public synchronized void lock() {
    if (lock) {
      try {
	wait();
      } catch (InterruptedException e) {}
    }
  }

  public synchronized void unlock() {
    if (lock) {
      lock = false;
      notify();
    }
  }

  public Hashtable getRow(int idx) {
    String label = (String) y_labels.elementAt(idx);
    return (label == null) ? null : (Hashtable)data.get(label);
  }

  public String getRowLabel(int idx) {
    return (String)y_labels.elementAt(idx);
  }

  public String getColLabel(int idx) {
    return (String)x_labels.elementAt(idx);
  }


  // number of unique rows/cols
  public int getRowSize() {
    return y_labels.size();
  }

  public int getColSize() {
    return x_labels.size();
  }


  // current scroll origin
  public int yStart() {
    return yidx;
  }

  public int xStart() {
    return xidx;
  }

  // merge:  input another chunk of data into our matrix

  public void merge (EOFInputStream file)
    throws java.io.IOException, java.io.EOFException
  {
    int data_read = 0;

    do {

      try {
	String xlabel = file.nextLine();	// eof is valid here
	String ylabel;	// x/y axis strings

	long value;	// this data point value
	int percent;	// size of point
	Color color;	// point color

	try {
	  ylabel = file.nextLine();
	  value = Long.parseLong(file.nextLine());
	  percent = Integer.parseInt(file.nextLine());
	  color = parseColor(file.nextLine());
	} catch (NumberFormatException e) {
	  // parse error
	  throw new java.io.IOException();
	} catch (java.io.EOFException e) {
	  System.out.println("bad data!!!");
	  throw new java.io.IOException(); // for invalid input
	}

	// the scatter plot is a hash list of hash lists, since it 
	// represents a variable sized 2d array

	// add this data to our list
	SPData cur = new SPData(value, percent, color);

	Hashtable row = (Hashtable)data.get(ylabel);
	if (row == null) {
	  // keep track of new labels in the order received
	  y_labels.addElement(ylabel);

	  // first entry in the row?  start a new one
	  row = new Hashtable(AVG_COLS);
	  data.put(ylabel, row);
	}

	// add this data value to the appropriate row
	row.put(xlabel, cur);

	if (x_values.get(xlabel) == null)
	  {
	    // keep track of unique x values
	    x_labels.addElement(xlabel);
	    x_values.put(xlabel, xlabel);
	  }

	n_data++;
	data_read++;

	if ((n_data % chunk_size) == 0)
	  break;

      } catch (java.io.EOFException e) {
	if (data_read == 0)
	  throw e;
	break;
      }

    } while (true);
  }

}
