/*
        Copyright(C) 1997 Network Flight Recorder, Inc.
        All rights reserved.

        Use and distribution of this software and its source code
        are governed by the terms and conditions of the
        Network Flight Recorder Software License ("LICENSE.TXT")

*/

package graph;



import java.awt.*;
import java.awt.image.*;
import java.util.*;



// ScaledTextVector implements a vector of strings that will
// generate images scaled to size, and rotated in one of four
// compass directions

class ScaledTextVector {

  public static final int R_0 = 0;
  public static final int R_90 = 90;
  public static final int R_180 = 180;
  public static final int R_270 = 270;

  String s_;
  Vector v_;
  Font f_;		// if we use a fixed font
  int rot_;
  int width_, height_;
  int max_len;		// length (chars) of longest string

  // there are about 8 variations to constructors depending
  // on whether you want roation and if you know what font
  // to use.  

  // if a font is specified, always use that font, and don't
  // actually do any scaling.  this is in place because netscape
  // on BSD/OS uses some sort of crucial resource for every font
  // you create and the app runs out of resource before it has
  // the sense to gc

  // also, this object can either take a vector or a string
  // XXX: 
  

  ScaledTextVector(Vector v, int rotation, Font f) {
    v_ = v;
    rot_ = rotation;
    f_ = f;
  }

  ScaledTextVector(String s, int rot, Font f) {
    s_ = s;
    rot_ = rot;
    f_ = f;
  }

/*
  ScaledTextVector(Vector v, int rot) {
    this(v, rot, null);
  }

  ScaledTextVector(String s, int rot) {
    this(s, rot, null);
  }

  ScaledTextVector(Vector v) {
    this(v, R_0);
  }

  ScaledTextVector(String s) {
    this(s, R_0);
  }
*/

  public int getCnt() {
    return ((v_ != null) ? v_.size() : 1);
  }

  public int getWidth() {
    return width_;
  }

  // return the length of the longest string
  private int getMaxLen() {

    if (max_len == 0) {
      if (v_ != null) {
	for (Enumeration e = v_.elements(); 
	     e.hasMoreElements(); ) {
	  max_len = Math.max(max_len, ((String)e.nextElement()).length());
	}
      } else if (s_ != null) {
	max_len = s_.length();
      }
    }

    return max_len;
  }

  private int calcWidth(FontMetrics fm) {

    int width = 0;

    if (v_ != null) {
      for (Enumeration e = v_.elements(); 
	   e.hasMoreElements(); ) {
	width = Math.max(width, fm.stringWidth((String)e.nextElement()));
      }
    } else if (s_ != null) {
      width = fm.stringWidth(s_);
    }

    return width;
  }


  // getCanvas(): return a canvas holding this text
  public Canvas getCanvas() {
    return new ScaledTextCanvas(this);
  }


  // get an image based off of the supplied graphics context
  public Image getImage(Component c, Graphics g, FontMetrics fm) {

    // generate an image from the given vector of strings

    // assume we have a font
    int lines = getCnt();

    int desc = fm.getMaxDescent();
    int ht = fm.getHeight();

    // tally up the sizes of the strings involved
    width_ = calcWidth(fm);
    height_ = ht*lines;
    f_ = fm.getFont();

    // shouldn't ever be 0...
    if (width_ == 0 || height_ == 0) return null;

    Image i = c.createImage(width_, height_);
    if (i == null) return null;

    Graphics ig = i.getGraphics();
    ig.setFont(f_);

    int base = ht - fm.getMaxDescent();	// get baseline;
    if (v_ != null) {
      // copy the strings into an "image"
      for (Enumeration e = v_.elements(); 
	   e.hasMoreElements(); ) {
	String str = (String)e.nextElement();
	ig.drawString(str, 0, base);

	base += ht;
      }
    } else {
	ig.drawString(s_, 0, base);
    }

    // rotate the image, as necessary
    if (rot_ != R_0) {
      Image ni = imageRotate(i, c, width_, height_, rot_);
      i.flush();
      i = ni;
    }

    return i;
  }

  public Image getImage(Component c, Graphics g, int width, int height) {
    FontMetrics fm = g.getFontMetrics(f_);

    return getImage(c, g, fm);
  }

  private Image imageRotate(Image i, Component c,
			    int width, int height, int rot) {

    int src_array[] = new int[width*height];
    PixelGrabber src = new PixelGrabber(i, 0, 0, width, height, 
					src_array,
					0, width);
    try {
      src.grabPixels();
    } catch (InterruptedException e) {
      return i;
    }

    int dst_array[] = new int[height*width];

    for (int x=0; x < width; x++) {
      for (int y=0; y < height; y++) {
	if (rot == R_90) {
	  dst_array[(height*(width-1-x))+y] =
	    src_array[(y*width)+x];
	} else if (rot == R_270) {
	  dst_array[(x*height)+((height-1)-y)] =
	    src_array[(y*width)+x];
	} else /* rot == R_180 */ {
	  dst_array[(y*width)+x] =
	    src_array[(((height-1)-y)*width)+x];
	}
      }
    }

    MemoryImageSource ms = new MemoryImageSource(height, width, 
						 dst_array, 0, height);
    return c.getToolkit().createImage(ms);
  }

}
