/*
        Copyright(C) 1997 Network Flight Recorder, Inc.
        All rights reserved.

        Use and distribution of this software and its source code
        are governed by the terms and conditions of the
        Network Flight Recorder Software License ("LICENSE.TXT")

*/

package graph;


import java.io.*;
import java.util.*;
import java.awt.*;
import nfr.misc.*;

//  ScatterPlot:  a class that implements the scatter plot version of the
//  abstract graph class.  

//  the constructor extracts header information from the data file; a viewer
//  client takes responsibility for reading subsequent data in, in "chunks"
//  and merging it into the growing scatter plot.  

//  the graph implements horizontal and vertical scrolling methods 
//  to track scroll right/left, up/down events, and the viewer
//  client calls "display" in order to generate a graph on a
//  given graphic tableau.  currently, "display" also takes a component
//  argument because I couldn't figure out any better way to induce creation
//  of an off-screen bitmap that to use the component's getImage() method.

//  axis labels are scaled to fit within the given space.  axes are rescaled
//  when their cached images have been set to null by an client resize routine.

public class ScatterPlot extends Graph {

  private static final int BORDER = 20;	//  pixel border around chart

  // NOTE: GAP must be < MIN_WIDTH

  private static final int MIN_SIZE=5;  //  min size of a point
  private static final int MAX_SIZE=40; //  max size of a point

  private static final int MIN_WIDTH=15;//  min size of a slot
  private static final int MAX_WIDTH=MAX_SIZE+GAP ;//  max size of a slot


  protected Font label_font;
  protected FontMetrics label_fm;

  // total number of displayed columns & rows
  int row_cnt;
  int col_cnt;

  // scroll values
  int xidx, yidx;		// ordinal position in scrolling 

  // input data
  String version;

  Vector titles;

  SPDataMatrix data;		// the actual bar graph data

  boolean new_image;		// data updated since last display

  public ScatterPlot (EOFInputStream file, int chunk_size) 
       throws java.io.IOException, java.io.EOFException
  {

    label_font = new Font("TimesRoman", Font.BOLD, 10);
    label_fm = null;

    xidx = 0;
    yidx = 0;
    row_cnt = 0;
    col_cnt = 0;

    // read initial boilerplate from the file, defining
    // titles, labels, etc.

    /* 
     * file should start "bar-graph", followed by version id,  #titles, the
     * titles, x and y axis labels, max y-value, min y-value,
     * and the data (read by BGData)
     */
    String line = file.nextLine();

    if (line.equals("Error:")) {
        line = file.nextLine();
        new popup_bummer(line, "Close" );
        throw new java.io.IOException();
    }

    if (! line.equals("scatter-plot")) {
        new nfr.misc.popup_bummer("web server sent "+line
                +", expected 'scatter-plot'", "Close" ); 
        throw new java.io.IOException();
    } 

    // read in the version, in case we have some data 
    // format change in the future
    version = file.nextLine();

    line = file.nextLine();

    int n_titles;
    try {
      n_titles = Integer.parseInt(line);
    } catch (NumberFormatException e) {
      System.err.println("Error in titles!");
      throw new java.io.IOException();
    }

    titles = new Vector();
    for (int i=0; i<n_titles; i++) {
      titles.addElement(file.nextLine());
    }

    new_image = true;
    data = new SPDataMatrix(chunk_size);
  }


  /****************************************

   abstract graph methods

   ***************************************/

  // getCanvas: return a canvas representing this data
  public Canvas getCanvas() {
    return new SPCanvas(this);
  }

  public boolean hasVscroll() {
    return true;
  }

  public boolean hasHscroll() {
    return true;
  }

  public int getHCount() {
    return (data == null) ? 0 : data.getColSize();
  }

  public int getVCount() {
    return (data == null) ? 0 : data.getRowSize();
  }

  public int getHVisible() {
    return col_cnt;
  }

  public int getVVisible() {
    return row_cnt;
  }

  public Vector getTitles() {
    return titles;
  }


  /********************************

   scatter plot methods

   ********************************/


  // updated: return whether image needs to be recalculated because
  // of a change in nature (e.g., scrolling or addition of new data).

  public synchronized boolean updated() {
    boolean needs_new = new_image;
    new_image = false;

    return needs_new;
  }

  public synchronized void merge (EOFInputStream file) 
       throws java.io.IOException, java.io.EOFException
  {
    // note that we must redisplay
    new_image = true;

    data.lock();
    try {
      data.merge(file);
    } finally {
      data.unlock();
    }
  }

  // display:  display as a bar graph within the space allotted
  public void display(Graphics gdisp, Component c) {

    // left to right
    if (data == null)
      return;

    int width = c.size().width;
    int height = c.size().height;

    // ensure we're not trying to update the data 
    // while display is going on
    data.lock();

    // make yourself a new context inside the border
    width -= 2*BORDER; height -= (2*BORDER);

    Graphics g = gdisp.create(BORDER, BORDER, width, height);

    g.setFont(label_font);

    try {

      // what is the visible size of our matrix?  take our current
      // scroll origin and compute a square from the number of
      // rows/cols left
      int nrow = data.getRowSize();	  // total # rows, cols
      int ncol = data.getColSize();

      if (nrow == 0 || ncol == 0)
	return;

      // don't take more than will fit on the screen
      int dim = Math.max(nrow, ncol);
      int max_dimension = Math.max(width-Y_AXIS_OFF, height-X_AXIS_OFF);
      dim = Math.min(dim, max_dimension/MIN_WIDTH);

      // allocate slot space, calculate space for each point
      int slot_size = Math.min((max_dimension/dim), MAX_WIDTH);
      int point_size = slot_size - GAP;		// size of biggest point
      int lab_height = Math.min(slot_size, LAB_HEIGHT);

      row_cnt = (height-X_AXIS_OFF)/slot_size;
      col_cnt = Math.min(ncol-xidx, ((width-Y_AXIS_OFF)/slot_size));

      if (label_fm == null) {
	label_font = gdisp.getFont();
	label_fm = gdisp.getFontMetrics();
      }

      // for each (row, col), extract some data
      for (int i = yidx; i > yidx-row_cnt && i >= 0; i--) {

	Hashtable row = data.getRow(i);

	if (row == null)
	  continue;

	int y = ((height-X_AXIS_OFF)-(row_cnt*slot_size))
		 + ((yidx-i) * slot_size);

	// print row label
	ScaledTextVector ysv = new ScaledTextVector(data.getRowLabel(i),
						    ScaledTextVector.R_0,
						    label_font);
	Image yim = ysv.getImage(c, g, label_fm);
	g.drawImage(yim, 0, y+(lab_height/2), null);
	yim.flush();

	for (int j=xidx, max_col=xidx+col_cnt; j < max_col; j++) {
	  String xlabel = data.getColLabel(j);
	
	  // see if we have data filled in for this slot
	  SPData cur = (SPData)row.get(xlabel);

	  int x = (Y_AXIS_OFF+GAP)+(j-xidx) * slot_size;


	  // if there's something there, plot it
	  if (cur != null) {
	    // make a point on the x,y coordinate of the
	    // screen matrix

	    g.setColor(cur.getColor());

	    // dot size varies between MIN_SIZE and point_size
	    int size = MIN_SIZE + 
	      (int)((double)(point_size-MIN_SIZE) * 
		    ((double)cur.getPercent()/100.0));

	    g.fillOval(x+((slot_size-size)/2), 
		       y+((slot_size-size)/2), size, size);
	  }

	  if (i == yidx) {
	    // first time through, draw x axis labels
	    // print row label
	    ScaledTextVector xsv = new ScaledTextVector(xlabel,
							ScaledTextVector.R_90,
							label_font);
	    Image xim = xsv.getImage(c, g, label_fm);
	    g.drawImage(xim, x+(lab_height/2), height-X_AXIS_OFF+GAP, null);
	    xim.flush();

	  }
	}
      }

    } finally {
      g.dispose();
      data.unlock();
    }
  }


  // hScroll: scroll the list left or right
  public synchronized boolean hScroll (int value) {

    if (data == null)
      return (false);

    // indeces are 0 based
    value -= 1;

    if (value == xidx || 
	 value < 0 ||
	 value >= getHCount())
      return (false);

    xidx = value;
    new_image = true;

    return (true);
  }

  // vScroll: scroll the list up or down
  public boolean vScroll (int value) {
    if (data == null)
      return (false);

    int max = getVCount();
    int visible = getVVisible();

    // indeces are 0 based
    // value -= 1;
    int new_yidx = max - value;
    if (yidx == new_yidx || 
	 new_yidx < 0 ||
	 new_yidx >= max)
      return (false);

    yidx = new_yidx;
    new_image = true;

    return (true);
  }

}
