/*
        Copyright(C) 1997 Network Flight Recorder, Inc.
        All rights reserved.

        Use and distribution of this software and its source code
        are governed by the terms and conditions of the
        Network Flight Recorder Software License ("LICENSE.TXT")

*/

package graph;


import java.applet.*;
import java.awt.*;
import java.io.*;
import java.net.URL;
import java.util.*;
import nfr.misc.*;

//  Viewer: Abstract class that provides the data acquisition
//  and thread environment for a graph viewer

public abstract class Viewer 
   extends Applet
   implements Runnable
{
  EOFInputStream file;  			// file with graph data

  public boolean is_dead = false;

  protected final static int MAX_CHUNK=40;	// max to read in one "chunk" 

  private Thread thread;		// file input thread
  protected Graph graph;		// the graph itself

  protected Scrollbar hscroll;		// our h scroll bar
  protected Scrollbar vscroll;		// our v scroll bar
  protected Panel panel;		// our display panel
  protected Canvas display;		// graph's main display
  private URL url;

  repTile r;

  Viewer (URL url) {
    super();
    this.url = url;
  }

  Viewer() {
    super();
  }

  // initialize the underlying graph
  protected abstract Graph initGraph (EOFInputStream file) 
       throws java.io.IOException;


  //  init: one-time initialization from the dataset.  open the 
  //  dataset, fire up the graph display code (which will
  //  read in certain parameters), arrange the applet components
  //  on the screen

  public void init() {

    try {
      file = new EOFInputStream(url.openStream());

    } catch (java.io.IOException e) {
      System.out.println("Error opening " + url.toString() 
		+ "  " + e.toString());
      is_dead = true;
    }
    try {

      // initialize the underlying graph boilerplate
      graph = initGraph(file);
    } catch (java.io.IOException e) {
      System.out.println("Error in initGraph " + url.toString() 
		+ "  " + e.toString());
      is_dead = true;

	return;
    }

    if (graph == null)
	{
	System.out.println("Viewer.java: graph is null.  Why?");
      is_dead = true;
	return;
	}

    setBackground(Color.white);

    // make a scaleable represntation of the major titles
    ScaledTextVector sv = new ScaledTextVector(graph.getTitles(),
	ScaledTextVector.R_0, nfr.misc.NFRInfo.default_font);

    // get the object to represent the main graph
    display = graph.getCanvas();

    // get us some scroll bars
    hscroll = graph.hasHscroll()
      ? new Scrollbar(Scrollbar.HORIZONTAL, 0, 0, 0, 0)
      : null;

    vscroll = graph.hasVscroll() 
      ? new Scrollbar(Scrollbar.VERTICAL, 0, 0, 0, 0)
      : null;

    // make a panel to hold the various graph components
    panel = new ViewPanel(graph,
			  sv.getCanvas(),
			  display,
			  hscroll,
			  vscroll);

    r = new repTile();
    setLayout(r);
    add(panel);
    r.setConstraints(0,0,0,0,"",panel);
    r.setStretchy(0,0);

    setScroll(graph);
  }


  //  start: fires up a thread that will read the applet data

  public void start() {

    if (thread == null) {
      thread = new Thread((Runnable)this);
      thread.start();
    }

  }

  //  stop:  close off your threads when you stop
  public void stop() {

    if (thread != null) {
      thread.stop();
    }

  }


  // reset the scroll region, based on how much data we have and
  // the number of columns/rows last painted

  public void setScroll(Graph graph) {

    if (hscroll != null) {
      int n_visible = Math.max(graph.getHVisible(), 1);
      int n_data = Math.max(graph.getHCount(), 1);
      int value = Math.max(hscroll.getValue(), 1);

      hscroll.setValues(value, n_visible, 1, n_data);
      graph.hScroll(value);
    }

    if (vscroll != null) {
      int n_visible = Math.max(graph.getVVisible(), 1);
      int n_data = Math.max(graph.getVCount(), 1);
      int value = Math.max(vscroll.getValue(), 1);
System.out.println("vs: value="+value+" n_vis="+n_visible);
      vscroll.setValues(value, n_visible, 1, n_data);
      graph.vScroll(value);
    }
  }

  boolean hmoving = false;
  boolean vmoving = false;

  // Handle gui events
  public boolean handleEvent(Event e) {

    if(hscroll != null &&
       e.target == hscroll) {

      // events within the horizontal scroll bar

      switch (e.id) {

      case Event.SCROLL_ABSOLUTE:
	// if you're scrolling, just make a note that you've started
	hmoving = true;
	break;

	// evidently, you get a mouse move when you end scrolling
      case Event.MOUSE_EXIT:
      case Event.MOUSE_MOVE:
	if (!hmoving)
	  break;	// normal mouse in the scrollbar

	// otherwise, fall through
	hmoving = false;

      case Event.SCROLL_PAGE_UP:
      case Event.SCROLL_PAGE_DOWN:
      case Event.SCROLL_LINE_UP:
      case Event.SCROLL_LINE_DOWN:
	// attempt scroll to the "value"th bar

	int value = hscroll.getValue();
	if (graph.hScroll(value)) {
	  // if we modified our position, repaint & adjust our
	  // scroll values
	  display.repaint();

	  setScroll(graph);
	}
      }
    } else if(vscroll != null &&
	      e.target == vscroll) {

      // events within the vertical scroll bar

      switch (e.id) {

      case Event.SCROLL_ABSOLUTE:
	// if you're scrolling, just make a note that you've started
	vmoving = true;
	break;

	// evidently, you get a mouse move when you end scrolling
      case Event.MOUSE_EXIT:
      case Event.MOUSE_MOVE:
	if (!vmoving)
	  break;	// normal mouse in the scrollbar

	// otherwise, fall through
	vmoving = false;

      case Event.SCROLL_PAGE_UP:
      case Event.SCROLL_PAGE_DOWN:
      case Event.SCROLL_LINE_UP:
      case Event.SCROLL_LINE_DOWN:
	// attempt scroll to the "value"th bar
	int value = vscroll.getValue();
	if (graph.vScroll(value)) {
	  // if we modified our position, repaint & adjust our
	  // scroll values
	  display.repaint();

	  setScroll(graph);
	}
      }
    }

    return super.handleEvent(e);
  }


        public void reshape ( int x, int y, int w , int h )
        {
                super.reshape(x,y,w,h);
System.out.println("View-ER reshape "+x+" "+y+" "+w+" "+h);
        }




}
