/*****************************************************************************
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT")


  File:   GoodFrame.java
  Author: Andrew Lambeth <andrew@nfr.net>

 *****************************************************************************/

package nfr.misc;

import nfr.misc.*;
import java.awt.*;
import java.util.*;


// GoodFrame adds register() and unregister() methods to Frame, so
// that a window hierarchy may be created.  Any windows which have
// been registered with a particular GoodFrame will be destroyed when
// that GoodFrame is destroyed.

// registration is accomplished automatically by calling
// GoodFrame(Component c) from any derived class's constructor.  The
// argument c is any component which belongs to the desired parent's
// heirarchy (usually the parent window itself should be passed in,
// but that isn't always available).

// GoodFrame also allows derived classes to prevent themselves from
// being closed or to perform cleanup before closing by overriding the
// beforeDisposal() method.  This method is called whenever the
// Frame's dispose() method is called.  There are three possible
// return values:

//     AGAIN - this prevents destroying the Frame and causes the 
//             beforeDisposal method to be called again
//     NOT_DISPOSABLE - this prevents destroying the Frame.
//     DISPOSABLE - this causes the Frame to be destroyed
//     Any other value is treated as if it were DISPOSABLE

// beforeDisposal() is run in a new thread so it is ok for it to block
// while waiting for user input (possibly calling MsgBox.wait()).  If
// you override beforeDisposal(), be sure to either call
// super.beforeDisposal() or to call unregister() in your version.

// This class would also be a good place to put things such as event
// handling of the window destroy event, when (if) we figure out how
// to get it.

public class GoodFrame extends Frame
{

public static final int AGAIN = -1;
public static final int NOT_DISPOSABLE = 0;
public static final int DISPOSABLE = 1;
  Vector mChildren = null;
  Window mParent = null;

public GoodFrame()
  {
    dbg.out("in GoodFrame()", 0);
    mChildren = new Vector();
  }

public GoodFrame(Component c)
  {
    dbg.out("in GoodFrame(c = " + c + ")", 0);

    mChildren = new Vector();

    // find the closest Window in c's ancestory
    while (c != null)
      {
	if( c instanceof Window)
	  {
	    mParent = (Window)c;
	    dbg.out("found an ancestral window", 0);
	    break;
	  }
	c= c.getParent();
      }

    // if my parent is also a GoodFrame, add myself to the hierarchy
    if((mParent != null) && (mParent instanceof GoodFrame))
      {
	dbg.out("my parent is also a GoodFrame", 0);
	((GoodFrame)mParent).register(this);
      }
    dbg.out("leaving GoodFrame()", 0);
  }

public int beforeDisposal()
  {
    if((mParent != null) && (mParent instanceof GoodFrame))
      ((GoodFrame)mParent).unregister(this);
    return DISPOSABLE;
  }

public void dispose()
  {
    Thread t = new Thread(new GoodFrameExit(this));
    t.start();
    System.out.println("spun off GoodFrameExit");
  }

  void super_dispose() { super.dispose(); }

public void register(Window w)
  {
    System.out.println("GoodFrame.register()");
    mChildren.addElement(w);
  }

public void unregister(Window w)
  {
    System.out.println("GoodFrame.unregister()");
    while(mChildren.removeElement(w))
      System.out.println("found it");
  }

public void destroyChildren()
  {
    System.out.println(mChildren.size() + " children to destroy");
    for(int i = 0; i < mChildren.size(); i++)
      {
	Window w = (Window)mChildren.elementAt(i);
	System.out.println("dispose window: " + w);
	w.dispose();
      }
    mChildren.removeAllElements();
  }
}

class GoodFrameExit implements Runnable
{
  GoodFrame mTarget = null;

  GoodFrameExit(GoodFrame gf)
  {
    mTarget = gf;
  }

public void run()
  {
    System.out.println("GoodFrameExit run()");
    switch(mTarget.beforeDisposal())
      {
      case GoodFrame.AGAIN:
	run();
	break;

      case GoodFrame.NOT_DISPOSABLE:
	break;

      case GoodFrame.DISPOSABLE:
      default:
	if((mTarget.mParent != null) && (mTarget.mParent instanceof GoodFrame))
	  ((GoodFrame)mTarget.mParent).unregister(mTarget);
	mTarget.destroyChildren();
	mTarget.super_dispose();
	break;

      }
  }
}
