/*****************************************************************************
  Copyright(C) 1996, 1997 Network Flight Recorder, Inc.
  All rights reserved.

  Use and distribution of this software and its source code
  are governed by the terms and conditions of the
  Network Flight Recorder Software License ("LICENSE.TXT")


  File:   SortView.java
  Author: Andrew Lambeth <andrew@nfr.net>

 *****************************************************************************/

//////////////////////////////////////////////////////////////////////////

// SortView is a class used for displaying a list of Objects in
// mutiple sort orders.  Each sort order is an instance of the
// ListView class.  The base ListView class is unsorted, and objects
// appear in the order in which they are inserted.  To create a sorted
// ListView, extend the class and override the compare(Object, Object)
// method.  ListViews are added to the SortView using addView(String,
// ListView), where string is the name to assign to this ListView.
// addView() creates a button with this name above the listbox
// display.  Pressing this button will cause the objects to be
// displayed in the order dictated by that ListView's compare()
// method.  The string which appears in the listbox is generated using
// the inserted object's toString() method.

//////////////////////////////////////////////////////////////////////////


package nfr.misc;

import java.awt.*;
import java.util.*;

public class SortView extends Panel
{
  Panel mViews = null;
  Panel mViewBtns = null;
  repTile mViewBtnsLayout = null;
  int mMinH = 0;
  int mMinW = 0;
  Button mCurSortBtn = null;

  Color mActiveBtnColor = Color.black;
  Color mInactiveBtnColor = Color.black;
  Color mListFg = Color.black;
  Color mListBg = Color.white;
  Font mActiveBtnFont = null;
  Font mInactiveBtnFont = null;

  //Object mCurSelection = null;
  Object[] mCurSelections = null;
  int mViewCount = 0;
  int[] mViewWidths;
  int mSpaceWidth = 0;
  char[] mSpaceVault;
  boolean mMultiSelect = true;

public SortView()
  {
    construct();
  }

public SortView(int minH, int minW)
  {
    mMinH = minH;
    mMinW = minW;
    construct();
  }

public int count()
  {
    // return the number in the largest View
    int max = 0;
    Component [] views = mViews.getComponents();
    for (int i = 0; i < mViews.countComponents(); i++)
      {
	if(((ListView)views[i]).count() > max)
	  max = ((ListView)views[i]).count();
      }    
    return max;
  }

public void setListColors(Color fg, Color bg)
  {
    mListFg = fg;
    mListBg = bg;
    Component [] views = mViews.getComponents();
    for (int i = 0; i < mViews.countComponents(); i++)
      {
	((ListView)views[i]).setForeground(mListFg);
	((ListView)views[i]).setBackground(mListBg);
      }
  }

void construct()
  {
    setFont(NFRInfo.font);

    
    mActiveBtnFont = new Font(NFRInfo.font.getName(),
			      Font.BOLD,
			      NFRInfo.font.getSize());

    mInactiveBtnFont = NFRInfo.font;

    repTile l = new repTile();
    setLayout(l);

    mViewWidths = new int[32];
    mViews = new Panel();
    mViews.setLayout(new CardLayout());
    mViewBtns = new Panel();
    mViewBtnsLayout = new repTile();
    mViewBtns.setLayout(mViewBtnsLayout);
      
    add(mViews);
    add(mViewBtns);

    l.setConstraints(0,0,0,0,"", mViewBtns);
    l.setConstraints(1,0,1,0,"", mViews);
    l.setStretchy(1,0);

    mSpaceWidth = CtoP.Width(" ", this);
    mSpaceVault = new char[512];
    for(int i = 0; i < mSpaceVault.length; i++)
      mSpaceVault[i] = ' ';
  }

public void build()
  {
    mViewBtnsLayout.setStretchy(-1,mViewCount - 1);
    // move the selected view to the front
    ((CardLayout)mViews.getLayout()).show(mViews, mCurSortBtn.getLabel());    
  }

public void addView(String name, ListView lv)
  {
    addView(name, lv, 0, false);
  }

public void addView(String name, ListView lv, int btnWidth)
  {
    addView(name, lv, btnWidth, false);
  }

public void addView(String name, ListView lv, int btnWidth, boolean select)
  {
    mViews.add(name, lv);
    lv.setMinSize(mMinH, mMinW);
    lv.setBackground(mListBg);
    lv.setForeground(mListFg);
    Button b = new Button(name);
    mViewBtns.add(b);
    mViewBtnsLayout.setConstraints(0,mViewCount, 0, mViewCount, "", b);
    mViewBtnsLayout.setMinSize(CtoP.Width(btnWidth,this),1,b);
    mViewWidths[mViewCount] = CtoP.Width(btnWidth,this);
    if((mCurSortBtn == null) || select)
      highlightButton(b);
    mViewCount++;
  }

public void insert(Object newThing)
  {
    String newThingString = null;
    if(newThing instanceof Formattable)
      {
	StringBuffer sbuf = new StringBuffer();
	String[] fields = ((Formattable)newThing).getFields();
	for (int i = 0; i < (fields.length - 1); i++)
	  {
	    sbuf.append(fields[i]);
	    int fill = mViewWidths[i] - CtoP.Width(fields[i], this);
	    fill = fill / mSpaceWidth;
	    
	    // bug: should do something here to preserve the column
	    // width when the field is too big
	    if(fill > 0)
	      sbuf.append(mSpaceVault, 0, fill);
	  }
	sbuf.append(fields[fields.length - 1]);
	newThingString = new String(sbuf);
      }
    else
      newThingString = String.valueOf(newThing);

    Component [] views = mViews.getComponents();
    for (int i = 0; i < mViews.countComponents(); i++)
      ((ListView)views[i]).insert(newThing, newThingString);
  }

public void remove(Object thing)
  {
    Component [] views = mViews.getComponents();
    for (int i = 0; i < mViews.countComponents(); i++)
      ((ListView)views[i]).remove(thing);
  }

//public Object getCurSelection() { return mCurSelection; }

public Object[] getCurSelections() { return mCurSelections; }

public void clear()
  {
    Component [] views = mViews.getComponents();
    for (int i = 0; i < mViews.countComponents(); i++)
      ((ListView)views[i]).clear();
    validate();
  }

public void setMinSize(int w, int h)
  {
    Component [] views = mViews.getComponents();
    for (int i = 0; i < mViews.countComponents(); i++)
      ((ListView)views[i]).setMinSize(w,h);
  }

  void highlightButton(Button b)
  {
    if(mCurSortBtn != null)
      {
	mCurSortBtn.setForeground(mInactiveBtnColor);
	mCurSortBtn.setFont(mInactiveBtnFont);
      }

    mCurSortBtn = b;

    mCurSortBtn.setForeground(mActiveBtnColor);
    mCurSortBtn.setFont(mActiveBtnFont);
  }

public boolean onDoubleClick(Event e)
  {
    System.out.println("got a double click");
    mCurSelections = ((ListView) e.target).getSelectedObjects();
    e.target = this;
    return false;
  }

public boolean handleEvent(Event e)
  {
    if(e.id == Event.LIST_SELECT)
      {
	mCurSelections = ((ListView) e.target).getSelectedObjects();
	return true;
      }
    else
      return super.handleEvent(e);
  }

public boolean action(Event e, Object arg)
  {
    //System.out.println("action() EVENT: " + e);
    if(e.target instanceof Button)
      {
	// change the appearance of the sort buttons to reflect the
	// new view selction
	highlightButton((Button)e.target);

	// move the selected view to the front
	((CardLayout)mViews.getLayout()).show(mViews,(String)arg);

	// now change the selected item in each view to reflect the
	// currently visible view, we only do it when the views are
	// shuffled since it doesn't really matter until now
	// bug: should really only do the one that's showing
	if(mCurSelections != null)
	  {
	    Component [] views = mViews.getComponents();
	    for (int i = 0; i < mViews.countComponents(); i++)
	      {
		for (int j = 0; j < mCurSelections.length; j++)
		  ((ListView)views[i]).select(mCurSelections[j]);    
	      }
	  }
	return true;
      }
    if(e.target instanceof ListView)
      {
	return onDoubleClick(e);
      }
    else
      return false;
  }

}
