/*
        Copyright(C) 1997 Network Flight Recorder, Inc.
        All rights reserved.

        Use and distribution of this software and its source code
        are governed by the terms and conditions of the
        Network Flight Recorder Software License ("LICENSE.TXT")

By Mark Sienkiewicz / NFR

*/


/*

The repTile layout manager lays things out in rows and columns.
- unlike the gridLayout, the rows and columns can be different sizes
- unlike gridBagLayout, it is pretty easy to get things to go where you
  want them.

You assign to each object a starting row/column and ending row/column.
repTile computes the size of each row and column from the preferred
sizes of the objects in those rows/columns.

There are really two treatements that objects get: 

- If an object starts and ends in the same column, it's preferred size
  is a minimum size for that column
- if an object ends in a different column than it starts in, it's size
  does not affect the size of the column

The same applies to rows.

One row and one column are designated "stretchy".  Use 
	r.setStretchy(row,col);
to select which one.  You can pass in -1 to mean you don't want a
stretchy row or column.  

A stretchy column grows to fill the space left over after all the 
other columns got their space allocated.  Usually, you would have
something like a TextArea span a stretchy column so it will grow
to the size of the window.  The same applies to rows.

Note that it is not required to have something in every row or in
every column.  Columns/rows can have size of 0, and certainly will
if some object does not begin and end in the same column/row.

When you set the start/end rows on an object, you can also set
flags that vary how the object it placed.  The flags are
passed in as a string.  Possible values are:
	ch	center object horizontally in space given to it
	cv	center object vertically in space given to it
	fl	object should be flush left in space given to it
	fr	object should be flush right
	ft	object should be flush to the top of space given to it
	fb	object should be flush bottom
You can concatenate flags:
	"ch ft" indicates centered horizontally but flush with the top

[ Note: ch is well tested and known to work; other flags are believed 
to work but have not actually been used much. ]


If making a new object:

	Panel p = new Panel();		// creates a panel object

	repTile r = new repTile();	// creat repTile object
	p.setLayout(r);			// make it the layout manager

If you want to use it in your applet:
	repTile r = new repTile();
	this.setLayout(r);

Now you make subordinate objects as you would normally, but you
use the repTile constraints to tell them where to go:
	Button b = new Button("hello");
	p.add(b);
	r.setConstraints(0,0,0,0,"",b);

The parameters to setConstraints() are:
	row
	col
		the starting row and column
	endrow
	endcol
		the ending row and column
	flags
		a string showing all the flags you want
	the actual object

You can set a minimum size for an object by
	r.setMinSize(width, height, b);
This will make the object at LEAST that number of pixels large.  It
may be bigger if the objects preferred size is bigger than what you
specify.

You shouldn't pass constants to this unless you absolutely have to.
If you do, you are making assumptions about the actual sizes of the
things in the run time environment.  If you're wrong, it can look really
bad.

You can specify a border around the whole area of the repTile by
using
	r.setborder(leftborder,rightborder,topborder,bottomborder);
It will leave the designated number of pixels of empty space on
each side.


Non-features:

I intended to allow borders around elements placed on the form,
but it didn't work right away and I decided I didn't need the
feature enough to fix it.

There are probably much more "java-like" ways to do some of this.
For example, the border might be better implemented with "insets",
but I don't feel like figuring out what they are and how to
use them.  

The name is a pun on the fact that it is a tiling-type layout manager
and that it is somewhat evolved from the more primitive layout managers
that are native to java.

*/

package nfr.misc;
import java.awt.*;
import java.util.*;

/*
* I found out that this is how you implement what would 
* be called a structure or record in other languages.
*/
class Constraints
{
	public int col;
	public int row;
	public int endrow;
	public int endcol;
	public boolean hcenter;
	public boolean vcenter;
	public boolean flushleft, flushright, flushtop, flushbottom;

	int minwidth, minheight;

	int preferredwidth, preferredheight;	// temp use only

	// doesn't work right - who knows why?  not used.
	public int leftborder, rightborder, topborder, bottomborder;
}

public class repTile implements LayoutManager
{
	Hashtable widgets = null;
	int maxrows = 0;
	int maxcols = 0;
	int stretchy_row = -1;
	int stretchy_col = -1;
	int leftborder, rightborder, topborder, bottomborder;

	public repTile()
	{
	}

        public void addLayoutComponent(String name, Component comp)
        {
        }

	public void removeLayoutComponent(Component comp)
	{
		widgets.remove(comp);
	}

	public void setConstraints(int row, int col, int endrow, int endcol,
		String flags, Component comp)
	{
		if (comp == null)
			return;
		if (widgets == null)
			widgets = new Hashtable(17);
		Constraints c = (Constraints)(widgets.get(comp));
		if (c == null) {
			c = new Constraints();
			widgets.put(comp,c);
		}
		c.row = row;
		c.col = col;
		c.endrow = endrow;
		c.endcol = endcol;
		if (endrow > maxrows - 1 )
			maxrows = endrow + 1;
		if (endcol > maxcols - 1 )
			maxcols = endcol + 1;
		c.hcenter=flags.indexOf("ch")>=0;
		c.vcenter=flags.indexOf("cv")>=0;
		c.flushleft=flags.indexOf("fl")>=0;
		c.flushright=flags.indexOf("fr")>=0;
		c.flushtop=flags.indexOf("ft")>=0;
		c.flushbottom=flags.indexOf("fb")>=0;
	}

	public void setMinSize(int width, int height, Component comp)
	{
		Constraints c = (Constraints)(widgets.get(comp));
		if (c == null) {
			c = new Constraints();
			widgets.put(comp,c);
		}
		c.minwidth=width;
		c.minheight=height;
	}

/* doesn't work:
	public void setObjectBorder(Component comp, int l, int r,
		int t, int b) {
                Constraints c = (Constraints)(widgets.get(comp));
                if (c == null) {
                        c = new Constraints();
                        widgets.put(comp,c);
                }
		c.leftborder = l;
		c.rightborder = r;
		c.topborder = t;
		c.bottomborder = b;
	}
*/

	public void setborder(int l, int r, int t, int b) {
		leftborder = l;
		rightborder = r;
		topborder = t;
		bottomborder=b;
	}

	public void setStretchy(int row, int col)
	{
		stretchy_row=row;
		stretchy_col=col;
	}

	public Constraints getConstraints(Component comp)
	{
		Constraints c = (Constraints)(widgets.get(comp));
		return c;
	}

	public Dimension minimumLayoutSize(Container target)
	{
		return layoutSize(target, false);
	}

	public Dimension preferredLayoutSize(Container target)
	{
		return layoutSize(target, false);
	}

	protected synchronized Dimension layoutSize(Container target, boolean reshape)
	{
		int stretch;

		Dimension dim;

		Insets inset;

		inset = target.insets();

		// how big is our parent?  note you may get 0,0 if it
		// hasn't been realized yet

		dim = target.size();
		int parent_width = dim.width;
		int parent_height = dim.height;

		// dim will be our returned size

		dim = new Dimension();

		// figure out how many widgets we have; allocate arrays
		// for cell boundaries

		int nwidgets = target.countComponents();

		int [] rowsize  = new int [ maxrows + 1];
		int [] rowstart  = new int [ maxrows + 1];
		int [] colsize  = new int [ maxcols + 1];
		int [] colstart  = new int [ maxcols + 1];

		// x and y count up how much space we've used to
		// get to the col/row we are on

		int x = leftborder + inset.left;
		int y = topborder + inset.top;

		// for each widget,
		//	find out how big it is
		//	if it only spans a single row/col, use it to
		//		increase that row/col size

		for (int i = 0 ; i < nwidgets ; i++)
			{
			Component comp = target.getComponent(i);
			Constraints c = (Constraints)(widgets.get(comp));

			Dimension d = comp.preferredSize();


			if ( c == null ) 
				{
				System.out.println(
					"object has no constraints!"
				 + comp.toString());
				}
			else 
				{
				if (d.width < c.minwidth)
					d.width = c.minwidth;
				if (d.height < c.minheight)
					d.height = c.minheight;
				c.preferredwidth = d.width;
				c.preferredheight= d.height;
				if ( c.row == c.endrow)
					if ( rowsize[c.row] < d.height)
						rowsize[c.row] = d.height;
				if (c.col == c.endcol)
					if ( colsize[c.col] < d.width)
						colsize[c.col] = d.width;
				}
			}

		// compute start of each row by adding sizes of previous rows

		for ( int i = 0; i < maxrows; i++)
			{
			rowstart[i] = y;
			y = y + rowsize[i];
			dim.height =Math.max(dim.width,rowstart[i]+rowsize[i]);
			}

		// if we didn't use up all the parent height,
		//	find a stretchy row and make it bigger until we do
		if (maxrows > 0)
			{
			stretch=parent_height-rowstart[maxrows-1]
				-rowsize[maxrows-1]-bottomborder-inset.bottom;
			if ((stretch > 0) && (stretchy_row >= 0))
				{
				rowsize[stretchy_row] += stretch;
				for (int i = stretchy_row + 1; i < maxrows; i++)
					rowstart[i] += stretch;
				}
			}

		// compute start of each col by adding sizes of previous cols

		for ( int i = 0; i < maxcols; i++)
			{
			colstart[i] = x;
			x = x + colsize[i];
			dim.width  =Math.max(dim.width,colstart[i]+colsize[i]);
			}

		// if we didn't use up all the parent width,
		//	find a stretchy col and make it bigger until we do

		if (maxcols > 0)
			{
			stretch=parent_width-colstart[maxcols-1]
				-colsize[maxcols-1]-rightborder-inset.right;
			if ((stretch > 0) && (stretchy_col >= 0))
				{
				colsize[stretchy_col] += stretch;
				for (int i = stretchy_col + 1; i < maxcols; i++)
					colstart[i] += stretch;
				}
			}


		// place each widget

		if (reshape)
			for ( int i = 0; i < nwidgets; i++)
				{
				Component comp=target.getComponent(i);
				Constraints c=(Constraints)(widgets.get(comp));
				if (c == null)
					continue;
				int xp,yp,w,h;
				if (c.hcenter) {
					w= c.preferredwidth;
					xp=colstart[c.col] + 
					(colstart[c.endcol]-colstart[c.col]
						+colsize[c.endcol]) / 2
						- w/2;
				} else if (c.flushleft) {
					w= c.preferredwidth;
					xp = colstart[c.col];
				} else if (c.flushright) {
					w= c.preferredwidth;
					xp = colstart[c.endcol] 
						+ colsize[c.endcol] - w;
				} else {
					w = colstart[c.endcol]-colstart[c.col]
						+colsize[c.endcol];
					xp = colstart[c.col];
				}
				if (c.vcenter) {
					h = c.preferredheight;
					yp=rowstart[c.row] + 
					(rowstart[c.endrow]-rowstart[c.row]
						+rowsize[c.endrow]) / 2
						+ h/2;
				} else if (c.flushtop) {
					h= c.preferredheight;
					yp = rowstart[c.row];
				} else if (c.flushbottom) {
					h= c.preferredheight;
					yp = rowstart[c.endrow] 
						+ colsize[c.endrow] - w;
				} else {
					h = rowstart[c.endrow]-rowstart[c.row]
						+rowsize[c.endrow];
					yp = rowstart[c.row];
				}
/*

doesn't work;  not interesting enough to figure out why:

				if (xp < (c.leftborder+colstart[c.col])) 
					xp =(c.leftborder+colstart[c.col]);
				if (yp < (c.topborder+rowstart[c.row])) 
					yp =(c.topborder+rowstart[c.row]);
				if (xp + w > (colstart[c.endcol]
					+ colsize[c.endcol]) - c.rightborder)
					w = xp+w-(c.rightborder+colstart
					    [c.endcol]+colsize[c.endcol]);
				if (yp + h > (rowstart[c.endrow]
					+ rowsize[c.endrow]) - c.bottomborder)
					h = yp+h-(c.bottomborder+rowstart
					    [c.endrow]+rowsize[c.endrow]);
*/
				comp.reshape( xp , yp, w, h);
				}

		return dim;
	}

	public void layoutContainer(Container target)
	{
		layoutSize(target, true);
	}
}
