/*
        Copyright(C) 1997 Network Flight Recorder, Inc.
        All rights reserved.

        Use and distribution of this software and its source code
        are governed by the terms and conditions of the
        Network Flight Recorder Software License ("LICENSE.TXT")

By Mark Sienkiewicz / NFR

*/


/*

tabthing - a multi page tab thing like Windows has

How to use:

	tabthing t = new tabthing(max_number_of_tabs);
		// you don't *have* to use them all.  it makes an array
		// that big for internal use

	t.addThing(object,"tab label");
		// each added object gets a new page.  the string is the
		// text to display on the tab label for this page

	t.compute();
		// tell it you are done adding things

It will use all the space given to it by the parent layout manager.

That's it.

bugs:  

Only allows one row of tabs currently.  

the way it coordinates the brighter/darker for 3d tabs is kludgey.

*/

package nfr.misc;

import java.awt.*;
import java.util.Hashtable;

class tabtop extends Canvas {
	String label;
	int textx, texty;
	boolean selected = false;

	int sideborder = 2;
	int topborder = 2;

	Color brighter = null, darker = null;

	static int corner_cut = 4;

	public tabtop ( String s ) {
		super();
		label=s;
	}

	/*
	* if they change the background, flush the cached
	* values.  I hope this works the way it LOOKS like it
	* should
	*/
	public void setBackground(Color c )
	{
		brighter=null;
		darker=null;
		super.setBackground(c);
	}


	/*
	* to make a brighter color, average in some white.  To
	* make a darker color, average in some black.
	*
	* note that java.awt.Color.brighter() doesn't work.  If
	* the background color is white, it returns you the same
	* color of white.
	*
	* Here, we look for the background color being too close to
	* white or too close to black.  If so, we pick some arbitrary
	* colors.
	*
	* I have arbitrarily decided that "too white" is 0xd0d0d0
	* or closer to white and "too black" is 0x606060 or darker.
	*
	* Otherwise, color_portion is a weighting to control how
	* much brighter or darker the background gets when you average
	* it with either black or white.  0 means the background
	* contributes nothing and 10 means it contributes everything.
	*
	*
	* the distance is calculated in 3 dimensions, where
	* a^2 + b^2 + c^2 = d^2.  I use a=b=c= 0xd0, so
	* d^s == a^2 * 3
	*
	* bug:  We end up using different highlighting colors for
	* the tabthing that Motif/Windows/whatever uses for buttons.
	* This kind of thing really should be coordinated the other 
	* types of 3d GUI objects.  Too bad there is no way to
	* do that with Java.
	*
	* bug: I haven't looked at what these constants look
	* like on black backgrounds. You may need to tweak them.
	* They look ok (but not fantastic) on white backgrounds.  
	* I conjecture that you just can't make this look good on 
 	* a white background.
	*
	*/
	static int max_distance = 0xd0 * 0xd0 * 3;
	static int min_distance = 0x60 * 0x60 * 3;
	static int color_portion = 3;

	public static Color make_brighter(Color c)
	{
		int r,g,b, distance;
		r=c.getRed();
		g=c.getGreen();
		b=c.getBlue();
		distance = r*r+g*g+b*b;
// System.out.println("rgb: " + r + " " + g + " " + b + "  dist " + distance + " max " + max_distance );
		if (distance > max_distance)
			return new Color(0xa0a0a0);
		if (distance < min_distance)
			return new Color(0xc0c0c0);
		return new Color(
			(r*color_portion + 255 * (10 - color_portion) )/10,
			(g*color_portion + 255 * (10 - color_portion) )/10,
			(b*color_portion + 255 * (10 - color_portion) )/10
			);
	} 

	public static Color make_darker(Color c)
	{
		int r,g,b, distance;
		r=c.getRed();
		g=c.getGreen();
		b=c.getBlue();
		distance = r*r+g*g+b*b;
		if (distance > max_distance)
			return new Color(0x000000);
		if (distance < min_distance)
			return new Color(0x606060);
		return new Color(
			(r*color_portion)/10,
			(g*color_portion)/10,
			(b*color_portion)/10
			);
	}

	/*
	* painting is really a matter of drawing a 2 line border.
	* The line is brighter on the left and top, then darker
	* on the right and bottom.
	*
	* pay attention to sideborder, which is the width of space
	* outside the lines on each side of the tab, and topborder
	* which is space above the tab.
	*/

	public void paint( Graphics gc ) {
		gc.drawString(label,textx, texty);
		Dimension d = size();
		Color fg = this.getBackground();
		if (brighter == null)
			brighter=make_brighter(this.getBackground());
		gc.setColor(brighter);

	/*
	* we are responsible for drawing the top of the card
	* where it cuts through our window.
	*/
	if (! selected)
		{
		/*
		* we are behind, so just draw it passing on
		*/
		gc.drawLine(0,d.height-2,d.width-1,d.height-2);
		gc.drawLine(0,d.height-1,d.width-1,d.height-1);
		}
	else
		{
		/*
		* we are in front, so clear the line where we might
		* not have been selected last time, then draw the
		* notched area on each side of the tab
		*/
		gc.clearRect(1,d.height-1,d.width-2,d.height-2);
		gc.drawLine(0,d.height-2,sideborder,d.height-2);
		gc.drawLine(0,d.height-1,sideborder,d.height-1);
		gc.drawLine(d.width-sideborder-2,d.height-2,
				d.width,d.height-2);
		gc.drawLine(d.width-sideborder-2,d.height-1,
				d.width,d.height-1);
		}

	// left side outer line
		gc.drawLine(sideborder,topborder+corner_cut,
			sideborder,d.height-1);

	// left side inner line
		gc.drawLine(sideborder+1,topborder+corner_cut,
			sideborder+1,d.height-1);

	// left corner outer line
		gc.drawLine(sideborder,corner_cut+topborder,
			corner_cut+sideborder,topborder);

	// left corner inner line
		gc.drawLine(sideborder+1,topborder+corner_cut,
			sideborder+corner_cut,topborder+1);

	// top outer line
		gc.drawLine(corner_cut+sideborder,topborder,
			d.width-1-corner_cut-sideborder,topborder);

	// top inner line
		gc.drawLine(sideborder+1+corner_cut,topborder+1,
			d.width-1-corner_cut-sideborder,topborder+1);

		if (darker == null)
			darker = make_darker(this.getBackground());
		gc.setColor(darker);

	// right corner outer line
		gc.drawLine(d.width-1-corner_cut-sideborder,topborder,
			d.width-1-sideborder,topborder+corner_cut);

	// right corner inner line
		gc.drawLine(d.width-1-corner_cut-sideborder,topborder+1,
			d.width-1-sideborder,corner_cut+topborder+1);

	// right side outer line
		gc.drawLine(d.width-1-sideborder,topborder+corner_cut,
			d.width-1-sideborder,d.height-3);

	// right side inner line
		gc.drawLine(d.width-1-sideborder-1,topborder+corner_cut,
				d.width-1-sideborder-1,d.height-3);

	}

	/*
	* watch carefully - at no time do my fingers leave my hands...
	*
	* The height we need is as high as the largest character + the
	* border at the top of the tab + some space for the 3d border
	* lines to be drawn in.  The 6 is because we draw a line 2 pixels
	* wide.  I account for 3 pixels on each side.
	*
	* The width we need is as wide as the label is in this font,
	* again with 3 pixels on each side for the 2 pixel wide line.
	*
	* If you are offended by magic numbers, knock yourself out
	* trying to abstract this 6 into something else.  Note that
	* you can derive the 6 from the arrangement of drawLine
	* calls in the paint() function.
	*/
	public Dimension preferredSize() {
		FontMetrics fm = this.getFontMetrics(this.getFont());
		Dimension mysize = new Dimension();
		texty = fm.getMaxAscent() + fm.getMaxDescent() + topborder;
		mysize.height = texty + topborder + 6;
		textx = sideborder + 6 ;
		mysize.width = textx + fm.stringWidth(label) + 6 + 
			sideborder;
		return mysize;
	}

	public boolean mouseDown(Event e, int x, int y ) {
		Event e1 = new Event(this, Event.ACTION_EVENT, this );
		this.getParent().handleEvent(e1);
		return true;
	}

}

public class tabthing extends Panel {
	repTile w;			// the layout manager
	int max_tabs;			// max tabs we can have
	int cur_tabs;			// current number of tabs
	int cur_selected=0;
	static int borderline = 20;

	/*
	* data on each tab:
	*  a name/label - displayed on the tab bar
	*  a component to put on the card - probably a composite type,
	*	but it's up to you to build it.
	*  a tabtop widget - the actual clickable creature.
	*/
	
	String [] names;		// label on each tab (must be unique)
	Component [] carray;		// component on card
	tabtop [] barray;		// button at the top

	/*
	* the body of the tab panel is a card layout on a panel
	*/
	Panel cardpanel;
	CardLayout card;


	Color brighter = null, darker = null;

	// You construct the tab thing with
	// tabthing t = new tabthing( max_number_of_tabs )

	public tabthing ( int n_tabs ) {
		names = new String[n_tabs];
		carray = new Component[n_tabs];
		barray = new tabtop[n_tabs];
		max_tabs = n_tabs;
		cur_tabs = 0;
		w=new repTile();
		setLayout(w);
		cardpanel = new Panel();
		add(cardpanel);

		// here theer be lameness - right now I only allow
		// one row of tabs at the top; splitting them into
		// multiple rows will not affect the interface, but
		// it will affect this code:

		w.setConstraints(1,0,1,n_tabs,"",cardpanel);
		w.setStretchy(1,n_tabs);
		card = new CardLayout();
		cardpanel.setLayout(card);
		w.setborder(borderline,borderline,borderline/2,borderline);
	}

	public void addThing( Component c, String s)  {
		int x;
		for (x=0; x< cur_tabs; x++) 
			{
			if (c == carray[x])
				break;
			}
		if (x < max_tabs)
			{
			carray[x]=c;
			names[x]=s;
			cur_tabs=x+1;
			}
		else
			{
			System.out.println("tabthing: added too many tabs:" + " name is " + s);
			// should throw an exception, but then everybody
			// has to write try/catch around this function...
			}
	}
	
	/*
	* call compute() after you have added all the tab elements you
	* want to have.
	*
	* It creates the top part of the cards.
	*/

	public void compute() {
		int x;
		tabtop b;
		Component c;
		for (x=0; x<cur_tabs; x++)
			{
			c=carray[x];
			b = new tabtop(names[x]);
			barray[x]=b;
			this.add(b);
			w.setConstraints(0,x,0,x,"xx",b);
			cardpanel.add(c);
			card.addLayoutComponent(names[x],c);
			}
		select_tab(0);
	}

	/*
	* paint basically draws the line around the card presently
	* displayed.
	*
	* right now, I use the functions in tabtop to compute the
	* brighter/darker colors.
	*
	* Note that the tabtop windows overlap the line across the
	* top of the card.  This is ok, because the tabtops draw the
	* top of the card that is in their window.  You still
	* have to draw it here because there is not a tabtop above
	* every position.
	*/
        public void paint(Graphics gc) {
/*
*  lt   rt
*tr +   +
*   |   |
*br +---+
*/
		Dimension d = cardpanel.size();
		Dimension me = this.size();
		Point p = cardpanel.location();
		int lt,rt,tr,br;
		lt = borderline/3;
		rt = me.width - borderline/2;
		tr = p.y - 1;
		br = me.height - borderline/2;

		if (brighter == null)
			brighter = tabtop.make_brighter(this.getBackground());

		gc.setColor(brighter);

		gc.drawLine(lt,tr,rt,tr);	// top of box; partly obscured
		gc.drawLine(lt,tr-1,rt,tr-1);

		gc.drawLine(lt,tr,lt,br);
		gc.drawLine(lt+1,tr-1,lt+1,br-1);

		if (darker == null)
			darker = tabtop.make_darker(this.getBackground());

		gc.setColor(darker);

		gc.drawLine(lt,br,rt,br);
		gc.drawLine(lt+1,br-1,rt-1,br-1);

		gc.drawLine(rt,br,rt,tr);
		gc.drawLine(rt-1,br-1,rt-1,tr+1);

	}

	/*
	* java 1.0 event function...
	*/

	public boolean action (Event e, Object arg) {
		if (e.target instanceof tabtop)
			for (int x = 0; x< cur_tabs; x++)
				{
				if (barray[x] == e.target)
					{
					select_tab(x);
					return true;
					}
				}
	return false;
	}

	public void select_tab(int which) {
		card.show(cardpanel,names[which]);

		barray[cur_selected].selected=false;
		barray[cur_selected].repaint();

		barray[which].selected=true;
		barray[which].repaint();

		cur_selected=which;
	}
}
